Package: vbpm
Type: Package
Title: Variational Bayes Psychometric Models
Version: 0.9.1
Authors@R: c(
    person("Jinsong", "Chen", email = "jinsong.chen@live.com",
           role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-0157-5469")),
    person("Yi", "Jin", role = "aut",
           comment = c(ORCID = "0000-0002-8604-5992")))
Description: Variational Bayes estimation for a family of psychometric
    measurement models. Two models are provided. Variational Bayes factor
    analysis (vbfa) is a regularized partially confirmatory factor model
    spanning the confirmatory-exploratory continuum via spike-and-slab
    priors on the loadings (Chen, Guo, Zhang, and Pan, 2021
    <doi:10.1037/met0000293>; Chen, 2023 <doi:10.3758/s13428-022-01884-7>;
    Jin and Chen, 2025 <doi:10.1080/10705511.2024.2432612>), with an optional
    dynamic (warm-started) regularization path, an orthogonal bifactor
    parameterization, and optional sparse residual (local dependence)
    estimation through a graphical spike-and-slab prior solved by QUIC
    (Jin, Chen, Yan, and Zhang, 2026 <doi:10.31234/osf.io/dehtv_v2>).
    Regularized MIMIC (vbmimic) extends this to multiple-indicators
    multiple-causes models, placing spike-and-slab priors on both the
    measurement and the structural part (Jin and Chen, 2025
    <doi:10.1080/00273171.2025.2483253>). Companion tools compute SEM-like
    fit statistics, and sweep a factor-count window to report candidate fit,
    criterion, and between-candidate loading-correspondence measurements
    without selecting a count (Chen and Jin, 2026
    <doi:10.48550/arXiv.2607.07159>). Data generators for either model family
    are also provided.
License: GPL-3
Encoding: UTF-8
LazyData: true
Depends: R (>= 4.1)
Imports: Rcpp, MASS, stats
LinkingTo: Rcpp, RcppArmadillo
Suggests: numDeriv, testthat (>= 3.0.0), knitr, rmarkdown
Config/testthat/edition: 3
VignetteBuilder: knitr
URL: https://github.com/Jinsong-Chen/vbpm
BugReports: https://github.com/Jinsong-Chen/vbpm/issues
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-08-24 18:44:24 UTC; HKU
Author: Jinsong Chen [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-0157-5469>),
  Yi Jin [aut] (ORCID: <https://orcid.org/0000-0002-8604-5992>)
Maintainer: Jinsong Chen <jinsong.chen@live.com>
Repository: CRAN
Date/Publication: 2026-09-05 14:00:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(coef,vbpm_fit)
S3method(fit_stats,vbfa)
S3method(fit_stats,vbmimic)
S3method(fit_stats,vbpm_fit)
S3method(plot,pefa)
S3method(print,pefa)
S3method(print,special_effects)
S3method(print,summary.pefa)
S3method(print,vbpm_fit)
S3method(summary,pefa)
export(fit_stats)
export(pefa)
export(sim_fa)
export(sim_lvm)
export(special_effects)
export(ssl)
export(vbfa)
export(vbmimic)
importFrom(MASS,mvrnorm)
importFrom(Rcpp,evalCpp)
importFrom(graphics,axis)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,mtext)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(stats,cor)
importFrom(stats,cov2cor)
importFrom(stats,plogis)
importFrom(stats,rbinom)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
useDynLib(vbpm, .registration = TRUE)

# vbpm

**Variational Bayes Psychometric Models**

`vbpm` is a home for variational Bayes estimation of psychometric measurement
models. Its models are *partially confirmatory*: they span the
confirmatory-to-exploratory continuum, with parameters on *specified* paths
anchored by a design matrix and parameters on *unspecified* paths selected from
the data by continuous spike-and-slab priors.

The package provides **variational Bayes factor analysis** (`vbfa`), where a
design matrix `Q` governs the item loadings, and **regularized MIMIC**
(`vbmimic`), which adds covariates predicting the factors and regularizes both
the measurement part (`Q_A`) and the structural part (`Q_B`).

## Features

- **`vbfa()`** — the factor-analysis estimator.
  - Spike-and-slab regularization of unspecified loadings.
  - **Dynamic exploration**: pass `v0` a decreasing vector for a warm-started
    regularization path (Ročková & George, 2018; Jin, Chen, Yan, & Zhang,
    2026); a scalar `v0` is a single fixed spike. `v0 = 0.001` reproduces the
    fixed-spike estimator exactly.
  - **Bifactor**: `bifactor = TRUE` fits one general factor plus `K`
    orthogonal group factors, with `Q` supplying only the group design --
    so `K` counts group factors, comparable to an oblique `K`-factor
    model. The `general` argument designs the general column (including
    bifactor-(S-1)-style zeros). `orthogonal = TRUE` remains available
    for plain orthogonal models (or hand-built bifactor designs).
  - **Local dependence**: set `ld = TRUE` to estimate a sparse residual
    precision by a graphical spike-and-slab prior (QUIC) — fully exploratory
    by default, or restricted by a residual design matrix `Qe`
    (Jin, Chen, Yan, & Zhang, 2026).
- **`vbmimic()`** — the MIMIC estimator.
  - Spike-and-slab regularization of **both** the measurement design (`Q_A`,
    items on factors) and the structural design (`Q_B`, factors on covariates).
  - Supports the same warm-started `v0` path. The defaults reproduce the
    published estimator exactly.
  - The ELBO is not yet available for this model, so `fit_stats()` returns a
    clearly marked limited result and `pefa()` sweeps `vbfa()` candidates only.
- **`fit_stats()`** — SEM-like fit statistics (RMSEA, SRMR, CFI, TLI, AIC, BIC)
  plus the model's objective. An S3 generic over fitted models: it builds the
  implied covariance from the residual covariance, so it is correct for
  diagonal *and* local-dependence `vbfa` fits, and it reads `orthogonal`/`ld`
  from the fit rather than asking you to repeat them. Information criteria use
  the nominal hard-selected parameter count by default; the numerical
  Jacobian-rank count is an explicit `rank_adjust = TRUE` sensitivity.
- **`special_effects()`** — post-process a bifactor fit: Schmid-Leiman
  proportionality CVs, approximate higher-order parameters (second-order
  loadings per group factor), and testlet/special effect sizes
  (Zhang & Chen, 2024, Eq. 16).
- **`pefa()`** — fit one partially exploratory model for every `K` in a fixed
  consecutive window and report what was measured. It selects no factor count
  and carries no cut, threshold, persistence horizon, collision policy, or
  stopping rule; composing the numbers into a conclusion is yours. Three
  tables:
  - `$sweep` — one row per candidate: `K`, `ELBO`, `AIC`, `BIC`,
    `RMSEA`, `SRMR`, `CFI`, `TLI`, `t`, `iter`, `converged`.
  - `$transitions` — one row per *adjacent* pair: percent-of-largest-positive
    ELBO and BIC gains, plus the seven structural facts `phi_min`, `rmsd`,
    `rmsd_max`, `ari`, `pip_rmsd`, `unmatched_ssl`, `collision`.
  - `$persistence` — three upper-triangular matrices, `phi` (holding
    `phi_min`), `rmsd` (holding `rmsd_max`), and a logical `collision` mask,
    covering *every* ordered pair. Read them directly as
    `x$persistence$phi`; there is no accessor, because selecting a stored
    triangle would only be a synonym for `$`.

  Columns correspond by **minimum sign-aligned squared distance**: each source
  column takes the target minimizing the squared length of `a - s b`, with
  `s = 1` when `a'b >= 0` and `-1` otherwise — Equation 17 of Chen (2023) made
  invariant to whole-column reflection. Backbone columns pair by position, each
  exploratory column then chooses independently, and congruence only breaks
  exact ties. The assignment is deliberately not one-to-one: when two columns
  choose the same target that reuse is a **collision**, reported as evidence
  and never repaired into a forced matching. Every ordered pair is compared
  directly from its two endpoints, never chained through the candidates
  between them. Candidate matrices live in K-named `$loadings` and `$pips`
  lists (`result$loadings[["4"]]`). Convergence is *recorded*, never used to
  suppress a number that is defined: a candidate that stopped at `max_it`
  keeps its matrices and all of its comparisons.
- **`ssl()`** — per-candidate sums of squared loadings, `colSums(Lam^2)`, on
  the same scale as `unmatched_ssl`. It screens nothing and enters no
  correspondence rule; it is what distinguishes a low `phi_min` caused by a
  near-empty column from one caused by structure actually moving. Compact
  `print()`, `summary()`, and `plot()` methods complete the set; none of them
  announces a selected count, because none is chosen.
- **`sim_fa()`** and **`sim_lvm()`** — two complementary data generators.
  `sim_fa()` is driven by the loading *pattern* (items per factor, alternating
  cross-loadings, minor factors) and can generate **higher-order/testlet**
  data directly (`gamma =` second-order loadings, via the Schmid-Leiman
  equivalence); `sim_lvm()` is driven by a loading *matrix* and covers
  predictor designs (observed and latent) and mixed response formats. Both
  simulate local dependence, categorical items, and missingness.

## Installation

### 1. A C++ toolchain (one-time setup)

`vbpm` ships C++ code — the QUIC solver behind the local-dependence branch —
so installing it compiles that code and a C++ toolchain must be present.
Many R users already have one (any package installed *from source* needs it);
check before installing anything:

```r
# install.packages("pkgbuild")
pkgbuild::has_build_tools(debug = TRUE)   # TRUE = you are ready, skip to step 2
```

If that returns `FALSE`:

| system | what to install | where |
|---|---|---|
| **Windows** | **Rtools**, matched to your R version — Rtools45 for R 4.5.x, Rtools44 for R 4.4.x (check with `getRversion()`). Run the installer with its defaults; ≈500 MB installed. | <https://cran.r-project.org/bin/windows/Rtools/> |
| **macOS** | Xcode **command line tools** (not the full Xcode app): run `xcode-select --install` in Terminal and accept the prompt; ≈1 GB. | <https://mac.r-project.org/tools/> |
| **Linux** | Your distribution's R development packages: `sudo apt install r-base-dev` (Debian/Ubuntu) or `sudo dnf install R-devel gcc-c++` (Fedora/RHEL); a few tens of MB. | — |

Restart R afterwards, then re-run the `has_build_tools()` check.

### 2. The package

```r
# install.packages("remotes")
remotes::install_github("Jinsong-Chen/vbpm", build_vignettes = TRUE)
```

Or clone and install from source:

```sh
git clone https://github.com/Jinsong-Chen/vbpm.git
R CMD INSTALL vbpm
```

Then:

```r
library(vbpm)
vignette("vbfa")         # start here
vignette("pefa")         # the factor-count sweep and its three tables
vignette("bifactor")     # bifactor structures, AO/AZ backbones, and sweeps
vignette("vbmimic")      # covariates predicting the factors
citation("vbpm")
```

Building the vignettes needs `knitr`, `rmarkdown`, and pandoc (bundled with
RStudio; from a plain shell, point `RSTUDIO_PANDOC` at it).

## Quick start

```r
library(vbpm)

sim <- sim_fa(N = 500, K = 3, ipf = 6, lam = .7, lac = .3)
Y   <- sim$dat

## partially confirmatory Q: 1 = anchor, 0 = fixed zero, -1 = estimate
Q <- matrix(-1, ncol(Y), 3)
Q[1:2, ] <- 0; Q[1:2, 1] <- 1        # two anchors on factor 1 only

fit <- vbfa(Y, Q)                    # dynamic path on by default; quiet
fit                                  # compact summary (S3 print method)
idx <- fit_stats(fit)                # nominal hard-selected count by default
```

Fits are ordinary named lists with a shared `vbpm_fit` class attached — fields
such as `fit$Lam`, `fit$pi`, and `fit$Phi` remain public API. Focused vignettes
cover `vbfa`, one `pefa()` sweep walked table by table under overextraction,
bifactor/higher-order/testlet structures with AO and AZ backbones, and
`vbmimic`.

MIMIC, with covariates predicting the factors:

```r
B <- matrix(0, 3, 9)                 # sparse structural design: 3 covariates
for (k in 1:3) B[k, (k*3-2):(k*3)] <- .3   # per factor

sim <- sim_lvm(N = 500, K = 3, J = 18, P = 9, b = B, phx = 0)
Y   <- sim$dat[, 1:18]               # items first ...
X   <- sim$dat[, 19:27]              # ... covariates last

## anchor two items per factor; leave the structural part exploratory.
## (With BOTH parts fully exploratory the solution can be rotationally
## ambiguous -- anchor at least one part.)
Q_A <- matrix(-1L, 18, 3)
for (k in 1:3) { a <- which(rep(1:3, each = 6) == k)[1:2]
                 Q_A[a, ] <- 0L; Q_A[a, k] <- 1L }
Q_B <- matrix(-1L, 3, 9)

fit <- vbmimic(Y, X, Q_A, Q_B)
round(fit$B, 2)                      # which covariates predict which factors
```

## References

- Partially confirmatory factor analysis with the Bayesian Lasso: Chen, J.,
  Guo, Z., Zhang, L., & Pan, J. (2021). A partially confirmatory approach to
  scale development with the Bayesian Lasso. *Psychological Methods*, 26(2),
  210–235. <https://doi.org/10.1037/met0000293>
- Generalized PCFA (mixed response formats; missing data under MAR): Chen, J.
  (2021). A generalized partially confirmatory factor analysis framework with
  mixed Bayesian Lasso methods. *Multivariate Behavioral Research*, 57(6),
  879–894. <https://doi.org/10.1080/00273171.2021.1925520>
- Bifactor, higher-order, and testlet effects within GPCFA (including effect
  sizes for testlet effects): Zhang, Y., & Chen, J. (2024). Accommodating and
  extending various models for special effects within the generalized
  partially confirmatory factor analysis framework. *Applied Psychological
  Measurement*, 48(4–5), 208–229. <https://doi.org/10.1177/01466216241261704>
- Fully and partially exploratory factor analysis: Chen, J. (2023). Fully and
  partially exploratory factor analysis with bi-level Bayesian regularization.
  *Behavior Research Methods*, 55(4), 2125–2142.
  <https://doi.org/10.3758/s13428-022-01884-7>
- Variational estimator: Jin, Y., & Chen, J. (2025). Regularized variational
  approximation for partially confirmatory factor analysis. *Structural
  Equation Modeling*, 32(3), 437–449.
  <https://doi.org/10.1080/10705511.2024.2432612>
- Warm-started regularization path and sparse residual/local-dependence
  estimation: Jin, Y., Chen, J., Yan, Z., & Zhang, Y. (2026). Sparse residual
  estimation in partially confirmatory factor analysis. *PsyArXiv preprint*.
  <https://doi.org/10.31234/osf.io/dehtv_v2>
- MIMIC: Jin, Y., & Chen, J. (2025). Regularized variational Bayesian
  approximations for variable selection in extended multiple-indicators
  multiple-causes models. *Multivariate Behavioral Research*.
  <https://doi.org/10.1080/00273171.2025.2483253>
- Fit assessment and factor-number selection: Chen, J., & Jin, Y. (2026).
  Recovering latent structures after variational Bayesian variable selection:
  Fit assessment and factor-number selection in partially exploratory factor
  analysis. *arXiv preprint* arXiv:2607.07159.
  <https://arxiv.org/abs/2607.07159>

## Known limitations

As of 0.9.1:

- **Response types.** The estimators model **continuous Gaussian responses**,
  and in-loop missing-data support covers that case in both `vbfa()` and
  `vbmimic()`. Categorical and mixed responses are **out of scope for the
  current line of releases** — supporting them means adding a threshold model
  and augmenting latent responses, which is a modelling extension rather than
  a feature. Simulating them is supported (`cati`/`noc` in `sim_fa()`,
  `ilvl` in `sim_lvm()`); estimating from them is not.
- **Bifactor `pefa()` sweeps are a research extension.** Their `K` counts group
  factors, and every comparison removes the labelled general column before
  applying exactly the loading/PIP correspondence an ordinary oblique sweep
  uses. General and group columns can nevertheless redistribute common variance
  as `K` changes, so a study that needs both modes should run them on the same
  data, backbone, and window rather than assume that either route must agree.
  The `bifactor` vignette shows one such pair of sweeps, where the bifactor
  criterion path and persistence triangle separate the candidates far less than
  the ordinary ones do. That is a property of that example, not a general
  ranking of the two modes.
- **No calibrated thresholds.** Nothing in the package tells you how large a
  congruence, how small an RMSD, or how steep a gain has to be. The vignettes
  show readings composed in local code and label them as such; none of those
  cuts has been validated against an error rate here.
- **No DIF paths in `vbmimic()`.** Covariates predict the *factors*
  (`Q_B`, i.e. impact); there are no direct covariate-to-item paths, so the
  model cannot currently express or detect differential item functioning.
- **Complete covariates.** `vbmimic()` requires complete `X`: covariates are
  conditioned on rather than modelled, so the model supplies no distribution
  to impute them from.
- **LD objective.** Local-dependence `vbfa()` fits return the
  terminal VECM objective and LD-aware fit statistics. `ELBO` is intentionally
  `NA`: calling the point-updated precision objective a joint variational lower
  bound would be misleading.
- **ELBO for MIMIC.** `vbmimic()` converges on a residual criterion rather than
  the bound, so `ELBO` is `NA`. `fit_stats()` describes the available parameter
  counts but covariance fit indices await a fully specified joint covariance.
- **Fully exploratory MIMIC.** With both `Q_A` and `Q_B` left entirely
  unspecified the model converges cleanly but the solution can be rotationally
  ambiguous. Anchor at least one part.

`vbfa()` reproduces its published reference output. A stored-reference
regression test checks that result during every `R CMD check`, at a strict
relative tolerance of `1e-8` that remains sensitive to material regression.
It is not asserted bitwise: the estimator runs through compiled BLAS/LAPACK,
whose SIMD and FMA choices differ between platforms, so the last one or two
ulps are a property of the build rather than of the estimator.

See `NEWS.md` for the per-version changelog.

## License

GPL-3.

## Authors

- Jinsong Chen (maintainer) — [\@Jinsong-Chen](https://github.com/Jinsong-Chen)
- Yi Jin — [\@jinyi1998](https://github.com/jinyi1998)

## SEM-like fit statistics for vbpm models. Hard selection (posterior
## inclusion probability >= tau) is the default for every statistic; soft
## variants carry an "_S" suffix. fit_stats() is the S3 generic.

## Shared validation for fit_stats() and sweep interfaces that expose the same
## parameter-count policy.  Keep this separate from the Jacobian path so a
## nominal-count request neither inspects nor needs the optional dependency.
.validate_rank_controls <- function(rank_adjust, rank_max_J) {
    if (!is.logical(rank_adjust) || length(rank_adjust) != 1L ||
        is.na(rank_adjust)) {
        stop("`rank_adjust` must be one non-missing logical value.",
             call. = FALSE)
    }
    if (!is.numeric(rank_max_J) || length(rank_max_J) != 1L ||
        is.na(rank_max_J) || !is.finite(rank_max_J) || rank_max_J <= 0 ||
        rank_max_J != floor(rank_max_J)) {
        stop("`rank_max_J` must be one finite positive whole number.",
             call. = FALSE)
    }
    invisible(NULL)
}

#' Fit statistics for vbpm models
#'
#' Computes structural-equation-model fit statistics for a fitted [vbfa()]
#' object under hard selection (an unspecified loading counts as active when its
#' posterior inclusion probability is at least `tau`). Absolute indices
#' (RMSEA, SRMR, CFI, TLI) and information criteria (AIC, BIC) are returned,
#' together with the model's ELBO.
#'
#' @param object A fitted `vbpm_fit` object.
#' @param Y The `N x J` data matrix the model was fit to.
#' @param Q The `J x K` loading design matrix used in the fit.
#' @param tau Hard-selection threshold on the posterior inclusion probability
#'   (default `0.5`).
#' @param gamma Reserved (extended BIC tuning); currently unused in the
#'   returned vector.
#' @param orthogonal Logical; whether the fit is an orthogonal (bifactor)
#'   model, in which case factor correlations are not free parameters. The
#'   default `NULL` reads the setting from the fit object itself, which is the
#'   recommended use. Supplying a value that contradicts the fit raises an
#'   error rather than silently miscounting parameters.
#' @param rank_adjust Logical; if `TRUE`, compute the effective parameter count
#'   as the rank of the Jacobian of the model-implied covariance. This explicit
#'   opt-in requires the suggested `numDeriv` package. The default `FALSE`
#'   uses the nominal parameter count.
#' @param rank_max_J Largest number of observed variables for which an explicit
#'   Jacobian-rank calculation is allowed (default `100`). If `rank_adjust =
#'   TRUE` and `J` exceeds this value, the method errors rather than silently
#'   substituting the nominal count.
#' @param ... Further arguments passed to methods. Unused by the methods
#'   supplied with the package.
#'
#' @return An object of class `vbpm_fit_stats` (a named numeric vector) with
#'   `t_nom`, `t`, `t_S` (parameter counts), `RMSEA`, `SRMR`, `CFI`, `TLI`,
#'   `AIC`, `BIC` (hard selection), `AIC_S`, `BIC_S` (soft selection), `ELBO`,
#'   and `objective`. The last is the model's comparability score: equal to
#'   `ELBO` for a diagonal fit, and the terminal VECM objective under local
#'   dependence, where `ELBO` is `NA`. It is comparable only across fits
#'   sharing the `objective_type` attribute, which is also attached.
#'
#'   The `vbmimic` method returns the same shape with `n_active_coef` in place
#'   of a full soft count and typed `NA` elsewhere; see [vbmimic()].
#'
#' @references
#' Chen, J., & Jin, Y. (2026). Recovering latent structures after variational
#' Bayesian variable selection: Fit assessment and factor-number selection in
#' partially exploratory factor analysis. *arXiv preprint* arXiv:2607.07159.
#'
#' Jin, Y., & Chen, J. (2025). Regularized variational approximation for
#' partially confirmatory factor analysis. *Structural Equation Modeling: A
#' Multidisciplinary Journal*, 32(3), 437-449.
#' \doi{10.1080/10705511.2024.2432612}
#'
#' @examples
#' sim <- sim_fa(N = 300, K = 3, ipf = 6, lam = .7, lac = .3, rseed = 1)
#' Q <- matrix(-1L, ncol(sim$dat), 3)
#' for (k in 1:3) { a <- which(rep(1:3, each = 6) == k)[1:2]
#'                  Q[a, ] <- 0L; Q[a, k] <- 1L }
#' fit <- vbfa(sim$dat, Q)
#' round(fit_stats(fit), 3)
#'
#' \donttest{
#' ## bifactor: orthogonal is read from the fit object, so the same call works
#' Qb <- cbind(1L, Q)
#' fb <- vbfa(sim$dat, Qb, orthogonal = TRUE)
#' round(fit_stats(fb), 3)
#' }
#'
#' @seealso [vbfa()], [pefa()]
#' @export
fit_stats <- function(object, ...) UseMethod("fit_stats")

#' @rdname fit_stats
#' @export
## `...` is required for S3 consistency with the generic fit_stats(object, ...)
fit_stats.vbfa <- function(object, Y = NULL, Q = object$Q, tau = 0.50, gamma = 0.5,
                     orthogonal = NULL, rank_adjust = FALSE, rank_max_J = 100,
                     ...) {
    .validate_rank_controls(rank_adjust, rank_max_J)
    fit <- object

    ## resolve `orthogonal` from the fit object; this used to default to FALSE,
    ## which silently miscounted K(K-1)/2 factor correlations for bifactor fits
    if (is.null(orthogonal)) {
        orthogonal <- isTRUE(fit$orthogonal)
    } else if (!is.null(fit$orthogonal) &&
               !identical(isTRUE(orthogonal), isTRUE(fit$orthogonal))) {
        stop("orthogonal = ", orthogonal, " contradicts the fit, which was ",
             "estimated with orthogonal = ", isTRUE(fit$orthogonal),
             ". Omit the argument to use the fit's own setting.",
             call. = FALSE)
    }

    N <- if (is.null(Y)) fit$nobs else nrow(Y)
    J <- nrow(fit$Lam); K <- ncol(Q); Ns <- N - 1
    m  <- J * (J + 1) / 2                    # unique covariance moments

    if (rank_adjust) {
        if (J > rank_max_J) {
            stop("`rank_adjust = TRUE` requires `J <= rank_max_J` (J = ", J,
                 ", rank_max_J = ", rank_max_J,
                 "). Raise `rank_max_J` or use `rank_adjust = FALSE`.",
                 call. = FALSE)
        }
        if (!requireNamespace("numDeriv", quietly = TRUE)) {
            stop("`rank_adjust = TRUE` requires the suggested package ",
                 "'numDeriv'. Install it or use `rank_adjust = FALSE`.",
                 call. = FALSE)
        }
    }

    ## ---- plug-in variational posterior summaries ----------------------
    Lam_h <- fit$Lam                    # Lambda_hat
    is_ld <- isTRUE(fit$ld)
    Wres  <- if (is_ld) fit$W else diag(1 / fit$PsiInv, J)
    if (is.null(Wres)) stop("Residual covariance is unavailable.", call. = FALSE)
    Phi   <- fit$Phi

    pi_ <- fit$pi;  pi_[Q == 1] <- 1;  pi_[Q == 0] <- 0   # pi_hat_{jk}
    aH  <- (Q == 1) | (Q == -1 & pi_ >= tau)             # hard-selection set
    Lam_H <- Lam_h * aH                                  # hard-selected
    Lam_S <- Lam_h * pi_                                 # soft-selected

    ## ---- model-implied covariances:  Sigma = Lam Phi Lam' + Psi -------
    sigma <- function(L) L %*% Phi %*% t(L) + Wres
    SgH <- sigma(Lam_H); SgS <- sigma(Lam_S)
    S <- if (!is.null(fit$sample_cov)) fit$sample_cov else stats::cor(Y)
    Sg0 <- diag(diag(S))

    ## ---- discrepancy F + plug-in log-likelihood -----------------------
    ## stable log-determinant (det() underflows to 0 for large, collinear S,
    ## which made log(det(S)) = -Inf and RMSEA/CFI degenerate at large J).
    logdet <- function(M) as.numeric(determinant(M, logarithm = TRUE)$modulus)
    F_ML <- function(Sg) {                   # log|Sg| + tr(S Sg^-1) - log|S| - J
        logdet(Sg) + sum(diag(S %*% solve(Sg))) - logdet(S) - J
    }
    ll <- function(Sg) -0.5 * Ns * (J * log(2 * pi) + logdet(Sg) +
                                    sum(diag(S %*% solve(Sg))))
    T_H <- Ns * F_ML(SgH);  T_0 <- Ns * F_ML(Sg0)

    ## ---- parameter counts and degrees of freedom ----------------------
    p_H   <- sum(aH)                          # required + active-unspecified
    n_phi <- if (orthogonal) 0L else K * (K - 1) / 2  # free factor correlations
    edge_H <- matrix(FALSE, J, J)
    edge_S <- 0
    if (is_ld) {
        qe_upper <- upper.tri(fit$Qe)
        edge_H <- qe_upper & (fit$Qe == 1 | (fit$Qe == -1 & fit$q_star >= tau))
        edge_S <- sum(fit$Qe[qe_upper] == 1) +
                  sum(fit$q_star[qe_upper & fit$Qe == -1])
    }
    n_edge <- sum(edge_H)
    t_nom <- p_H + J + n_phi + n_edge

    if (rank_adjust) {
        ## Rank-adjusted count: t = rank(d vech(Sigma) / d theta').  All
        ## derivative inputs and closures deliberately live in this explicit
        ## opt-in branch; the nominal path below does no Jacobian setup.
        resid0 <- if (is_ld) c(diag(fit$Psi), fit$Psi[edge_H]) else diag(Wres)
        th0 <- c(Lam_H[aH], if (n_phi > 0) Phi[lower.tri(Phi)], resid0)
        rebuild_vech <- function(th) {
            L <- matrix(0, J, K); L[aH] <- th[seq_len(p_H)]
            P <- diag(K)
            if (n_phi > 0) {
                P[lower.tri(P)] <- th[p_H + seq_len(n_phi)]
                P[upper.tri(P)] <- t(P)[upper.tri(P)]
            }
            pos <- p_H + n_phi
            rd <- th[pos + seq_len(J)]
            if (is_ld) {
                Prec <- diag(rd, J)
                if (n_edge > 0) {
                    ev <- th[pos + J + seq_len(n_edge)]
                    Prec[edge_H] <- ev
                    Prec[t(edge_H)] <- ev
                }
                Rcov <- solve(Prec)
            } else Rcov <- diag(rd)
            Sg <- L %*% P %*% t(L) + Rcov
            Sg[lower.tri(Sg, diag = TRUE)]
        }
        Delta  <- numDeriv::jacobian(rebuild_vech, th0)
        sv     <- svd(Delta, nu = 0, nv = 0)$d
        t_H    <- sum(sv > 1e-6 * max(sv))    # rank count; == t_nom when identified
    } else {
        t_H    <- t_nom                       # nominal count (rank check skipped)
    }
    t_S <- J + n_phi + sum(Q == 1) + sum(pi_[Q == -1]) + edge_S

    df_H <- m - t_H
    df_0 <- J * (J - 1) / 2

    ## ---- absolute / incremental fit (hard-selected) -------------------
    RMSEA <- if (df_H > 0)
                 sqrt(max(0, (T_H - df_H) / (df_H * Ns)))
             else NA_real_
    si <- 1 / sqrt(diag(S))
    R0 <- si * S   * rep(si, each = J)
    Rh <- si * SgH * rep(si, each = J)
    SRMR <- sqrt(2 * sum((R0 - Rh)[lower.tri(R0, diag = TRUE)]^2) /
                 (J * (J + 1)))
    denom <- max(T_H - df_H, T_0 - df_0, 0)
    ## When both model and baseline discrepancies vanish, the relative-fit
    ## denominator is zero and the selected model is conventionally perfect.
    CFI <- if (denom > 0) 1 - max(T_H - df_H, 0) / denom else 1
    TLI <- if (df_H > 0) (T_0 / df_0 - T_H / df_H) / (T_0 / df_0 - 1) else NA_real_

    ## ---- relative fit: information criteria ---------------------------
    llH <- ll(SgH); llS <- ll(SgS)
    AIC   <- -2 * llH + 2 * t_H
    BIC   <- -2 * llH + log(Ns) * t_H
    AIC_S <- -2 * llS + 2 * t_S
    BIC_S <- -2 * llS + log(Ns) * t_S

    out <- c(t_nom = t_nom, t = t_H, t_S = t_S,
      RMSEA = RMSEA, SRMR = SRMR,
      CFI = CFI, TLI = TLI,
      AIC = AIC, BIC = BIC,
      AIC_S = AIC_S, BIC_S = BIC_S,
      ELBO = { e <- as.numeric(fit$ELBO); if (length(e) && is.finite(e)) e else NA_real_ },
      ## The model's own comparability score. For a diagonal fit this equals
      ## ELBO; under LD it is the terminal VECM objective, where ELBO is NA.
      ## Comparable only across fits sharing the objective_type attribute.
      objective = { o <- as.numeric(fit$objective)
                    if (length(o) && is.finite(o)) o else NA_real_ })
    class(out) <- c("vbpm_fit_stats", "numeric")
    attr(out, "model") <- "vbfa"
    attr(out, "objective_type") <- if (!is.null(fit$objective_type))
        fit$objective_type else if (is_ld) "vecm" else "elbo"
    out
}

#' @rdname fit_stats
#' @export
fit_stats.vbmimic <- function(object, Y = NULL, Q = NULL, tau = 0.5, ...) {
    ## Soft-selected COEFFICIENTS only: the measurement loadings plus the
    ## structural coefficients. This is deliberately not reported as `t_S`,
    ## which elsewhere means a full soft parameter count -- it omits the
    ## residual precisions, the factor covariance, and the inclusion rates.
    ## It must not be fed to an information criterion.
    nA <- sum(object$Q_A == 1) + sum(object$pi_A[object$Q_A == -1])
    nB <- sum(object$Q_B == 1) + sum(object$pi_B[object$Q_B == -1])
    out <- c(t_nom = NA_real_, t = NA_real_, t_S = NA_real_,
             n_active_coef = nA + nB,
             RMSEA = NA_real_, SRMR = NA_real_, CFI = NA_real_, TLI = NA_real_,
             AIC = NA_real_, BIC = NA_real_, AIC_S = NA_real_, BIC_S = NA_real_,
             ELBO = NA_real_, objective = NA_real_)
    class(out) <- c("vbpm_fit_stats", "numeric")
    attr(out, "model") <- "vbmimic"
    attr(out, "objective_type") <- "none"
    attr(out, "note") <- paste(
        "SEM-like fit statistics require the MIMIC objective and joint",
        "covariance derivation, neither of which is available yet.",
        "n_active_coef counts soft-selected measurement and structural",
        "coefficients only and is not a full parameter count.")
    out
}

#' @rdname fit_stats
#' @export
fit_stats.vbpm_fit <- function(object, ...)
    stop("No fit_stats() method is defined for model class: ", class(object)[1],
         call. = FALSE)

## Core smoke tests for vbpm, ported from the analysis/vbfa_r2 suite.

## small oblique DGP with two anchors per factor
make_dat <- function(N = 400, K = 3, seed = 1, ld_pairs = NULL, ld_rho = .5) {
  set.seed(seed)
  J   <- K * 8
  grp <- rep(1:K, each = 8)
  Phi <- matrix(.3, K, K); diag(Phi) <- 1
  Lam <- matrix(0, J, K); Lam[cbind(1:J, grp)] <- .65
  cross <- seq(3, J, by = 5); Lam[cbind(cross, (grp[cross] %% K) + 1)] <- .35
  psi_var <- pmax(1 - rowSums((Lam %*% chol(Phi))^2), .25)
  Sig_e <- diag(psi_var)
  if (!is.null(ld_pairs)) for (r in seq_len(nrow(ld_pairs))) {
    i <- ld_pairs[r, 1]; j <- ld_pairs[r, 2]
    Sig_e[i, j] <- Sig_e[j, i] <- ld_rho * sqrt(psi_var[i] * psi_var[j])
  }
  eta <- matrix(rnorm(N * K), N, K) %*% chol(Phi)
  E   <- matrix(rnorm(N * J), N, J) %*% chol(Sig_e)
  Q <- matrix(-1L, J, K)
  for (k in 1:K) { a <- which(grp == k)[1:2]; Q[a, ] <- 0L; Q[a, k] <- 1L }
  list(Y = eta %*% t(Lam) + E, Q = Q, Lam = Lam, grp = grp)
}

congruence <- function(Le, Lt) {
  K <- ncol(Lt); cc <- abs(cor(Le, Lt)); perm <- integer(K); used <- integer(0)
  for (k in order(-apply(cc, 2, max))) {
    cand <- order(-cc[, k]); cand <- cand[!(cand %in% used)]
    perm[k] <- cand[1]; used <- c(used, cand[1])
  }
  vapply(1:K, function(k) abs(sum(Le[, perm[k]] * Lt[, k])) /
           sqrt(sum(Le[, perm[k]]^2) * sum(Lt[, k]^2)), numeric(1))
}

test_that("vbfa(v0 = 0.001) reproduces the frozen r1.1 estimator", {
  ## THE regression test protecting the paper-1 artifact. The fixture stores
  ## both the data and the reference outputs of vbfa_r1.1.R (CS repo), so this
  ## cannot be invalidated by generator drift; regenerate only via
  ## dev/make_r11_fixture.R and only if the reference itself is re-frozen.
  ##
  ## Numerical agreement is asserted at a tolerance rather than bitwise. The
  ## estimator runs through compiled BLAS/LAPACK, whose SIMD and FMA choices
  ## differ between platforms, so the last one or two ulps are not portable --
  ## a reference frozen under one BLAS will not reproduce exactly under
  ## another. The tolerance is strict in relative terms and remains sensitive
  ## to material regression, so the guard keeps its power.
  ref <- readRDS(test_path("fixtures", "vbfa_r11_ref.rds"))
  f <- vbfa(ref$Y, ref$Q, v0 = 0.001, max_it = 1500)
  expect_equal(unname(f$Lam), unname(ref$Lam), tolerance = 1e-8)
  expect_equal(unname(f$Phi), unname(ref$Phi), tolerance = 1e-8)
  expect_equal(unname(f$PsiInv), unname(ref$PsiInv), tolerance = 1e-8)
  expect_equal(f$ELBO, ref$ELBO, tolerance = 1e-8)
  ## The iteration count is discrete and must match exactly: a different
  ## stopping point means the path itself changed, not just its rounding.
  expect_identical(f$iter, ref$iter)     # 79, matching CS smoke test T1
})


test_that("scalar v0 = 0.001 reproduces the fixed-spike estimator deterministically", {
  d <- make_dat()
  f1 <- vbfa(d$Y, d$Q, v0 = 0.001, max_it = 1500, convChk = FALSE)
  f2 <- vbfa(d$Y, d$Q, v0 = 0.001, max_it = 1500, convChk = FALSE)
  expect_identical(f1$Lam, f2$Lam)            # same seed -> identical
  expect_equal(f1$iter, f2$iter)
})

test_that("default dynamic path converges and recovers loadings", {
  d <- make_dat()
  f <- vbfa(d$Y, d$Q, convChk = FALSE)
  expect_equal(f$flag, 1)
  expect_true(is.finite(f$ELBO))
  expect_gt(min(congruence(f$Lam, d$Lam)), 0.85)
})

test_that("orthogonal bifactor returns Phi = I", {
  d  <- make_dat()
  Qb <- cbind(1L, d$Q)
  f  <- vbfa(d$Y, Qb, orthogonal = TRUE, convChk = FALSE)
  expect_equal(f$flag, 1)
  expect_identical(f$Phi, diag(1, ncol(Qb), ncol(Qb)))
})

test_that("LD mode recovers planted residual pairs and exposes VECM objective", {
  pairs <- rbind(c(4, 9), c(11, 20), c(15, 23))
  d <- make_dat(seed = 2, ld_pairs = pairs, ld_rho = .5)
  Qe <- matrix(-1L, 24, 24); diag(Qe) <- 0L
  f <- vbfa(d$Y, d$Q, ld = TRUE, Qe = Qe, convChk = FALSE,
            tolVal = 1e-3, max_it = 300)
  expect_true(is.na(f$ELBO))
  expect_identical(f$objective_type, "vecm")
  expect_true(is.finite(f$objective))
  expect_equal(sum(f$objective_terms), f$objective, tolerance = 1e-10)
  Poff <- abs(f$Psi); Poff[lower.tri(Poff, diag = TRUE)] <- 0
  top3 <- order(-Poff)[1:3]
  found <- apply(t(apply(cbind(row(Poff)[top3], col(Poff)[top3]), 1, sort)), 1,
                 paste, collapse = "-")
  planted <- apply(pairs, 1, paste, collapse = "-")
  expect_gte(sum(planted %in% found), 2)

  ## Sparsity, not just ranking. Measured on this scenario the package selects
  ## exactly the 3 planted edges while the V4 reference ld-vb_r.R selects 24.
  ## The cause is upstream of the LD machinery, which is identical in both:
  ## the reference attenuates the loadings (mean primary .475 against a true
  ## .65), so more common variance is left in the residual covariance and QUIC
  ## reads it as local dependence. The packaged branch takes its measurement
  ## updates from r1.1 instead, recovers the loadings (.640), and leaves little
  ## for QUIC to find. Guarding the false-positive rate is therefore the
  ## meaningful regression, not numerical agreement with the reference.
  expect_lte(sum(f$q_star[upper.tri(f$q_star)] >= .5), 6)
})

test_that("incomplete continuous data use deterministic joint moments", {
  d <- make_dat(N = 180); Y <- d$Y
  Y[cbind(1:20, rep(1:4, each = 5))] <- NA
  Y[31, 1:3] <- NA
  f1 <- vbfa(Y, d$Q, v0 = .001, max_it = 300, convChk = FALSE)
  f2 <- vbfa(Y, d$Q, v0 = .001, max_it = 300, convChk = FALSE)
  expect_false(anyNA(f1$Lam))
  expect_identical(f1$Lam, f2$Lam)
  expect_equal(f1$preprocess$n_missing, 23)
  expect_identical(f1$preprocess$missing_mask, is.na(Y))
  expect_true(is.finite(f1$objective))
  Y_bad <- Y; Y_bad[, 1] <- NA
  expect_error(vbfa(Y_bad, d$Q), "at least one observed")
  Y_bad <- Y; Y_bad[1, ] <- NA
  expect_error(vbfa(Y_bad, d$Q), "all responses missing")
})

test_that("the ld switch governs Qe", {
  d <- make_dat()
  Qe <- matrix(-1L, 24, 24); diag(Qe) <- 0L

  ## Qe without ld = TRUE is ignored, with a warning, and the fit is diagonal
  expect_warning(f <- vbfa(d$Y, d$Q, Qe = Qe, convChk = FALSE), "ignored")
  expect_false(f$ld)
  expect_false(is.null(f$PsiInv))
  expect_true(is.finite(f$ELBO))

  ## ld = TRUE with Qe = NULL means fully exploratory local dependence
  f2 <- vbfa(d$Y, d$Q, ld = TRUE, convChk = FALSE,
             tolVal = 1e-3, max_it = 100)
  expect_true(f2$ld)
  expect_null(f2$PsiInv)
  expect_true(is.na(f2$ELBO))
  expect_equal(dim(f2$Psi), c(24L, 24L))
})

test_that("fit_stats is generic and supports LD fits", {
  d <- make_dat()
  f <- vbfa(d$Y, d$Q, convChk = FALSE)
  fs <- fit_stats(f)
  expect_named(fs, c("t_nom", "t", "t_S", "RMSEA", "SRMR", "CFI", "TLI",
                     "AIC", "BIC", "AIC_S", "BIC_S", "ELBO", "objective"))
  ## for a diagonal fit the objective IS the ELBO; under LD it is the VECM
  ## objective, where ELBO is NA. The README promises this element.
  expect_equal(unname(fs["objective"]), unname(fs["ELBO"]))
  fl <- vbfa(d$Y, d$Q, ld = TRUE, convChk = FALSE, max_it = 300, tolVal = 1e-3)
  fsl <- fit_stats(fl)
  expect_true(all(is.finite(fsl[c("RMSEA", "SRMR", "CFI", "TLI", "BIC")])))
  expect_identical(attr(fsl, "objective_type"), "vecm")
})

test_that("fit_stats reads orthogonal from the fit (the bifactor footgun)", {
  d  <- make_dat()
  Qb <- cbind(1L, d$Q)
  fb <- vbfa(d$Y, Qb, orthogonal = TRUE, convChk = FALSE)

  ## the default call now counts the bifactor correctly: no free factor
  ## correlations. Before the fix it silently added K(K-1)/2 of them.
  fs_default  <- fit_stats(fb, Y = d$Y, Q = Qb)
  fs_explicit <- fit_stats(fb, Y = d$Y, Q = Qb, orthogonal = TRUE)
  expect_identical(fs_default, fs_explicit)
  K <- ncol(Qb)
  expect_equal(unname(fs_default["t_nom"]),
               sum((Qb == 1) | (Qb == -1 & {p <- fb$pi; p[Qb == 1] <- 1
                                            p[Qb == 0] <- 0; p >= .5})) + 24)

  ## a contradictory value errors instead of miscounting
  expect_error(fit_stats(fb, Y = d$Y, Q = Qb, orthogonal = FALSE), "contradicts")
  fo <- vbfa(d$Y, d$Q, convChk = FALSE)
  expect_error(fit_stats(fo, Y = d$Y, Q = d$Q, orthogonal = TRUE), "contradicts")
})

test_that("pefa returns evidence over the window and selects nothing", {
  d <- make_dat()
  Q0 <- d$Q[, 1:2, drop = FALSE]
  r <- pefa(Q0, d$Y, Kmin = 2, Kmax = 4, verbose = FALSE,
               max_it = 1500, v0 = 0.001)
  expect_s3_class(r, "pefa")
  expect_false(inherits(r, "vbpm_sweep"))
  expect_true(all(r$sweep$converged))

  ## The evidence-only contract: nothing here decides a count.
  expect_null(r$selected_K)
  expect_null(r$boundary)
  expect_false(any(grepl("_gain", names(r$sweep))))

  expect_identical(nrow(r$transitions), nrow(r$sweep) - 1L)
  ## 0.9.0 stores all-pairs evidence as three upper-triangular W x W matrices
  ## rather than one long pair table.
  W <- nrow(r$sweep)
  expect_identical(names(r$persistence), c("phi", "rmsd", "collision"))
  for (m in r$persistence) expect_identical(dim(m), c(W, W))
  expect_true(all(vapply(r$loadings, is.matrix, logical(1))))
  expect_true(all(vapply(r$pips, is.matrix, logical(1))))

  s <- summary(r)
  expect_s3_class(s, "summary.pefa")
  expect_identical(s$ssl, ssl(r))
  expect_identical(s$settings, r$settings)
  expect_output(print(r), "PEFA sweep \\(ordinary\\)")
  ## No display names a chosen count, because the package owns no count rule.
  expect_false(any(grepl("selected", capture.output(print(r)),
                         ignore.case = TRUE)))
  expect_output(print(s), "Candidates:")
})

test_that("fits carry the vbpm_fit class; print and coef dispatch", {
  d <- make_dat()
  f <- vbfa(d$Y, d$Q, convChk = FALSE)
  expect_s3_class(f, "vbfa")
  expect_s3_class(f, "vbpm_fit")
  expect_identical(f$Q, d$Q)                    # design stored in the fit
  out <- capture.output(print(f))
  expect_true(any(grepl("vbfa", out)))
  expect_true(any(grepl("converged", out)))
  expect_identical(coef(f), f$Lam)
  ## $ access is untouched by the class
  expect_identical(f$Lam, unclass(f)$Lam)
})

test_that("renamed output fields are present and old names are gone", {
  d <- make_dat()
  f <- vbfa(d$Y, d$Q, convChk = FALSE)
  expect_true(all(c("Lam_var", "eta_cov", "Phi_inv_mean") %in% names(f)))
  expect_false(any(c("sigsq.q.Lam", "PHI.q.eta", "M.q.PHIiver", "plotDat")
                   %in% names(f)))
  expect_equal(dim(f$Lam_var), dim(f$Lam))
  expect_equal(dim(f$eta_cov), c(ncol(d$Q), ncol(d$Q)))
})

test_that("vbfa is quiet by default and messages when convChk = TRUE", {
  d <- make_dat(N = 150)
  ## max_it = 5 will not converge, so silence the expected warning and test
  ## only the message channel
  expect_no_message(suppressWarnings(
    vbfa(d$Y, d$Q, v0 = 0.001, max_it = 5)))
  expect_message(suppressWarnings(
    vbfa(d$Y, d$Q, v0 = 0.001, max_it = 5, convChk = TRUE)),
    "relative error")
})

test_that("sim_fa produces data with planted local dependence", {
  s <- sim_fa(N = 300, K = 3, ipf = 6, lam = .7, lac = .3, ecr = .3, rseed = 7)
  expect_equal(dim(s$dat), c(300L, 18L))
  expect_false(anyNA(s$dat))
  expect_gt(nrow(s$ofd_ind), 0)
})

test_that("bifactor mode equals the hand-built general column bit-for-bit", {
  d  <- make_dat()
  fb <- vbfa(d$Y, d$Q, bifactor = TRUE)
  fh <- vbfa(d$Y, cbind(1L, d$Q), orthogonal = TRUE)
  expect_identical(fb$Lam, fh$Lam)
  expect_identical(fb$ELBO, fh$ELBO)
  expect_identical(fb$pi, fh$pi)
  ## bookkeeping fields
  expect_true(fb$bifactor); expect_identical(fb$n_general, 1L)
  expect_identical(fb$Q, cbind(rep(1L, nrow(d$Q)), d$Q, deparse.level = 0))
  expect_true(fb$orthogonal)
  ## print shows the general/group split
  expect_output(print(fb), "1 general [+] 3 group factors")
})

test_that("bifactor = TRUE overrides orthogonal = FALSE, with a message", {
  d <- make_dat()
  expect_message(
    f <- suppressWarnings(          # max_it deliberately tiny; iteration
      vbfa(d$Y, d$Q, bifactor = TRUE, orthogonal = FALSE, max_it = 50)),
    "overridden")
  expect_true(f$orthogonal)
  ## no message when orthogonal is left at its default
  expect_no_message(suppressWarnings(
    vbfa(d$Y, d$Q, bifactor = TRUE, max_it = 50)))
})

test_that("bifactor mode validates general and rejects a general column in Q", {
  d <- make_dat()
  expect_error(vbfa(d$Y, cbind(1L, d$Q), bifactor = TRUE), "all-specified")
  expect_error(vbfa(d$Y, d$Q, bifactor = TRUE, general = 2), "-1, 0, or 1")
  expect_error(vbfa(d$Y, d$Q, bifactor = TRUE, general = 0), "at least one")
  ## bifactor-(S-1)-style: zero general loadings where general = 0
  g <- rep(1L, ncol(d$Y)); g[1:4] <- 0L
  f <- suppressWarnings(            # truncated run; only the zeros matter
    vbfa(d$Y, d$Q, bifactor = TRUE, general = g, max_it = 200))
  expect_true(all(f$Lam[1:4, 1] == 0))
})

test_that("pefa bifactor sweep counts group factors without storing K_total", {
  d  <- make_dat()
  Q0 <- d$Q[, 1:2, drop = FALSE]
  r  <- pefa(Q0, d$Y, Kmin = 2, Kmax = 3, bifactor = TRUE, v0 = .001,
             max_it = 500, verbose = FALSE)
  expect_true(isTRUE(r$settings$bifactor))
  expect_identical(r$sweep$K, 2:3)
  expect_false("K_total" %in% names(r$sweep))
  expected_dims <- stats::setNames(3:4, c("2", "3"))
  expect_identical(vapply(r$loadings, ncol, integer(1)), expected_dims)
  expect_identical(vapply(r$pips, ncol, integer(1)), expected_dims)
  ## The adjacent row carries all seven structural facts; $persistence
  ## projects three of them over every direct pair.
  expect_true(all(c("phi_min", "rmsd", "rmsd_max", "ari", "pip_rmsd",
                    "unmatched_ssl", "collision") %in% names(r$transitions)))
  expect_type(r$persistence$collision, "logical")
  expect_true(is.finite(r$persistence$phi["2", "3"]))
  expect_true(is.finite(r$persistence$rmsd["2", "3"]))
  expect_output(print(r), "group factors [(][+]1 general")
  ## K is the group-factor count everywhere; no total-factor column is stored
  ## or displayed.
  expect_false(any(grepl("K_total", capture.output(print(summary(r))))))
})

test_that("the fit object stores each quantity exactly once", {
  d <- make_dat()
  f <- vbfa(d$Y, d$Q, v0 = .001, max_it = 500, convChk = FALSE)
  ## the mirror lists removed in 0.7.0 stay gone
  for (nm in c("coefficients", "design", "posterior", "settings"))
    expect_null(f[[nm]])
  ## complete data: no mask stored, but the count is still reported
  expect_null(f$preprocess$missing_mask)
  expect_equal(f$preprocess$n_missing, 0)
  ## incomplete data: the mask is stored
  Y <- d$Y; Y[1:5, 1] <- NA
  fm <- vbfa(Y, d$Q, v0 = .001, max_it = 500, convChk = FALSE)
  expect_identical(fm$preprocess$missing_mask, is.na(Y))
  expect_equal(fm$preprocess$n_missing, 5)
  ## no two top-level components hold identical large content
  big <- Filter(function(z) is.matrix(z) && length(z) > 100, f)
  if (length(big) > 1) {
    pairs <- utils::combn(seq_along(big), 2)
    dupes <- apply(pairs, 2, function(ij) identical(big[[ij[1]]], big[[ij[2]]]))
    expect_false(any(dupes))
  }
  ## pefa keeps only compact loading/PIP matrices, never full candidate fits
  r <- pefa(d$Q[, 1:2, drop = FALSE], d$Y, Kmin = 2, Kmax = 3, v0 = .001,
            max_it = 500, verbose = FALSE)
  expect_null(r[["fit", exact = TRUE]])
  expect_null(r[["selected_fit", exact = TRUE]])
  expect_null(r[["fits", exact = TRUE]])
  expect_true(all(vapply(r$loadings, is.matrix, logical(1))))
  expect_true(all(vapply(r$pips, is.matrix, logical(1))))
})

## Tests for the MIMIC estimator and the sim_lvm generator.

## MIMIC design used throughout: K factors, each predicted by 3 of the P
## covariates, in disjoint blocks.
mimic_B <- function(K = 3, P = 9, val = .3) {
  B <- matrix(0, K, P)
  for (k in seq_len(K)) B[k, ((k - 1) * 3 + 1):(k * 3)] <- val
  B
}

## two anchors per factor -- the partially confirmatory idiom shared with vbfa()
anchor_QA <- function(J = 18, K = 3) {
  Q <- matrix(-1L, J, K); grp <- rep(seq_len(K), each = J / K)
  for (k in seq_len(K)) { a <- which(grp == k)[1:2]; Q[a, ] <- 0L; Q[a, k] <- 1L }
  Q
}

test_that("vbmimic recovers the structural pattern up to factor permutation", {
  B <- mimic_B()
  s <- sim_lvm(N = 600, K = 3, J = 18, P = 9, b = B, phx = 0, rseed = 11)
  Y <- s$dat[, 1:18]; X <- s$dat[, 19:27]
  f <- vbmimic(Y, X, anchor_QA(), matrix(-1L, 3, 9))

  expect_equal(f$flag, 1)
  expect_true(is.na(f$ELBO))                     # documented: no ELBO yet
  expect_equal(dim(f$A), c(18L, 3L))
  expect_equal(dim(f$B), c(3L, 9L))
  expect_s3_class(f, "vbmimic")
  expect_s3_class(f, "vbpm_fit")
  out <- capture.output(print(f))
  expect_true(any(grepl("MIMIC", out)))
  expect_named(coef(f), c("A", "B"))

  ## same covariate blocks, allowing for label switching
  truth <- apply(B != 0, 1, function(r) paste(which(r), collapse = ","))
  est   <- apply(abs(f$B) > .15, 1, function(r) paste(which(r), collapse = ","))
  expect_setequal(est, truth)
})

test_that("a fully exploratory Q_A AND Q_B still converges", {
  ## With nothing anchored in either part the solution is rotationally
  ## ambiguous: convergence is guaranteed to be clean, recovery is not
  ## (observed both recovering and merging factors, depending on
  ## standardization and seed). So only convergence is asserted here; the
  ## anchored test above carries the recovery claim, and the docs carry the
  ## "anchor at least one part" advice.
  B <- mimic_B()
  s <- sim_lvm(N = 600, K = 3, J = 18, P = 9, b = B, phx = 0, rseed = 11)
  f <- vbmimic(s$dat[, 1:18], s$dat[, 19:27],
               matrix(-1L, 18, 3), matrix(-1L, 3, 9))
  expect_equal(f$flag, 1)
})

test_that("vbmimic is deterministic (consumes no RNG)", {
  s <- sim_lvm(N = 300, K = 3, J = 18, P = 9, b = mimic_B(), phx = 0, rseed = 5)
  Y <- s$dat[, 1:18]; X <- s$dat[, 19:27]
  QA <- matrix(-1L, 18, 3); QB <- matrix(-1L, 3, 9)
  set.seed(1);   f1 <- vbmimic(Y, X, QA, QB)
  set.seed(999); f2 <- vbmimic(Y, X, QA, QB)
  expect_identical(f1$A, f2$A)
  expect_identical(f1$B, f2$B)
  expect_equal(f1$iter, f2$iter)
})

test_that("vbmimic validates its inputs", {
  s <- sim_lvm(N = 200, K = 3, J = 18, P = 9, b = mimic_B(), phx = 0, rseed = 2)
  Y <- s$dat[, 1:18]; X <- s$dat[, 19:27]
  QA <- matrix(-1L, 18, 3); QB <- matrix(-1L, 3, 9)

  ## NA in Y is supported as of 0.4.0 (see test-missing-ld.R); NA in X is not,
  ## because X is conditioned on rather than modelled.
  Yna <- Y; Yna[1, 1] <- NA
  expect_no_error(vbmimic(Yna, X, QA, QB))
  Xna <- X; Xna[1, 1] <- NA
  expect_error(vbmimic(Y, Xna, QA, QB), "missing covariates")
  expect_error(vbmimic(Y, X[1:10, ], QA, QB), "same number of rows")
  expect_error(vbmimic(Y, X, matrix(-1L, 17, 3), QB), "nrow\\(Q_A\\)")
  expect_error(vbmimic(Y, X, QA, matrix(-1L, 3, 8)), "ncol\\(Q_B\\)")
  expect_error(vbmimic(Y, X, QA, matrix(-1L, 2, 9)), "nrow\\(Q_B\\)")
  expect_error(vbmimic(Y, X, matrix(2L, 18, 3), QB), "Q_A entries")
  expect_error(vbmimic(Y, X, QA, matrix(2L, 3, 9)), "Q_B entries")
  expect_error(vbmimic(Y, X, QA, QB, v0 = -1), "v0 must be positive")
})

test_that("anchored entries of Q_B are estimated and fixed zeros stay zero", {
  B <- mimic_B()
  s <- sim_lvm(N = 500, K = 3, J = 18, P = 9, b = B, phx = 0, rseed = 4)
  QB <- matrix(0L, 3, 9)          # everything fixed to zero except one anchor
  QB[1, 1] <- 1L
  f <- vbmimic(s$dat[, 1:18], s$dat[, 19:27], matrix(-1L, 18, 3), QB)
  expect_true(all(f$B[QB == 0] == 0))
  expect_true(f$B[1, 1] != 0)
})

test_that("the v0 path runs and a scalar v0 is a single stage", {
  s <- sim_lvm(N = 300, K = 3, J = 18, P = 9, b = mimic_B(), phx = 0, rseed = 6)
  Y <- s$dat[, 1:18]; X <- s$dat[, 19:27]
  QA <- matrix(-1L, 18, 3); QB <- matrix(-1L, 3, 9)
  f_sc <- vbmimic(Y, X, QA, QB, v0 = 0.001)
  f_pa <- vbmimic(Y, X, QA, QB, v0 = c(0.01, 0.005, 0.001))
  expect_equal(f_sc$path$n_stage, 1L)
  expect_equal(f_pa$path$n_stage, 3L)
  expect_length(f_pa$path$stage_iters, 3)
  expect_equal(f_pa$flag, 1)
})

test_that("fit_stats covers vbmimic with explicitly limited indices", {
  s <- sim_lvm(N = 300, K = 3, J = 18, P = 9, b = mimic_B(), phx = 0, rseed = 8)
  f <- vbmimic(s$dat[, 1:18], s$dat[, 19:27],
              matrix(-1L, 18, 3), matrix(-1L, 3, 9))
  fs <- fit_stats(f)
  expect_s3_class(fs, "vbpm_fit_stats")
  ## t_S is NA for MIMIC: the soft coefficient count is reported separately
  ## as n_active_coef, because it omits residual and factor-covariance
  ## parameters and must not be fed to an information criterion.
  expect_true(is.na(fs["t_S"]))
  expect_true(is.finite(fs["n_active_coef"]))
  expect_true(all(is.na(fs[c("RMSEA", "SRMR", "CFI", "TLI", "ELBO")])))
})

## ---- sim_lvm ------------------------------------------------------------

test_that("sim_lvm confines missingness to the item block", {
  s <- sim_lvm(N = 200, K = 3, J = 18, P = 5, misp = .3, rseed = 9)
  expect_equal(ncol(s$dat), 23L)
  expect_gt(sum(is.na(s$dat[, 1:18])), 0)
  expect_equal(sum(is.na(s$dat[, 19:23])), 0)   # regression: LAWBL recycled here
})

test_that("phd correlates the structural disturbances", {
  B <- mimic_B(K = 5, P = 15)
  s0 <- sim_lvm(N = 800, K = 5, J = 30, P = 15, b = B, phx = 0, phd = 0, rseed = 1)
  s3 <- sim_lvm(N = 800, K = 5, J = 30, P = 15, b = B, phx = 0, phd = .3, rseed = 1)
  off <- function(m) mean(m[upper.tri(m)])
  expect_lt(abs(off(s0$PHI)), .10)
  expect_gt(off(s3$PHI), .10)
})

test_that("sim_lvm supports latent predictors and rejects bad coefficients", {
  s <- sim_lvm(N = 300, K = 4, J = 20, K1 = 2, b1 = .4, rseed = 3)
  expect_equal(dim(s$mb1), c(2L, 2L))
  ## dense b over-explains the factors -> negative disturbance variance
  expect_error(sim_lvm(N = 100, K = 5, J = 30, P = 15, b = .3),
               "disturbance variance")
  expect_error(sim_lvm(N = 100, K = 3, J = 18, K1 = 3), "K1")
})

test_that("sim_lvm's plain factor path is unchanged and still simulates LD", {
  s <- sim_lvm(N = 300, K = 3, J = 18, ecr = .3, rseed = 7)
  expect_equal(dim(s$dat), c(300L, 18L))
  expect_false(anyNA(s$dat))
  expect_gt(nrow(s$ofd_ind), 0)
})

## Focused regression tests for the 0.8.2 parameter-count policy.

make_rank_fit_082 <- function(J = 3L) {
  Q <- matrix(0L, J, 1L)
  Q[1L, 1L] <- 1L
  structure(
    list(
      Q = Q,
      Lam = matrix(c(0.7, rep(0, J - 1L)), J, 1L),
      pi = matrix(c(1, rep(0, J - 1L)), J, 1L),
      orthogonal = FALSE,
      nobs = 120L,
      ld = FALSE,
      PsiInv = rep(2, J),
      Phi = matrix(1, 1L, 1L),
      sample_cov = diag(J),
      ELBO = -20,
      objective = -20,
      objective_type = "elbo"
    ),
    class = c("vbfa", "vbpm_fit")
  )
}

test_that("nominal counting is the deterministic default", {
  fit <- make_rank_fit_082()

  ## Shadow requireNamespace() only for a copy of the S3 method. If the
  ## nominal branch inspects the optional package, this sentinel fails.
  method <- getS3method("fit_stats", "vbfa")
  method_env <- new.env(parent = environment(method))
  method_env$requireNamespace <- function(...) {
    stop("nominal path touched the optional namespace", call. = FALSE)
  }
  environment(method) <- method_env

  stats <- method(fit)
  expect_identical(unname(stats[["t"]]), unname(stats[["t_nom"]]))
})

test_that("rank controls reject malformed values", {
  fit <- make_rank_fit_082()

  bad_adjust <- list(NULL, NA, 1, "TRUE", logical(), c(TRUE, FALSE))
  for (value in bad_adjust) {
    expect_error(
      fit_stats(fit, rank_adjust = value),
      "rank_adjust.*one non-missing logical"
    )
  }

  bad_max <- list(NULL, NA_real_, NaN, Inf, -Inf, 0, -1, 1.5, TRUE,
                  "100", numeric(), c(50, 100))
  for (value in bad_max) {
    expect_error(
      fit_stats(fit, rank_max_J = value),
      "rank_max_J.*finite positive whole number"
    )
  }

  expect_silent(fit_stats(fit, rank_adjust = FALSE, rank_max_J = 1L))
})

test_that("an unavailable numDeriv package is an explicit opt-in error", {
  fit <- make_rank_fit_082()

  ## Give a private copy of the method a deterministic unavailable-package
  ## result without changing the package namespace or the machine libraries.
  method <- getS3method("fit_stats", "vbfa")
  method_env <- new.env(parent = environment(method))
  method_env$requireNamespace <- function(package, quietly = FALSE) FALSE
  environment(method) <- method_env

  expect_error(
    method(fit, rank_adjust = TRUE),
    "rank_adjust = TRUE.*requires.*numDeriv"
  )
})

test_that("large explicit rank requests do not silently fall back", {
  fit <- make_rank_fit_082(J = 4L)

  expect_error(
    fit_stats(fit, rank_adjust = TRUE, rank_max_J = 3L),
    "J <= rank_max_J.*Raise.*rank_max_J.*rank_adjust = FALSE"
  )
})

test_that("explicit rank adjustment computes the Jacobian rank", {
  skip_if_not_installed("numDeriv")
  fit <- make_rank_fit_082()

  nominal <- fit_stats(fit)
  adjusted <- fit_stats(fit, rank_adjust = TRUE)

  ## With one indicator, its loading and uniqueness contribute to the same
  ## variance moment, so this deliberately underidentified example loses one
  ## degree of Jacobian rank.
  expect_identical(unname(nominal[["t"]]), 4)
  expect_identical(unname(adjusted[["t_nom"]]), 4)
  expect_identical(unname(adjusted[["t"]]), 3)
})

## Internal wrapper for the compiled QUIC solver (src/quic_impl.cpp), used by
## vbfa()'s local-dependence branch. Returns a J x J x 2 array: slice 1 is the
## residual precision Psi, slice 2 is its inverse W.
.quic_rcpp <- function(J, S, Rho, eps, max_iter_quic)
  my_quic_cpp(J, S, Rho, eps, max_iter_quic)

#' @title Simulating data with Latent Variable Modeling
#'
#' @description \code{sim_fa} can simulate data based on factor analysis or
#' item response models with different response formats (continuous or categorical),
#' loading patterns and residual covariance (local dependence) structures.
#'
#' \code{sim_fa()} and [sim_lvm()] are complementary rather than
#' interchangeable: \code{sim_fa()} is the *structure-side* generator, driven
#' by the loading pattern itself -- items per factor (`ipf`), alternating-sign
#' cross-loadings (`alt_sign`), or minor factors (`minor`, definition-based
#' per Auerswald & Moshagen, 2019) -- and is the generator for [vbfa()].
#' [sim_lvm()] is the *model-side* generator, adding observed/latent
#' predictors (`P`/`b`, `K1`/`ph1`/`b1`) and mixed response formats (`ilvl`),
#' and is the generator for [vbmimic()]. Both generators also accept a
#' population loading matrix directly (`mla`) with identical semantics.
#'
#' @name sim_fa
#'
#' @param N Sample size.
#'
#' @param mla Population loading matrix (J x K). If supplied, the pattern
#' arguments (\code{ipf}, \code{cpf}, \code{lam}, \code{lac}, \code{alt_sign},
#' \code{add_*}, \code{minor}, \code{zero_it}) are ignored and \code{K},
#' \code{J} are taken from it -- the same convention as [sim_lvm()].
#'
#' @param K Number of factors.
#'
#' @param ipf Items per factor.
#'
#' @param cpf Cross-loadings per factor.
#'
#' @param lam Number of formal iterations for posterior sampling.
#'
#' @param lac Number of iterations to update the sampling information.
#'
#' @param phi Homogeneous correlations between any two factors.
#'
#' @param ph12 Correlation between factor 1 and 2 (if it's different from
#' \code{phi}). Shared with [sim_lvm()], which uses \code{ph1} for a
#' different purpose (correlation among latent predictors); see [sim_lvm()]
#' for that distinction.
#'
#' @param gamma Optional second-order loadings for direct \strong{higher-order}
#' generation: a scalar or length-\code{K} vector in \eqn{(0, 1)}. When
#' supplied, each group factor \eqn{\eta_k} is driven by a general factor
#' \eqn{\xi} through \eqn{\eta_k = \gamma_k \xi + \zeta_k}, and the data are
#' generated from the equivalent Schmid-Leiman orthogonal bifactor matrix:
#' general loadings \eqn{\lambda_j \gamma_k} and group loadings
#' \eqn{\lambda_j \sqrt{1 - \gamma_k^2}} (a testlet model is the same model).
#' The returned \code{MLA} is this \eqn{J \times (K+1)} bifactor matrix
#' (general factor first), \code{PHI} is the identity, and the first-order
#' matrix and \code{gamma} are returned as \code{MLA1} and \code{gamma}.
#' Incompatible with \code{mla}, \code{cpf > 0}, \code{add_ind},
#' \code{add1_ind}, \code{minor}, and \code{zero_it}; \code{phi}/\code{ph12}
#' are ignored (the SL solution is orthogonal by construction).
#'
#' @param ecr Residual correlation (local dependence).
#'
#' @param ome_out Output factor score or not.
#'
#' @param cati The set of categorical (polytomous) items in sequence number (i.e., 1 to \eqn{J});
#' \code{NULL} for no and -1 for all (default is \code{NULL}).
#'
#' @param noc Number of categories for categorical items
#'
#' @param misp Proportion of missingness.
#'
#' @param rseed An integer for the random seed.
#'
#' @param necb Number of between-factor local dependence.
#'
#' @param necw Number of within-factor local dependence.
#'
#' @param add_ind (Additional) minor factor with cross-loadings.
#'
#' @param add_la Value of cross-loadings on (Additional) minor factor.
#'
#' @param add_phi Correlations between (Additional) minor factor and other factors.
#'
#' @param alt_sign Logical; alternate the sign of the cross-loadings across the
#' cross-loading items (default \code{TRUE}).
#'
#' @param add1_ind A second (additional) minor factor with cross-loadings,
#' given as item indices; \code{c()} for none.
#'
#' @param add1_la Value of the cross-loadings on the second minor factor.
#'
#' @param add1_phi Correlations between the second minor factor and other
#' factors.
#'
#' @param minor Optional character vector requesting orthogonal minor factors
#' loading on every indicator, each \code{"weak"} or \code{"moderate"}
#' (Auerswald & Moshagen, 2019); \code{NULL} for none.
#'
#' @param zero_it Surplus items with zero loading.
#'
#' @param digits Number of significant digits to print when printing numeric values.
#'
#' @return An object of class \code{list} containing the data, loading, and factorial correlation matrix.
#'
#'
#' @importFrom MASS mvrnorm
#'
#' @export
#'
#' @examples
#'
#' # for continuous data with cross-loadings and local dependence effect .3
#' out <- sim_fa(N=1000,K=3,ipf=6,lam = .7, lac=.3,ecr=.3)
#' summary(out$dat)
#' out$MLA
#' out$ofd_ind
#'
#' # for categorical data with cross-loadings .4 and 10% missingness
#' out <- sim_fa(N=1000,K=3,ipf=6,lam = .7, lac=.4,cati=-1,noc=4,misp=.1)
#' summary(out$dat)
#' out$MLA
#' out$ofd_ind
#'
#' ## matrix-driven: supply the population loading matrix directly
#' mla <- matrix(0, 12, 2)
#' mla[1:6, 1] <- .7; mla[7:12, 2] <- c(.8, .7, .6, .7, .8, .6)
#' s <- sim_fa(N = 300, mla = mla, phi = .3, rseed = 4)
#' s$MLA                      # the generating matrix, as supplied
#'
#' ## higher-order (testlet) data, generated directly: 3 group factors driven
#' ## by one general factor with second-order loadings .8, .7, .6
#' s <- sim_fa(N = 500, K = 3, ipf = 6, lam = .7, gamma = c(.8, .7, .6))
#' round(s$MLA, 2)            # J x 4 Schmid-Leiman bifactor matrix (general first)
#' s$gamma                    # the second-order loadings
#'
#' @importFrom MASS mvrnorm
#' @export
sim_fa <- function(N = 1000, mla = NULL, K = 3, ipf = 8, cpf = 2, lam = 0.7, lac = 0.3, alt_sign = TRUE, phi = 0.5, ph12 = -1,
        gamma = NULL, ecr = .0, ome_out = FALSE,cati = NULL, noc = c(4), misp = 0, rseed = 333,
        necw=K,necb=K,add_ind=c(),add_la=0,add_phi=0,
        add1_ind=c(), add1_la=0, add1_phi=0, zero_it=0, minor=NULL, digits = 4) {

    ## `minor`: an optional character vector requesting orthogonal minor factors,
    ## each loading on EVERY indicator, following Auerswald & Moshagen (2019,
    ## Psychological Methods) / Lorenzo-Seva et al. (2011):
    ##   "weak"     -> loadings ~ U(-.10, .10)            (~0.33% explained var)
    ##   "moderate" -> loadings ~ +/- U(.09, .11)          (~1% explained var)
    ## e.g. minor = c("weak","weak","moderate","moderate") adds 2 weak + 2 moderate
    ## minor factors. Minor factors are uncorrelated with the major factors and
    ## with one another. This is the modern, definition-based alternative to the
    ## ad hoc add_ind/add1_ind mechanism above.

    if (exists(".Random.seed", .GlobalEnv))
        oldseed <- .GlobalEnv$.Random.seed else oldseed <- NULL
        set.seed(rseed)

    oo <- options()       # code line i
    on.exit(options(oo))  # code line i+1
    # old_digits <- getOption("digits")
    options(digits = digits)

    use_ho <- !is.null(gamma)
    if (use_ho) {
        if (!is.null(mla))
            stop("gamma (higher-order generation) cannot be combined with a supplied mla.", call. = FALSE)
        if ((!missing(cpf) && cpf > 0) || length(add_ind) > 0 ||
            length(add1_ind) > 0 || length(minor) > 0 || zero_it > 0)
            stop("gamma (higher-order generation) is incompatible with cpf > 0, add_ind, add1_ind, minor, and zero_it.", call. = FALSE)
        cpf <- 0                      # simple structure at the first order
        if (!is.numeric(gamma) || any(!is.finite(gamma)) ||
            any(gamma <= 0) || any(gamma >= 1))
            stop("gamma must be numeric in (0, 1).", call. = FALSE)
        if (length(gamma) != 1L && length(gamma) != K)
            stop("gamma must be a scalar or a length-K vector.", call. = FALSE)
        gamma <- rep_len(as.numeric(gamma), K)
    }

    use_mla <- !is.null(mla)
    if (use_mla) {
        ## ---- matrix-driven: supplied mla IS the population loading matrix.
        ## Pattern arguments (ipf, cpf, lam, lac, alt_sign, add_*, minor,
        ## zero_it) are ignored; K, J, ipf, cpf follow sim_lvm's convention.
        ## Determined here (before PHI is built) so PHI comes out K x K.
        mla <- as.matrix(mla)
        K   <- ncol(mla)
        J   <- nrow(mla)
        K1  <- K
        ipf <- J / K
        cpf <- 0
        if (ecr > 0 && J %% K != 0)
            stop("Local dependence (ecr > 0) with a supplied mla assumes J is divisible by K; use ecr = 0 with an unbalanced mla.", call. = FALSE)
    }

    PHI <- matrix(phi, K, K)
    diag(PHI) <- 1
    if (ph12 > -1 && ph12 < 1) PHI[1,2] <- PHI[2,1]<-ph12

    if (!use_mla) {
    if (alt_sign) {
      cross_vals <- rep(c(lac, -lac), length.out = cpf)
    } else {
      cross_vals <- rep(lac, cpf)
    }
    lam0 <- c(rep(lam, ipf), rep(0, ipf - cpf), cross_vals, rep(0, (K - 2) * ipf))
    # lam0 <- c(rep(lam, ipf), rep(0, ipf - cpf), rep(lac, cpf), rep(0, (K - 2) * ipf))

    lam1 <- matrix(lam0, ipf, K)

    mla <- c()
    for (i in K:1) {
        # i <- 1
        ind <- (c(1:K) + i)%%K
        ind[ind == 0] <- K
        mla <- rbind(mla, lam1[, ind])
    }

    J <- K * ipf
    K1<-K

    ## higher-order generation: replace the first-order structure by its
    ## Schmid-Leiman orthogonal bifactor equivalent. For item j on group
    ## factor k with first-order loading lambda_j and second-order loading
    ## gamma_k: general loading = lambda_j * gamma_k, group loading =
    ## lambda_j * sqrt(1 - gamma_k^2). With cpf = 0 each row of mla has a
    ## single nonzero entry, so mla %*% gamma gives the general column.
    if (use_ho) {
        mla1 <- mla                                  # first-order J x K
        gen  <- as.vector(mla1 %*% gamma)
        grp  <- sweep(mla1, 2, sqrt(1 - gamma^2), "*")
        mla  <- cbind(gen, grp, deparse.level = 0)
        K1   <- K + 1
        PHI  <- diag(K1)                             # orthogonal by construction
    }

    if(length(add_ind)>0){
        # len<-length(add_ind)
        K1<-K+1
        mla<-cbind(mla,0)
        mla[add_ind,K1]<-add_la
        PHI<-cbind(PHI,add_phi)
        PHI<-rbind(PHI,add_phi)
        PHI[K1,K1]<-1
    }

    if(length(add1_ind)>0){
      # len<-length(add_ind)
      K1<-K1+1
      mla<-cbind(mla,0)
      mla[add1_ind,K1]<-add1_la
      PHI<-cbind(PHI,add1_phi)
      PHI<-rbind(PHI,add1_phi)
      PHI[K1,K1]<-1
    }

    ## modern minor-factor mechanism (definition-based; see signature comment)
    if (length(minor) > 0) {
        Jm <- nrow(mla)                       # current number of items (= K*ipf)
        for (mt in minor) {
            load_col <- switch(mt,
                weak     = runif(Jm, -0.10, 0.10),
                moderate = sample(c(-1, 1), Jm, replace = TRUE) * runif(Jm, 0.09, 0.11),
                stop("minor must contain only 'weak' or 'moderate'"))
            K1  <- K1 + 1
            mla <- cbind(mla, load_col)       # minor factor loads on every item
            PHI <- cbind(PHI, 0); PHI <- rbind(PHI, 0); PHI[K1, K1] <- 1  # orthogonal
        }
    }

    if (zero_it>0){
        J <- J + zero_it
        mla <- rbind(mla,matrix(0,zero_it,K1))
    }
    }

    ecm <- matrix(0, J, J)
    if (ecr > 0) {
        iecb <- iecw <- NULL
        li <- c(1:K) * ipf - cpf
        for (i in 1:necw) {
            if(necw>0) ecm[li[i], li[i] - 1] <- ecm[li[i] - 1,li[i]] <- ecr
        }
        for (i in 1:necb) {
            i1 <- i%%K + 1
            if(necb>0) ecm[li[i] - 2, li[i1] - 3] <- ecm[li[i1] - 3,li[i] - 2]<- ecr
        }
    }

    # iecb <- iecw <- NULL
    # li <- c(1:K) * ipf - cpf
    # for (i in 1:necw) {
    #     if(necw>0) iecw <- rbind(iecw, c(li[i], li[i] - 1))
    #     # i1 <- i%%K + 1
    #     # iecb <- rbind(iecb, c(li[i] - 2, li[i1] - 3))
    # }
    # for (i in 1:necb) {
    #     # iecw <- rbind(iecw, c(li[i], li[i] - 1))
    #     i1 <- i%%K + 1
    #     if(necb>0) iecb <- rbind(iecb, c(li[i] - 2, li[i1] - 3))
    # }

    # ecm <- matrix(0, J, J)
    #
    # if (ecr > 0) {
    #     ecm[iecw] <- ecm[iecw[, 2:1]] <- ecr
    #     ecm[iecb] <- ecm[iecb[, 2:1]] <- ecr
    # }

    evr <- rep(0, J)
    for (j in 1:J) {
        evr[j] = 1 - t(mla[j, ]) %*% PHI %*% mla[j, ]
    }  #end j

    diag(ecm) <- evr
    er <- mvrnorm(N, rep(0, J), ecm)
    eta <- mvrnorm(N, rep(0, K1), PHI)
    y <- t(mla %*% t(eta)) + er
    scale <- apply(y, 2, sd)
    eigen <- diag(crossprod(mla))

    out <- list(N = N, PHI = PHI, MLA = mla, Eigen = eigen, PSX = ecm, scale = scale)
    if (use_ho) {
        out$MLA1  <- mla1                 # first-order loading matrix (J x K)
        out$gamma <- gamma                # second-order loadings
    }
    if (ome_out)
        out$OME <- eta

    pos <- lower.tri(ecm)
    ind <- which(pos, arr.ind = T)
    rind <- which(ecm[pos] > 0)
    out$ofd_ind <- ind[rind, ]

    Jp <- length(cati)
    if (Jp > 0)
        {
            if (Jp == 1 && cati == -1) {
                cati <- c(1:J)
                Jp <- J
            }
            yc <- y[, cati]
            len <- length(noc)
            if (Jp%%len != 0)
                stop("cati is not divisable by length(noc).", call. = F)
            for (i in 1:len) {
                M <- noc[i]  #categories
                UL <- 3
                LL <- -3
                stp <- LL + c(1:(M - 1)) * (UL - LL)/M  #M-1 step points
                i0 <- (i - 1) * Jp/len + 1
                i1 <- i * Jp/len
                for (j in i0:i1) {
                  tmp <- yc[, j] > matrix(stp, N, (M - 1), byrow = T)
                  yc[, j] <- rowSums(tmp) + 1  # value starting from 1
                }
            }  #end len
            y[, cati] <- yc
            out$noc <- noc
            out$cati <- cati
        }  #end Jp

    mind <- matrix(rbinom(N * J, 1, misp), N, J)
    y[mind == 1] <- NA
    out$misp <- misp
    out$dat <- y

    if (!is.null(oldseed))
        .GlobalEnv$.Random.seed <- oldseed else rm(".Random.seed", envir = .GlobalEnv)

    return(out)
}

## special_effects(): SL check, higher-order parameters, Eq.-16 effect sizes.

ho_fit <- function(N = 800, bifactor = TRUE) {
  sim <- sim_fa(N = N, K = 3, ipf = 6, lam = .8, gamma = c(.8, .7, .6),
                rseed = 11)
  Q <- matrix(-1L, 18, 3)
  for (k in 1:3) { a <- which(rep(1:3, each = 6) == k)[1:2]
                   Q[a, ] <- 0L; Q[a, k] <- 1L }
  if (bifactor) vbfa(sim$dat, Q, bifactor = TRUE, v0 = .001)
  else          vbfa(sim$dat, cbind(1L, Q), orthogonal = TRUE, v0 = .001)
}

test_that("special_effects reproduces the vignette's plain-R computations", {
  fit <- ho_fit()
  se  <- special_effects(fit)
  cl  <- rep(1:3, each = 6)

  ## independent reimplementation (the vignette's code, known cluster ids)
  rt  <- sapply(1:18, function(j) abs(fit$Lam[j, cl[j] + 1]) / abs(fit$Lam[j, 1]))
  cvs <- sapply(1:3, function(k) { r <- rt[cl == k]; sd(r) / mean(r) })
  rbar <- sapply(1:3, function(k) mean(rt[cl == k]))
  Dk  <- sapply(1:3, function(k) {
    col <- k + 1
    inc <- (fit$Q[, col] == 1) | (fit$Q[, col] == -1 & fit$pi[, col] > .5)
    sum(fit$Lam[inc, col]^2) / sum(inc) })

  expect_equal(se$groups$cv, cvs)
  expect_equal(se$mean_cv, sum(rep(cvs, each = 6)) / 18)
  expect_equal(unname(se$gamma), 1 / sqrt(1 + rbar^2))
  expect_equal(se$groups$D, Dk)
  expect_equal(se$groups$n_items, rep(6L, 3))
  ## lambda = |b_gen| / gamma of the item's group
  expect_equal(se$items$lambda,
               abs(fit$Lam[, 1]) / (1 / sqrt(1 + rbar^2))[cl])
  expect_length(se$unassigned, 0)
  expect_identical(se$source, "bifactor")
  expect_output(print(se), "Item-weighted mean CV")
})

test_that("legacy hand-built fits give identical numbers", {
  se_b <- special_effects(ho_fit(bifactor = TRUE))
  se_l <- special_effects(ho_fit(bifactor = FALSE))
  expect_identical(se_l$source, "legacy")
  expect_equal(se_b$groups, se_l$groups)
  expect_equal(se_b$mean_cv, se_l$mean_cv)
})

test_that("special_effects rejects non-bifactor fits with guidance", {
  sim <- sim_fa(N = 300, K = 2, ipf = 5, cpf = 0, lam = .7, rseed = 3)
  Q <- matrix(-1L, 10, 2)
  Q[1:2, ] <- 0L; Q[1:2, 1] <- 1L; Q[6:7, ] <- 0L; Q[6:7, 2] <- 1L
  f_obl <- vbfa(sim$dat, Q, v0 = .001, max_it = 500, convChk = FALSE)
  expect_error(special_effects(f_obl), "oblique")
  f_ort <- vbfa(sim$dat, Q, orthogonal = TRUE, v0 = .001, max_it = 500,
                convChk = FALSE)
  expect_error(special_effects(f_ort), "all-specified")
  expect_error(special_effects(ho_fit(), tau = 2), "tau")
})

test_that("membership flags: unassigned, crossloaded, no_general", {
  fit <- ho_fit()
  ## doctor a copy: item 3 loses its group loading entirely
  f2 <- fit
  f2$pi[3, 2:4] <- 0.01
  se <- special_effects(f2)
  expect_true(3 %in% se$unassigned)
  expect_equal(se$groups$n_items[1], 5L)
  expect_true(is.na(se$items$ratio[3]))
  ## item 4 active on two group factors -> assigned to larger |loading|
  f3 <- fit
  f3$pi[4, 3] <- 0.9
  se3 <- special_effects(f3)
  expect_true(4 %in% se3$crossloaded)
  expect_identical(se3$items$group[4], "group1")   # own loading is larger
  ## bifactor-(S-1)-style: general fixed to 0 for cluster 1's items
  g <- rep(1L, 18); g[1:6] <- 0L
  fs1 <- suppressWarnings(
    vbfa(sim_fa(N = 400, K = 3, ipf = 6, lam = .8, gamma = c(.8, .7, .6),
                rseed = 11)$dat,
         { Q <- matrix(-1L, 18, 3)
           for (k in 1:3) { a <- which(rep(1:3, each = 6) == k)[1:2]
                            Q[a, ] <- 0L; Q[a, k] <- 1L }; Q },
         bifactor = TRUE, general = g, v0 = .001, max_it = 500))
  ses1 <- special_effects(fs1)
  expect_identical(ses1$no_general, 1:6)
  expect_true(all(is.na(ses1$items$ratio[1:6])))
  expect_true(all(is.na(ses1$items$lambda[1:6])))
  expect_output(print(ses1), "no general loading")
})

## S3 methods for fitted vbpm models. The class was introduced at 0.3.0; see
## the vbpm_fit help page for the full rationale and what the class does and
## does not change for existing code.

#' The vbpm_fit class: what it is and why it exists
#'
#' Every estimator in `vbpm` returns its results with an S3 class attached:
#' [vbfa()] returns `c("vbfa", "vbpm_fit")` and [vbmimic()] returns
#' `c("vbmimic", "vbpm_fit")`. This page explains what that means in
#' practice, because "S3 class" sounds like more machinery than it is.
#'
#' @section What an S3 class actually is:
#' An S3 class is nothing more than a **label attached to an ordinary R
#' object**. A `vbpm_fit` is still a plain named list -- the class attribute
#' changes none of its contents and none of the ways you already use it:
#'
#' * `fit$Lam`, `fit$eta`, `fit$ELBO`, ... work exactly as before;
#' * `names(fit)` lists every component;
#' * `str(fit)` shows the full structure;
#' * `unclass(fit)` strips the label and gives back the bare list, should you
#'   ever want it.
#'
#' The label does exactly two things. First, generic functions such as
#' [print()] can now **dispatch**: typing `fit` at the console no longer
#' floods the screen with every matrix in the object, but shows a compact
#' summary (model type, dimensions, convergence, ELBO, active loadings).
#' Second, functions receiving a fit can **recognize what it is**: this is
#' how [fit_stats()] knows a bifactor fit from an oblique one without being told
#' (the fit carries its own `orthogonal` and `ld` settings), and how it sends a
#' [vbmimic()] fit to its own method. That method, `fit_stats.vbmimic()`,
#' returns a populated but deliberately limited result -- `NA` for every
#' SEM-like index, plus a single `n_active_coef` count of soft-selected
#' measurement and structural coefficients -- because the MIMIC objective and
#' joint covariance derivation those indices need are not available yet. It is
#' the family fallback `fit_stats.vbpm_fit()` that stops instead of computing
#' something meaningless, and it applies to classes with no method of their
#' own.
#'
#' @section Why the family shares a parent class:
#' The specific class (`"vbfa"`, `"vbmimic"`) comes first, the family class
#' `"vbpm_fit"` second. Methods written for `vbpm_fit` therefore apply to
#' every model in the family unless a model overrides them -- one `print`
#' method serves all estimators, and future family members (e.g. further
#' variational models) inherit sensible behaviour on day one.
#'
#' @section Why two classes rather than one class plus a model field:
#' A natural-looking alternative is a single `"vbpm_fit"` class with a field
#' saying which model it is. The field exists -- every fit carries
#' `$model` (`"vbfa"` or `"vbmimic"`), and it is the right thing to *read*
#' when a script wants to branch on model type. But it cannot replace the
#' class vector, because the class vector is what R's method dispatch runs
#' on. With the two-element class, model-specific behaviour is added by
#' writing a method (`fit_stats.vbmimic()`), the family default by writing
#' `fit_stats.vbpm_fit()`, and a future model (say `vblvm`) joins by
#' returning `c("vblvm", "vbpm_fit")` and adding only the methods where it
#' differs -- including from outside this package. With a single class, every
#' generic would have to contain a hand-written `switch(object$model, ...)`
#' that this package alone can extend, and `inherits(fit, "vbfa")` would stop
#' working. This is the same convention base R uses: a `glm` object has class
#' `c("glm", "lm")`, not class `"lm"` with a family field. So: dispatch on the
#' classes, read `$model` for display and bookkeeping.
#'
#' @section What is deliberately NOT hidden:
#' Some R packages wrap results in opaque objects whose internals are
#' discouraged territory. `vbpm` does the opposite: the list components are
#' documented in each estimator's help page (`?vbfa`, `?vbmimic`) and are
#' part of the public API. The class adds convenience on top; it takes
#' nothing away.
#'
#' @name vbpm_fit
#' @aliases vbpm_fit-class
#' @seealso [vbfa()], [vbmimic()], [print.vbpm_fit()]
NULL


#' Compact display of a fitted vbpm model
#'
#' Prints a short summary -- model type, dimensions, convergence, timing, and
#' selection results -- instead of dumping the full list. Access the parts as
#' usual (`fit$Lam` etc.; see [vbpm_fit]); `str(fit)` shows everything.
#'
#' @param x A fit returned by [vbfa()] or [vbmimic()].
#' @param tau Threshold on the posterior inclusion probability used only for
#'   the "active loadings" line (default `0.5`, matching [fit_stats()]).
#' @param ... Ignored (present for compatibility with the [print()] generic).
#'
#' @return `x`, invisibly (the R convention for print methods, so `print(fit)`
#'   can be used in pipes without losing the object).
#'
#' @examples
#' sim <- sim_fa(N = 300, K = 3, ipf = 6, lam = .7, lac = .3, rseed = 1)
#' Q <- matrix(-1L, ncol(sim$dat), 3)
#' for (k in 1:3) { a <- which(rep(1:3, each = 6) == k)[1:2]
#'                  Q[a, ] <- 0L; Q[a, k] <- 1L }
#' fit <- vbfa(sim$dat, Q)
#' fit               # dispatches here: compact summary, not a list dump
#' fit$Lam[1:3, ]    # the object is still an ordinary list underneath
#'
#' @export
print.vbpm_fit <- function(x, tau = 0.5, ...) {
  is_mimic <- inherits(x, "vbmimic")

  if (is_mimic) {
    J <- nrow(x$A); K <- ncol(x$A); P <- ncol(x$B); N <- nrow(x$eta)
    cat("vbmimic: regularized VB MIMIC model\n")
    cat(sprintf("  %d items, %d factors, %d covariates, N = %d\n",
                J, K, P, N))
  } else {
    J <- nrow(x$Lam); K <- ncol(x$Lam); N <- nrow(x$eta)
    cat("vbfa: VB partially confirmatory factor analysis\n")
    if (isTRUE(x$bifactor))
      cat(sprintf("  %d items, 1 general + %d group factors, N = %d  (bifactor%s)\n",
                  J, K - 1L, N, if (isTRUE(x$ld)) ", local dependence" else ""))
    else
      cat(sprintf("  %d items, %d factors, N = %d  (%s%s)\n", J, K, N,
                  if (isTRUE(x$orthogonal)) "orthogonal" else "oblique",
                  if (isTRUE(x$ld)) ", local dependence" else ""))
  }

  n_stage <- x$path$n_stage
  cat(sprintf("  %s in %d iterations%s (%.2f secs)\n",
              if (isTRUE(x$flag == 1)) "converged" else "NOT converged",
              x$iter,
              if (!is.null(n_stage) && n_stage > 1)
                sprintf(" over %d v0 stages", n_stage) else "",
              as.numeric(x$time, units = "secs")))

  if (!is.null(x$ELBO))
    cat("  ELBO:", if (is.na(x$ELBO)) "NA (not available for this model)"
        else format(x$ELBO, digits = 8), "\n")

  if (is_mimic) {
    unsA <- x$Q_A == -1
    unsB <- x$Q_B == -1
    if (any(unsA))
      cat(sprintf("  active unspecified loadings    (PIP >= %.2g): %d of %d\n",
                  tau, sum(x$pi_A[unsA] >= tau), sum(unsA)))
    if (any(unsB))
      cat(sprintf("  active structural coefficients (PIP >= %.2g): %d of %d\n",
                  tau, sum(x$pi_B[unsB] >= tau), sum(unsB)))
  } else {
    uns <- x$Q == -1
    if (any(uns))
      cat(sprintf("  active unspecified loadings (PIP >= %.2g): %d of %d\n",
                  tau, sum(x$pi[uns] >= tau), sum(uns)))
  }

  cat("\nComponents: ", paste(names(x), collapse = ", "), "\n", sep = "")
  cat("Access them with $ as usual; see ?vbpm_fit.\n")
  invisible(x)
}


#' Extract coefficients from a fitted vbpm model
#'
#' For a [vbfa()] fit, returns the loading matrix. For a [vbmimic()] fit,
#' returns a list with the measurement loadings `A` and the structural
#' coefficients `B`.
#'
#' @param object A fit returned by [vbfa()] or [vbmimic()].
#' @param ... Ignored.
#' @return A matrix (vbfa) or a list of two matrices (vbmimic).
#' @examples
#' sim <- sim_fa(N = 300, K = 3, ipf = 6, lam = .7, lac = .3, rseed = 1)
#' Q <- matrix(-1L, ncol(sim$dat), 3)
#' for (k in 1:3) { a <- which(rep(1:3, each = 6) == k)[1:2]
#'                  Q[a, ] <- 0L; Q[a, k] <- 1L }
#' fit <- vbfa(sim$dat, Q)
#' round(coef(fit), 2)
#' @export
coef.vbpm_fit <- function(object, ...) {
  if (inherits(object, "vbmimic")) list(A = object$A, B = object$B)
  else object$Lam
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(vbpm)

## ----sweep--------------------------------------------------------------------
sim <- sim_fa(N = 500, K = 3, ipf = 6, lam = .7, lac = .3, rseed = 11)
Y <- sim$dat
groups <- rep(1:3, each = 6)

Q0 <- matrix(-1L, ncol(Y), 2)
for (k in 1:2) Q0[which(groups == k)[1:2], k] <- 1L
Q0[1:8, ]

p <- pefa(Q0, Y, Kmin = 2, Kmax = 7, verbose = FALSE)
p

## ----object-------------------------------------------------------------------
names(p)
class(p)
names(p$settings)

## ----matrices-----------------------------------------------------------------
names(p$loadings)
names(p$pips)
round(p$loadings[["3"]], 2)
round(p$pips[["3"]][1:6, ], 2)

## ----sweep-table--------------------------------------------------------------
p$sweep

## ----transitions--------------------------------------------------------------
p$transitions

## ----persistence--------------------------------------------------------------
names(p$persistence)
round(p$persistence$phi, 3)
round(p$persistence$rmsd, 3)
p$persistence$collision

## ----k2-----------------------------------------------------------------------
round(p$loadings[["2"]][c(1, 7, 8, 13, 17, 18), ], 2)

## ----ssl----------------------------------------------------------------------
ssl(p)

## ----k7-----------------------------------------------------------------------
round(p$loadings[["7"]][13:18, 3:5], 2)

## ----reading------------------------------------------------------------------
profile_specs <- data.frame(
  profile = c(".85/r1", ".80/r2", ".70/r3"),
  phi_cut = c(.85, .80, .70), r = 1:3,
  practical_default = c(FALSE, TRUE, FALSE)
)
profile_sources <- 2:4

read_sweep <- function(x, phi_cut, r, sources) {
  K <- as.integer(x$sweep$K)
  converged <- x$sweep$converged

  read_count <- function(score) {
    usable <- length(K) >= 2L && all(is.finite(K)) && all(diff(K) == 1L) &&
      length(score) == length(K) && all(is.finite(score)) &&
      is.logical(converged) && all(converged %in% TRUE)
    if (!usable) return(list(usable = FALSE, Khat = NA_integer_))
    gain <- diff(score)
    gain_max <- max(gain)
    if (!(gain_max > 0)) return(list(usable = TRUE, Khat = NA_integer_))
    peak <- max(which(gain == gain_max))
    hit <- which(seq_along(gain) > peak & gain < .20 * gain_max)
    list(usable = TRUE,
         Khat = if (length(hit)) K[hit[1L]] else NA_integer_)
  }

  ELBO <- read_count(x$sweep$ELBO)
  BIC <- read_count(-x$sweep$BIC)
  C20 <- sort(unique(c(ELBO$Khat, BIC$Khat)[
    is.finite(c(ELBO$Khat, BIC$Khat))]))

  phi <- x$persistence$phi
  collision <- x$persistence$collision
  conv <- setNames(converged, K)
  state <- vapply(sources, function(k) {
    targets <- k + seq_len(r)
    edge <- vapply(targets, function(target) {
      from <- as.character(k); to <- as.character(target)
      if (!from %in% rownames(phi) || !to %in% colnames(phi) ||
          !isTRUE(conv[[from]]) || !isTRUE(conv[[to]]) ||
          is.na(collision[from, to])) return(NA_integer_)
      if (isTRUE(collision[from, to])) return(0L)
      value <- phi[from, to]
      if (!is.finite(value)) NA_integer_ else as.integer(value >= phi_cut)
    }, integer(1))
    if (any(edge %in% 0L)) 0L else
      if (length(edge) && all(edge %in% 1L)) 1L else NA_integer_
  }, integer(1))

  persistent <- which(state == 1L)
  if (!length(state)) {
    resolution <- "unresolved"
    Kp <- NA_integer_
  } else if (!length(persistent)) {
    resolution <- if (all(state %in% 0L)) "nonpersistent" else "unresolved"
    Kp <- NA_integer_
  } else {
    top <- max(persistent)
    higher <- if (top < length(state)) state[(top + 1L):length(state)] else 0L
    resolution <- if (all(higher %in% 0L)) "persistent" else "unresolved"
    Kp <- if (resolution == "persistent") sources[top] else NA_integer_
  }

  hit_ELBO <- is.finite(Kp) && is.finite(ELBO$Khat) && ELBO$Khat == Kp
  hit_BIC <- is.finite(Kp) && is.finite(BIC$Khat) && BIC$Khat == Kp
  support <- if (hit_ELBO && hit_BIC) "both" else if (hit_ELBO) "ELBO only" else
    if (hit_BIC) "BIC only" else "none"
  any_count_usable <- ELBO$usable || BIC$usable
  layer <- if (resolution == "persistent" && Kp %in% C20) "L1" else
    if (resolution == "persistent" && any_count_usable) "L2" else
    if (resolution == "nonpersistent") "L3" else
      "unclassified"
  gap <- if (resolution == "persistent" && length(C20))
    min(abs(C20 - Kp)) else NA_real_

  list(Khat_ELBO = ELBO$Khat, Khat_BIC = BIC$Khat, C20 = C20,
       count_usable = c(ELBO = ELBO$usable, BIC = BIC$usable),
       support = support, Kp = Kp, source_state = setNames(state, sources),
       resolution = resolution, layer = layer, g_CP = gap)
}

profile_row <- function(x, spec) {
  z <- read_sweep(x, spec$phi_cut, spec$r, profile_sources)
  data.frame(profile = spec$profile, practical_default = spec$practical_default,
             Khat_ELBO = z$Khat_ELBO, Khat_BIC = z$Khat_BIC,
             C20 = if (length(z$C20)) paste0("{", paste(z$C20,
               collapse = ","), "}") else "{}",
             support = z$support, Kp = z$Kp, layer = z$layer,
             g_CP = z$g_CP, row.names = NULL)
}

profile_results <- do.call(rbind, lapply(seq_len(nrow(profile_specs)),
  function(i) profile_row(p, profile_specs[i, ])))
profile_results

## A declared source at the upper edge cannot persist without its endpoint.
short_window <- read_sweep(p, phi_cut = .85, r = 1L, sources = 7L)
data.frame(profile = ".85/r1", declared_source = 7L,
           required_endpoint = 8L, resolution = short_window$resolution,
           layer = short_window$layer)

## ----summary------------------------------------------------------------------
summary(p)

## ----plots, fig.width=7, fig.height=4.5---------------------------------------
plot(p, type = "objective", criterion = "ELBO")
plot(p, type = "gain", criterion = "BIC", pct = TRUE)
plot(p, type = "fit")

## Internal: Gaussian conditional moments of the missing responses in one row.
## Deterministic analogue of the MCMC data augmentation in LAWBL/PCFA. Kept
## separate so it can be tested against the textbook MVN conditional directly.
##
##   W        J x J residual covariance (diagonal or full)
##   mean_row length-J factor-model mean eta_i Lam'
##   y_row    length-J response row (missing entries ignored)
##   M, O     integer indices of missing and observed entries
## Returns list(mean = E[y_M | y_O], cov = Var[y_M | y_O]).
.cond_moments <- function(W, mean_row, y_row, M, O) {
  if (!length(O))
    return(list(mean = mean_row[M], cov = W[M, M, drop = FALSE]))
  WOO_inv_r <- solve(W[O, O, drop = FALSE], y_row[O] - mean_row[O])
  list(mean = mean_row[M] + as.vector(W[M, O, drop = FALSE] %*% WOO_inv_r),
       cov  = W[M, M, drop = FALSE] -
              W[M, O, drop = FALSE] %*%
              solve(W[O, O, drop = FALSE], W[O, M, drop = FALSE]))
}

## Internal: log-normalizing constant of the inverse generalized Wishart,
## used in the oblique ELBO.
logC_igw <- function(xi, Lam, P) {
  0.5 * (xi - P + 1) * log(det(Lam)) -
    0.5 * P * (xi - P + 1) * log(2) -
    0.25 * P * (P - 1) * log(pi) -
    sum(lgamma((xi - P - (1:P)) / 2 + 1))
}


#' Variational Bayes partially confirmatory factor analysis
#'
#' Fits a partially confirmatory factor-analytic measurement model by
#' regularized mean-field variational Bayes. Loadings on *specified* factors
#' are anchored by the design matrix `Q`; loadings on *unspecified* factors are
#' selected from the data through continuous spike-and-slab priors. Optional
#' features are a warm-started dynamic regularization path, an orthogonal
#' bifactor parameterization, and sparse residual (local-dependence)
#' estimation.
#'
#' @param Y Numeric `N x J` data matrix (observations by items). Standardized
#'   internally. Missing continuous responses are handled by deterministic
#'   Gaussian conditional-moment augmentation inside the variational loop.
#' @param Q Integer `J x K` design matrix for the loadings: `1` = specified
#'   (anchored) loading, `0` = fixed zero, `-1` = unspecified (estimated by
#'   spike-and-slab). `K` is the number of factors.
#' @param ld Logical switch for local-dependence (sparse residual) estimation.
#'   `FALSE` (default) fits diagonal residuals and ignores `Qe` (with a warning
#'   if one is supplied). `TRUE` estimates a sparse residual precision by a
#'   graphical spike-and-slab prior solved by QUIC, with the search space given
#'   by `Qe`.
#' @param Qe Residual design matrix (`J x J`), read only when `ld = TRUE`.
#'   Entries `1` (freely estimated residual dependence), `-1` (uncertain;
#'   selected by spike-and-slab), `0` (shrunk to zero). Symmetric; its diagonal
#'   is ignored. The default `NULL` means fully exploratory local dependence,
#'   i.e. `Qe = matrix(-1, J, J)`.
#' @param bifactor Logical. If `TRUE`, fit an orthogonal bifactor model with
#'   `Q` supplying only the **group** design (`J x K` for `K` group factors):
#'   the general column is added internally according to `general`, so the
#'   user-facing `K` counts group factors and is directly comparable to an
#'   oblique `K`-factor model. Implies orthogonal factors --
#'   `bifactor = TRUE` overrides `orthogonal = FALSE` (with a message).
#'   Supplying a `Q` that already contains an all-specified column is an
#'   error: with `bifactor = TRUE` the general column must come from
#'   `general`, not from `Q`.
#' @param general Design of the general column when `bifactor = TRUE`: a
#'   scalar or length-`J` vector with the usual codes (`1` specified, `0`
#'   fixed zero, `-1` unspecified). The default `1` gives the standard
#'   bifactor (every item loads on the general factor). A vector with `0`
#'   entries yields bifactor-(S-1)-style designs in which some items carry
#'   no general loading; at least one entry must be `1`.
#' @param orthogonal Logical. If `TRUE`, the factor correlation is fixed at
#'   the identity (a plain orthogonal factor model; also the engine under a
#'   hand-built bifactor design). Default `FALSE` (oblique; factor
#'   correlation estimated). For bifactor models prefer `bifactor = TRUE`,
#'   which implies orthogonality and keeps `K` counting group factors.
#' @param v0 Loading-spike variance. A decreasing **vector** gives a
#'   warm-started dynamic regularization path (each stage runs to convergence
#'   and warm-starts the next); a **scalar** gives a single fixed spike. The
#'   default path ends at `0.001`; a scalar `v0 = 0.001` reproduces the
#'   fixed-spike estimator exactly.
#'
#'   For a **gentler path that aids convergence**, pass a finer, longer,
#'   log-spaced descent starting from a larger `v0`, e.g.
#'   `v0 = exp(seq(log(0.05), log(0.001), length.out = 10))`. Each stage is
#'   then a small perturbation of the previous one, so its warm start is a good
#'   initialization and it converges quickly and stably. Prefer this when a fit
#'   fails to converge within `max_it`, when the loading solution looks
#'   unstable, or for near-singular / strongly regularized models: it navigates
#'   the non-convex spike-and-slab landscape better than a single hard fit, and
#'   is usually preferable to (and often faster than) simply raising `max_it`.
#'   Keep the endpoint at your target; log spacing suits the multiplicative
#'   spike.
#' @param max_it Maximum variational iterations **per stage**. Default `5000`.
#' @param convChk Logical; report per-iteration relative error via
#'   [message()] (so it can be silenced with [suppressMessages()]). Default
#'   `FALSE` (quiet).
#' @param tolVal Convergence tolerance on the maximum relative error.
#' @param ld_control A named list of local-dependence controls, read only when
#'   `ld = TRUE`: `xi0` (spike-penalty path, units of `N`; default
#'   `seq(0.1, 1, length.out = 5)`), `xi1` (slab penalty, units of `N`;
#'   default `0.01`), `diag_penalty` (`1` penalizes the precision diagonal with
#'   `xi1`, `0` leaves it unpenalized), `quic_eps`, `quic_max_it`, and the Beta
#'   prior parameters `a1`, `b1` on the local-dependence proportion.
#'
#' @return An object of class `c("vbfa", "vbpm_fit")` -- a named list (see
#'   [vbpm_fit] for what the class does and does not change) with elements:
#'   \describe{
#'     \item{Lam}{`J x K` posterior mean loadings.}
#'     \item{pi}{Posterior inclusion probabilities of the unspecified loadings
#'       (specified entries are reported as coded in `Q`).}
#'     \item{eta}{`N x K` posterior mean factor scores.}
#'     \item{Phi}{Factor correlation matrix; the identity when
#'       `orthogonal = TRUE`.}
#'     \item{PsiInv}{Residual precisions; `NULL` when `ld = TRUE`.}
#'     \item{ELBO}{Evidence lower bound; `NA` when `ld = TRUE`.}
#'     \item{rho}{Posterior inclusion rate of the unspecified loadings.}
#'     \item{Lam_var}{Posterior variances of the loadings.}
#'     \item{eta_cov}{Posterior covariance of the factor scores.}
#'     \item{Phi_inv_mean}{Posterior mean of the factor precision,
#'       E\[Phi^-1\].}
#'     \item{iter, flag, time}{Total iterations, `1` if the final stage
#'       converged within `max_it` (else `0`), and elapsed time.}
#'     \item{Q, Qe, orthogonal, ld}{The design matrices and settings the model
#'       was fit with (`Qe` is `NULL` when `ld = FALSE`); read by [fit_stats()]
#'       and [print.vbpm_fit()] so they never need repeating.}
#'     \item{bifactor, n_general, general}{Bifactor bookkeeping: whether the
#'       fit declared a bifactor structure, how many general columns it has,
#'       and the general design (`NULL` unless `bifactor = TRUE`).}
#'     \item{Psi, W, q_star, xi_star, tau}{Local-dependence results (`NULL`
#'       when `ld = FALSE`): residual precision, its inverse, residual-edge
#'       inclusion probabilities, edge shrinkage, and the LD proportion.}
#'     \item{path}{The `v0` schedule (and `xi0` under LD) with per-stage
#'       iteration counts.}
#'     \item{preprocess}{Centering and scaling used, `n_missing`, the response
#'       type, and `missing_mask` -- the mask is stored only when the data
#'       actually contain missing values, and is `NULL` otherwise.}
#'     \item{sample_cov}{The sample correlation matrix the fit statistics use.}
#'   }
#'
#'   Every quantity appears exactly once: as of 0.7.0 the fit no longer
#'   carries `$coefficients`, `$design`, `$posterior`, or `$settings`, which
#'   duplicated the elements above.
#'
#' @section Notes:
#' The diagonal-residual estimator is **deterministic** (fixed initialization,
#' no random numbers), so results are reproducible without a seed; there is no
#' `rseed` argument. (The local-dependence branch's QUIC solver shuffles its
#' active set with the RNG, but that is a convex solve, so the shuffle order
#' does not change the solution.)
#'
#' Missing continuous responses are supported (see `Y`). Support for
#' categorical or mixed responses is not yet implemented.
#'
#' Under local dependence `ELBO` is `NA` by design, and this is not a
#' placeholder: the residual precision and the mixing proportion are point
#' updates rather than variational factors, so no single joint mean-field
#' bound over all unknowns is defined. The quantities that *are* defined are
#' returned under their own names --- `objective` (the terminal VECM objective,
#' with `objective_type = "vecm"`) and `ELBO_conditional` (the conditional
#' bound given the terminal precision and mixing proportion). Use `objective`
#' for model comparison, and only across fits sharing one `objective_type`.
#'
#' @references
#' Chen, J., Guo, Z., Zhang, L., & Pan, J. (2021). A partially confirmatory
#' approach to scale development with the Bayesian Lasso. *Psychological
#' Methods*, 26(2), 210-235. \doi{10.1037/met0000293}
#'
#' Chen, J. (2021). A generalized partially confirmatory factor analysis
#' framework with mixed Bayesian Lasso methods. *Multivariate Behavioral
#' Research*, 57(6), 879-894. \doi{10.1080/00273171.2021.1925520}
#'
#' Jin, Y., & Chen, J. (2025). Regularized variational approximation for
#' partially confirmatory factor analysis. *Structural Equation Modeling: A
#' Multidisciplinary Journal*, 32(3), 437-449.
#' \doi{10.1080/10705511.2024.2432612}
#'
#' Jin, Y., Chen, J., Yan, Z., & Zhang, Y. (2026). Sparse residual estimation
#' in partially confirmatory factor analysis. *PsyArXiv preprint*.
#' \doi{10.31234/osf.io/dehtv_v2} --- the warm-started `v0` path and the
#' `ld = TRUE` sparse-residual (local dependence) branch.
#'
#' Rockova, V., & George, E. I. (2018). The spike-and-slab LASSO.
#' *Journal of the American Statistical Association*, 113(521), 431-444.
#' \doi{10.1080/01621459.2016.1260469}
#'
#' @examples
#' ## partially confirmatory: two anchors per factor, the rest exploratory
#' sim <- sim_fa(N = 200, K = 3, ipf = 6, lam = .7, lac = .3, rseed = 1)
#' Q <- matrix(-1L, ncol(sim$dat), 3)
#' for (k in 1:3) { a <- which(rep(1:3, each = 6) == k)[1:2]; Q[a, ] <- 0; Q[a, k] <- 1 }
#' fit <- vbfa(sim$dat, Q)
#' fit$flag                    # 1 if converged
#' round(fit$Lam, 2)           # loadings
#' round(fit$pi, 2)            # posterior inclusion probabilities
#'
#' \donttest{
#' ## orthogonal bifactor, preferred form: Q holds only the GROUP design and
#' ## K counts group factors (comparable to an oblique K-factor model)
#' fb <- vbfa(sim$dat, Q, bifactor = TRUE, v0 = .001)  # fixed spike: fast demo
#' fb                          # prints "1 general + 3 group factors"
#' fb$Phi                      # identity by construction
#' ## equivalent long form: vbfa(sim$dat, cbind(1L, Q), orthogonal = TRUE)
#'
#' ## local dependence: simulate correlated residuals, then set ld = TRUE
#' simLD <- sim_fa(N = 300, K = 3, ipf = 6, lam = .7, lac = .3, ecr = .3,
#'                 rseed = 2)
#' fLD <- vbfa(simLD$dat, Q, ld = TRUE, max_it = 300,
#'             tolVal = 1e-3)
#' ## largest recovered residual edges vs the planted ones
#' Poff <- abs(fLD$Psi); Poff[lower.tri(Poff, diag = TRUE)] <- 0
#' which(Poff >= sort(Poff, decreasing = TRUE)[3], arr.ind = TRUE)
#' simLD$ofd_ind
#'
#' ## empirical illustration: NLSY 1997, 27 items treated as continuous.
#' ## A 500-row subset keeps the example quick; the vignette fits all 3,458.
#' data(nlsy27)
#' Yn <- as.matrix(nlsy27$dat)[1:500, ]
#' sum(is.na(Yn))            # incomplete cells, imputed in-loop
#' fn <- vbfa(Yn, nlsy27$Q)
#' round(fn$Lam, 2)
#' }
#'
#' @export
vbfa <- function(Y, Q, ld = FALSE, Qe = NULL, orthogonal = FALSE,
                 bifactor = FALSE, general = 1,
                 v0 = c(0.01, 0.005, 0.002, 0.001),
                 max_it = 5000, convChk = FALSE, tolVal = 1e-4,
                 ld_control = list()) {

  call <- match.call()

  ## ---- input checks --------------------------------------------------
  Y <- as.matrix(Y)
  missing_mask <- is.na(Y)
  has_missing <- any(missing_mask)
  if (has_missing && any(colSums(!missing_mask) == 0L))
    stop("Every item must have at least one observed response.", call. = FALSE)
  if (has_missing && any(rowSums(!missing_mask) == 0L))
    stop("Rows with all responses missing are not supported.", call. = FALSE)
  Q <- as.matrix(Q)
  if (nrow(Q) != ncol(Y))
    stop("The numbers of items in data and Q are unequal.", call. = FALSE)
  if (!all(Q %in% c(-1, 0, 1)))
    stop("Q entries must be in {-1, 0, 1}.", call. = FALSE)

  ## ---- bifactor mode: Q is the GROUP design; the general column is built
  ## here, so user-facing K counts group factors (comparable to an oblique
  ## K-factor model). bifactor = TRUE overrides orthogonal = FALSE.
  bifactor <- isTRUE(bifactor)
  if (bifactor) {
    if (!is.numeric(general) || anyNA(general) || !all(general %in% c(-1, 0, 1)))
      stop("general must contain only -1, 0, or 1 (scalar or length-J vector).",
           call. = FALSE)
    general <- rep_len(as.integer(general), ncol(Y))
    if (!any(general == 1))
      stop("general must specify (1) the general loading for at least one item.",
           call. = FALSE)
    if (any(apply(Q == 1, 2, all)))
      stop("Q contains an all-specified column. With bifactor = TRUE, supply ",
           "only the GROUP design in Q; the general column is added ",
           "internally (see the `general` argument).", call. = FALSE)
    if (!missing(orthogonal) && !isTRUE(orthogonal))
      message("bifactor = TRUE implies orthogonal factors; ",
              "orthogonal = FALSE is overridden.")
    orthogonal <- TRUE
    Q <- cbind(general, Q, deparse.level = 0)
  }

  ld <- isTRUE(ld)
  if (!ld) {
    if (!is.null(Qe)) {
      warning("Qe is ignored because ld = FALSE; set ld = TRUE to enable ",
              "local-dependence estimation.", call. = FALSE)
      Qe <- NULL
    }
  } else {
    if (is.null(Qe))
      Qe <- matrix(-1, ncol(Y), ncol(Y))   # fully exploratory local dependence
    Qe <- as.matrix(Qe)
    if (nrow(Qe) != ncol(Y) || ncol(Qe) != ncol(Y))
      stop("Qe must be J x J (J = ncol(Y)).", call. = FALSE)
    if (!all(Qe %in% c(-1, 0, 1)))
      stop("Qe entries must be in {-1, 0, 1}.", call. = FALSE)
    if (!isSymmetric(unname(Qe)))
      stop("Qe must be symmetric (its diagonal is ignored).", call. = FALSE)
    quic_fun <- .quic_rcpp   # package-internal QUIC solver (src/quic_impl.cpp)
  }

  ## ---- ld_control defaults -------------------------------------------
  J <- ncol(Y)
  N <- nrow(Y)
  P <- ncol(Q)
  ctrl <- list(xi0 = seq(0.1, 1, length.out = 5),  # units of N
               xi1 = 0.01,                          # units of N
               diag_penalty = 1,
               quic_eps = 1e-3, quic_max_it = 1000,
               a1 = 1, b1 = J)
  if (length(ld_control)) {
    unknown <- setdiff(names(ld_control), names(ctrl))
    if (length(unknown))
      stop("Unknown ld_control entries: ", paste(unknown, collapse = ", "),
           call. = FALSE)
    ctrl[names(ld_control)] <- ld_control
  }

  ## ---- dynamic-path staging -------------------------------------------
  ## stages = max(length(v0), length(xi0)); shorter held at its last value.
  v0_seq <- as.numeric(v0)
  if (any(v0_seq <= 0)) stop("v0 must be positive.", call. = FALSE)
  if (ld) {
    xi0_seq <- as.numeric(ctrl$xi0)
    n_stage <- max(length(v0_seq), length(xi0_seq))
    v0_seq  <- c(v0_seq,  rep(v0_seq[length(v0_seq)],
                              n_stage - length(v0_seq)))
    xi0_seq <- c(xi0_seq, rep(xi0_seq[length(xi0_seq)],
                              n_stage - length(xi0_seq)))
  } else {
    n_stage <- length(v0_seq)
    xi0_seq <- rep(NA_real_, n_stage)
  }

  ## ---- RNG hygiene ---------------------------------------------------
  ## The diagonal-residual estimator consumes no random numbers (deterministic).
  ## The only RNG use is the QUIC active-set shuffle in the local-dependence
  ## branch -- a convex solve, so the order does not change the solution. We
  ## save and restore the caller's RNG state so a fit leaves it untouched.
  if (exists(".Random.seed", .GlobalEnv)) {
    oldseed <- .GlobalEnv$.Random.seed
    on.exit(assign(".Random.seed", oldseed, envir = .GlobalEnv))
  }

  ## ---- shared setup and initialization -------------------------------
  center <- colMeans(Y, na.rm = TRUE)
  scale_ <- apply(Y, 2, stats::sd, na.rm = TRUE)
  if (any(!is.finite(scale_) | scale_ <= 0))
    stop("Every item must have a positive observed-data standard deviation.",
         call. = FALSE)
  ## Preserve the historical complete-data arithmetic bit-for-bit. The
  ## explicit observed-data transformation is used only when it is needed.
  Y <- if (has_missing) sweep(sweep(Y, 2, center, "-"), 2, scale_, "/") else
       scale(Y)
  Y[missing_mask] <- 0   # observed-scale mean; updated inside the VB loop
  missing_rows <- if (has_missing) which(rowSums(missing_mask) > 0L) else integer()
  missing_cov_correction <- matrix(0, J, J)

  flag <- 0
  xi.PHI  <- 2 * P + 1
  Lam.PHI <- diag(1, P, P)

  mu.q.Lam <- mu.q.Lam_sq <- matrix(0.8, J, P)
  mu.q.Lam[Q == -1] <- mu.q.Lam_sq[Q == -1] <- 0.2
  mu.q.Lam[Q == 0]  <- mu.q.Lam_sq[Q == 0]  <- 0
  ## Fixed-zero loadings have no posterior, so their variance must be 0. The
  ## update loops skip Q == 0, so this initial value persists; leaving it at 1
  ## (as ld-vb_r.R does) silently adds sum_p E[eta_p^2] * 1 to the expected
  ## residual covariance diagonal of every item with fixed-zero entries in the
  ## LD branch -- ~ +1 per zero column after the /N -- inflating anchored
  ## items' residual variances ~5x. Found 2026-07-23; the V4 reference shares
  ## the defect.
  sigsq.q.Lam <- matrix(1, J, P)
  sigsq.q.Lam[Q == 0] <- 0

  mu.q.eta <- mu.q.eta_sq <- matrix(1, N, P)
  mu.q.d <- matrix(1, J, P)
  v1 <- 1
  inv_v1 <- 1 / v1
  h_0 <- 1
  gamma.q.d <- rep(1, J)

  PHI.q.eta <- M.q.LamTLam <- matrix(1, P, P)
  M.q.PHIiver <- diag(1, P, P)   # orthogonal: stays I throughout (Phi = I)

  xi.q.PHI <- N + xi.PHI

  a <- 1
  b <- 0.1
  a.q.psi <- rep(N / 2 + a, J)
  b.q.psi <- rep(b, J)
  mu.q.psiInv <- rep(1, J)

  a0 <- 1
  b0 <- 1

  q <- Q
  rho <- 0.5

  mu_tem2 <- matrix(0, J, P)

  ## ---- LD-specific state ------------------------------------------------
  if (ld) {
    Psi <- W <- diag(1, J)       # full residual PRECISION (Psi) and its
    tau <- 0.5                   # inverse W; tau = LD proportion
    q_star <- xi_star <- matrix(0.5, J, J)
    xi1_abs <- ctrl$xi1 * N
    a1 <- ctrl$a1
    b1 <- ctrl$b1
    upper_tri_idx  <- which(upper.tri(Psi), arr.ind = TRUE)
    uncertain_upper <- upper.tri(Qe) & Qe == -1
  }

  ## ---- staged VB loop ----------------------------------------------------
  itNum <- 0
  stage_iters <- integer(n_stage)
  MFVBstartTime <- Sys.time()

  for (stage in seq_len(n_stage)) {
    v0 <- v0_seq[stage]
    inv_v0 <- 1 / v0
    if (ld) xi0_abs <- xi0_seq[stage] * N
    converged <- FALSE
    parVecHatPrev <- NULL
    stage_it <- 0
    flag <- 0        # reset per stage; reflects the FINAL stage on return
    if (convChk && n_stage > 1)
      message("Stage ", stage, "/", n_stage, ": v0 = ", signif(v0, 4),
              if (ld) paste0("; xi0 = ", signif(xi0_seq[stage], 4), " * N"))

    while (!converged) {
      itNum <- itNum + 1
      stage_it <- stage_it + 1

      ## -- missing-response conditional moments ------------------------
      ## Deterministic analogue of LAWBL's Gaussian MCMC data augmentation.
      ## Multiple missing responses in a row are updated jointly.
      if (has_missing) {
        W_now <- if (ld) W else diag(1 / mu.q.psiInv, J)
        mean_now <- mu.q.eta %*% t(mu.q.Lam)
        missing_cov_correction[,] <- 0
        for (ii in missing_rows) {
          M <- which(missing_mask[ii, ])
          O <- which(!missing_mask[ii, ])
          cmv <- .cond_moments(W_now, mean_now[ii, ], Y[ii, ], M, O)
          Y[ii, M] <- cmv$mean
          missing_cov_correction[M, M] <-
            missing_cov_correction[M, M, drop = FALSE] + cmv$cov
        }
      }

      ## -- eta update ---------------------------------------------------
      if (!ld) {
        M.q.LamTpsiInvLam <- t(mu.q.Lam) %*% diag(mu.q.psiInv, J, J) %*% mu.q.Lam
        diag(M.q.LamTpsiInvLam) <- apply(mu.q.Lam_sq * mu.q.psiInv, 2, sum)
        PHI.q.eta <- solve(M.q.LamTpsiInvLam + M.q.PHIiver)
        mu.q.eta <- t(PHI.q.eta %*% t(Y %*% (mu.q.Lam * mu.q.psiInv)))
      } else {
        M.q.LamTpsiInvLam <- t(mu.q.Lam) %*% Psi %*% mu.q.Lam
        diag(M.q.LamTpsiInvLam) <- diag(M.q.LamTpsiInvLam) +
          apply(diag(Psi) * (mu.q.Lam_sq - mu.q.Lam^2), 2, sum)
        PHI.q.eta <- solve(M.q.LamTpsiInvLam + M.q.PHIiver)
        mu.q.eta <- t(PHI.q.eta %*% t(Y %*% Psi %*% mu.q.Lam))
      }

      ## standardize factor variance to 1 via the posterior second moment
      ## E_q[eta_p^2] = mean(eta_p^2) + Var_q(eta_p)   (both modes)
      scale_k <- sqrt(colMeans(mu.q.eta^2) + diag(PHI.q.eta))
      mu.q.eta <- t(t(mu.q.eta) / scale_k)
      PHI.q.eta <- PHI.q.eta / outer(scale_k, scale_k)
      mu.q.eta_sq <- sweep(mu.q.eta^2, 2, diag(PHI.q.eta), '+')

      E_eta_eta <- t(mu.q.eta) %*% mu.q.eta + N * PHI.q.eta

      if (!orthogonal) {               # oblique: inverse-Wishart update
        Lam.q.PHI <- E_eta_eta + Lam.PHI
        M.q.PHIiver <- (xi.q.PHI - P + 1) * solve(Lam.q.PHI)
      }                                # orthogonal: M.q.PHIiver stays I

      etaY <- t(mu.q.eta) %*% Y

      ## -- loading (and, non-LD, uniqueness) updates ----------------------
      if (!ld) {
        for (j in 1:J) {
          E_lam_lam_j <- tcrossprod(mu.q.Lam[j, ])
          diag(E_lam_lam_j) <- mu.q.Lam_sq[j, ]
          RSS_j <- sum(Y[, j]^2) + missing_cov_correction[j, j] -
            2 * sum(etaY[, j] * mu.q.Lam[j, ]) +
            sum(E_eta_eta * E_lam_lam_j)
          b.q.psi[j] <- RSS_j / 2 + b
          mu.q.psiInv[j] <- a.q.psi[j] / b.q.psi[j]
          for (p in 1:P) {
            cross_part <- sum(E_eta_eta[p, -p] * mu.q.Lam[j, -p])
            if (Q[j, p] == 1) {
              sigsq.q.Lam[j, p] <- (E_eta_eta[p, p] * mu.q.psiInv[j] + h_0)^(-1)
              sigsq.q.Lam[j, p] <- max(sigsq.q.Lam[j, p], 1e-6)
              sigsq.q.Lam[j, p] <- min(sigsq.q.Lam[j, p], 1e6)
              mu.q.Lam[j, p] <- sigsq.q.Lam[j, p] * mu.q.psiInv[j] *
                (etaY[p, j] - cross_part)
              mu.q.Lam_sq[j, p] <- mu.q.Lam[j, p]^2 + sigsq.q.Lam[j, p]
              q[j, p] <- 1
            } else if (Q[j, p] == -1) {
              tem <- 1/2 * log(inv_v1 / inv_v0) + log(rho / (1 - rho)) +
                1/2 * mu.q.Lam_sq[j, p] * (inv_v0 - inv_v1)
              tem <- pmin(tem, 100)
              q[j, p] <- exp(tem) / (1 + exp(tem))
              mu_tem2[j, p] <- (1 - q[j, p]) * inv_v0 + q[j, p] * inv_v1
              sigsq.q.Lam[j, p] <- (E_eta_eta[p, p] * mu.q.psiInv[j] +
                                      mu_tem2[j, p])^(-1)
              sigsq.q.Lam[j, p] <- max(sigsq.q.Lam[j, p], 1e-6)
              sigsq.q.Lam[j, p] <- min(sigsq.q.Lam[j, p], 1e6)
              mu.q.Lam[j, p] <- sigsq.q.Lam[j, p] * mu.q.psiInv[j] *
                (etaY[p, j] - cross_part)
              mu.q.Lam_sq[j, p] <- mu.q.Lam[j, p]^2 + sigsq.q.Lam[j, p]
            }
          }
        }
      } else {
        ## LD: loading update gains a cross-item term (tem2) because the
        ## residual precision Psi is non-diagonal
        for (j in 1:J) {
          for (p in 1:P) {
            if (Q[j, p] == 0) next
            if (Q[j, p] == 1) {
              sigsq.q.Lam[j, p] <- (E_eta_eta[p, p] * Psi[j, j] + h_0)^(-1)
              sigsq.q.Lam[j, p] <- max(sigsq.q.Lam[j, p], 1e-6)
              sigsq.q.Lam[j, p] <- min(sigsq.q.Lam[j, p], 1e6)
              tem1 <- Psi[j, j] * (etaY[p, j] -
                                     sum(E_eta_eta[p, -p] * mu.q.Lam[j, -p]))
              tem2_vec <- etaY[p, -j] -
                as.vector(mu.q.Lam[-j, , drop = FALSE] %*% E_eta_eta[p, ])
              tem2 <- Psi[j, -j] %*% matrix(tem2_vec, ncol = 1)
              mu.q.Lam[j, p] <- sigsq.q.Lam[j, p] * (tem1 + tem2)
              mu.q.Lam_sq[j, p] <- mu.q.Lam[j, p]^2 + sigsq.q.Lam[j, p]
              q[j, p] <- 1
            } else if (Q[j, p] == -1) {
              tem <- 1/2 * log(inv_v1 / inv_v0) + log(rho / (1 - rho)) +
                1/2 * mu.q.Lam_sq[j, p] * (inv_v0 - inv_v1)
              tem <- pmin(pmax(tem, -100), 100)
              q[j, p] <- 1 / (1 + exp(-tem))
              mu_tem2[j, p] <- (1 - q[j, p]) * inv_v0 + q[j, p] * inv_v1
              sigsq.q.Lam[j, p] <- (E_eta_eta[p, p] * Psi[j, j] +
                                      mu_tem2[j, p])^(-1)
              sigsq.q.Lam[j, p] <- max(sigsq.q.Lam[j, p], 1e-6)
              sigsq.q.Lam[j, p] <- min(sigsq.q.Lam[j, p], 1e6)
              tem1 <- Psi[j, j] * (etaY[p, j] -
                                     sum(E_eta_eta[p, -p] * mu.q.Lam[j, -p]))
              tem2_vec <- etaY[p, -j] -
                as.vector(mu.q.Lam[-j, , drop = FALSE] %*% E_eta_eta[p, ])
              tem2 <- Psi[j, -j] %*% matrix(tem2_vec, ncol = 1)
              mu.q.Lam[j, p] <- sigsq.q.Lam[j, p] * (tem1 + tem2)
              mu.q.Lam_sq[j, p] <- mu.q.Lam[j, p]^2 + sigsq.q.Lam[j, p]
            }
          }
        }
      }

      ## -- rho update (shared) --------------------------------------------
      a0_q_rho <- a0 + sum(q[Q == -1])
      b0_q_rho <- b0 + sum(Q == -1) - sum(q[Q == -1])
      rho <- a0_q_rho / (sum(Q == -1) + a0 + b0)

      ## -- LD block: q_star / tau / S / QUIC -------------------------------
      if (ld) {
        for (i in seq_len(nrow(upper_tri_idx))) {
          jr <- upper_tri_idx[i, 1]
          jc <- upper_tri_idx[i, 2]
          if (Qe[jr, jc] == -1) {
            tem <- plogis(log(tau / (1 - tau)) + log(xi1_abs / xi0_abs) +
                            (xi0_abs - xi1_abs) * abs(Psi[jr, jc]))
          } else if (Qe[jr, jc] == 1) {
            tem <- 1
          } else {
            tem <- 0
          }
          q_star[jr, jc] <- tem
          q_star[jc, jr] <- tem
        }

        xi_star <- xi1_abs * q_star + xi0_abs * (1 - q_star)
        if (ctrl$diag_penalty == 0) {
          diag(xi_star) <- 0
        } else {
          diag(xi_star) <- xi1_abs
        }

        if (sum(uncertain_upper) > 0) {
          tau <- (a1 - 1 + sum(q_star[uncertain_upper])) /
            (a1 + b1 - 2 + sum(uncertain_upper))
        }

        xi_star <- xi_star / N

        ## expected residual second-moment matrix S
        tem2m <- mu.q.Lam %*% E_eta_eta %*% t(mu.q.Lam)
        diag(tem2m) <- diag(tem2m) +
          rowSums(sigsq.q.Lam * matrix(diag(E_eta_eta), J, P, byrow = TRUE))
        cross_term <- t(Y) %*% mu.q.eta %*% t(mu.q.Lam)
        S <- (tem2m + t(Y) %*% Y + missing_cov_correction -
              cross_term - t(cross_term)) / N

        res_quic <- quic_fun(J, S, xi_star,
                             eps = ctrl$quic_eps,
                             max_iter_quic = ctrl$quic_max_it)
        Psi <- res_quic[, , 1]
        W   <- res_quic[, , 2]
      }

      ## -- convergence check -----------------------------------------------
      ## (the two expressions coincide when Psi is diagonal; the non-LD form
      ## is kept verbatim for bit-exact reproduction of the fixed-spike fit)
      if (!ld) {
        parVecHat <- c(t(t(abs(Y - mu.q.eta %*% t(mu.q.Lam))) * mu.q.psiInv))
      } else {
        parVecHat <- c(abs(Y - mu.q.eta %*% t(mu.q.Lam)) %*% Psi)
      }

      if (!is.null(parVecHatPrev)) {
        relErrTemp <- abs(parVecHat - parVecHatPrev)
        relErrTemp[is.nan(relErrTemp)] <- 0
        relErr <- max(relErrTemp)
        if (convChk)
          message("Iteration ", itNum, "; MFVB relative error: ",
                  signif(relErr, 5))
        if (relErr < tolVal) {
          converged <- TRUE
          flag <- 1
        }
      }
      if (stage_it >= max_it) {
        converged <- TRUE
        warning("Maximum number of MFVB iterations exceeded",
                if (n_stage > 1) paste0(" at stage ", stage), ".\n")
      }
      parVecHatPrev <- parVecHat
    }
    stage_iters[stage] <- stage_it
  }

  ## ---- ELBO ---------------------------------------------------------------
  if (!ld) {
    q_uns <- q[Q == -1]
    q_uns <- pmin(pmax(q_uns, 1e-12), 1 - 1e-12)
    if (orthogonal) {
      eta_phi_term <- -0.5 * sum(diag(E_eta_eta)) +
        0.5 * N * log(det(PHI.q.eta)) +
        0.5 * N * P
    } else {
      eta_phi_term <- 0.5 * N * log(det(PHI.q.eta)) +
        0.5 * N * P +
        logC_igw(xi.PHI, Lam.PHI, P) -
        logC_igw(xi.q.PHI, Lam.q.PHI, P)
    }
    ELBO <- (
      -0.5 * N * J * log(2 * pi) +
        sum(a * log(b) - lgamma(a) -
              a.q.psi * log(b.q.psi) + lgamma(a.q.psi)) +
        eta_phi_term +
        0.5 * sum(
          log(sigsq.q.Lam[Q == 1] / h_0) + 1 -
            mu.q.Lam_sq[Q == 1] / h_0
        ) +
        0.5 * sum(
          log(sigsq.q.Lam[Q == -1]) + 1 -
            (1 - q_uns) * log(v0) -
            q_uns * log(v1) -
            mu.q.Lam_sq[Q == -1] *
            ((1 - q_uns) / v0 + q_uns / v1)
        ) +
        lbeta(a0_q_rho, b0_q_rho) -
        lbeta(a0, b0) -
        sum(q_uns * log(q_uns) + (1 - q_uns) * log(1 - q_uns))
    )
  } else {
    ## Manuscript-V4-consistent terminal VECM objective. Psi and tau are point
    ## estimates, so this is deliberately not labelled a joint ELBO.
    q_uns <- pmin(pmax(q[Q == -1], 1e-12), 1 - 1e-12)
    if (orthogonal) {
      eta_phi_term <- -0.5 * sum(diag(E_eta_eta)) +
        0.5 * N * as.numeric(determinant(PHI.q.eta, logarithm = TRUE)$modulus) +
        0.5 * N * P
    } else {
      eta_phi_term <-
        0.5 * N * as.numeric(determinant(PHI.q.eta, logarithm = TRUE)$modulus) +
        0.5 * N * P + logC_igw(xi.PHI, Lam.PHI, P) -
        logC_igw(xi.q.PHI, Lam.q.PHI, P)
    }
    loading_term <-
      0.5 * sum(log(sigsq.q.Lam[Q == 1] / h_0) + 1 -
                  mu.q.Lam_sq[Q == 1] / h_0) +
      0.5 * sum(log(sigsq.q.Lam[Q == -1]) + 1 -
                  (1 - q_uns) * log(v0) - q_uns * log(v1) -
                  mu.q.Lam_sq[Q == -1] *
                  ((1 - q_uns) / v0 + q_uns / v1)) +
      lbeta(a0_q_rho, b0_q_rho) - lbeta(a0, b0) -
      sum(q_uns * log(q_uns) + (1 - q_uns) * log(1 - q_uns))

    logdet_psi <- as.numeric(determinant(Psi, logarithm = TRUE)$modulus)
    likelihood_term <- -0.5 * N * J * log(2 * pi) +
      0.5 * N * logdet_psi - 0.5 * N * sum(S * Psi)

    up <- upper.tri(Psi)
    uncertain <- up & Qe == -1
    fixed_slab <- up & Qe == 1
    fixed_spike <- up & Qe == 0
    qs <- pmin(pmax(q_star[uncertain], 1e-12), 1 - 1e-12)
    abs_unc <- abs(Psi[uncertain])
    edge_term <- sum(qs * (log(tau) + log(xi1_abs / 2) - xi1_abs * abs_unc) +
                     (1 - qs) * (log(1 - tau) + log(xi0_abs / 2) -
                                 xi0_abs * abs_unc) -
                     qs * log(qs) - (1 - qs) * log(1 - qs)) +
      sum(log(xi1_abs / 2) - xi1_abs * abs(Psi[fixed_slab])) +
      sum(log(xi0_abs / 2) - xi0_abs * abs(Psi[fixed_spike]))
    diag_term <- if (ctrl$diag_penalty == 0) 0 else
      J * log(xi1_abs) - xi1_abs * sum(diag(Psi))
    tau_term <- (a1 - 1) * log(tau) + (b1 - 1) * log(1 - tau) -
      lbeta(a1, b1)
    VECM_objective <- likelihood_term + eta_phi_term + loading_term +
      edge_term + diag_term + tau_term
    objective_terms <- c(likelihood = likelihood_term,
                         latent = eta_phi_term,
                         loadings = loading_term,
                         edges = edge_term,
                         diagonal = diag_term,
                         tau = tau_term)
    ELBO <- NA_real_
  }

  ## ---- sign check ----------------------------------------------------
  mean_mu.q.Lam <- apply(mu.q.Lam, 2, mean)
  sign_flag <- which(mean_mu.q.Lam < 0)
  if (length(sign_flag) > 0) {
    mu.q.Lam[, sign_flag] <- -mu.q.Lam[, sign_flag]
    mu.q.eta[, sign_flag] <- -mu.q.eta[, sign_flag]
    sign_vec <- rep(1, ncol(mu.q.Lam))
    sign_vec[sign_flag] <- -1
    PHI.q.eta <- PHI.q.eta * outer(sign_vec, sign_vec)
    M.q.PHIiver <- M.q.PHIiver * outer(sign_vec, sign_vec)
    if (!orthogonal) Lam.q.PHI <- Lam.q.PHI * outer(sign_vec, sign_vec)
  }

  MFVBendTime <- Sys.time()
  MFVBtime <- MFVBendTime - MFVBstartTime

  ## Output names are the public API and use plain-language names; the VB
  ## notation (mu.q.*, PHI.q.eta, ...) stays inside the estimator, where it
  ## mirrors the papers' derivations. Renamed at 0.3.0:
  ##   sigsq.q.Lam -> Lam_var, PHI.q.eta -> eta_cov,
  ##   M.q.PHIiver -> Phi_inv_mean; plotDat dropped.
  out <- list(model = "vbfa", call = call,
              nobs = N, nitem = J, nfactor = P,
              converged = isTRUE(flag == 1),
              Lam = mu.q.Lam, pi = q, eta = mu.q.eta,
              Phi = if (orthogonal) diag(1, P, P) else cov2cor(Lam.q.PHI),
              PsiInv = if (!ld) mu.q.psiInv else NULL,
              ELBO = ELBO,
              ELBO_conditional = if (ld) likelihood_term + eta_phi_term +
                                  loading_term else ELBO,
              objective = if (ld) VECM_objective else ELBO,
              objective_type = if (ld) "vecm" else "elbo", rho = rho,
              objective_terms = if (ld) objective_terms else c(ELBO = ELBO),
              Lam_var = sigsq.q.Lam,        # posterior variances of loadings
              eta_cov = PHI.q.eta,          # posterior covariance of scores
              Phi_inv_mean = M.q.PHIiver,   # E[Phi^-1] under q
              iter = itNum, flag = flag, time = MFVBtime,
              ## settings the model was fit with, so downstream consumers
              ## (fit_stats, print) need not be told twice
              Q = Q, Qe = if (ld) Qe else NULL,
              orthogonal = orthogonal, ld = ld,
              bifactor = bifactor,
              n_general = if (bifactor) 1L else 0L,
              general = if (bifactor) general else NULL,
              ## `missing_mask` is stored only when something is missing: an
              ## all-FALSE J x N matrix was the single largest component of a
              ## complete-data fit (65% of it at N = 3000, J = 27).
              preprocess = list(center = center, scale = scale_,
                                missing_mask = if (has_missing) missing_mask else NULL,
                                n_missing = sum(missing_mask),
                                response_type = "continuous"),
              sample_cov = stats::cov2cor((crossprod(Y) +
                                            missing_cov_correction) / (N - 1)),
              ## LD extras (NULL when ld = FALSE)
              Psi = if (ld) Psi else NULL,
              W = if (ld) W else NULL,
              q_star = if (ld) q_star else NULL,
              xi_star = if (ld) xi_star else NULL,
              tau = if (ld) tau else NULL,
              ## dynamic-path metadata
              path = list(v0 = v0_seq,
                          xi0 = if (ld) xi0_seq else NULL,
                          n_stage = n_stage, stage_iters = stage_iters))
  class(out) <- c("vbfa", "vbpm_fit")
  out
}

## Verification for the three features added after 0.1.0: in-loop missing-data
## handling in vbfa(), the local-dependence objective, and LD-aware fit
## statistics. These close verification items 2, 3 and 6 of
## the 0.4.0 extension work, which were previously only covered
## behaviourally.

make_dat <- function(N = 300, K = 3, seed = 1) {
  set.seed(seed)
  J <- K * 6
  grp <- rep(1:K, each = 6)
  Phi <- matrix(.3, K, K); diag(Phi) <- 1
  Lam <- matrix(0, J, K); Lam[cbind(1:J, grp)] <- .7
  psi <- pmax(1 - rowSums((Lam %*% chol(Phi))^2), .25)
  eta <- matrix(rnorm(N * K), N, K) %*% chol(Phi)
  Y <- eta %*% t(Lam) + matrix(rnorm(N * J), N, J) %*% diag(sqrt(psi))
  Q <- matrix(-1L, J, K)
  for (k in 1:K) { a <- which(grp == k)[1:2]; Q[a, ] <- 0L; Q[a, k] <- 1L }
  list(Y = Y, Q = Q, J = J, K = K)
}

## ---- item 3: conditional moments vs a direct MVN calculation -------------

test_that(".cond_moments matches the textbook Gaussian conditional (full W)", {
  set.seed(7)
  J <- 6
  A <- matrix(rnorm(J * J), J, J)
  W <- crossprod(A) + diag(J)          # arbitrary SPD residual covariance
  mean_row <- rnorm(J)
  y_row <- rnorm(J)
  M <- c(2L, 5L); O <- setdiff(seq_len(J), M)

  got <- vbpm:::.cond_moments(W, mean_row, y_row, M, O)

  ## direct formulas: mu_M + W_MO W_OO^-1 (y_O - mu_O), W_MM - W_MO W_OO^-1 W_OM
  WMO <- W[M, O, drop = FALSE]; WOO <- W[O, O, drop = FALSE]
  ref_mean <- mean_row[M] + as.vector(WMO %*% solve(WOO) %*% (y_row[O] - mean_row[O]))
  ref_cov  <- W[M, M, drop = FALSE] - WMO %*% solve(WOO) %*% W[O, M, drop = FALSE]

  expect_equal(got$mean, ref_mean)
  expect_equal(unname(got$cov), unname(ref_cov))
  ## conditioning must reduce uncertainty
  expect_true(all(diag(got$cov) < diag(W[M, M, drop = FALSE])))
})

test_that(".cond_moments reduces to the factor-model mean under diagonal W", {
  set.seed(8)
  J <- 5
  W <- diag(runif(J, .3, 1))
  mean_row <- rnorm(J); y_row <- rnorm(J)
  M <- c(1L, 4L); O <- setdiff(seq_len(J), M)
  got <- vbpm:::.cond_moments(W, mean_row, y_row, M, O)
  ## Equation (2): with W_MO = 0 the conditional mean is mu_M and the
  ## conditional covariance is W_MM -- observed responses carry no residual
  ## information about the missing ones.
  expect_equal(got$mean, mean_row[M])
  expect_equal(unname(got$cov), unname(W[M, M, drop = FALSE]))
})

test_that(".cond_moments handles a row with no observed responses", {
  W <- diag(3); mean_row <- c(1, 2, 3); y_row <- rep(NA_real_, 3)
  got <- vbpm:::.cond_moments(W, mean_row, y_row, M = 1:3, O = integer())
  expect_equal(got$mean, mean_row)
  expect_equal(unname(got$cov), W)
})

test_that("joint updating keeps the uncertainty that cell-by-cell imputation destroys", {
  ## The design note requires conditioning jointly on the *observed* subset,
  ## rather than treating an already-imputed neighbour as evidence.
  ##
  ## Note what this does and does not change. The conditional MEAN is the same
  ## either way: conditional expectation is linear, so plugging in E[y2 | y3]
  ## and re-conditioning reproduces E[y1 | y3] exactly (tower property). The
  ## difference is entirely in the second moment, which is what feeds the
  ## residual sufficient statistics -- treating an imputed cell as observed
  ## asserts zero uncertainty about it.
  W <- matrix(c(1, .8, .5,
                .8, 1, .4,
                .5, .4, 1), 3, 3)
  mean_row <- c(0, 0, 0); y_row <- c(NA, NA, 2)

  joint <- vbpm:::.cond_moments(W, mean_row, y_row, M = 1:2, O = 3L)
  step1 <- vbpm:::.cond_moments(W, mean_row, y_row, M = 2L, O = 3L)$mean
  y2 <- y_row; y2[2] <- step1
  seq1  <- vbpm:::.cond_moments(W, mean_row, y2, M = 1L, O = c(2L, 3L))

  ## means agree, as the tower property requires
  expect_equal(joint$mean[1], seq1$mean, tolerance = 1e-12)

  ## but the joint covariance is a full 2x2 that keeps the two missing
  ## responses correlated, and is strictly more uncertain about item 1 than
  ## the sequential route believes
  expect_equal(dim(joint$cov), c(2L, 2L))
  expect_true(abs(joint$cov[1, 2]) > 1e-8)
  expect_gt(joint$cov[1, 1], seq1$cov[1, 1])
})

## ---- item 2: mask integrity ----------------------------------------------

test_that("observed cells are never overwritten and missing cells are filled", {
  d <- make_dat()
  Y <- d$Y
  mask <- matrix(FALSE, nrow(Y), ncol(Y))
  mask[cbind(1:15, rep(1:3, each = 5))] <- TRUE
  Y[mask] <- NA
  f <- vbfa(Y, d$Q, v0 = .001, max_it = 200)

  ## the fit records exactly the mask it was given
  expect_identical(f$preprocess$missing_mask, mask)
  expect_equal(f$preprocess$n_missing, sum(mask))

  ## Observed entries must drive the fit: perturbing only the *values already
  ## marked missing* in the input cannot change anything, because they are
  ## replaced before first use.
  Y2 <- Y; Y2[mask] <- NA          # same mask, values were NA either way
  f2 <- vbfa(Y2, d$Q, v0 = .001, max_it = 200)
  expect_identical(f$Lam, f2$Lam)
})

test_that("row order does not change the fit (permutation invariance)", {
  d <- make_dat(N = 200)
  Y <- d$Y
  Y[cbind(1:12, rep(1:4, each = 3))] <- NA
  f  <- vbfa(Y, d$Q, v0 = .001, max_it = 200)

  set.seed(99); perm <- sample(nrow(Y))
  fp <- vbfa(Y[perm, ], d$Q, v0 = .001, max_it = 200)

  ## Item-level parameters are row-order invariant ...
  expect_equal(f$Lam, fp$Lam, tolerance = 1e-6)
  expect_equal(f$PsiInv, fp$PsiInv, tolerance = 1e-6)
  ## ... and factor scores follow the permutation.
  expect_equal(f$eta[perm, ], fp$eta, tolerance = 1e-6)
})

test_that("complete data is unaffected by the missing-data code path", {
  d <- make_dat()
  f <- vbfa(d$Y, d$Q, v0 = .001, max_it = 300)
  expect_equal(f$preprocess$n_missing, 0)
  expect_null(f$preprocess$missing_mask)   # not stored for complete data
  expect_true(is.finite(f$objective))
})

## ---- LD objective and item 6: LD-aware fit statistics --------------------

ld_dat <- function(N = 400, seed = 3, rho = .5) {
  set.seed(seed)
  K <- 3; J <- 18
  grp <- rep(1:K, each = 6)
  Phi <- matrix(.3, K, K); diag(Phi) <- 1
  Lam <- matrix(0, J, K); Lam[cbind(1:J, grp)] <- .7
  psi <- pmax(1 - rowSums((Lam %*% chol(Phi))^2), .25)
  Sig_e <- diag(psi)
  pairs <- rbind(c(2, 8), c(5, 14))
  for (r in seq_len(nrow(pairs))) {
    i <- pairs[r, 1]; j <- pairs[r, 2]
    Sig_e[i, j] <- Sig_e[j, i] <- rho * sqrt(psi[i] * psi[j])
  }
  eta <- matrix(rnorm(N * K), N, K) %*% chol(Phi)
  Y <- eta %*% t(Lam) + matrix(rnorm(N * J), N, J) %*% chol(Sig_e)
  Q <- matrix(-1L, J, K)
  for (k in 1:K) { a <- which(grp == k)[1:2]; Q[a, ] <- 0L; Q[a, k] <- 1L }
  list(Y = Y, Q = Q, pairs = pairs, J = J, K = K)
}

test_that("the LD objective is labelled honestly and ELBO stays NA", {
  d <- ld_dat()
  f <- vbfa(d$Y, d$Q, ld = TRUE, max_it = 150, tolVal = 1e-3)
  expect_identical(f$objective_type, "vecm")
  expect_true(is.finite(f$objective))
  ## ELBO is deliberately NA: Psi and tau are point updates, not variational
  ## factors, so no single joint mean-field bound is defined.
  expect_true(is.na(f$ELBO))
  ## the conditional bound of Eq. (3) is available separately
  expect_true(is.finite(f$ELBO_conditional))
  ## a diagonal fit reports the ordinary ELBO instead
  fd <- vbfa(d$Y, d$Q, v0 = .001, max_it = 200)
  expect_identical(fd$objective_type, "elbo")
  expect_equal(fd$objective, fd$ELBO)
})

test_that("LD fit_stats uses the residual covariance W, not the precision", {
  d <- ld_dat()
  f <- vbfa(d$Y, d$Q, ld = TRUE, max_it = 150, tolVal = 1e-3)
  fs <- fit_stats(f, d$Y, d$Q)

  expect_s3_class(fs, "vbpm_fit_stats")
  expect_true(all(is.finite(fs[c("RMSEA", "SRMR", "CFI", "TLI")])))

  ## W and Psi are genuinely different matrices here, so a fit-statistic
  ## computed from the wrong one would be detectably different. Rebuild the
  ## implied covariance from W and check SRMR agrees with the reported value.
  expect_false(isTRUE(all.equal(f$W, f$Psi)))
  pi_ <- f$pi; pi_[d$Q == 1] <- 1; pi_[d$Q == 0] <- 0
  aH <- (d$Q == 1) | (d$Q == -1 & pi_ >= .5)
  LamH <- f$Lam * aH
  Sg <- LamH %*% f$Phi %*% t(LamH) + f$W
  S  <- f$sample_cov
  si <- 1 / sqrt(diag(S))
  R0 <- si * S  * rep(si, each = ncol(S))
  Rh <- si * Sg * rep(si, each = ncol(S))
  srmr <- sqrt(2 * sum((R0 - Rh)[lower.tri(R0, diag = TRUE)]^2) /
                 (ncol(S) * (ncol(S) + 1)))
  expect_equal(unname(fs["SRMR"]), srmr, tolerance = 1e-8)
})

test_that("LD parameter counts grow with the selected residual edges", {
  d <- ld_dat()
  f  <- vbfa(d$Y, d$Q, ld = TRUE, max_it = 150, tolVal = 1e-3)
  fs <- fit_stats(f, d$Y, d$Q)
  fd <- vbfa(d$Y, d$Q, v0 = .001, max_it = 200)
  fsd <- fit_stats(fd, d$Y, d$Q)
  ## an LD model estimates residual edges on top of the diagonal model, so its
  ## nominal count must be at least as large
  expect_gte(unname(fs["t_nom"]), unname(fsd["t_nom"]))
})

test_that("missing data and local dependence compose", {
  d <- ld_dat(N = 350)
  Y <- d$Y
  Y[cbind(1:10, rep(c(2, 8), each = 5))] <- NA   # inside a planted LD pair
  f <- vbfa(Y, d$Q, ld = TRUE, max_it = 150, tolVal = 1e-3)
  expect_equal(f$preprocess$n_missing, 10)
  expect_false(anyNA(f$Lam))
  expect_true(is.finite(f$objective))
  expect_identical(f$objective_type, "vecm")
})

## ---- missing responses in vbmimic ---------------------------------------

test_that("vbmimic handles missing responses in Y but rejects missing X", {
  B <- matrix(0, 3, 9); for (k in 1:3) B[k, ((k - 1) * 3 + 1):(k * 3)] <- .3
  s <- sim_lvm(N = 500, K = 3, J = 18, P = 9, b = B, phx = 0, rseed = 21)
  Y <- s$dat[, 1:18]; X <- s$dat[, 19:27]
  QA <- matrix(-1L, 18, 3)
  for (k in 1:3) { a <- which(rep(1:3, each = 6) == k)[1:2]
                   QA[a, ] <- 0L; QA[a, k] <- 1L }
  QB <- matrix(-1L, 3, 9)

  Ym <- Y
  Ym[cbind(1:20, rep(1:4, each = 5))] <- NA
  f <- vbmimic(Ym, X, QA, QB)
  expect_equal(f$flag, 1)
  expect_false(anyNA(f$A))
  expect_false(anyNA(f$B))
  expect_equal(f$preprocess$n_missing, 20)
  expect_identical(f$preprocess$missing_mask, is.na(Ym))

  ## deterministic with missing data too
  f2 <- vbmimic(Ym, X, QA, QB)
  expect_identical(f$A, f2$A)

  ## structural recovery survives a modest amount of missingness
  truth <- apply(B != 0, 1, function(r) paste(which(r), collapse = ","))
  est   <- apply(abs(f$B) > .15, 1, function(r) paste(which(r), collapse = ","))
  expect_setequal(est, truth)

  ## covariates must be complete
  Xm <- X; Xm[1, 1] <- NA
  expect_error(vbmimic(Y, Xm, QA, QB), "missing covariates")

  ## degenerate masks are rejected
  Ybad <- Y; Ybad[, 1] <- NA
  expect_error(vbmimic(Ybad, X, QA, QB), "at least one observed")
  Ybad <- Y; Ybad[1, ] <- NA
  expect_error(vbmimic(Ybad, X, QA, QB), "all responses missing")
})

test_that("complete data is unchanged by the vbmimic missing-data path", {
  B <- matrix(0, 3, 9); for (k in 1:3) B[k, ((k - 1) * 3 + 1):(k * 3)] <- .3
  s <- sim_lvm(N = 300, K = 3, J = 18, P = 9, b = B, phx = 0, rseed = 22)
  f <- vbmimic(s$dat[, 1:18], s$dat[, 19:27],
               matrix(-1L, 18, 3), matrix(-1L, 3, 9))
  expect_equal(f$preprocess$n_missing, 0)
  expect_null(f$preprocess$missing_mask)   # not stored for complete data
})

## ---- pefa plot method ----------------------------------------------------

test_that("plot.pefa draws its three compact displays without error", {
  d <- make_dat(N = 250)
  r <- pefa(d$Q[, 1:2, drop = FALSE], d$Y, Kmin = 2, Kmax = 3,
            verbose = FALSE, v0 = .001, max_it = 300)
  tmp <- tempfile(fileext = ".png")
  grDevices::png(tmp, width = 600, height = 400)
  on.exit({ grDevices::dev.off(); unlink(tmp) }, add = TRUE)

  ## objective: three criterion paths; gain: two; fit: the four descriptive
  ## indices in one two-panel display.
  for (crit in c("elbo", "aic", "bic")) {
    expect_silent(plot(r, type = "objective", criterion = crit))
  }
  for (crit in c("elbo", "bic")) {
    expect_silent(plot(r, type = "gain", criterion = crit))
    expect_silent(plot(r, type = "gain", criterion = crit, pct = TRUE))
  }
  expect_silent(plot(r, type = "fit"))

  ## No reference line of any kind is drawable: 0.9.0 dropped the soft
  ## criteria, the cut threshold and the selected-K marker together, so the
  ## method neither imports nor calls abline().
  expect_false("abline" %in% unlist(getNamespaceImports("vbpm"),
                                    use.names = FALSE))
  expect_false(any(grepl("abline", deparse(body(vbpm:::plot.pefa)))))

  ## Soft criteria and the stability view are gone; the gain view has no AIC.
  for (crit in c("aic_s", "bic_s", "sabic")) {
    expect_error(plot(r, type = "gain", criterion = crit), "unavailable")
  }
  expect_error(plot(r, type = "gain", criterion = "aic"), "unavailable")
  expect_error(plot(r, type = "stability"), "arg")
  expect_error(plot(r, pct = NA), "pct must be TRUE or FALSE")
})

## ---- item 5b: the terminal VECM objective ------------------------------

test_that("the VECM objective is monotone within a single stage", {
  ## Empirical half of verification item 5b. The analytical half -- a
  ## term-by-term check against Eq. (Qpsi) of the V4 manuscript -- is recorded
  ## by the 0.4.0 LD audit. A single stage is forced by
  ## a scalar v0 AND a single xi0, since n_stage is the max of the two lengths.
  ## Across a stage boundary a decrease would be legitimate (v0 changes there).
  d <- ld_dat(N = 350, seed = 5)
  ctrl <- list(xi0 = 1)
  its <- c(5, 15, 30, 60, 100)
  obj <- vapply(its, function(m) {
    f <- suppressWarnings(vbfa(d$Y, d$Q, ld = TRUE, v0 = 0.001,
                               ld_control = ctrl, max_it = m, tolVal = 1e-12))
    expect_equal(f$path$n_stage, 1L)
    f$objective
  }, numeric(1))
  expect_true(all(is.finite(obj)))
  ## non-decreasing up to floating point at convergence
  expect_true(all(diff(obj) > -1e-6))
})

test_that("objective_terms decompose the objective exactly", {
  d <- ld_dat(N = 300, seed = 6)
  f <- vbfa(d$Y, d$Q, ld = TRUE, max_it = 120, tolVal = 1e-3)
  expect_named(f$objective_terms,
               c("likelihood", "latent", "loadings", "edges", "diagonal", "tau"))
  expect_true(all(is.finite(f$objective_terms)))
  expect_equal(sum(f$objective_terms), f$objective)
  ## ELBO_conditional is the VE-step part: likelihood + latent + loadings,
  ## i.e. Eq. (3); the objective adds the Psi/tau prior terms on top.
  expect_equal(unname(sum(f$objective_terms[c("likelihood", "latent", "loadings")])),
               f$ELBO_conditional)
})

## ---- audit follow-up: input validation ----------------------------------
## (the checkpoint-provenance tests were removed with the checkpoint feature
## in 0.8.0, and the select_K_elbow() tests with the selector itself in 0.9.0;
## pefa() performs one in-memory sweep, never resumes, and chooses nothing.)

test_that("pefa validates its window", {
  d <- make_dat(N = 120)
  Q0 <- d$Q[, 1:2, drop = FALSE]
  ## a reversed window would silently become a descending sequence
  expect_error(pefa(Q0, d$Y, Kmin = 4, Kmax = 2, verbose = FALSE),
               "non-decreasing")
  expect_error(pefa(Q0, d$Y, Kmin = 2.5, Kmax = 4, verbose = FALSE),
               "whole number")
  expect_error(pefa(Q0, d$Y, Kmin = 2, Kmax = 3, tau = 1.5, verbose = FALSE),
               "tau")

  ## 0.9.0 removed `...` along with every selection control, so neither a mode
  ## control nor a count rule can reach the call at all: it fails to match
  ## rather than being caught and rejected. Every control that changes a
  ## returned number is now a named formal recorded in $settings, which is a
  ## stronger guarantee than the old screen.
  expect_false(any(c("...", "ld", "Qe", "orthogonal", "cuts", "sustain",
                     "stability_eps") %in% names(formals(pefa))))
  for (bad in list(list(ld = TRUE), list(orthogonal = TRUE),
                   list(Qe = diag(ncol(d$Y))), list(cuts = c(primary = 10)),
                   list(stability_eps = 0))) {
    expect_error(
      do.call(pefa, c(list(Q0 = Q0, Y = d$Y, Kmin = 2, Kmax = 3,
                           verbose = FALSE), bad)),
      "unused argument"
    )
  }
})

test_that("anchored items' residual variances are not inflated under LD", {
  ## Regression for the fixed-zero variance leak: sigsq.q.Lam[Q == 0] held its
  ## initialization value 1 and the LD branch's expected residual covariance
  ## summed loading variances over ALL columns, adding ~1 per zero column to
  ## anchored items' S diagonal (~5x variance inflation; loadings unaffected).
  ## Inherited from ld-vb_r.R, which shares the defect.
  d <- ld_dat(N = 400, seed = 2)
  anch <- which(rowSums(d$Q == 0L) > 0)          # anchored rows have Q == 0 cells
  f  <- vbfa(d$Y, d$Q, ld = TRUE, max_it = 300, tolVal = 1e-3)
  fd <- vbfa(d$Y, d$Q, v0 = .001, max_it = 500)
  ## LD residual variances at anchored items must be commensurate with the
  ## diagonal fit's, not multiples of it (pre-fix ratio was ~5)
  ratio <- diag(f$W)[anch] / (1 / fd$PsiInv[anch])
  expect_true(all(ratio < 2))
  ## and the reported posterior variance of a fixed-zero loading is 0
  expect_true(all(f$Lam_var[d$Q == 0L] == 0))
  expect_true(all(fd$Lam_var[d$Q == 0L] == 0))
})

## Partially exploratory factor analysis: a fixed-window candidate sweep.
##
## pefa() fits every K in a consecutive window and reports candidate and
## between-candidate measurements.  It chooses no factor count, threshold,
## horizon, collision policy, or stopping rule; those belong to the analysis
## that consumes the object.
##
## The failure contract is deliberately simple.  A candidate that cannot be
## fitted, or whose required quantities are malformed, aborts the whole call
## with an error naming the K: there is no status/reason taxonomy and no
## partially populated object.  A candidate that merely failed to converge is
## retained, because its returned quantities are still numbers.


# ---- canonical naming -------------------------------------------------

## Item names come from the data when usable, otherwise from a fixed pattern.
## A supplied rownames(Q0) must agree rather than silently winning.
.pefa_item_names <- function(Y, Q0, J) {
  nm <- colnames(Y)
  nm <- if (!is.null(nm) && length(nm) == J && !anyNA(nm) &&
            all(nzchar(nm)) && !anyDuplicated(nm)) {
    as.character(nm)
  } else {
    paste0("item_", seq_len(J))
  }
  qn <- rownames(Q0)
  if (!is.null(qn) && !identical(as.character(qn), nm)) {
    stop("rownames(Q0) must match the item names taken from colnames(Y).",
         call. = FALSE)
  }
  nm
}

.pefa_backbone_names <- function(Q0, K0) {
  nm <- colnames(Q0)
  if (!is.null(nm) && length(nm) == K0 && !anyNA(nm) &&
      all(nzchar(nm)) && !anyDuplicated(nm)) {
    return(as.character(nm))
  }
  if (K0 == 0L) character(0) else paste0("F", seq_len(K0))
}

## Group-block factor names for candidate K: backbone first, then appended
## exploratory columns continuing the F-numbering.  Bifactor prepends "G".
.pefa_dimnames <- function(item_names, backbone_names, K0, K, bifactor) {
  extra <- if (K > K0) paste0("F", seq.int(K0 + 1L, K)) else character(0)
  cols <- c(backbone_names, extra)
  if (isTRUE(bifactor)) cols <- c("G", cols)
  list(item_names, cols)
}


# ---- one candidate ----------------------------------------------------

.pefa_stop_K <- function(K, what) {
  stop(sprintf("PEFA candidate K = %d: %s.", K, what), call. = FALSE)
}

## "Required" means the loading and PIP matrices plus the criterion scalars a
## downstream gain or count consumes.  RMSEA/SRMR/CFI/TLI are descriptive and
## may legitimately be NA -- RMSEA and TLI are undefined at df <= 0 -- so they
## are normalized rather than treated as candidate failures.
.pefa_required_stats <- c("AIC", "BIC", "t")
.pefa_descriptive_stats <- c("RMSEA", "SRMR", "CFI", "TLI")

.pefa_matrix_or_stop <- function(x, dn, kind, K) {
  if (is.null(x)) .pefa_stop_K(K, paste("the", kind, "matrix is missing"))
  if (!is.matrix(x) || !is.numeric(x) || is.complex(x)) {
    .pefa_stop_K(K, paste("the", kind, "matrix is not a real numeric matrix"))
  }
  if (!identical(dim(x), c(length(dn[[1L]]), length(dn[[2L]])))) {
    .pefa_stop_K(K, sprintf("the %s matrix is %d by %d, expected %d by %d",
                            kind, nrow(x), ncol(x),
                            length(dn[[1L]]), length(dn[[2L]])))
  }
  value <- matrix(as.double(x), nrow(x), ncol(x), dimnames = dn)
  if (!all(is.finite(value))) {
    .pefa_stop_K(K, paste("the", kind, "matrix has nonfinite entries"))
  }
  if (identical(kind, "PIP") && (any(value < 0) || any(value > 1))) {
    .pefa_stop_K(K, "the PIP matrix has values outside [0, 1]")
  }
  value
}

.pefa_candidate <- function(fit, fs, K, dn) {
  lam <- .pefa_matrix_or_stop(fit[["Lam"]], dn, "loading", K)
  pip <- .pefa_matrix_or_stop(fit[["pi"]], dn, "PIP", K)

  elbo <- fit[["ELBO"]]
  if (is.null(elbo) || length(elbo) != 1L || !is.numeric(elbo) ||
      is.complex(elbo) || !is.finite(elbo)) {
    .pefa_stop_K(K, "ELBO is not one finite real number")
  }

  need <- c(.pefa_required_stats, .pefa_descriptive_stats)
  if (!is.numeric(fs) || is.complex(fs) || is.null(names(fs)) ||
      !all(need %in% names(fs))) {
    .pefa_stop_K(K, "fit_stats() did not return the required named scalars")
  }
  value <- vapply(need, function(f) as.double(unname(fs[[f]])), numeric(1))
  if (!all(is.finite(value[.pefa_required_stats]))) {
    .pefa_stop_K(K, "AIC, BIC, or t is nonfinite")
  }
  ## t counts parameters, so a negative or fractional value is a malformed
  ## candidate rather than an unusual one.
  if (value[["t"]] < 0 || value[["t"]] != round(value[["t"]])) {
    .pefa_stop_K(K, "t is not a nonnegative whole number")
  }
  ## A descriptive index may be NA when its own definition does not apply --
  ## RMSEA and TLI are undefined at df <= 0.  That legitimate NA is preserved.
  ## NaN or an infinity is a malformed engine result rather than an undefined
  ## quantity, so it is named and refused instead of being laundered into an
  ## NA that would read as "undefined".
  descriptive <- value[.pefa_descriptive_stats]
  ## is.na() is TRUE for NaN, so the two must be separated explicitly:
  ## a literal NA_real_ is kept, a NaN is refused.
  bad <- is.nan(descriptive) |
    (!is.na(descriptive) & !is.finite(descriptive))
  if (any(bad)) {
    .pefa_stop_K(K, paste0(paste(names(descriptive)[bad], collapse = ", "),
                           " is NaN or infinite"))
  }

  flag <- fit[["flag"]]
  converged <- length(flag) == 1L && is.numeric(flag) && !is.na(flag) &&
    is.finite(flag) && flag == 1
  ## An absent iteration count is NA_integer_; a reported one that is not a
  ## representable nonnegative whole number is malformed, not merely unusual,
  ## so it is refused rather than coerced into -1L or 0L.
  iter <- fit[["iter"]]
  if (is.null(iter)) {
    iter <- NA_integer_
  } else if (length(iter) == 1L && is.numeric(iter) && !is.complex(iter) &&
             is.na(iter) && !is.nan(iter)) {
    iter <- NA_integer_
  } else if (length(iter) == 1L && is.numeric(iter) && !is.complex(iter) &&
             is.finite(iter) && iter >= 0 && iter == round(iter) &&
             iter <= .Machine$integer.max) {
    iter <- as.integer(iter)
  } else {
    .pefa_stop_K(K, "iter is not a representable nonnegative whole number")
  }

  row <- data.frame(
    K = as.integer(K),
    ELBO = as.double(elbo),
    AIC = value[["AIC"]], BIC = value[["BIC"]],
    RMSEA = descriptive[["RMSEA"]], SRMR = descriptive[["SRMR"]],
    CFI = descriptive[["CFI"]], TLI = descriptive[["TLI"]],
    t = value[["t"]],
    iter = iter,
    converged = converged,
    stringsAsFactors = FALSE
  )
  list(row = row, loading = lam, pip = pip)
}


# ---- derived views ----------------------------------------------------

## Percent of the largest positive gain on that path.  Dividing before scaling
## keeps gains near the largest finite double representable.
.pefa_gain_pct <- function(gain) {
  if (!length(gain)) return(numeric(0))
  gmax <- max(gain)
  if (!(gmax > 0)) return(rep(NA_real_, length(gain)))
  100 * (gain / gmax)
}

.pefa_fact_names <- c("phi_min", "rmsd", "rmsd_max", "ari", "pip_rmsd",
                      "unmatched_ssl", "collision")

.pefa_transitions <- function(sweep, pair_facts) {
  n <- nrow(sweep) - 1L
  if (n <= 0L) {
    out <- data.frame(K_from = integer(0), K_to = integer(0),
                      ELBO_gain_pct = numeric(0), BIC_gain_pct = numeric(0),
                      stringsAsFactors = FALSE)
    for (f in .pefa_fact_names) {
      out[[f]] <- if (f == "collision") logical(0) else numeric(0)
    }
    return(out)
  }
  from <- seq_len(n)
  to <- from + 1L

  ## Positive favors K_to on both paths.
  g_elbo <- as.double(sweep$ELBO[to] - sweep$ELBO[from])
  g_bic <- as.double(sweep$BIC[from] - sweep$BIC[to])
  if (!all(is.finite(g_elbo)) || !all(is.finite(g_bic))) {
    stop("PEFA adjacent criterion differences must be finite; a ",
         "nonrepresentable difference is a construction error rather than a ",
         "partially populated path.", call. = FALSE)
  }

  out <- data.frame(
    K_from = as.integer(sweep$K[from]), K_to = as.integer(sweep$K[to]),
    ELBO_gain_pct = .pefa_gain_pct(g_elbo),
    BIC_gain_pct = .pefa_gain_pct(g_bic),
    stringsAsFactors = FALSE
  )
  ## The adjacent rows reuse the same in-memory pair results the all-pairs
  ## view projects; matching never runs a second way for the two outputs.
  key <- paste(out$K_from, out$K_to, sep = "-")
  for (f in .pefa_fact_names) {
    out[[f]] <- vapply(key, function(k) pair_facts[[k]][[f]],
                       if (f == "collision") logical(1) else numeric(1),
                       USE.NAMES = FALSE)
  }
  out
}

## $rmsd stores rmsd_max, the weakest-column value, while $transitions$rmsd is
## the pooled one.  The two names differ on purpose and must not be crossed.
.pefa_persistence <- function(Ks, pair_facts) {
  W <- length(Ks)
  nm <- as.character(Ks)
  blank <- function(fill) matrix(fill, W, W, dimnames = list(nm, nm))
  out <- list(phi = blank(NA_real_), rmsd = blank(NA_real_),
              collision = blank(NA))
  src <- c(phi = "phi_min", rmsd = "rmsd_max", collision = "collision")
  for (i in seq_len(W)) {
    for (j in seq_len(W)) {
      if (i >= j) next
      f <- pair_facts[[paste(Ks[i], Ks[j], sep = "-")]]
      for (m in names(src)) out[[m]][i, j] <- f[[src[[m]]]]
    }
  }
  out
}


# ---- argument validation ----------------------------------------------

.pefa_whole <- function(value, nm, min_value) {
  if (length(value) != 1L || !is.numeric(value) || is.complex(value) ||
      is.na(value) || !is.finite(value) || value < min_value ||
      value != round(value) || value > .Machine$integer.max) {
    stop(nm, " must be a single whole number in [", min_value, ", ",
         .Machine$integer.max, "].", call. = FALSE)
  }
  as.integer(value)
}

.pefa_flag <- function(value, nm) {
  if (length(value) != 1L || !is.logical(value) || is.na(value)) {
    stop(nm, " must be TRUE or FALSE.", call. = FALSE)
  }
  as.logical(value)
}

.pefa_bounded <- function(value, nm, lo, hi) {
  if (length(value) != 1L || !is.numeric(value) || is.complex(value) ||
      is.na(value) || !is.finite(value) || value < lo || value > hi) {
    stop(nm, " must be a single finite number in [", lo, ", ", hi, "].",
         call. = FALSE)
  }
  as.double(value)
}


# ---- constructor ------------------------------------------------------

#' Partially exploratory factor analysis over a factor-count window
#'
#' Fits [vbfa()] for every integer `K` in a fixed consecutive window, holding
#' the partially specified backbone `Q0` fixed and appending regularized
#' exploratory columns. It reports candidate summaries and between-candidate
#' loading measurements.
#'
#' `pefa()` **chooses nothing**: no factor count, threshold, persistence
#' horizon, collision policy, or stopping rule. Those belong to the analysis
#' that consumes the object.
#'
#' Ordinary candidates are oblique with diagonal residuals. In bifactor mode
#' `K` counts group factors and [vbfa()] adds one general column, so a stored
#' matrix has `K + 1` columns; every between-candidate comparison removes that
#' column first and works on the `K` group columns.
#'
#' One private comparator produces every between-candidate number, and both
#' `transitions` and `persistence` are read from its stored results. The
#' accessors do no matching of their own: [ssl()], `print()`, `summary()`, and
#' `plot()` only describe what is already stored, and none of them rematches
#' columns or applies an analysis rule. There is no `persistence()` accessor:
#' `x$persistence$phi`, `$rmsd`, and `$collision` are already the three
#' stored matrices.
#'
#' @section The returned object:
#' A list of class `"pefa"` with exactly these components, in order.
#' \describe{
#'   \item{`sweep`}{One row per candidate, in column order. `K` is integer;
#'     `ELBO`, `AIC`, and `BIC` are required finite doubles; `RMSEA`, `SRMR`,
#'     `CFI`, and `TLI` are descriptive doubles; `t` is a double; `iter` is a
#'     nonnegative integer, or `NA_integer_` when a well-formed engine result
#'     reports none; `converged` is a nonmissing logical. `AIC` and `BIC` are
#'     hard-selection values at the resolved `tau`, and `t` is their effective
#'     parameter count -- a finite, nonnegative, whole-valued count stored as
#'     a double -- rank-adjusted when `rank_adjust = TRUE`. Each descriptive
#'     index is `NA_real_` exactly when its own definition does not apply, and
#'     is never silently substituted.}
#'   \item{`transitions`}{One row per adjacent pair: the integers `K_from` and
#'     `K_to`, the doubles `ELBO_gain_pct` and `BIC_gain_pct`, then the seven
#'     structural facts -- the doubles `phi_min`, `rmsd`, `rmsd_max`, `ari`,
#'     `pip_rmsd`, `unmatched_ssl` and the logical `collision`. Positive gains
#'     favor `K_to`. Each percentage divides by the largest positive gain on
#'     its own path, or is `NA_real_` when no positive gain exists. Raw
#'     criteria stay in `sweep`.}
#'   \item{`persistence`}{Three upper-triangular `W` by `W` matrices over the
#'     fitted `K` values, named and ordered `phi`, `rmsd`, `collision`: `phi`
#'     holds `phi_min`, `rmsd` holds **`rmsd_max`**, and `collision` is the
#'     logical reuse mask. Diagonal and lower triangle are `NA`. Every cell is
#'     a direct endpoint comparison, never a chain through intervening
#'     candidates.}
#'   \item{`loadings`, `pips`}{Named `K`-indexed lists of candidate matrices
#'     with items in rows and factors in columns: `J` by `K`, or `J` by
#'     `K + 1` with the general column first in bifactor mode. Loadings are
#'     finite and PIPs lie in `[0, 1]`.}
#'   \item{`Q0`, `settings`}{The canonical `J` by `K0` integer backbone, and
#'     the resolved controls `bifactor`, `general`, `v0`, `max_it`, `convChk`,
#'     `tolVal`, `tau`, `rank_adjust`, `rank_max_J`, in that order. `general`
#'     is `NULL` outside bifactor mode. `settings` records fitting and
#'     fit-statistic controls only: no `verbose`, and no analysis rule.}
#' }
#'
#' The loading and PIP matrices and the criteria `ELBO`, `AIC`, `BIC`, and `t`
#' are required. A fitting failure or a malformed required quantity aborts the
#' call with an error naming the `K`; there is no status taxonomy and no
#' partially populated object. A candidate that merely failed to converge is
#' retained with `converged = FALSE`, and one aggregate warning names every
#' affected `K`.
#'
#' @section Loading correspondence:
#' For source column `a` and target column `b`, let `s = 1` when `a'b >= 0` and
#' `-1` otherwise, and let `d(a, b) = ||a - s b||^2`. Backbone columns pair by
#' position. Each exploratory source independently selects the exploratory
#' target minimizing `d`, which is Equation 17 of Chen (2023) made invariant to
#' whole-column reflection. Ties break by larger finite absolute Tucker
#' congruence, then by smaller target index.
#'
#' The assignment is independent, not one-to-one, so two sources may select the
#' same target. That reuse is a **collision**: it is recorded rather than
#' repaired, the pair keeps every otherwise defined measurement, and the target
#' left unselected still counts toward `unmatched_ssl`. There is no salience
#' screen: every column participates, and a comparison always contains exactly
#' `K_from` scored pairs.
#'
#' A zero-norm column is no exception. Its aligned distance is defined, so it
#' takes part in the assignment and in `rmsd` and `rmsd_max`, but its Tucker
#' congruence is `0/0` and therefore undefined. An undefined congruence is
#' reported as `NA_real_` and makes that comparison's `phi_min` `NA_real_`;
#' it is never dropped by an `na.rm`.
#'
#' Because `phi_min` is a minimum over *every* column, an overextracted
#' candidate carrying a near-empty column will report a low `phi_min` and often
#' `collision = TRUE`. Read that with [ssl()]: it usually means the candidate
#' contains a column with nothing in it, not that the retained structure moved.
#'
#' @section Identification:
#' `pefa()` imposes no upper bound on `K` relative to `J`. The Ledermann bound
#' `K <= (2J + 1 - sqrt(8J + 1)) / 2` orients the choice of `Kmax`, but it is
#' derived for an unrestricted factor model and does not constrain a
#' regularized partially confirmatory candidate, so applying it as a package
#' check would be wrong.
#'
#' Beyond that bound expect candidates that do not converge and fit indices
#' that become `NA`: `RMSEA` and `TLI` are undefined at non-positive degrees of
#' freedom. Both are reported evidence about the window rather than failures.
#' `rank_adjust = TRUE` is the sharper diagnostic, since it counts the rank of
#' the Jacobian of the unique covariance elements and therefore tests local
#' identification directly; that count cannot exceed `J(J + 1)/2`, so its
#' degrees of freedom saturate at zero instead of going negative.
#'
#' @section Removed in 0.9.0:
#' `select_K_elbow()`, the `pefa()` arguments `cuts`, `sustain`, and
#' `stability_eps`, and the `$selected_K` and `$boundary` components existed
#' in 0.8.3 and are gone. That is an intentional pre-1.0 breaking change rather
#' than a deprecation cycle: the package owns no count rule, so it can own
#' neither a selector nor a stability cutoff. The replacement is ordinary
#' analysis code over the stored path, as in the last chunk of the Examples,
#' and such code has to say which variant it is, because look-ahead and
#' boundary-fallback variants disagree on the same path. `NEWS.md` lists the
#' removals in full.
#'
#' @param Q0 A `J` by `K0` backbone with entries `1`, `0`, or `-1`. Zero
#'   columns (`K0 = 0`, fully exploratory) are allowed.
#' @param Y A numeric `N` by `J` matrix of observed values, optionally with
#'   literal `NA`; `NaN` and infinite values are rejected. Every row and every
#'   item needs an observed value, and every item a positive finite SD.
#' @param Kmin,Kmax Inclusive window, whole numbers with
#'   `1 <= Kmin <= Kmax <= .Machine$integer.max` and `Kmin >= ncol(Q0)`.
#' @param bifactor Fit an orthogonal bifactor parameterization.
#' @param general Scalar or length-`J` general design in `{-1, 0, 1}` with at
#'   least one `1`, used only when `bifactor = TRUE`.
#' @param v0,max_it,convChk,tolVal Optimizer controls passed to [vbfa()]:
#'   finite positive `v0`, positive whole `max_it`, `TRUE`/`FALSE` `convChk`,
#'   and finite `tolVal >= 0`.
#' @param tau Loading-inclusion threshold in `[0, 1]`, used by [fit_stats()].
#' @param rank_adjust,rank_max_J Rank-adjustment controls for [fit_stats()].
#'   `rank_adjust = TRUE` requires `J <= rank_max_J` and the suggested
#'   `numDeriv` package.
#' @param verbose Print one progress line per candidate.
#'
#' @return An object of class `"pefa"`; see *The returned object* below.
#'
#' @references
#' Chen, J., & Jin, Y. (2026). Recovering latent structures after variational
#' Bayesian variable selection: Fit assessment and factor-number selection in
#' partially exploratory factor analysis. *arXiv preprint* arXiv:2607.07159.
#'
#' Chen, J. (2023). Fully and partially exploratory factor analysis with
#' bi-level Bayesian regularization. *Behavior Research Methods*, 55(4),
#' 2125--2142. \doi{10.3758/s13428-022-01884-7}
#'
#' @examples
#' \donttest{
#' sim <- sim_fa(N = 300, K = 3, ipf = 4, lam = .7, lac = .3, rseed = 1)
#' Q0 <- matrix(-1L, ncol(sim$dat), 2)
#' groups <- rep(1:3, each = 4)
#' for (k in 1:2) Q0[which(groups == k)[1:2], k] <- 1L
#'
#' fit <- pefa(Q0, sim$dat, 2, 4, verbose = FALSE)
#' fit$sweep
#' fit$transitions[, c("K_from", "K_to", "ELBO_gain_pct", "phi_min")]
#' fit$persistence$phi
#' ssl(fit)
#'
#' ## A count rule is the analysis's, never the package's. This one stops at
#' ## the first ELBO gain under 20% and looks no further ahead; a
#' ## full-look-ahead variant can answer differently on the same path.
#' g <- fit$transitions$ELBO_gain_pct
#' fit$transitions$K_from[which(g < 20)[1]]
#' }
#'
#' @seealso [vbfa()], [fit_stats()], [ssl()]
#' @export
pefa <- function(Q0, Y, Kmin, Kmax,
                 bifactor = FALSE, general = 1,
                 v0 = c(.01, .005, .002, .001), max_it = 10000,
                 convChk = FALSE, tolVal = 1e-4, tau = .50,
                 rank_adjust = FALSE, rank_max_J = 100, verbose = TRUE) {
  ## Check Q0's type on the ORIGINAL object.  as.matrix() would turn a data
  ## frame or factor into something acceptable, and the %in% test below
  ## coerces, so character "1"/"0"/"-1" and logicals would otherwise pass.
  if (!is.matrix(Q0) || !is.numeric(Q0) || is.complex(Q0)) {
    stop("Q0 must be a real numeric or integer matrix.", call. = FALSE)
  }
  Y <- as.matrix(Y)
  J <- ncol(Y)
  K0 <- ncol(Q0)
  if (!is.numeric(Y) || is.complex(Y) || nrow(Y) < 1L || J < 1L) {
    stop("Y must be a nonempty numeric N by J matrix.", call. = FALSE)
  }
  if (any(is.nan(Y)) || any(is.infinite(Y))) {
    stop("Y may contain literal NA but not NaN or infinite values.",
         call. = FALSE)
  }
  if (nrow(Q0) != J) stop("nrow(Q0) must equal ncol(Y).", call. = FALSE)
  if (K0 > 0L && (anyNA(Q0) || !all(Q0 %in% c(-1, 0, 1)))) {
    stop("Q0 entries must be -1, 0, or 1 with no missing values.",
         call. = FALSE)
  }

  Kmin <- .pefa_whole(Kmin, "Kmin", 1)
  Kmax <- .pefa_whole(Kmax, "Kmax", 1)
  max_it <- .pefa_whole(max_it, "max_it", 1)
  rank_max_J <- .pefa_whole(rank_max_J, "rank_max_J", 1)
  if (Kmax < Kmin) {
    stop(sprintf("Kmax (%d) < Kmin (%d): the window must be non-decreasing.",
                 Kmax, Kmin), call. = FALSE)
  }
  if (Kmin < K0) {
    stop(sprintf("Kmin (%d) < K0 (%d): the window must reach the backbone.",
                 Kmin, K0), call. = FALSE)
  }
  tau <- .pefa_bounded(tau, "tau", 0, 1)
  tolVal <- .pefa_bounded(tolVal, "tolVal", 0, Inf)
  bifactor <- .pefa_flag(bifactor, "bifactor")
  convChk <- .pefa_flag(convChk, "convChk")
  rank_adjust <- .pefa_flag(rank_adjust, "rank_adjust")
  verbose <- .pefa_flag(verbose, "verbose")
  if (!is.numeric(v0) || is.complex(v0) || !length(v0) || anyNA(v0) ||
      any(!is.finite(v0)) || any(v0 <= 0)) {
    stop("v0 must be a nonempty vector of finite positive numbers.",
         call. = FALSE)
  }
  v0 <- as.double(unname(v0))
  .validate_rank_controls(rank_adjust, rank_max_J)
  if (isTRUE(rank_adjust)) {
    if (J > rank_max_J) {
      stop("rank_adjust = TRUE requires J <= rank_max_J (", J, " > ",
           rank_max_J, ").", call. = FALSE)
    }
    if (!requireNamespace("numDeriv", quietly = TRUE)) {
      stop("rank_adjust = TRUE requires the optional numDeriv package.",
           call. = FALSE)
    }
  }

  general_design <- NULL
  if (isTRUE(bifactor)) {
    if (!is.numeric(general) || is.complex(general) || anyNA(general) ||
        !length(general) || !all(general %in% c(-1, 0, 1)) ||
        !(length(general) %in% c(1L, J))) {
      stop("general must be a scalar or length-J vector of -1, 0, or 1.",
           call. = FALSE)
    }
    general_design <- rep_len(as.integer(general), J)
    if (!any(general_design == 1L)) {
      stop("general must specify at least one general loading (1).",
           call. = FALSE)
    }
    if (K0 > 0L && any(apply(Q0 == 1, 2L, all))) {
      stop("With bifactor = TRUE, Q0 supplies only the group backbone.",
           call. = FALSE)
    }
  }

  item_names <- .pefa_item_names(Y, Q0, J)
  backbone_names <- .pefa_backbone_names(Q0, K0)
  Q0 <- matrix(as.integer(Q0), J, K0,
               dimnames = list(item_names, backbone_names))

  Ks <- seq.int(Kmin, Kmax)
  settings <- list(bifactor = bifactor, general = general_design, v0 = v0,
                   max_it = max_it, convChk = convChk, tolVal = tolVal,
                   tau = tau, rank_adjust = rank_adjust,
                   rank_max_J = rank_max_J)

  rows <- vector("list", length(Ks))
  loadings <- vector("list", length(Ks))
  pips <- vector("list", length(Ks))

  for (i in seq_along(Ks)) {
    K <- Ks[i]
    dn <- .pefa_dimnames(item_names, backbone_names, K0, K, bifactor)
    Q <- if (K > K0) cbind(Q0, matrix(-1L, J, K - K0)) else Q0
    args <- list(Y = Y, Q = Q, v0 = v0, max_it = max_it,
                 convChk = convChk, tolVal = tolVal)
    if (isTRUE(bifactor)) {
      args <- c(args, list(bifactor = TRUE, general = general_design))
    }

    ## The per-candidate iteration-limit warning is muffled: nonconvergence is
    ## a recorded fact reported once at the end, not a repeated console event.
    fit <- tryCatch(
      withCallingHandlers(
        do.call(vbfa, args),
        warning = function(w) {
          if (startsWith(conditionMessage(w),
                         "Maximum number of MFVB iterations exceeded")) {
            invokeRestart("muffleWarning")
          }
        }
      ),
      error = function(e) {
        .pefa_stop_K(K, paste("vbfa() failed:", conditionMessage(e)))
      }
    )
    fs <- tryCatch(
      fit_stats(fit, Y = Y, Q = fit$Q, tau = tau,
                rank_adjust = rank_adjust, rank_max_J = rank_max_J),
      error = function(e) {
        .pefa_stop_K(K, paste("fit_stats() failed:", conditionMessage(e)))
      }
    )

    candidate <- .pefa_candidate(fit, fs, K, dn)
    rows[[i]] <- candidate$row
    loadings[[i]] <- candidate$loading
    pips[[i]] <- candidate$pip
    if (isTRUE(verbose)) {
      cat(sprintf("K=%2d  ELBO=%-12s BIC=%-10s CFI=%-7s conv=%s\n", K,
                  format(candidate$row$ELBO, digits = 7),
                  format(candidate$row$BIC, digits = 7),
                  format(candidate$row$CFI, digits = 3),
                  candidate$row$converged))
    }
  }

  sweep <- do.call(rbind, rows)
  rownames(sweep) <- NULL
  names(loadings) <- as.character(Ks)
  names(pips) <- as.character(Ks)

  ## One comparison per ordered pair, computed directly from its endpoints.
  ## Both derived views read these same results.
  offset <- as.integer(isTRUE(bifactor))
  group <- function(m) if (offset == 0L) m else m[, -1L, drop = FALSE]
  reg <- function(K) {
    Q <- if (K > K0) cbind(Q0, matrix(-1L, J, K - K0)) else Q0
    matrix(Q == -1L, J, K)
  }
  pair_facts <- list()
  for (i in seq_along(Ks)) {
    for (j in seq_along(Ks)) {
      if (i >= j) next
      a <- as.character(Ks[i])
      b <- as.character(Ks[j])
      pair_facts[[paste(Ks[i], Ks[j], sep = "-")]] <- .pefa_compare_pair(
        lam_from = group(loadings[[a]]), lam_to = group(loadings[[b]]),
        pip_from = group(pips[[a]]), pip_to = group(pips[[b]]),
        reg_from = reg(Ks[i]), reg_to = reg(Ks[j]), K0 = K0
      )
    }
  }

  out <- list(sweep = sweep,
              transitions = .pefa_transitions(sweep, pair_facts),
              persistence = .pefa_persistence(Ks, pair_facts),
              loadings = loadings, pips = pips, Q0 = Q0,
              settings = settings)
  class(out) <- "pefa"

  bad <- sweep$K[!sweep$converged]
  if (length(bad)) {
    warning("PEFA candidates did not converge at K = ",
            paste(bad, collapse = ", "),
            "; their returned quantities are retained.", call. = FALSE)
  }
  out
}

## The 0.9.0 measurement kernel: sign-invariant Equation-17 correspondence and
## the seven structural facts (technical note sections 4-5, plan section 9.1).
##
## Every fixture calls the private comparator directly, so no model is fitted
## and every expected number below is hand-computable.  The comparator receives
## group-block matrices only: pefa() strips a bifactor general column before it
## gets here, so nothing in this file needs to know about bifactor mode.

## One ordered pair of candidate matrices.  PIPs and regularization masks
## default to values that keep the loading-based facts the subject of the test.
.cmp <- function(lam_from, lam_to, K0 = 0,
                 pip_from = NULL, pip_to = NULL,
                 reg_from = NULL, reg_to = NULL) {
  J <- nrow(lam_from)
  if (is.null(pip_from)) pip_from <- matrix(0.5, J, ncol(lam_from))
  if (is.null(pip_to)) pip_to <- matrix(0.5, J, ncol(lam_to))
  if (is.null(reg_from)) reg_from <- matrix(TRUE, J, ncol(lam_from))
  if (is.null(reg_to)) reg_to <- matrix(TRUE, J, ncol(lam_to))
  vbpm:::.pefa_compare_pair(lam_from, lam_to, pip_from, pip_to,
                            reg_from, reg_to, K0)
}

.facts <- c("phi_min", "rmsd", "rmsd_max", "ari", "pip_rmsd",
            "unmatched_ssl", "collision")

## Independent unsigned Equation 17: argmin_j sum((a_i - b_j)^2), with no sign
## alignment and no tie machinery.  On a fixture whose selected inner products
## are all nonnegative this must agree with the shipped rule (note invariant 1).
.eq17_unsigned <- function(lam_from, lam_to, K0) {
  J <- nrow(lam_from)
  k <- ncol(lam_from)
  l <- ncol(lam_to)
  src <- if (k > K0) seq.int(K0 + 1L, k) else integer(0)
  tgt <- if (l > K0) seq.int(K0 + 1L, l) else integer(0)
  assigned <- vapply(src, function(i) {
    d <- vapply(tgt, function(j) sum((lam_from[, i] - lam_to[, j])^2),
                numeric(1))
    tgt[which.min(d)]
  }, numeric(1))
  ps <- c(seq_len(K0), src)
  pt <- c(seq_len(K0), assigned)
  d <- lam_from[, ps, drop = FALSE] - lam_to[, pt, drop = FALSE]
  phi <- vapply(seq_len(k), function(p) {
    a <- lam_from[, ps[p]]
    b <- lam_to[, pt[p]]
    sum(a * b) / sqrt(sum(a^2) * sum(b^2))
  }, numeric(1))
  list(assigned = assigned,
       rmsd = sqrt(sum(d^2) / (J * k)),
       rmsd_max = max(sqrt(colSums(d^2) / J)),
       phi_min = min(phi),
       unmatched_ssl = max(colSums(lam_to[, setdiff(tgt, assigned),
                                          drop = FALSE]^2)))
}


# ---- the defining behavior --------------------------------------------

test_that("minimum aligned SSE decides, not maximum congruence", {
  ## The release discriminator (note invariant 3).  `collinear` is a perfect
  ## rescaling of the source, so its congruence is exactly 1 and no target can
  ## beat it; `near_miss` is nearly the source itself, so its aligned squared
  ## distance is 100x smaller.  The two rules choose different targets.
  source_col <- c(1, 0)
  collinear <- c(2, 0)     # congruence 1,    aligned SSE 1
  near_miss <- c(1, 0.1)   # congruence < 1,  aligned SSE 0.01

  sse <- function(b) sum((source_col - b)^2)
  phi <- function(b) {
    abs(sum(source_col * b)) / sqrt(sum(source_col^2) * sum(b^2))
  }
  expect_gt(sse(collinear), sse(near_miss))     # SSE prefers near_miss
  expect_gt(phi(collinear), phi(near_miss))     # congruence prefers collinear

  f <- .cmp(cbind(source_col), cbind(collinear, near_miss))

  ## The SSE target wins.  Three independent readings pin the choice:
  ## the reported congruence is near_miss's, the pooled RMSD is near_miss's,
  ## and the leftover SSL is collinear's 2^2 rather than near_miss's 1.01.
  expect_equal(f$phi_min, 1 / sqrt(1.01))
  expect_equal(f$rmsd, sqrt(0.01 / 2))
  expect_equal(f$rmsd_max, sqrt(0.01 / 2))
  expect_equal(f$unmatched_ssl, 4)
  expect_false(f$collision)

  ## The comparator returns exactly the seven documented facts, in order.
  expect_named(f, .facts)
  expect_type(f$collision, "logical")
  expect_true(all(vapply(f[setdiff(.facts, "collision")], is.double,
                         logical(1))))
})

test_that("the rule reproduces unsigned Equation 17 when signs do not bind", {
  lam_from <- cbind(c(.9, .8, .1, 0, .1, 0),
                    c(.1, 0, .8, .7, 0, .1),
                    c(0, .1, .1, 0, .9, .8))
  lam_to <- cbind(c(.85, .75, .1, 0, .1, 0),
                  c(.05, 0, .85, .75, 0, .1),
                  c(0, .1, .1, 0, .85, .8),
                  c(.3, -.2, .1, .4, -.1, .2))
  ref <- .eq17_unsigned(lam_from, lam_to, K0 = 1)
  ## Every scored inner product is positive here, so s = +1 throughout and the
  ## shipped rule must equal the plain unsigned minimizer computed above.
  expect_identical(ref$assigned, c(2, 3))

  f <- .cmp(lam_from, lam_to, K0 = 1)
  expect_equal(f$rmsd, ref$rmsd)
  expect_equal(f$rmsd_max, ref$rmsd_max)
  expect_equal(f$phi_min, ref$phi_min)
  expect_equal(f$unmatched_ssl, ref$unmatched_ssl)
})


# ---- reflection invariance --------------------------------------------

test_that("whole-column reflection changes no fact", {
  lam_from <- cbind(c(.9, .8, .1, 0, .1, 0),
                    c(.1, 0, .8, .7, 0, .1),
                    c(0, .1, .1, 0, .9, .8))
  lam_to <- cbind(c(.85, .75, .1, 0, .1, 0),
                  c(.05, 0, .85, .75, 0, .1),
                  c(0, .1, .1, 0, .85, .8),
                  c(.3, -.2, .1, .4, -.1, .2))
  pip_from <- matrix(seq(0.05, 0.95, length.out = 18), 6, 3)
  pip_to <- matrix(seq(0.02, 0.98, length.out = 24), 6, 4)

  base <- .cmp(lam_from, lam_to, K0 = 1, pip_from = pip_from,
               pip_to = pip_to)
  expect_true(is.finite(base$phi_min))
  expect_gt(base$rmsd, 0)

  ## Every scored inner product is nonzero on this fixture, so the note
  ## requires bitwise -- not merely approximate -- invariance.  Source columns
  ## (backbone and exploratory), target columns (matched and unmatched), and
  ## then all four at once.
  for (cc in seq_len(ncol(lam_from))) {
    a <- lam_from
    a[, cc] <- -a[, cc]
    expect_identical(.cmp(a, lam_to, K0 = 1, pip_from = pip_from,
                          pip_to = pip_to), base)
  }
  for (cc in seq_len(ncol(lam_to))) {
    b <- lam_to
    b[, cc] <- -b[, cc]
    expect_identical(.cmp(lam_from, b, K0 = 1, pip_from = pip_from,
                          pip_to = pip_to), base)
  }
  a <- lam_from
  b <- lam_to
  a[, c(1L, 3L)] <- -a[, c(1L, 3L)]
  b[, c(2L, 4L)] <- -b[, c(2L, 4L)]
  expect_identical(.cmp(a, b, K0 = 1, pip_from = pip_from, pip_to = pip_to),
                   base)
})

test_that("reflection at a zero inner product holds to tolerance", {
  ## s = +1 at a'b = 0 is an arbitrary convention, so the note promises
  ## rounding-level rather than bitwise agreement here.  Both backbone pairs
  ## below are exactly orthogonal.
  lam_from <- cbind(c(1, 0, 0, 0), c(0, 1, 0, 0))
  lam_to <- cbind(c(0, 1, 0, 0), c(1, 0, 0, 0), c(0, 0, 1, 0))
  base <- .cmp(lam_from, lam_to, K0 = 2)
  expect_equal(base$phi_min, 0)

  for (cc in seq_len(ncol(lam_from))) {
    a <- lam_from
    a[, cc] <- -a[, cc]
    expect_equal(.cmp(a, lam_to, K0 = 2), base)
  }
  for (cc in seq_len(ncol(lam_to))) {
    b <- lam_to
    b[, cc] <- -b[, cc]
    expect_equal(.cmp(lam_from, b, K0 = 2), base)
  }
})


# ---- the tie rule -----------------------------------------------------

test_that("an exact distance tie breaks by congruence, then by index", {
  ## With ||a|| = 1 the aligned distance is 1 + ||b||^2 - 2 phi ||b||, so
  ## targets of different norm and different congruence can tie exactly.  All
  ## four targets below sit at aligned SSE exactly 1.
  a <- c(1, 0, 0, 0)
  zero <- c(0, 0, 0, 0)          # phi undefined
  half <- c(.5, .5, .5, .5)      # phi 0.5
  oneA <- c(1, 1, 0, 0)          # phi 1/sqrt(2)
  oneB <- c(1, 0, 1, 0)          # phi 1/sqrt(2), same norm, different pattern
  targets <- cbind(zero, half, oneA, oneB)
  expect_identical(apply(targets, 2L, function(b) sum((a - b)^2)),
                   c(zero = 1, half = 1, oneA = 1, oneB = 1))

  ## PIP column 4 differs from the source's, so pip_rmsd reads out whether the
  ## third or the fourth target was taken.
  pip_to <- cbind(rep(.5, 4), rep(.5, 4), rep(.5, 4), rep(.9, 4))
  f <- .cmp(cbind(a), targets, pip_to = pip_to)

  ## Largest finite congruence wins, so the two 1/sqrt(2) targets beat both
  ## the undefined zero column and the 0.5 column ...
  expect_equal(f$phi_min, 1 / sqrt(2))
  ## ... and the remaining tie goes to the lower original target index.
  expect_equal(f$pip_rmsd, 0)
  ## Leftovers are zero (0), half (1) and oneB (2): the max is oneB's.
  expect_equal(f$unmatched_ssl, 2)
})

test_that("a finite congruence outranks an undefined one in a tie", {
  a <- c(1, 0, 0, 0)
  zero <- c(0, 0, 0, 0)
  half <- c(.5, .5, .5, .5)
  expect_identical(sum((a - zero)^2), sum((a - half)^2))

  ## The zero column sits at the lower index, so only the congruence
  ## preference can move the choice off it.
  f <- .cmp(cbind(a), cbind(zero, half))
  expect_equal(f$phi_min, 0.5)
  expect_identical(f$unmatched_ssl, 0)

  ## When every tied congruence is undefined, the index rule settles it: the
  ## first zero column is taken and the second is the leftover.
  f0 <- .cmp(cbind(a), cbind(zero, zero))
  expect_identical(f0$phi_min, NA_real_)
  expect_identical(f0$unmatched_ssl, 0)
})

test_that("a near tie is not a tie", {
  ## Same fixture as the exact tie, except the high-congruence target is moved
  ## by 1e-9 so its aligned SSE is strictly larger.  No tolerance may widen
  ## that into a tie: the 0.5-congruence target must now win on distance.
  a <- c(1, 0, 0, 0)
  nudged <- c(1, 1 + 1e-9, 0, 0)
  half <- c(.5, .5, .5, .5)
  expect_gt(sum((a - nudged)^2), sum((a - half)^2))

  f <- .cmp(cbind(a), cbind(nudged, half))
  expect_equal(f$phi_min, 0.5)
  expect_equal(f$unmatched_ssl, sum(nudged^2))
})


# ---- no eligibility screen --------------------------------------------

test_that("weak, singleton and exact-zero columns all participate", {
  ## A salience screen of any size would drop the 1e-8 column, and phi_min
  ## would come back as 1 from the two clean pairs.  Every column is scored,
  ## so the dead column's arbitrary congruence is what is reported.
  weak <- c(1e-8, 0, 0, 0)
  lam_from <- cbind(c(.9, .9, 0, 0), weak, c(0, 0, 0, 1))
  lam_to <- cbind(c(.9, .9, 0, 0), c(0, 0, 1e-9, 0), c(0, 0, 0, 1),
                  c(.4, .4, .4, .4))

  f <- .cmp(lam_from, lam_to)
  ## The near-dead source matches the smallest-norm target, exactly as the
  ## note predicts, and drags phi_min to 0.
  expect_equal(f$phi_min, 0)
  ## Exactly k = 3 pairs are scored: pooled RMSD is the 1e-8 column's residual
  ## spread over J * k cells, which no screened-out column could produce.
  expect_equal(f$rmsd, sqrt((1e-16 + 1e-18) / 12))
  expect_equal(f$rmsd_max, sqrt((1e-16 + 1e-18) / 4))
  ## The only unselected target is the dense one.
  expect_equal(f$unmatched_ssl, 4 * 0.16)
  expect_false(f$collision)
})

test_that("a zero-norm source keeps RMSD finite and types phi_min as NA", {
  lam_from <- cbind(c(1, 0, 0, 0), c(0, 0, 0, 0))
  lam_to <- cbind(c(1, 0, 0, 0), c(0, 1, 0, 0), c(0, 0, 1, 0))
  f <- .cmp(lam_from, lam_to)

  ## NA_real_, not NaN: the undefined congruence is coerced rather than
  ## dropped with na.rm.
  expect_identical(f$phi_min, NA_real_)
  expect_false(is.nan(f$phi_min))
  ## Aligned distance is defined for a zero column, so both RMSDs stay finite.
  expect_equal(f$rmsd, sqrt(1 / 8))
  expect_equal(f$rmsd_max, sqrt(1 / 4))
  ## All three unit targets are equidistant from the dead column, so the index
  ## rule sends it onto the first one and raises a collision.
  expect_true(f$collision)
  expect_equal(f$unmatched_ssl, 1)
})


# ---- collision and leftovers ------------------------------------------

test_that("target reuse is recorded and never repaired", {
  ## Both sources are nearest to target 1, so the independent assignment
  ## reuses it instead of pushing the second source onto a runner-up.
  lam_from <- cbind(c(1, 0, 0, 0), c(.9, 0, 0, 0))
  lam_to <- cbind(c(1, 0, 0, 0), c(0, 3, 0, 0), c(0, 0, 2, 0))
  f <- .cmp(lam_from, lam_to)

  expect_true(f$collision)
  ## Both leftovers count, and the reported value is the largest SSL (9), not
  ## the sum (13) and not the smaller one (4).
  expect_equal(f$unmatched_ssl, 9)
  ## A collision blanks nothing: every numeric fact is still defined.
  expect_equal(f$phi_min, 1)
  expect_equal(f$rmsd, sqrt(0.01 / 8))
  expect_equal(f$rmsd_max, sqrt(0.01 / 4))
})

test_that("a clean one-to-one map reports collision FALSE", {
  lam_from <- cbind(c(1, 0, 0, 0), c(0, 1, 0, 0))
  lam_to <- cbind(c(1, 0, 0, 0), c(0, 1, 0, 0), c(0, 0, 1, 0))
  f <- .cmp(lam_from, lam_to)
  expect_false(f$collision)
  expect_identical(f$rmsd, 0)
  expect_identical(f$rmsd_max, 0)
  expect_equal(f$phi_min, 1)
  expect_equal(f$unmatched_ssl, 1)
})

test_that("an unused exact-zero target gives unmatched_ssl exactly zero", {
  f <- .cmp(cbind(c(1, 0)), cbind(c(1, 0), c(0, 0)))
  expect_identical(f$unmatched_ssl, 0)
  expect_false(f$collision)
})


# ---- the remaining metrics --------------------------------------------

test_that("ari compares max-loading partitions with lower-index ties", {
  ## Rows 3 and 4 are all-zero at the source and row 4 is all-zero at the
  ## target, so each is an exact loading tie resolved to the lowest column.
  lam_from <- cbind(c(1, 0, 0, 0), c(0, 1, 0, 0))
  lam_to <- cbind(c(1, 0, 0, 0), c(0, 1, 0, 0), c(0, 0, 1, 0))
  expect_identical(max.col(abs(lam_from), ties.method = "first"),
                   c(1L, 2L, 1L, 1L))
  expect_identical(max.col(abs(lam_to), ties.method = "first"),
                   c(1L, 2L, 3L, 1L))
  ## Contingency counts give A = 1, B = 3, C = 1, E = 1/2, D = 3/2.
  expect_equal(.cmp(lam_from, lam_to)$ari, (1 - 0.5) / 1.5)

  ## ARI reads every endpoint group column, not just the matched ones.
  ## Emptying the unselected third target leaves the assignment and both RMSDs
  ## untouched, yet moves the target partition and therefore the index.
  lam_dead <- cbind(c(1, 0, 0, 0), c(0, 1, 0, 0), c(0, 0, 0, 0))
  dead <- .cmp(lam_from, lam_dead)
  expect_identical(dead$rmsd, .cmp(lam_from, lam_to)$rmsd)
  expect_identical(dead$ari, 1)

  ## A negative index is reachable and is reported as such.
  lam_to2 <- cbind(c(1, 0, 0, 0), c(0, 0, 0, 0), c(0, 0, 1, 0))
  expect_equal(.cmp(lam_from, lam_to2)$ari, (1 - 1.5) / 1.5)
})

test_that("a one-item comparison returns the documented ari of 1", {
  f <- .cmp(matrix(1, 1, 1), matrix(c(1, 2), 1, 2))
  expect_identical(f$ari, 1)
})

test_that("pip_rmsd uses only cells regularized at both endpoints", {
  lam_from <- cbind(c(1, 1, 0, 0), c(0, 0, 1, 0))
  lam_to <- cbind(c(1, 1, 0, 0), c(0, 0, 1, 0), c(0, 0, 0, 1))
  ## Rows 1-2 of the backbone are fixed anchors at both endpoints, so only
  ## 2 backbone cells and 4 exploratory cells are jointly regularized.
  reg_from <- cbind(c(FALSE, FALSE, TRUE, TRUE), rep(TRUE, 4))
  reg_to <- cbind(c(FALSE, FALSE, TRUE, TRUE), rep(TRUE, 4), rep(TRUE, 4))
  pip_from <- cbind(c(.9, .9, .1, .2), c(.05, .05, .95, .10))
  pip_to <- cbind(c(.8, .8, .3, .2), c(.05, .15, .85, .10),
                  c(.5, .5, .5, .5))

  sel <- reg_from & reg_to[, c(1L, 2L)]
  expect_identical(sum(sel), 6L)
  d <- pip_from[sel] - pip_to[, c(1L, 2L)][sel]
  expect_equal(sum(d^2), 0.06)

  f <- .cmp(lam_from, lam_to, K0 = 1, pip_from = pip_from, pip_to = pip_to,
            reg_from = reg_from, reg_to = reg_to)
  expect_equal(f$pip_rmsd, sqrt(0.06 / 6))
  ## Column 3 of the target is unmatched, so its PIPs never enter.
  expect_equal(f$unmatched_ssl, 1)
})

test_that("pip_rmsd is NA when no cell is regularized at both endpoints", {
  lam_from <- cbind(c(1, 0), c(0, 1))
  lam_to <- cbind(c(1, 0), c(0, 1), c(1, 1))
  f <- .cmp(lam_from, lam_to,
            reg_from = matrix(FALSE, 2, 2), reg_to = matrix(TRUE, 2, 3))
  expect_identical(f$pip_rmsd, NA_real_)
  ## The loading facts stay defined.
  expect_identical(f$rmsd, 0)
})

test_that("pooled rmsd never exceeds the columnwise maximum", {
  fixtures <- list(
    list(cbind(c(1, 0, 0, 0), c(0, 1, 0, 0)),
         cbind(c(1, 0, 0, 0), c(0, 0, 0, 0), c(0, 0, 1, 0))),
    list(cbind(c(.9, .8, .1, 0), c(.1, 0, .8, .7)),
         cbind(c(.85, .75, .1, 0), c(.05, 0, .85, .75), c(.3, -.2, .1, .4))),
    list(cbind(c(1, 0, 0, 0), c(.9, 0, 0, 0)),
         cbind(c(1, 0, 0, 0), c(0, 3, 0, 0), c(0, 0, 2, 0))),
    list(cbind(c(1, 1, 1, 1)), cbind(c(1, 1, 1, 1), c(0, 0, 0, 1)))
  )
  for (fx in fixtures) {
    f <- .cmp(fx[[1L]], fx[[2L]])
    ## A mean cannot exceed a maximum in exact arithmetic, but two equal
    ## columns can invert the pair by one ulp, so this is a tolerance claim.
    expect_lte(f$rmsd, f$rmsd_max + 1e-12)
  }

  ## Unequal columns separate the two by a hand-computable margin: aligned
  ## squared distances 0 and 1 over J = 4 give sqrt(1/8) pooled and sqrt(1/4)
  ## as the weakest column.
  f <- .cmp(fixtures[[1L]][[1L]], fixtures[[1L]][[2L]])
  expect_equal(f$rmsd, sqrt(1 / 8))
  expect_equal(f$rmsd_max, sqrt(1 / 4))
  expect_lt(f$rmsd, f$rmsd_max)
})


# ---- backbone handling ------------------------------------------------

test_that("backbone columns pair by position, never by fit", {
  ## The target's first two columns are the source's in swapped order, so a
  ## free matcher would swap them and report RMSD 0.  Position pairing must
  ## instead report the orthogonal-pair distance.
  lam_from <- cbind(c(1, 0, 0, 0), c(0, 1, 0, 0))
  lam_to <- cbind(c(0, 1, 0, 0), c(1, 0, 0, 0), c(0, 0, 1, 0))
  f <- .cmp(lam_from, lam_to, K0 = 2)

  expect_equal(f$rmsd, sqrt(0.5))
  expect_equal(f$rmsd_max, sqrt(0.5))
  expect_equal(f$phi_min, 0)
  ## k = K0 is a legal edge: there are no exploratory sources at all, so no
  ## reuse is possible and the single spare target is the leftover.
  expect_false(f$collision)
  expect_equal(f$unmatched_ssl, 1)
})

test_that("backbone position pairs are immune to a better exploratory target", {
  ## Column 3 of the target is an exact copy of backbone source column 1, yet
  ## the backbone still pairs 1 with 1.
  lam_from <- cbind(c(1, 0, 0, 0), c(0, 0, 1, 0))
  lam_to <- cbind(c(.5, 0, 0, 0), c(0, 0, 1, 0), c(1, 0, 0, 0))
  f <- .cmp(lam_from, lam_to, K0 = 1)
  ## Backbone pair distance 0.25, exploratory pair distance 0.
  expect_equal(f$rmsd, sqrt(0.25 / 8))
  expect_equal(f$rmsd_max, sqrt(0.25 / 4))
  ## The exploratory source took target 2, so target 3 is the leftover.
  expect_equal(f$unmatched_ssl, 1)
})


# ---- direct, never chained --------------------------------------------

test_that("a direct comparison ignores what an intervening candidate matched", {
  ## L1 -> L2 sends the single source onto column 1 of L2; L2 -> L3 sends that
  ## column onto column 2 of L3.  Composing those two assignments would land
  ## on L3 column 2, but the direct L1 -> L3 comparison lands on column 1.
  L1 <- cbind(c(1, 0, 0))
  L2 <- cbind(c(1, 1, 0), c(0, 0, 1))
  L3 <- cbind(c(1, 0, 0), c(1, 1, 0), c(0, 0, 1))

  ## Leg 1: the source is closer to L2 column 1 than to column 2.
  expect_equal(.cmp(L1, L2)$rmsd, sqrt(1 / 3))
  ## Leg 2: that column is an exact copy of L3 column 2, so it goes there.
  expect_identical(.cmp(cbind(L2[, 1L]), L3)$rmsd, 0)

  direct <- .cmp(L1, L3)
  ## Direct: an exact copy of the source sits at L3 column 1.
  expect_identical(direct$rmsd, 0)
  expect_identical(direct$rmsd_max, 0)
  ## The leftovers are L3 columns 2 and 3, whose SSLs are 2 and 1.  Had the
  ## chained answer been used the leftovers would be columns 1 and 3, so
  ## unmatched_ssl would have been 1.
  expect_equal(direct$unmatched_ssl, 2)
  expect_equal(.cmp(L1, cbind(L3[, 2L]))$rmsd, sqrt(1 / 3))
})


# ---- numerical robustness ---------------------------------------------

test_that("extreme finite columns do not overflow into a manufactured tie", {
  ## Both unscaled aligned distances overflow to Inf, which without rescaling
  ## would tie and hand the choice to the tie rule.  On a common power-of-two
  ## basis the nearer target is still strictly nearer.
  a <- c(1e200, 0)
  near <- c(0, 1e200)
  far <- c(0, 2e200)
  expect_identical(sum((a - near)^2), Inf)
  expect_identical(sum((a - far)^2), Inf)

  f <- .cmp(cbind(a), cbind(near, far))
  ## sqrt(2) * 1e200 spread over J = 2 cells is exactly 1e200.
  expect_equal(f$rmsd, 1e200)
  expect_equal(f$rmsd_max, 1e200)
  expect_true(is.finite(f$rmsd))
  expect_false(f$collision)

  ## Reversing the target order reverses nothing: the choice is a distance,
  ## not an index accident.
  f2 <- .cmp(cbind(a), cbind(far, near))
  expect_equal(f2$rmsd, f$rmsd)
})

test_that("extreme finite columns do not underflow into a manufactured tie", {
  ## The mirror image of the overflow case.  Both unscaled aligned distances
  ## underflow to exactly 0, and both congruences are exactly 1, so without
  ## rescaling the tie rule would fall through to the index and take the first
  ## -- strictly farther -- target.  On a common power-of-two basis the nearer
  ## target is still strictly nearer.
  a <- c(1e-200, 0)
  far <- c(3e-200, 0)     # aligned distance (2e-200)^2, 4x the other
  near <- c(2e-200, 0)    # aligned distance (1e-200)^2
  expect_identical(sum((a - far)^2), 0)
  expect_identical(sum((a - near)^2), 0)
  ## The same two distances on a common basis, where they are representable
  ## and strictly ordered 4 to 1.
  sse <- function(b) sum(((a - b) / 2^-664)^2)
  expect_gt(sse(far), sse(near))

  f <- .cmp(cbind(a), cbind(far, near))
  ## Both leftover SSLs underflow to 0, so the RMSDs are what read out the
  ## choice: 1e-200 / sqrt(2) for `near`, twice that for `far`.
  expect_equal(f$rmsd, 1e-200 / sqrt(2))
  expect_equal(f$rmsd_max, 1e-200 / sqrt(2))
  expect_false(f$rmsd == 2e-200 / sqrt(2))
  expect_false(f$collision)

  ## Reversing the target order reverses nothing: the choice is a distance,
  ## not an index accident.
  expect_identical(.cmp(cbind(a), cbind(near, far))$rmsd, f$rmsd)
})

test_that("a huge but representable RMSD is formed instead of Inf", {
  ## The aligned sum of squares is 2e600 and cannot be held in a double, but
  ## the RMSD that comes out of it can, so scaling has to happen before the
  ## division rather than after it.
  a <- c(1e300, 0, 0, 0)
  targets <- cbind(c(0, 1e300, 0, 0), c(0, 0, 9e300, 0))
  f <- .cmp(cbind(a), targets)
  expect_true(is.finite(f$rmsd))
  expect_equal(f$rmsd, sqrt(2) / 2 * 1e300)
  expect_equal(f$rmsd_max, sqrt(2) / 2 * 1e300)
})

## The 0.9.0 pefa() object contract: the seven top-level components, the three
## derived tables, the failure and validation rules, and the public surface
## that survived the lean redesign (technical note section 6, plan 9.2/9.3).
##
## Fits are deliberately tiny -- J = 8, N = 150, a single v0 -- because nothing
## here depends on estimation quality, only on shape, type and bookkeeping.

.mk <- function(N = 150, K = 2, ipf = 4, seed = 11) {
  set.seed(seed)
  J <- K * ipf
  grp <- rep(seq_len(K), each = ipf)
  Lam <- matrix(0, J, K)
  Lam[cbind(seq_len(J), grp)] <- .75
  Phi <- matrix(.3, K, K)
  diag(Phi) <- 1
  psi <- pmax(1 - rowSums((Lam %*% chol(Phi))^2), .25)
  eta <- matrix(rnorm(N * K), N, K) %*% chol(Phi)
  Y <- eta %*% t(Lam) + matrix(rnorm(N * J), N, J) %*% diag(sqrt(psi))
  Q0 <- matrix(-1L, J, K)
  for (k in seq_len(K)) {
    a <- which(grp == k)[1:2]
    Q0[a, ] <- 0L
    Q0[a, k] <- 1L
  }
  list(Y = Y, Q0 = Q0, J = as.integer(J), K0 = as.integer(K))
}

.fit3 <- function(d = .mk()) {
  pefa(d$Q0, d$Y, 2, 4, v0 = .001, max_it = 400, verbose = FALSE)
}

## The regularization mask pefa() rebuilds for candidate K from the stored
## backbone.  Reassembly and the direct-comparison checks both need it.
.reg_of <- function(Q0, K) {
  J <- nrow(Q0)
  K0 <- ncol(Q0)
  Q <- if (K > K0) cbind(Q0, matrix(-1L, J, K - K0)) else Q0
  matrix(Q == -1L, J, K)
}

## Group-block view: bifactor candidates carry a leading general column that
## every between-candidate comparison removes first.
.grp <- function(m, bifactor) if (bifactor) m[, -1L, drop = FALSE] else m

.recompare <- function(x, K_from, K_to) {
  bif <- isTRUE(x$settings$bifactor)
  a <- as.character(K_from)
  b <- as.character(K_to)
  vbpm:::.pefa_compare_pair(
    lam_from = .grp(x$loadings[[a]], bif), lam_to = .grp(x$loadings[[b]], bif),
    pip_from = .grp(x$pips[[a]], bif), pip_to = .grp(x$pips[[b]], bif),
    reg_from = .reg_of(x$Q0, K_from), reg_to = .reg_of(x$Q0, K_to),
    K0 = ncol(x$Q0))
}

.fact_names <- c("phi_min", "rmsd", "rmsd_max", "ari", "pip_rmsd",
                 "unmatched_ssl", "collision")

## One synthetic pair result, so the derived-view builders can be exercised on
## values chosen by hand rather than produced by a fit.
.hand_fact <- function(phi_min, rmsd, rmsd_max, ari, pip_rmsd, unmatched_ssl,
                       collision) {
  list(phi_min = phi_min, rmsd = rmsd, rmsd_max = rmsd_max, ari = ari,
       pip_rmsd = pip_rmsd, unmatched_ssl = unmatched_ssl,
       collision = collision)
}


# ---- top-level shape --------------------------------------------------

test_that("the object holds exactly the seven documented components", {
  r <- .fit3()
  expect_identical(names(r),
                   c("sweep", "transitions", "persistence", "loadings",
                     "pips", "Q0", "settings"))
  expect_identical(class(r), "pefa")
  expect_type(r, "list")
  expect_length(r, 7L)
})

test_that("every component removed in 0.9.0 is absent", {
  r <- .fit3()
  gone <- c("provenance", "selected_K", "boundary", "profile", "evidence",
            "checkpoint", "resume", "lineage", "hashes", "hash", "build_id",
            "fit", "fits", "selected_fit", "status", "reason", "n_pairs",
            "stability_eps", "pair_detail", "pairs", "decisions")
  for (nm in gone) expect_null(r[[nm, exact = TRUE]])

  ## Neither a soft criterion nor a status taxonomy survives in any table.
  expect_false(any(c("t_S", "AIC_S", "BIC_S", "status", "reason") %in%
                     names(r$sweep)))
  expect_false(any(c("status", "reason", "n_pairs", "assignment",
                     "collision_location", "unmatched_n") %in%
                     names(r$transitions)))
  ## No speculative sweep superclass.
  expect_false(inherits(r, "vbpm_sweep"))
})

test_that("$settings holds exactly the resolved fit controls, in order", {
  d <- .mk()
  r <- .fit3(d)
  s <- r$settings
  expect_identical(names(s),
                   c("bifactor", "general", "v0", "max_it", "convChk",
                     "tolVal", "tau", "rank_adjust", "rank_max_J"))
  expect_identical(s$bifactor, FALSE)
  expect_null(s$general)
  expect_identical(s$v0, 0.001)
  expect_identical(s$max_it, 400L)
  expect_identical(s$convChk, FALSE)
  expect_identical(s$tolVal, 1e-4)
  expect_identical(s$tau, 0.5)
  expect_identical(s$rank_adjust, FALSE)
  expect_identical(s$rank_max_J, 100L)
  ## Unnamed scalars, and no analysis policy or verbosity value.
  expect_null(names(s$bifactor))
  expect_null(names(s$v0))
  expect_false(any(c("verbose", "cuts", "sustain", "delta", "eps",
                     "stability_eps") %in% names(s)))

  ## Bifactor mode resolves `general` to a length-J integer design.
  rb <- pefa(d$Q0, d$Y, 2, 3, bifactor = TRUE, v0 = .001, max_it = 300,
             verbose = FALSE)
  expect_identical(rb$settings$bifactor, TRUE)
  expect_identical(rb$settings$general, rep(1L, d$J))
})

test_that("$Q0, $loadings and $pips are the documented candidate primitives", {
  d <- .mk()
  r <- .fit3(d)
  expect_true(is.matrix(r$Q0))
  expect_identical(dim(r$Q0), c(d$J, 2L))
  expect_type(r$Q0, "integer")
  expect_true(all(r$Q0 %in% c(-1L, 0L, 1L)))
  expect_identical(colnames(r$Q0), c("F1", "F2"))

  keys <- as.character(2:4)
  expect_identical(names(r$loadings), keys)
  expect_identical(names(r$pips), keys)
  for (k in keys) {
    L <- r$loadings[[k]]
    P <- r$pips[[k]]
    expect_true(is.matrix(L) && is.double(L))
    expect_true(is.matrix(P) && is.double(P))
    expect_identical(dim(L), c(d$J, as.integer(k)))
    expect_identical(dim(P), dim(L))
    expect_identical(dimnames(P), dimnames(L))
    expect_identical(rownames(L), rownames(r$Q0))
    expect_true(all(is.finite(L)))
    expect_true(all(is.finite(P)) && all(P >= 0) && all(P <= 1))
  }
  ## Backbone names lead the factor dimnames; exploratory columns continue it.
  expect_identical(colnames(r$loadings[["4"]]), c("F1", "F2", "F3", "F4"))
})


# ---- $sweep -----------------------------------------------------------

test_that("$sweep has exactly the documented columns and types", {
  r <- .fit3()
  expect_s3_class(r$sweep, "data.frame")
  expect_identical(names(r$sweep),
                   c("K", "ELBO", "AIC", "BIC", "RMSEA", "SRMR", "CFI", "TLI",
                     "t", "iter", "converged"))
  expect_identical(nrow(r$sweep), 3L)
  expect_identical(r$sweep$K, 2:4)
  expect_type(r$sweep$K, "integer")
  expect_type(r$sweep$iter, "integer")
  expect_type(r$sweep$converged, "logical")
  for (nm in c("ELBO", "AIC", "BIC", "RMSEA", "SRMR", "CFI", "TLI", "t")) {
    expect_type(r$sweep[[nm]], "double")
  }
  ## Required candidate facts are finite; t is a whole nonnegative count that
  ## is nonetheless stored as a double.
  expect_true(all(is.finite(r$sweep$ELBO)))
  expect_true(all(is.finite(r$sweep$AIC)))
  expect_true(all(is.finite(r$sweep$BIC)))
  expect_true(all(is.finite(r$sweep$t)))
  expect_true(all(r$sweep$t >= 0))
  expect_identical(r$sweep$t, round(r$sweep$t))
  expect_false(anyNA(r$sweep$converged))
  expect_identical(rownames(r$sweep), as.character(1:3))
})


# ---- $transitions -----------------------------------------------------

test_that("$transitions has exactly the documented columns and types", {
  r <- .fit3()
  tr <- r$transitions
  expect_s3_class(tr, "data.frame")
  expect_identical(names(tr),
                   c("K_from", "K_to", "ELBO_gain_pct", "BIC_gain_pct",
                     .fact_names))
  expect_identical(ncol(tr), 11L)
  expect_identical(nrow(tr), nrow(r$sweep) - 1L)
  expect_identical(tr$K_from, 2:3)
  expect_identical(tr$K_to, 3:4)
  expect_type(tr$K_from, "integer")
  expect_type(tr$K_to, "integer")
  expect_type(tr$collision, "logical")
  for (nm in c("ELBO_gain_pct", "BIC_gain_pct",
               setdiff(.fact_names, "collision"))) {
    expect_type(tr[[nm]], "double")
  }
})

test_that("a one-candidate window keeps the typed zero-row schema", {
  d <- .mk()
  r <- pefa(d$Q0, d$Y, 2, 2, v0 = .001, max_it = 300, verbose = FALSE)
  expect_identical(nrow(r$sweep), 1L)
  tr <- r$transitions
  expect_identical(nrow(tr), 0L)
  expect_identical(names(tr),
                   c("K_from", "K_to", "ELBO_gain_pct", "BIC_gain_pct",
                     .fact_names))
  expect_type(tr$K_from, "integer")
  expect_type(tr$K_to, "integer")
  expect_type(tr$collision, "logical")
  for (nm in c("ELBO_gain_pct", "BIC_gain_pct",
               setdiff(.fact_names, "collision"))) {
    expect_type(tr[[nm]], "double")
  }
  ## The persistence triangles degenerate to a single typed NA cell.
  expect_identical(dim(r$persistence$phi), c(1L, 1L))
  expect_identical(r$persistence$phi[1L, 1L], NA_real_)
  expect_identical(r$persistence$collision[1L, 1L], NA)
})


# ---- $persistence -----------------------------------------------------

test_that("$persistence is three plain upper-triangular K-keyed matrices", {
  r <- .fit3()
  p <- r$persistence
  expect_type(p, "list")
  expect_identical(names(p), c("phi", "rmsd", "collision"))
  keys <- as.character(2:4)
  for (nm in names(p)) {
    m <- p[[nm]]
    expect_true(is.matrix(m))
    expect_identical(dim(m), c(3L, 3L))
    expect_identical(dimnames(m), list(keys, keys))
    ## Ordinary matrices only: no custom class, no smuggled attribute.
    expect_identical(names(attributes(m)), c("dim", "dimnames"))
    expect_identical(class(m), c("matrix", "array"))
    ## Diagonal and lower triangle are typed NA.
    expect_true(all(is.na(m[!upper.tri(m)])))
    expect_true(all(!is.na(m[upper.tri(m)])))
  }
  expect_type(p$phi, "double")
  expect_type(p$rmsd, "double")
  expect_type(p$collision, "logical")
  expect_identical(p$phi[2L, 1L], NA_real_)
  expect_identical(p$collision[3L, 3L], NA)
})

test_that("every adjacent persistence cell is its transition field", {
  r <- .fit3()
  tr <- r$transitions
  p <- r$persistence
  for (i in seq_len(nrow(tr))) {
    a <- as.character(tr$K_from[i])
    b <- as.character(tr$K_to[i])
    expect_identical(p$phi[a, b], tr$phi_min[i])
    ## The mapping is rmsd_max, NOT the pooled $transitions$rmsd.
    expect_identical(p$rmsd[a, b], tr$rmsd_max[i])
    expect_identical(p$collision[a, b], tr$collision[i])
  }
  ## The two RMSD definitions really do differ on this sweep, so the check
  ## above could not have passed by accident.
  expect_true(any(tr$rmsd != tr$rmsd_max))
  expect_false(isTRUE(all.equal(p$rmsd["2", "3"], tr$rmsd[1L])))
})

test_that("the derived views read one shared pair result, keyed correctly", {
  ## Hand-chosen facts with rmsd and rmsd_max deliberately unequal, so a
  ## crossed mapping cannot survive.
  Ks <- 2:4
  facts <- list(
    "2-3" = .hand_fact(0.91, 0.11, 0.21, 0.31, 0.41, 0.51, FALSE),
    "3-4" = .hand_fact(0.92, 0.12, 0.22, 0.32, 0.42, 0.52, TRUE),
    "2-4" = .hand_fact(0.93, 0.13, 0.23, 0.33, 0.43, 0.53, TRUE)
  )
  sweep <- data.frame(K = Ks, ELBO = c(-100, -50, -60), BIC = c(300, 280, 290))

  tr <- vbpm:::.pefa_transitions(sweep, facts)
  expect_identical(tr$K_from, 2:3)
  expect_identical(tr$phi_min, c(0.91, 0.92))
  expect_identical(tr$rmsd, c(0.11, 0.12))
  expect_identical(tr$rmsd_max, c(0.21, 0.22))
  expect_identical(tr$ari, c(0.31, 0.32))
  expect_identical(tr$pip_rmsd, c(0.41, 0.42))
  expect_identical(tr$unmatched_ssl, c(0.51, 0.52))
  expect_identical(tr$collision, c(FALSE, TRUE))

  p <- vbpm:::.pefa_persistence(Ks, facts)
  expect_identical(p$phi["2", "3"], 0.91)
  expect_identical(p$phi["2", "4"], 0.93)
  expect_identical(p$rmsd["2", "3"], 0.21)      # rmsd_max, not 0.11
  expect_identical(p$rmsd["3", "4"], 0.22)
  expect_identical(p$rmsd["2", "4"], 0.23)
  ## A collision on a non-adjacent pair reaches the mask on its own cell.
  expect_identical(p$collision["2", "4"], TRUE)
  expect_identical(p$collision["2", "3"], FALSE)
})

test_that("all-pairs cells are direct endpoint comparisons", {
  r <- .fit3()
  Ks <- r$sweep$K
  for (i in seq_along(Ks)) {
    for (j in seq_along(Ks)) {
      if (i >= j) next
      f <- .recompare(r, Ks[i], Ks[j])
      a <- as.character(Ks[i])
      b <- as.character(Ks[j])
      expect_identical(r$persistence$phi[a, b], f$phi_min)
      expect_identical(r$persistence$rmsd[a, b], f$rmsd_max)
      expect_identical(r$persistence$collision[a, b], f$collision)
    }
  }
})

test_that("a chained composition would give a different all-pairs answer", {
  ## Three synthetic candidates in which composing the adjacent assignments
  ## disagrees with the direct one: the single K = 1 source matches column 1
  ## of the K = 2 candidate, that column matches column 2 of the K = 3
  ## candidate, but the source itself matches column 1 of the K = 3 candidate.
  L <- list("1" = cbind(c(1, 0, 0)),
            "2" = cbind(c(1, 1, 0), c(0, 0, 1)),
            "3" = cbind(c(1, 0, 0), c(1, 1, 0), c(0, 0, 1)))
  P <- lapply(L, function(m) matrix(0.5, nrow(m), ncol(m)))
  Q0 <- matrix(0L, 3L, 0L)
  cmp <- function(a, b) {
    vbpm:::.pefa_compare_pair(L[[a]], L[[b]], P[[a]], P[[b]],
                              .reg_of(Q0, as.integer(a)),
                              .reg_of(Q0, as.integer(b)), 0L)
  }
  facts <- list("1-2" = cmp("1", "2"), "2-3" = cmp("2", "3"),
                "1-3" = cmp("1", "3"))
  p <- vbpm:::.pefa_persistence(1:3, facts)

  ## Direct: an exact copy of the source, so the weakest-column RMSD is 0.
  expect_identical(p$rmsd["1", "3"], 0)
  ## Chained would have gone through the K = 2 column and landed on the
  ## K = 3 column whose distance is sqrt(1/3).
  expect_equal(facts[["1-2"]]$rmsd_max, sqrt(1 / 3))
  expect_gt(facts[["1-2"]]$rmsd_max, p$rmsd["1", "3"])
})


# ---- gains ------------------------------------------------------------

test_that("gains are oriented toward K_to and scaled by their own maximum", {
  Ks <- 2:5
  facts <- stats::setNames(
    rep(list(.hand_fact(1, 0, 0, 1, 0, 0, FALSE)), 3L),
    c("2-3", "3-4", "4-5"))
  sweep <- data.frame(K = Ks,
                      ELBO = c(-100, -50, -60, -40),
                      BIC = c(300, 280, 290, 285))
  tr <- vbpm:::.pefa_transitions(sweep, facts)

  ## ELBO gain is ELBO_to - ELBO_from: 50, -10, 20, largest positive 50.
  expect_identical(tr$ELBO_gain_pct, c(100, -20, 40))
  ## BIC gain is BIC_from - BIC_to: 20, -10, 5, largest positive 20.
  expect_identical(tr$BIC_gain_pct, c(100, -50, 25))
})

test_that("a path with no positive gain is entirely NA", {
  Ks <- 2:4
  facts <- stats::setNames(
    rep(list(.hand_fact(1, 0, 0, 1, 0, 0, FALSE)), 2L), c("2-3", "3-4"))
  sweep <- data.frame(K = Ks, ELBO = c(-100, -150, -200),
                      BIC = c(300, 280, 260))
  tr <- vbpm:::.pefa_transitions(sweep, facts)
  expect_identical(tr$ELBO_gain_pct, c(NA_real_, NA_real_))
  ## The BIC path is unaffected: normalization is per path.
  expect_identical(tr$BIC_gain_pct, c(100, 100))

  ## A real all-nonpositive sweep reaches the same place.
  r <- .fit3()
  expect_true(all(diff(r$sweep$ELBO) < 0))
  expect_true(all(is.na(r$transitions$ELBO_gain_pct)))
  expect_true(all(is.na(r$transitions$BIC_gain_pct)))
})

test_that("a nonrepresentable adjacent difference is a controlled error", {
  facts <- list("2-3" = .hand_fact(1, 0, 0, 1, 0, 0, FALSE))
  sweep <- data.frame(K = 2:3, ELBO = c(-1e308, 1e308), BIC = c(1, 2))
  expect_error(vbpm:::.pefa_transitions(sweep, facts), "must be finite")
  sweep2 <- data.frame(K = 2:3, ELBO = c(1, 2), BIC = c(-1e308, 1e308))
  expect_error(vbpm:::.pefa_transitions(sweep2, facts), "must be finite")
})


# ---- reassembly -------------------------------------------------------

test_that("both derived views rebuild from the five source components", {
  r <- .fit3()
  local_mocked_bindings(
    vbfa = function(...) {
      stop("vbfa() must not be called when rebuilding a derived view.")
    },
    .package = "vbpm")

  Ks <- r$sweep$K
  K0 <- ncol(r$Q0)
  bif <- isTRUE(r$settings$bifactor)
  facts <- list()
  for (i in seq_along(Ks)) {
    for (j in seq_along(Ks)) {
      if (i >= j) next
      a <- as.character(Ks[i])
      b <- as.character(Ks[j])
      facts[[paste(Ks[i], Ks[j], sep = "-")]] <- vbpm:::.pefa_compare_pair(
        lam_from = .grp(r$loadings[[a]], bif),
        lam_to = .grp(r$loadings[[b]], bif),
        pip_from = .grp(r$pips[[a]], bif), pip_to = .grp(r$pips[[b]], bif),
        reg_from = .reg_of(r$Q0, Ks[i]), reg_to = .reg_of(r$Q0, Ks[j]),
        K0 = K0)
    }
  }
  expect_identical(vbpm:::.pefa_transitions(r$sweep, facts), r$transitions)
  expect_identical(vbpm:::.pefa_persistence(Ks, facts), r$persistence)
})


# ---- failure contract -------------------------------------------------

test_that("a negative or fractional t aborts the call naming the K", {
  ## t counts parameters, so these are malformed candidates rather than
  ## unusual ones (technical note 6.1, plan 9.2).
  d <- .mk()
  real <- vbpm::fit_stats
  for (bad in c(-5, 12.5)) {
    local({
      value <- bad
      local_mocked_bindings(
        fit_stats = function(object, ...) {
          out <- real(object, ...)
          if (ncol(object$Lam) == 3L) out[["t"]] <- value
          out
        },
        .package = "vbpm")
      expect_error(pefa(d$Q0, d$Y, 2, 3, v0 = .001, max_it = 200,
                        verbose = FALSE),
                   "PEFA candidate K = 3: t is not a nonnegative whole number")
    })
  }
})

test_that("a nonfinite required criterion aborts the call naming the K", {
  d <- .mk()
  real <- vbpm::fit_stats
  for (broken in c("AIC", "BIC", "t")) {
    local({
      field <- broken
      local_mocked_bindings(
        fit_stats = function(object, ...) {
          out <- real(object, ...)
          if (ncol(object$Lam) == 3L) out[[field]] <- Inf
          out
        },
        .package = "vbpm")
      expect_error(pefa(d$Q0, d$Y, 2, 3, v0 = .001, max_it = 200,
                        verbose = FALSE),
                   "PEFA candidate K = 3: AIC, BIC, or t is nonfinite")
    })
  }
})

test_that("a nonfinite ELBO aborts the call naming the K", {
  d <- .mk()
  real <- vbpm::vbfa
  local_mocked_bindings(
    vbfa = function(...) {
      f <- real(...)
      if (ncol(f$Lam) == 3L) f$ELBO <- NaN
      f
    },
    .package = "vbpm")
  expect_error(pefa(d$Q0, d$Y, 2, 3, v0 = .001, max_it = 200,
                    verbose = FALSE),
               "PEFA candidate K = 3: ELBO is not one finite real number")
})

test_that("a malformed candidate matrix aborts the call naming the K", {
  ## fit_stats() is stubbed out here so the loading matrix itself is the only
  ## thing that can fail: the point is the candidate-normalization check, not
  ## how a fit statistic reacts to a broken matrix.
  d <- .mk()
  real <- vbpm::vbfa
  local_mocked_bindings(
    vbfa = function(...) {
      f <- real(...)
      if (ncol(f$Lam) == 3L) f$Lam[1L, 1L] <- Inf
      f
    },
    fit_stats = function(...) {
      c(AIC = 1, BIC = 2, t = 3, RMSEA = 0, SRMR = 0, CFI = 1, TLI = 1)
    },
    .package = "vbpm")
  expect_error(pefa(d$Q0, d$Y, 2, 3, v0 = .001, max_it = 200,
                    verbose = FALSE),
               "PEFA candidate K = 3: the loading matrix has nonfinite")
})

test_that("a dimensionally inconsistent candidate matrix names its K", {
  d <- .mk()
  real <- vbpm::vbfa
  local_mocked_bindings(
    vbfa = function(...) {
      f <- real(...)
      if (ncol(f$Lam) == 2L) f$Lam <- f$Lam[, 1L, drop = FALSE]
      f
    },
    fit_stats = function(...) {
      c(AIC = 1, BIC = 2, t = 3, RMSEA = 0, SRMR = 0, CFI = 1, TLI = 1)
    },
    .package = "vbpm")
  expect_error(pefa(d$Q0, d$Y, 2, 3, v0 = .001, max_it = 200,
                    verbose = FALSE),
               "PEFA candidate K = 2: the loading matrix is 8 by 1")
})

test_that("a PIP outside [0, 1] aborts the call naming the K", {
  d <- .mk()
  real <- vbpm::vbfa
  local_mocked_bindings(
    vbfa = function(...) {
      f <- real(...)
      if (ncol(f$Lam) == 2L) f$pi[1L, 1L] <- 1.5
      f
    },
    .package = "vbpm")
  expect_error(pefa(d$Q0, d$Y, 2, 3, v0 = .001, max_it = 200,
                    verbose = FALSE),
               "PEFA candidate K = 2: the PIP matrix has values outside")
})

test_that("a vbfa() failure aborts the call naming the K", {
  d <- .mk()
  local_mocked_bindings(
    vbfa = function(Y, Q, ...) {
      if (ncol(Q) == 3L) stop("synthetic engine failure")
      stop("unexpected K")
    },
    .package = "vbpm")
  expect_error(pefa(d$Q0, d$Y, 3, 3, v0 = .001, max_it = 200,
                    verbose = FALSE),
               "PEFA candidate K = 3: vbfa\\(\\) failed: synthetic engine")
})

test_that("a legitimately undefined descriptive index is kept as NA_real_", {
  ## NA means "this statistic's own definition does not apply" -- RMSEA and
  ## TLI are undefined at df <= 0. It is preserved, not treated as a defect.
  d <- .mk()
  real <- vbpm::fit_stats
  local_mocked_bindings(
    fit_stats = function(object, ...) {
      out <- real(object, ...)
      out[["RMSEA"]] <- NA_real_
      out[["TLI"]] <- NA_real_
      out
    },
    .package = "vbpm")
  r <- pefa(d$Q0, d$Y, 2, 3, v0 = .001, max_it = 200, verbose = FALSE)
  expect_identical(r$sweep$RMSEA, rep(NA_real_, 2L))
  expect_identical(r$sweep$TLI, rep(NA_real_, 2L))
  expect_false(any(is.nan(r$sweep$RMSEA)))
  expect_true(all(is.finite(r$sweep$SRMR)))
  expect_true(all(is.finite(r$sweep$AIC)))
})

test_that("a malformed descriptive index is named and refused", {
  ## NaN and the infinities are engine defects, not undefined quantities, so
  ## they are not laundered into an NA that would read as "undefined".
  d <- .mk()
  real <- vbpm::fit_stats
  for (spec in list(list(f = "RMSEA", v = NaN), list(f = "TLI", v = Inf),
                    list(f = "CFI", v = -Inf), list(f = "SRMR", v = NaN))) {
    local({
      field <- spec$f
      value <- spec$v
      local_mocked_bindings(
        fit_stats = function(object, ...) {
          out <- real(object, ...)
          if (ncol(object$Lam) == 3L) out[[field]] <- value
          out
        },
        .package = "vbpm")
      expect_error(pefa(d$Q0, d$Y, 2, 3, v0 = .001, max_it = 200,
                        verbose = FALSE),
                   paste0("PEFA candidate K = 3: ", field,
                          " is NaN or infinite"))
    })
  }
})

test_that("iter is validated before coercion", {
  d <- .mk()
  real_vbfa <- vbpm::vbfa
  for (value in list(-1, .5, Inf, NaN, 2^31, TRUE, 1 + 0i)) {
    local({
      v <- value
      local_mocked_bindings(
        vbfa = function(...) {
          out <- real_vbfa(...)
          out$iter <- v
          out
        },
        .package = "vbpm")
      expect_error(pefa(d$Q0, d$Y, 2, 2, v0 = .001, max_it = 200,
                        verbose = FALSE),
                   "iter is not a representable nonnegative whole number")
    })
  }
  ## A genuinely absent count is NA_integer_, not an error.
  local_mocked_bindings(
    vbfa = function(...) {
      out <- real_vbfa(...)
      out$iter <- NULL
      out
    },
    .package = "vbpm")
  r <- pefa(d$Q0, d$Y, 2, 2, v0 = .001, max_it = 200, verbose = FALSE)
  expect_identical(r$sweep$iter, NA_integer_)
})

test_that("Q0 storage type is enforced before coercion", {
  d <- .mk()
  J <- nrow(d$Q0)
  chr <- matrix(as.character(d$Q0), J, ncol(d$Q0))
  lgl <- matrix(d$Q0 == 1L, J, ncol(d$Q0))
  cpx <- matrix(as.complex(d$Q0), J, ncol(d$Q0))
  for (bad in list(chr, lgl, cpx, as.data.frame(d$Q0),
                   factor(as.character(d$Q0)))) {
    expect_error(pefa(bad, d$Y, 2, 2, v0 = .001, max_it = 200,
                      verbose = FALSE),
                 "Q0 must be a real numeric or integer matrix")
  }
  ## Integer and double matrices of -1/0/1 remain acceptable, as does a
  ## correctly typed zero-column backbone.
  expect_s3_class(pefa(d$Q0, d$Y, 2, 2, v0 = .001, max_it = 200,
                       verbose = FALSE), "pefa")
  expect_s3_class(pefa(matrix(as.double(d$Q0), J, ncol(d$Q0)), d$Y, 2, 2,
                       v0 = .001, max_it = 200, verbose = FALSE), "pefa")
  expect_s3_class(pefa(matrix(0L, J, 0L), d$Y, 1, 2, v0 = .001, max_it = 200,
                       verbose = FALSE), "pefa")
})

test_that("nonconverged candidates are retained under one aggregate warning", {
  d <- .mk()
  seen <- character(0)
  r <- withCallingHandlers(
    pefa(d$Q0, d$Y, 2, 3, v0 = .001, max_it = 3, verbose = FALSE),
    warning = function(w) {
      seen <<- c(seen, conditionMessage(w))
      invokeRestart("muffleWarning")
    })
  ## Exactly one warning for the whole call, naming every affected K, and the
  ## per-candidate iteration-limit warnings are muffled.
  expect_length(seen, 1L)
  expect_match(seen, "did not converge at K = 2, 3")
  ## The candidates themselves survive with their numbers intact.
  expect_s3_class(r, "pefa")
  expect_identical(r$sweep$converged, c(FALSE, FALSE))
  expect_true(all(is.finite(r$sweep$ELBO)))
  expect_true(all(is.finite(r$sweep$BIC)))
  expect_identical(nrow(r$transitions), 1L)
  expect_identical(summary(r)$nonconverged_K, 2:3)
  expect_output(print(r), "nonconverged K: 2, 3")
})


# ---- argument validation ----------------------------------------------

test_that("the K window is validated before anything is fitted", {
  d <- .mk()
  expect_error(pefa(d$Q0, d$Y, 4, 2, verbose = FALSE), "non-decreasing")
  whole <- "must be a single whole number"
  expect_error(pefa(d$Q0, d$Y, 2.5, 4, verbose = FALSE),
               paste("Kmin", whole))
  expect_error(pefa(d$Q0, d$Y, 2, 3.5, verbose = FALSE),
               paste("Kmax", whole))
  expect_error(pefa(d$Q0, d$Y, Inf, 4, verbose = FALSE),
               paste("Kmin", whole))
  expect_error(pefa(d$Q0, d$Y, NA_integer_, 4, verbose = FALSE), "Kmin")
  expect_error(pefa(d$Q0, d$Y, NaN, 4, verbose = FALSE), "Kmin")
  expect_error(pefa(d$Q0, d$Y, 2 + 0i, 4, verbose = FALSE), "Kmin")
  expect_error(pefa(d$Q0, d$Y, "2", 4, verbose = FALSE), "Kmin")
  expect_error(pefa(d$Q0, d$Y, c(2, 3), 4, verbose = FALSE), "Kmin")
  expect_error(pefa(d$Q0, d$Y, 0, 4, verbose = FALSE), "Kmin")
  expect_error(pefa(d$Q0, d$Y, 2, 2^31, verbose = FALSE), "Kmax")
  ## The window must reach the backbone.
  expect_error(pefa(d$Q0, d$Y, 1, 3, verbose = FALSE),
               "must reach the backbone")
})

test_that("Q0 is validated", {
  d <- .mk()
  bad <- d$Q0
  bad[1L, 1L] <- 2L
  expect_error(pefa(bad, d$Y, 2, 3, verbose = FALSE), "-1, 0, or 1")
  bad2 <- d$Q0
  bad2[1L, 1L] <- NA_integer_
  expect_error(pefa(bad2, d$Y, 2, 3, verbose = FALSE), "missing values")
  expect_error(pefa(d$Q0[-1L, , drop = FALSE], d$Y, 2, 3, verbose = FALSE),
               "nrow\\(Q0\\) must equal ncol\\(Y\\)")
  ## Supplied item names must agree with the data rather than silently win.
  named <- d$Q0
  rownames(named) <- paste0("z", seq_len(nrow(named)))
  expect_error(pefa(named, d$Y, 2, 3, verbose = FALSE), "rownames\\(Q0\\)")
})

test_that("Y is validated", {
  d <- .mk()
  bad <- d$Y
  bad[1L, 1L] <- Inf
  expect_error(pefa(d$Q0, bad, 2, 3, verbose = FALSE), "not NaN or infinite")
  bad[1L, 1L] <- NaN
  expect_error(pefa(d$Q0, bad, 2, 3, verbose = FALSE), "not NaN or infinite")
  chr <- matrix(as.character(d$Y), nrow(d$Y), ncol(d$Y))
  expect_error(pefa(d$Q0, chr, 2, 3, verbose = FALSE),
               "nonempty numeric N by J matrix")
  ## An all-missing item is rejected downstream, and the K is still named.
  gap <- d$Y
  gap[, 1L] <- NA
  expect_error(pefa(d$Q0, gap, 2, 2, v0 = .001, verbose = FALSE),
               "PEFA candidate K = 2")
})

test_that("the remaining fit controls are validated", {
  d <- .mk()
  expect_error(pefa(d$Q0, d$Y, 2, 3, tau = 1.5, verbose = FALSE), "tau")
  expect_error(pefa(d$Q0, d$Y, 2, 3, tau = NA_real_, verbose = FALSE), "tau")
  expect_error(pefa(d$Q0, d$Y, 2, 3, tolVal = -1, verbose = FALSE), "tolVal")
  expect_error(pefa(d$Q0, d$Y, 2, 3, v0 = 0, verbose = FALSE), "v0")
  expect_error(pefa(d$Q0, d$Y, 2, 3, v0 = numeric(0), verbose = FALSE), "v0")
  expect_error(pefa(d$Q0, d$Y, 2, 3, v0 = c(.01, NA), verbose = FALSE), "v0")
  expect_error(pefa(d$Q0, d$Y, 2, 3, max_it = 0, verbose = FALSE), "max_it")
  expect_error(pefa(d$Q0, d$Y, 2, 3, rank_max_J = -1, verbose = FALSE),
               "rank_max_J")
  expect_error(pefa(d$Q0, d$Y, 2, 3, convChk = NA, verbose = FALSE),
               "convChk must be TRUE or FALSE")
  expect_error(pefa(d$Q0, d$Y, 2, 3, verbose = "yes"),
               "verbose must be TRUE or FALSE")
  expect_error(pefa(d$Q0, d$Y, 2, 3, bifactor = TRUE, general = 2,
                    verbose = FALSE), "-1, 0, or 1")
  expect_error(pefa(d$Q0, d$Y, 2, 3, bifactor = TRUE, general = 0,
                    verbose = FALSE), "at least one general loading")
  expect_error(pefa(d$Q0, d$Y, 2, 3, bifactor = TRUE,
                    general = rep(1L, 3L), verbose = FALSE), "length-J")
})


# ---- public surface ---------------------------------------------------

test_that("the formal surface matches the technical note exactly", {
  f <- formals(pefa)
  expect_identical(names(f)[1:4], c("Q0", "Y", "Kmin", "Kmax"))
  expect_identical(names(f),
                   c("Q0", "Y", "Kmin", "Kmax", "bifactor", "general", "v0",
                     "max_it", "convChk", "tolVal", "tau", "rank_adjust",
                     "rank_max_J", "verbose"))
  expect_identical(f$bifactor, FALSE)
  expect_identical(eval(f$general), 1)
  expect_identical(eval(f$v0), c(.01, .005, .002, .001))
  expect_identical(eval(f$max_it), 10000)
  expect_identical(f$convChk, FALSE)
  expect_identical(eval(f$tolVal), 1e-4)
  expect_identical(eval(f$tau), .50)
  expect_identical(f$rank_adjust, FALSE)
  expect_identical(eval(f$rank_max_J), 100)
  expect_identical(f$verbose, TRUE)

  ## No selector, cut, stability, checkpoint, save or dots argument survives.
  expect_false(any(c("...", "cuts", "sustain", "delta", "stability_eps",
                     "checkpoint", "resume", "save", "file", "select",
                     "ld", "Qe", "orthogonal") %in% names(f)))
})

test_that("every removed API is gone from the namespace and the exports", {
  ns <- asNamespace("vbpm")
  removed <- c("select_K_elbow", "verify_pefa", "pair_detail", "extend_pefa",
               "persistence", "persistence.pefa", "pefa_provenance",
               "pefa_evidence")
  for (nm in removed) {
    expect_false(exists(nm, envir = ns, inherits = FALSE),
                 label = paste0("namespace object ", nm))
    expect_false(nm %in% getNamespaceExports("vbpm"),
                 label = paste0("export ", nm))
  }
  ## The exported surface is exactly the delivered one.
  expect_identical(sort(getNamespaceExports("vbpm")),
                   sort(c("fit_stats", "pefa", "sim_fa", "sim_lvm",
                          "special_effects", "ssl", "vbfa", "vbmimic")))
  ## The retired matcher/provenance internals are gone too.
  for (nm in c(".pefa_pair_facts", ".pefa_aligned_rmsd", ".pefa_provenance",
               ".pefa_verify", ".pefa_hash", ".pefa_status")) {
    expect_false(exists(nm, envir = ns, inherits = FALSE), label = nm)
  }
})

test_that("ssl() is exactly colSums(L^2) and never screens", {
  d <- .mk()
  r <- .fit3(d)
  s <- ssl(r)
  expect_type(s, "list")
  expect_identical(names(s), names(r$loadings))
  for (k in names(s)) {
    expect_type(s[[k]], "double")
    expect_identical(s[[k]], colSums(r$loadings[[k]]^2))
    expect_identical(names(s[[k]]), colnames(r$loadings[[k]]))
  }
  ## Bifactor keeps the general column first and reports K + 1 entries.
  rb <- pefa(d$Q0, d$Y, 2, 3, bifactor = TRUE, v0 = .001, max_it = 300,
             verbose = FALSE)
  sb <- ssl(rb)
  expect_identical(lengths(sb), c("2" = 3L, "3" = 4L))
  expect_identical(names(sb[["3"]]), c("G", "F1", "F2", "F3"))
  expect_identical(sb[["3"]], colSums(rb$loadings[["3"]]^2))

  expect_error(ssl(unclass(r)), "must be a \"pefa\" object")
})


# ---- covered fitting cases --------------------------------------------

test_that("a fully exploratory window (K0 = 0) sweeps and compares", {
  d <- .mk()
  r <- pefa(matrix(0L, d$J, 0L), d$Y, 1, 3, v0 = .001, max_it = 300,
            verbose = FALSE)
  expect_identical(dim(r$Q0), c(d$J, 0L))
  expect_identical(r$sweep$K, 1:3)
  expect_identical(nrow(r$transitions), 2L)
  expect_true(all(is.finite(r$transitions$rmsd)))
  ## With no backbone every source column is exploratory, so a positive
  ## ELBO gain exists and the percentage path is populated.
  expect_true(any(is.finite(r$transitions$ELBO_gain_pct)))
  expect_identical(max(r$transitions$ELBO_gain_pct), 100)
  ## Every all-pairs cell is still a direct comparison.
  expect_identical(r$persistence$phi["1", "3"], .recompare(r, 1, 3)$phi_min)
})

test_that("a backbone-only source window is a legal edge", {
  ## Kmin == K0, so the K = 2 candidate has no exploratory source column and
  ## every pair leaving it is pure position matching.
  d <- .mk()
  r <- .fit3(d)
  expect_identical(ncol(r$Q0), 2L)
  expect_identical(min(r$sweep$K), 2L)
  expect_false(r$transitions$collision[1L])
  expect_true(is.finite(r$transitions$unmatched_ssl[1L]))
})

test_that("a bifactor sweep strips the general column without storing less", {
  d <- .mk()
  r <- pefa(d$Q0, d$Y, 2, 4, bifactor = TRUE, v0 = .001, max_it = 300,
            verbose = FALSE)
  expect_identical(r$sweep$K, 2:4)
  ## K counts group factors; the stored matrices keep the general column.
  expect_identical(vapply(r$loadings, ncol, integer(1)),
                   c("2" = 3L, "3" = 4L, "4" = 5L))
  expect_identical(vapply(r$pips, ncol, integer(1)),
                   c("2" = 3L, "3" = 4L, "4" = 5L))
  expect_identical(colnames(r$loadings[["3"]]), c("G", "F1", "F2", "F3"))
  ## Comparisons operate on the group block only, which is what makes them
  ## the same measurement as an ordinary sweep's.
  expect_identical(r$transitions$phi_min[1L], .recompare(r, 2, 3)$phi_min)
  expect_identical(r$persistence$rmsd["2", "4"], .recompare(r, 2, 4)$rmsd_max)
  ## Stripping is done on a copy: the stored matrices still have K + 1 columns.
  expect_identical(ncol(r$loadings[["4"]]), 5L)
})


# ---- display methods --------------------------------------------------

test_that("print, summary and print.summary read only stored fields", {
  d <- .mk()
  r <- .fit3(d)
  expect_output(print(r), "PEFA sweep \\(ordinary\\): K = 2:4, K0 = 2")
  expect_output(print(r), "sweep 3 x 11 \\| transitions 2 x 11")
  expect_output(print(r), "all candidates converged")
  invisible(capture.output(shown <- withVisible(print(r))))
  expect_false(shown$visible)

  s <- summary(r)
  expect_s3_class(s, "summary.pefa")
  expect_identical(names(s),
                   c("window", "sweep", "transitions", "persistence", "ssl",
                     "settings", "nonconverged_K"))
  expect_identical(s$window, c(Kmin = 2L, Kmax = 4L, K0 = 2L))
  expect_identical(s$sweep, r$sweep)
  expect_identical(s$transitions, r$transitions)
  expect_identical(s$persistence, r$persistence)
  expect_identical(s$ssl, ssl(r))
  expect_identical(s$settings, r$settings)
  expect_identical(s$nonconverged_K, integer(0))

  out <- paste(capture.output(print(s)), collapse = "\n")
  expect_match(out, "Candidates:")
  expect_match(out, "SS loadings")
  expect_match(out, "Transitions:")
  expect_match(out, "Persistence")
  ## phi is displayed before rmsd.
  expect_lt(regexpr("phi \\(phi_min\\)", out),
            regexpr("rmsd \\(rmsd_max\\)", out))
  ## No method claims a count or a threshold.
  expect_false(grepl("selected|recommend|threshold|cut", out,
                     ignore.case = TRUE))
})

test_that("a collided persistence cell is marked for display only", {
  r <- .fit3()
  marked <- r
  marked$persistence$collision["2", "3"] <- TRUE
  out <- paste(capture.output(print(summary(marked))), collapse = "\n")
  expect_match(out, "\\*")
  expect_match(out, "target reuse")
  ## The stored triangles keep their storage modes: the marker is text only.
  expect_type(marked$persistence$phi, "double")
  expect_type(marked$persistence$collision, "logical")
  expect_identical(marked$persistence$phi, r$persistence$phi)

  ## With no collision anywhere the legend line is not printed.
  clean <- paste(capture.output(print(summary(r))), collapse = "\n")
  expect_false(grepl("target reuse", clean))
})

test_that("the compact displays run on bifactor and one-candidate objects", {
  d <- .mk()
  rb <- pefa(d$Q0, d$Y, 2, 3, bifactor = TRUE, v0 = .001, max_it = 300,
             verbose = FALSE)
  expect_output(print(rb), "bifactor")
  expect_output(print(rb), "group factors \\(\\+1 general\\)")
  expect_output(print(summary(rb)), "SS loadings .*general column first")

  r1 <- pefa(d$Q0, d$Y, 2, 2, v0 = .001, max_it = 300, verbose = FALSE)
  expect_output(print(r1), "sweep 1 x 11 \\| transitions 0 x 11")
  expect_output(print(summary(r1)), "single candidate")

  tmp <- tempfile(fileext = ".png")
  grDevices::png(tmp, width = 500, height = 400)
  on.exit({
    grDevices::dev.off()
    unlink(tmp)
  }, add = TRUE)
  ## A one-candidate gain view is an empty panel rather than an error.
  expect_silent(plot(r1, type = "gain", criterion = "ELBO"))
  expect_silent(plot(rb, type = "objective", criterion = "BIC"))
  expect_silent(plot(rb, type = "gain", criterion = "BIC", pct = TRUE))
  expect_silent(plot(rb, type = "fit"))
  ## The gain view has no AIC path, and no removed criterion is accepted.
  expect_error(plot(rb, type = "gain", criterion = "AIC"), "unavailable")
  expect_error(plot(rb, type = "gain", criterion = "BIC_S"), "unavailable")
  expect_error(plot(rb, type = "stability"), "arg")
})

## Latent-variable-model data generator, absorbed from LAWBL::sim_lvm (Chen,
## CRAN 1.5.0) as part of retiring LAWBL into vbpm. Changes on absorption:
##   * `phd` added -- correlated structural disturbances, so the MIMIC generator
##     of Jin & Chen (2025) is covered (its factor residual covariance U is
##     unrestricted; LAWBL's was always diagonal).
##   * missingness bug fixed -- the N x J mask was applied to an N x (J + P)
##     matrix, so R recycled it and NAs bled into the covariate block.
##   * `P > 0` no longer forces the covariates to share `phi` with the factors;
##     `phx` controls them separately (default = `phi` for back-compatibility).
##   * input validation, namespace imports, no partial-argument reliance.
## The generating equations are otherwise unchanged.

#' Simulate data from a latent variable model
#'
#' Generates factor-analytic or item-response data with optional observed and
#' latent predictors (MIMIC / structural models), cross-loadings, local
#' dependence, categorical or mixed response formats, and missingness.
#'
#' `sim_lvm()` and [sim_fa()] are complementary rather than interchangeable.
#' Both accept a population loading matrix directly (`mla`, identical
#' semantics in either function); where they differ is what drives the
#' pattern-based path when `mla = NULL`, and what else each one models:
#'
#' \describe{
#'   \item{`sim_lvm()`}{is the *model-side* generator. Use it when the design
#'     involves **predictors** -- observed covariates (`P`, `b`) or latent
#'     predictors (`K1`, `ph1`, `b1`) -- or **mixed response formats** per
#'     item (`ilvl`). This is the generator for [vbmimic()].}
#'   \item{[sim_fa()]}{is the *structure-side* generator. Use it when the
#'     question is about the loading pattern itself: items per factor (`ipf`),
#'     alternating-sign cross-loadings (`alt_sign`), or minor factors
#'     (`minor`, definition-based per Auerswald & Moshagen, 2019). This is the
#'     generator for [vbfa()].}
#' }
#'
#' The two generators share `ph12` for the factor 1-2 correlation. `ph1`
#' exists only in `sim_lvm()`, where it means something different: the
#' correlation **among latent predictors** (`K1`), not the factor 1-2
#' correlation.
#'
#' @param N Sample size.
#' @param mla Population loading matrix. If supplied, `K`, `J` and `cpf` are
#'   taken from it.
#' @param K Number of factors (used when `mla = NULL`). Must divide `J`.
#' @param J Number of items (used when `mla = NULL`).
#' @param cpf Number of cross-loadings per factor (used when `mla = NULL`).
#' @param lam Value of the primary loadings.
#' @param lac Value of the cross-loadings.
#' @param phi Homogeneous correlation between any two factors.
#' @param ph12 Correlation between factors 1 and 2, when it differs from `phi`;
#'   `-1` (default) to use `phi`.
#' @param phx Correlation among the observed covariates. Defaults to `phi`,
#'   which is LAWBL's behaviour; set `phx = 0` for uncorrelated covariates.
#' @param ecr Residual correlation (local dependence).
#' @param P Number of observed predictors (MIMIC model); `0` for none.
#' @param b Coefficients of the observed predictors: a scalar, or a
#'   `(K - K1) x P` matrix.
#' @param K1 Number of latent predictors (MIMIC / structural model).
#' @param ph1 Correlation among the latent predictors.
#' @param b1 Coefficients of the latent predictors: a scalar, or a
#'   `(K - K1) x K1` matrix.
#' @param phd Correlation among the structural disturbances (the residuals of
#'   the regression of factors on predictors). `0` (default) gives independent
#'   disturbances, which is LAWBL's behaviour; a positive value reproduces
#'   designs with a freely correlated factor residual covariance. Ignored when
#'   `P = 0` and `K1 = 0`.
#' @param ilvl Levels of every item, length `J + P`; values below 2 mean
#'   continuous. `NULL` to use `cati`/`noc` instead.
#' @param cati Polytomous items by index; `NULL` for none, `-1` for all. Used
#'   only when `ilvl = NULL`.
#' @param noc Number of categories for the polytomous items.
#' @param misp Proportion of missingness, applied to the `J` items only (never
#'   to the covariate block).
#' @param ome_out Return the factor scores.
#' @param necw Number of within-factor local dependencies.
#' @param necb Number of between-factor local dependencies.
#' @param add_ind Indices of items loading on an additional minor factor.
#' @param add_la Loading value on the additional minor factor.
#' @param add_phi Correlation between the additional minor factor and the
#'   others.
#' @param zero_it Number of surplus items with zero loadings.
#' @param rseed Random seed.
#' @param digits Significant digits used while the function is running.
#'
#' @return A list with the data (`dat`; items first, then any covariates),
#'   the loading matrix (`MLA`), factor correlations (`PHI`), residual
#'   covariance (`PSX`), the indices of the local-dependence pairs
#'   (`ofd_ind`), and, when predictors are present, the coefficient matrices
#'   (`mb`, `mb1`) and covariate correlation (`PHX`).
#'
#' @references
#' Chen, J. (2022). *LAWBL: Latent (variable) analysis with Bayesian
#' learning* (R package version 1.5.0).
#'
#' Jin, Y., & Chen, J. (2025). Regularized variational Bayesian approximations
#' for variable selection in extended multiple-indicators multiple-causes
#' models. *Multivariate Behavioral Research*.
#' \doi{10.1080/00273171.2025.2483253}
#'
#' @seealso [sim_fa()] for the structure-side generator, [vbmimic()] and
#'   [vbfa()] for the estimators.
#'
#' @examples
#' ## MIMIC data: 18 items on 3 factors, 9 covariates in disjoint blocks of 3.
#' ## Supply `b` as a matrix for a sparse structural design; a scalar `b` makes
#' ## every covariate predict every factor, which quickly leaves no disturbance
#' ## variance (the function stops if it does).
#' B <- matrix(0, 3, 9)
#' for (k in 1:3) B[k, ((k - 1) * 3 + 1):(k * 3)] <- .3
#'
#' s <- sim_lvm(N = 500, K = 3, J = 18, P = 9, b = B, phx = 0, rseed = 1)
#' dim(s$dat)      # 500 x 27 -- items first, covariates last
#' s$mb            # the generating structural coefficients
#'
#' ## correlated disturbances (unrestricted factor residual covariance)
#' s2 <- sim_lvm(N = 500, K = 3, J = 18, P = 9, b = B, phx = 0, phd = .3,
#'               rseed = 1)
#' round(s2$PHI, 2)
#'
#' ## matrix-driven: supply the population loading matrix directly
#' mla <- matrix(0, 12, 2)
#' mla[1:6, 1] <- .7; mla[7:12, 2] <- c(.8, .7, .6, .7, .8, .6)
#' s3 <- sim_lvm(N = 300, mla = mla, phi = .3, rseed = 4)
#' s3$MLA                     # the generating matrix, as supplied
#'
#' @importFrom MASS mvrnorm
#' @export
sim_lvm <- function(N = 1000, mla = NULL, K = 3, J = 18, cpf = 0,
                    lam = 0.7, lac = 0.3, phi = 0.3, ph12 = -1, phx = NULL,
                    ecr = 0, P = 0, b = 0.3, K1 = 0, ph1 = 0.2, b1 = 0.3,
                    phd = 0, ilvl = NULL, cati = NULL, noc = c(4), misp = 0,
                    ome_out = FALSE, necw = K, necb = K, add_ind = c(),
                    add_la = 0.5, add_phi = 0, zero_it = 0, rseed = 333,
                    digits = 4) {

  if (is.null(mla) && J %% K != 0)
    stop("J should be a multiple of K.", call. = FALSE)
  if (K1 > 0 && K1 >= K)
    stop("K1 (latent predictors) must be smaller than K.", call. = FALSE)
  if (phd < 0 || phd >= 1)
    stop("phd must be in [0, 1).", call. = FALSE)
  if (is.null(phx)) phx <- phi

  if (exists(".Random.seed", .GlobalEnv))
    oldseed <- .GlobalEnv$.Random.seed
  else oldseed <- NULL
  set.seed(rseed)

  oo <- options()
  on.exit(options(oo))
  options(digits = digits)

  if (!is.null(mla)) {
    mla <- as.matrix(mla)
    K   <- ncol(mla)
    J   <- nrow(mla)
    cpf <- 0
  }

  x <- NULL

  if (P == 0 && K1 == 0) {
    ## ---- plain factor model ------------------------------------------
    PHI <- matrix(phi, K, K)
    diag(PHI) <- 1
    if (ph12 > -1 && ph12 < 1) PHI[1, 2] <- PHI[2, 1] <- ph12
    PHX <- NULL
    eta <- mvrnorm(N, rep(0, K), PHI)
  } else {
    ## ---- MIMIC / structural model -------------------------------------
    mb <- mb1 <- PHX <- NULL
    tmp <- eta <- 0
    Kd  <- K - K1                 # factors that are regressed on predictors

    if (P > 0) {
      PHX <- matrix(phx, P, P)
      diag(PHX) <- 1
      x <- mvrnorm(N, rep(0, P), PHX)
      mb <- if (length(b) == 1) matrix(b, Kd, P) else as.matrix(b)
      if (nrow(mb) != Kd || ncol(mb) != P)
        stop("b must be a scalar or a (K - K1) x P matrix.", call. = FALSE)
      tmp <- mb %*% PHX %*% t(mb)
      eta <- t(mb %*% t(x))
    }

    if (K1 > 0) {
      PH1 <- matrix(ph1, K1, K1)
      diag(PH1) <- 1
      ome <- mvrnorm(N, rep(0, K1), PH1)
      mb1 <- if (length(b1) == 1) matrix(b1, Kd, K1) else as.matrix(b1)
      if (nrow(mb1) != Kd || ncol(mb1) != K1)
        stop("b1 must be a scalar or a (K - K1) x K1 matrix.", call. = FALSE)
      tmp <- tmp + mb1 %*% PH1 %*% t(mb1)
      eta <- eta + t(mb1 %*% t(ome))
    }

    ## structural disturbances: variances make each factor unit-variance;
    ## phd > 0 correlates them (an unrestricted factor residual covariance).
    dvar <- 1 - diag(tmp)
    if (any(dvar <= 0))
      stop("Predictor coefficients imply non-positive disturbance variance; ",
           "reduce b/b1.", call. = FALSE)
    if (Kd > 1) {
      DIS <- diag(dvar)
      if (phd > 0) {
        sdv <- sqrt(dvar)
        DIS <- phd * outer(sdv, sdv)
        diag(DIS) <- dvar
      }
    } else {
      DIS <- dvar
    }
    eta <- eta + mvrnorm(N, rep(0, Kd), DIS)
    if (K1 > 0) eta <- cbind(eta, ome)
  }

  ## ---- loading matrix ---------------------------------------------------
  ipf <- J / K
  if (is.null(mla)) {
    if (K > 1) {
      lam0 <- c(rep(lam, ipf), rep(0, ipf - cpf), rep(lac, cpf),
                rep(0, (K - 2) * ipf))
    } else {
      lam0 <- lam
    }
    lam1 <- matrix(lam0, ipf, K)
    if (K == 1) {
      mla <- lam1
    } else {
      mla <- c()
      for (i in K:1) {
        ind <- (c(1:K) + i) %% K
        ind[ind == 0] <- K
        mla <- rbind(mla, lam1[, ind])
      }
    }
    if (P == 0 && K1 == 0) {
      Kp1 <- K
      if (!is.null(add_ind)) {
        Kp1 <- K + 1
        mla <- cbind(mla, 0)
        mla[add_ind, Kp1] <- add_la
        PHI <- cbind(PHI, add_phi)
        PHI <- rbind(PHI, add_phi)
        PHI[Kp1, Kp1] <- 1
      }
      if (zero_it > 0) {
        J <- J + zero_it
        mla <- rbind(mla, matrix(0, zero_it, Kp1))
      }
      eta <- mvrnorm(N, rep(0, Kp1), PHI)
    }
  }

  ## ---- residual covariance (local dependence) ---------------------------
  ecm <- matrix(0, J, J)
  if (ecr > 0) {
    li <- c(1:K) * ipf - cpf
    for (i in 1:necw) {
      if (necw > 0) ecm[li[i], li[i] - 1] <- ecm[li[i] - 1, li[i]] <- ecr
    }
    for (i in 1:necb) {
      i1 <- i %% K + 1
      if (necb > 0)
        ecm[li[i] - 2, li[i1] - 3] <- ecm[li[i1] - 3, li[i] - 2] <- ecr
    }
  }

  PHI <- cor(eta)
  evr <- 1 - diag(mla %*% PHI %*% t(mla))
  diag(ecm) <- evr
  er <- mvrnorm(N, rep(0, J), ecm)
  y  <- t(mla %*% t(eta)) + er
  if (P > 0) y <- cbind(y, x)

  scale <- apply(y, 2, sd)
  eigen <- diag(crossprod(mla))
  out <- list(N = N, PHI = PHI, MLA = mla, Eigen = eigen, PSX = ecm,
              scale = scale)
  if (ome_out) out$OME <- eta

  pos  <- lower.tri(ecm)
  ind  <- which(pos, arr.ind = TRUE)
  rind <- which(ecm[pos] > 0)
  out$ofd_ind <- ind[rind, ]

  ## ---- categorization ----------------------------------------------------
  UL <- 3
  LL <- -3
  if (is.null(ilvl)) {
    Jp <- length(cati)
    if (Jp > 0) {
      if (Jp == 1 && cati == -1) {
        cati <- c(1:J)
        Jp <- J
      }
      yc  <- y[, cati]
      len <- length(noc)
      if (Jp %% len != 0)
        stop("cati is not divisable by length(noc).", call. = FALSE)
      for (i in 1:len) {
        M   <- noc[i]
        stp <- LL + c(1:(M - 1)) * (UL - LL) / M
        i0  <- (i - 1) * Jp / len + 1
        i1  <- i * Jp / len
        for (j in i0:i1) {
          tmp2 <- yc[, j] > matrix(stp, N, (M - 1), byrow = TRUE)
          yc[, j] <- rowSums(tmp2) + 1
        }
      }
      y[, cati] <- yc
      out$noc  <- noc
      out$cati <- cati
    }
  } else {
    if (length(ilvl) != J + P)
      stop("number of item levels should be J + P.", call. = FALSE)
    for (i in 1:(J + P)) {
      M <- ilvl[i]
      if (M >= 2) {
        M    <- round(M)
        stp  <- LL + c(1:(M - 1)) * (UL - LL) / M
        tmp2 <- y[, i] > matrix(stp, N, (M - 1), byrow = TRUE)
        y[, i] <- rowSums(tmp2) + 1
      }
    }
  }

  ## ---- missingness -------------------------------------------------------
  ## applied to the J items only. LAWBL built an N x J mask and indexed an
  ## N x (J + P) matrix with it, so the logical vector was recycled and NAs
  ## appeared in the covariate block; the explicit column restriction below
  ## is the fix.
  if (misp > 0) {
    mind <- matrix(rbinom(N * J, 1, misp), N, J)
    yi   <- y[, 1:J, drop = FALSE]
    yi[mind == 1] <- NA
    y[, 1:J] <- yi
  }
  out$misp <- misp
  out$dat  <- y

  if (P != 0 || K1 != 0) {
    out$mb  <- mb
    out$mb1 <- mb1
    out$PHX <- PHX
  }

  if (!is.null(oldseed))
    .GlobalEnv$.Random.seed <- oldseed
  else rm(".Random.seed", envir = .GlobalEnv)

  out
}

## Description and display for the object returned by pefa().  Everything in
## this file reads stored fields: nothing here refits, rematches columns, or
## applies an analysis rule, so a display can never disagree with $transitions
## or $persistence.

#' Sums of squared loadings for a PEFA sweep
#'
#' Descriptive column sizes for every candidate in a [pefa()] sweep:
#' `colSums(Lam^2)` on the stored loading matrix of each candidate. They are
#' not Gram eigenvalues, they take part in no correspondence or selection
#' rule, and no cutoff is applied to them. `summary()` calls this function, so
#' there is one definition.
#'
#' A low `phi_min` in `$transitions` or `$persistence$phi` is a weakest-link
#' reading over every source column, including one the fit has driven to
#' nearly zero. Such a column matches the smallest-norm target and carries an
#' essentially arbitrary congruence. `ssl()` is what separates "this candidate
#' contains a column with nothing in it" from "the retained structure moved".
#'
#' @param x A result returned by [pefa()].
#'
#' @return A list with one element per candidate, named as `x$loadings` is
#'   (the candidate `K` as a character string). Each element is the double
#'   vector `colSums(L^2)` of that candidate's loading matrix and keeps the
#'   stored loading-column names when the matrix has them. In bifactor mode
#'   each vector covers all `K + 1` stored columns with the general factor
#'   first, so the group-block scale used by every between-candidate
#'   comparison is `ssl(x)[[k]][-1]`.
#'
#' @seealso [pefa()]
#'
#' @export
ssl <- function(x) {
  if (!inherits(x, "pefa")) {
    stop("`x` must be a \"pefa\" object returned by pefa().", call. = FALSE)
  }
  loadings <- x[["loadings"]]
  if (!is.list(loadings)) {
    stop("`x$loadings` must be the named list of candidate loading matrices.",
         call. = FALSE)
  }
  lapply(loadings, function(m) colSums(m^2))
}


# ---- private display helpers ------------------------------------------

.pefa_disp_window <- function(x) {
  K <- x[["sweep"]][["K"]]
  c(Kmin = as.integer(min(K)), Kmax = as.integer(max(K)),
    K0 = as.integer(ncol(x[["Q0"]])))
}

## which() rather than a logical index so a defensive NA cannot silently
## widen the reported set.
.pefa_disp_nonconverged <- function(sweep) {
  as.integer(sweep[["K"]][which(!sweep[["converged"]])])
}

.pefa_disp_mode <- function(settings) {
  if (isTRUE(settings[["bifactor"]])) "bifactor" else "ordinary"
}

.pefa_disp_general <- function(settings) {
  if (isTRUE(settings[["bifactor"]])) " group factors (+1 general)" else ""
}

## Display copy only.  The stored numeric and logical triangles keep their
## own storage modes, so a "*" marker can never leak back into the object.
.pefa_disp_tri <- function(values, collision, digits) {
  out <- matrix("", nrow(values), ncol(values),
                dimnames = dimnames(values))
  upper <- upper.tri(values)
  if (any(upper)) {
    flag <- collision[upper]
    out[upper] <- paste0(formatC(values[upper], format = "f",
                                 digits = digits),
                         ifelse(!is.na(flag) & flag, "*", ""))
  }
  out
}

.pefa_disp_persistence <- function(persistence, digits) {
  phi <- persistence[["phi"]]
  rmsd <- persistence[["rmsd"]]
  collision <- persistence[["collision"]]
  cat("\nPersistence (direct K_from x K_to comparisons):\n")
  cat("  phi (phi_min):\n")
  print(.pefa_disp_tri(phi, collision, digits), quote = FALSE, right = TRUE)
  cat("  rmsd (rmsd_max):\n")
  print(.pefa_disp_tri(rmsd, collision, digits), quote = FALSE, right = TRUE)
  ## One legend line, and only when the reader can actually see a marker.
  if (any(upper.tri(collision) & !is.na(collision) & collision)) {
    cat("  * = target reuse (collision)\n")
  }
  invisible(NULL)
}

## A single K gives range() a degenerate interval, which is not a usable
## plotting window.
.pefa_disp_xlim <- function(K) {
  if (length(K) == 1L) return(c(K - 0.5, K + 0.5))
  range(K)
}

.pefa_disp_range <- function(values, top = 0.20) {
  values <- values[is.finite(values)]
  if (!length(values)) return(c(0, 1))
  r <- range(values)
  span <- diff(r)
  if (span == 0) span <- abs(r[1L]) * 0.01 + 1
  c(r[1L] - 0.05 * span, r[2L] + top * span)
}

## The formal default carries every objective criterion, so a gain view has
## to narrow an untouched default rather than reject it.
.pefa_disp_criterion <- function(criterion, type) {
  allowed <- if (identical(type, "gain")) {
    c("ELBO", "BIC")
  } else c("ELBO", "AIC", "BIC")
  if (!is.character(criterion) || !length(criterion) || anyNA(criterion)) {
    stop("criterion must be a criterion name.", call. = FALSE)
  }
  chosen <- toupper(criterion)
  if (length(chosen) > 1L) {
    if (!identical(chosen, c("ELBO", "AIC", "BIC"))) {
      stop("criterion must be a single criterion name.", call. = FALSE)
    }
    return(allowed[[1L]])
  }
  if (!(chosen %in% allowed)) {
    stop(sprintf(
      "criterion = \"%s\" is unavailable for type = \"%s\"; use one of %s.",
      criterion, type, paste0("\"", allowed, "\"", collapse = ", ")),
      call. = FALSE)
  }
  chosen
}


# ---- methods ----------------------------------------------------------

#' Methods for a PEFA factor-count sweep
#'
#' Descriptive readers for the object returned by [pefa()]. They report what
#' was fitted and measured and nothing else: no method selects, recommends, or
#' marks a factor count, and none draws a threshold, because the package owns
#' no count rule. Count rules, persistence thresholds, and collision policies
#' belong to the analysis that consumes this object.
#'
#' @param object,x A result returned by [pefa()]; for
#'   `print.summary.pefa()`, the object returned by `summary()`.
#' @param ... Further arguments; for `plot()`, graphical arguments passed to
#'   the first base plot in the selected display.
#'
#' @return `print()` and `plot()` return their input invisibly.
#'   `summary()` returns a list of class `"summary.pefa"` with exactly the
#'   components `window` (named integer `Kmin`, `Kmax`, `K0`), `sweep`,
#'   `transitions`, `persistence`, `ssl`, `settings`, and `nonconverged_K`
#'   (an integer vector of the K values whose candidate did not converge,
#'   `integer(0)` when all converged).
#'
#' @seealso [pefa()], [ssl()]
#'
#' @rdname pefa-methods
#' @export
print.pefa <- function(x, ...) {
  w <- .pefa_disp_window(x)
  settings <- x[["settings"]]
  cat(sprintf("PEFA sweep (%s): K = %d:%d%s, K0 = %d\n",
              .pefa_disp_mode(settings), w[["Kmin"]], w[["Kmax"]],
              .pefa_disp_general(settings), w[["K0"]]))
  phi <- x[["persistence"]][["phi"]]
  cat(sprintf(
    "  sweep %d x %d | transitions %d x %d | persistence 3 x %d x %d\n",
    nrow(x[["sweep"]]), ncol(x[["sweep"]]),
    nrow(x[["transitions"]]), ncol(x[["transitions"]]),
    nrow(phi), ncol(phi)))
  bad <- .pefa_disp_nonconverged(x[["sweep"]])
  if (length(bad)) {
    cat("  nonconverged K: ", paste(bad, collapse = ", "), "\n", sep = "")
  } else {
    cat("  all candidates converged\n")
  }
  invisible(x)
}

#' @rdname pefa-methods
#' @export
summary.pefa <- function(object, ...) {
  out <- list(
    window = .pefa_disp_window(object),
    sweep = object[["sweep"]],
    transitions = object[["transitions"]],
    persistence = object[["persistence"]],
    ssl = ssl(object),
    settings = object[["settings"]],
    nonconverged_K = .pefa_disp_nonconverged(object[["sweep"]])
  )
  class(out) <- "summary.pefa"
  out
}

#' @param digits Number of decimals used when printing measured quantities.
#'
#' @rdname pefa-methods
#' @export
print.summary.pefa <- function(x, digits = 3, ...) {
  settings <- x[["settings"]]
  w <- x[["window"]]
  cat(sprintf("PEFA sweep summary (%s): K = %d:%d%s, K0 = %d\n",
              .pefa_disp_mode(settings), w[["Kmin"]], w[["Kmax"]],
              .pefa_disp_general(settings), w[["K0"]]))

  cat("\nCandidates:\n")
  candidates <- x[["sweep"]]
  ## Criteria and indices live on different scales; rounding each to its own
  ## readable width keeps the table narrow without hiding a real difference.
  for (nm in intersect(c("ELBO", "AIC", "BIC"), names(candidates))) {
    candidates[[nm]] <- round(candidates[[nm]], 1L)
  }
  for (nm in intersect(c("RMSEA", "SRMR", "CFI", "TLI"),
                       names(candidates))) {
    candidates[[nm]] <- round(candidates[[nm]], digits)
  }
  print(candidates, row.names = FALSE)

  cat("\nSS loadings (colSums(Lam^2)",
      if (isTRUE(settings[["bifactor"]])) "; general column first" else "",
      "):\n", sep = "")
  for (nm in names(x[["ssl"]])) {
    cat(sprintf("  K=%s: %s\n", nm,
                paste(sprintf("%.2f", x[["ssl"]][[nm]]), collapse = " ")))
  }

  cat("\nTransitions:\n")
  transitions <- x[["transitions"]]
  if (nrow(transitions)) {
    num <- vapply(transitions, is.numeric, logical(1))
    transitions[num] <- lapply(transitions[num], round, digits)
    print(transitions, row.names = FALSE)
  } else {
    cat("  none: the window holds a single candidate\n")
  }

  .pefa_disp_persistence(x[["persistence"]], digits)
  invisible(x)
}

#' @param type Display: criterion trajectories (`"objective"`), one gain
#'   trajectory (`"gain"`), or descriptive absolute and incremental fit
#'   trajectories (`"fit"`).
#' @param criterion Criterion path. `"ELBO"`, `"AIC"`, or `"BIC"` for
#'   `type = "objective"`; `"ELBO"` or `"BIC"` for `type = "gain"`, which has
#'   no AIC path. Ignored when `type = "fit"`.
#' @param pct Gain view only. `FALSE` (the default) computes the raw oriented
#'   gains `ELBO_to - ELBO_from` or `BIC_from - BIC_to` from `$sweep`;
#'   `TRUE` reads the stored `ELBO_gain_pct` or `BIC_gain_pct` column of
#'   `$transitions`, which is `NA` for a path with no positive gain.
#'
#' @rdname pefa-methods
#' @importFrom graphics axis legend lines mtext par plot
#' @export
plot.pefa <- function(x, type = c("objective", "gain", "fit"),
                      criterion = c("ELBO", "AIC", "BIC"),
                      pct = FALSE, ...) {
  type <- match.arg(type)
  ## The fit view has no criterion path, so it ignores the argument outright
  ## rather than rejecting a value it would not have read.
  if (type != "fit") criterion <- .pefa_disp_criterion(criterion, type)
  if (length(pct) != 1L || !is.logical(pct) || is.na(pct)) {
    stop("pct must be TRUE or FALSE.", call. = FALSE)
  }
  d <- x[["sweep"]]
  K <- d[["K"]]
  col1 <- "#1f77b4"
  col2 <- "#d62728"
  xlim <- .pefa_disp_xlim(K)
  xlab <- if (isTRUE(x[["settings"]][["bifactor"]])) {
    "Number of group factors (K)"
  } else "Number of factors (K)"

  if (type == "objective") {
    better <- if (criterion == "ELBO") {
      "higher is better"
    } else "lower is better"
    plot(K, d[[criterion]], type = "b", pch = 16, col = col1, xlim = xlim,
         ylim = .pefa_disp_range(d[[criterion]]), xaxt = "n", xlab = xlab,
         ylab = paste0(criterion, " (", better, ")"),
         main = paste("PEFA", criterion, "path"), ...)
    axis(1, at = K)
  } else if (type == "gain") {
    ylab <- if (pct) {
      "Gain (% of largest positive gain)"
    } else "Oriented gain (positive favors K_to)"
    main <- paste("PEFA", criterion, if (pct) "gain (%)" else "gain", "path")
    if (nrow(d) < 2L) {
      ## A one-candidate window has no adjacent pair.  That is a legal
      ## window, so the gain view stays an empty panel instead of erroring.
      plot(K, rep(NA_real_, length(K)), type = "n", xlim = xlim,
           ylim = c(0, 1), xaxt = "n", xlab = xlab, ylab = ylab,
           main = main, ...)
      axis(1, at = K)
      mtext("no adjacent pair in a one-candidate window", side = 3,
            line = 0.2, cex = 0.9)
    } else {
      transitions <- x[["transitions"]]
      gains <- if (pct) {
        transitions[[paste0(criterion, "_gain_pct")]]
      } else if (criterion == "ELBO") {
        diff(d[["ELBO"]])
      } else -diff(d[["BIC"]])
      at <- if (pct) transitions[["K_to"]] else K[-1L]
      ## No reference line is drawn anywhere in this method, so zero is held
      ## inside the window instead: the sign of a gain is what a reader needs.
      plot(at, gains, type = "b", pch = 16, col = col1, xlim = xlim,
           ylim = .pefa_disp_range(c(gains, 0)), xaxt = "n", xlab = xlab,
           ylab = ylab, main = main, ...)
      axis(1, at = K)
    }
  } else {
    op <- par(mfrow = c(1, 2), mar = c(5, 4.2, 3, 1) + 0.1,
              oma = c(0, 0, 2, 0))
    on.exit(par(op))
    plot(K, d[["RMSEA"]], type = "b", pch = 16, col = col1, xlim = xlim,
         ylim = .pefa_disp_range(c(d[["RMSEA"]], d[["SRMR"]])), xaxt = "n",
         xlab = xlab, ylab = "Index (lower is better)",
         main = "Absolute misfit", ...)
    axis(1, at = K)
    lines(K, d[["SRMR"]], type = "b", pch = 17, lty = 2, col = col2)
    legend("top", c("RMSEA", "SRMR"), col = c(col1, col2), lty = c(1, 2),
           pch = c(16, 17), bty = "n", cex = 0.9)

    plot(K, d[["CFI"]], type = "b", pch = 16, col = col1, xlim = xlim,
         ylim = .pefa_disp_range(c(d[["CFI"]], d[["TLI"]])), xaxt = "n",
         xlab = xlab, ylab = "Index (higher is better)",
         main = "Incremental fit")
    axis(1, at = K)
    lines(K, d[["TLI"]], type = "b", pch = 17, lty = 2, col = col2)
    legend("bottom", c("CFI", "TLI"), col = c(col1, col2), lty = c(1, 2),
           pch = c(16, 17), bty = "n", cex = 0.9)
    mtext("PEFA descriptive fit trajectories", outer = TRUE, font = 2)
  }
  invisible(x)
}

## Special-effect summaries for orthogonal bifactor fits: the Schmid-Leiman
## proportionality check, the (approximate) higher-order parameterization,
## and testlet/special effect sizes. Pure post-processing of a vbfa fit;
## no re-estimation.

#' Special-effect summaries of a bifactor fit
#'
#' Post-processes an orthogonal bifactor [vbfa()] fit into the three
#' summaries used throughout the `bifactor` vignette: (i) the
#' **Schmid-Leiman (SL) proportionality check** -- the within-group
#' coefficient of variation (CV) of the group-to-general loading ratio, per
#' group factor and item-weighted overall; (ii) the **(approximate)
#' higher-order parameterization** -- second-order loadings
#' \eqn{\hat\gamma_k = 1/\sqrt{1 + \bar r_k^2}} from each group's mean ratio
#' \eqn{\bar r_k}, and first-order loadings
#' \eqn{\hat\lambda_j = |b^{gen}_j| / \hat\gamma_{k(j)}}; and (iii)
#' **testlet/special effect sizes** by Eq. 16 of Zhang and Chen (2024): the
#' average squared loading on each group factor, counting specified loadings
#' always and unspecified ones only when their posterior inclusion
#' probability exceeds `tau`.
#'
#' All output is **descriptive**. Near-zero CVs signal SL proportionality --
#' the fitted bifactor is then (approximately) a higher-order model, and by
#' the same token a testlet model, so the higher-order parameters are a
#' faithful re-expression. Values near `.1` are conventionally read as
#' approximately higher-order, but the cutoff is suggestive, not a test; the
#' higher-order parameters and effect sizes are reported regardless, and
#' with clearly non-constant ratios they are a deliberately lossy summary of
#' a genuinely richer bifactor. The ratio inherits the sampling softness of
#' the general/group split, so CVs are noisy below (roughly) a few thousand
#' observations -- see the `bifactor` vignette for demonstrations, caveats,
#' and the covariance geometry behind all three summaries.
#'
#' **Which fits are accepted.** A fit with `bifactor = TRUE` is used as-is
#' (general column first, group columns after). A legacy hand-built fit is
#' also accepted when it has `orthogonal = TRUE` and exactly one
#' all-specified column in `Q`, which is taken as the general factor. Group
#' membership of each item is the group column where it is anchored
#' (`Q == 1`) or, for unspecified rows, its single active group column
#' (inclusion probability above `tau`). Items with no active group loading
#' are excluded from the ratio summaries and reported; items active on
#' several group columns are assigned to the anchored one (else the largest
#' absolute loading) and reported. Ratios use absolute loadings; items whose
#' general loading is fixed at zero (bifactor-(S-1)-style designs) carry no
#' ratio and are reported.
#'
#' @param object A fitted [vbfa()] object with a bifactor structure (see
#'   Details).
#' @param tau Hard-selection threshold on the posterior inclusion
#'   probability (default `0.5`), used both for group-membership inference
#'   and for the Eq.-16 effect sizes.
#'
#' @return An object of class `special_effects`: a list with
#'   \item{groups}{A data frame with one row per group factor: `n_items`
#'     (assigned items), `mean_ratio`, `sd_ratio`, `cv`, `gamma`
#'     (second-order loading), and `D` (Eq.-16 effect size).}
#'   \item{mean_cv}{The item-weighted mean CV across group factors.}
#'   \item{items}{A data frame with one row per item: assigned `group`
#'     (`NA` if unassigned), `b_gen`, `b_grp`, `ratio`, and `lambda` (the
#'     first-order loading implied by the higher-order parameterization).}
#'   \item{gamma}{Named vector of second-order loadings (same as in
#'     `groups`).}
#'   \item{unassigned, crossloaded, no_general}{Integer vectors of item
#'     indices flagged by the membership rules above.}
#'   \item{tau, source, call}{Bookkeeping (`source` is `"bifactor"` for
#'     native fits, `"legacy"` for hand-built ones).}
#'
#' @references
#' Schmid, J., & Leiman, J. M. (1957). The development of hierarchical
#' factor solutions. *Psychometrika*, 22(1), 53-61.
#'
#' Yung, Y.-F., Thissen, D., & McLeod, L. D. (1999). On the relationship
#' between the higher-order factor model and the hierarchical factor model.
#' *Psychometrika*, 64(2), 113-128.
#'
#' Zhang, Y., & Chen, J. (2024). Accommodating and extending various models
#' for special effects within the generalized partially confirmatory factor
#' analysis framework. *Applied Psychological Measurement*, 48(4-5),
#' 208-229. \doi{10.1177/01466216241261704}
#'
#' @examples
#' ## higher-order data: the ratios are near-constant and gamma is recovered
#' sim <- sim_fa(N = 500, K = 3, ipf = 6, lam = .8, gamma = c(.8, .7, .6),
#'               rseed = 11)
#' Q <- matrix(-1L, 18, 3)
#' for (k in 1:3) { a <- which(rep(1:3, each = 6) == k)[1:2]
#'                  Q[a, ] <- 0L; Q[a, k] <- 1L }
#' fit <- vbfa(sim$dat, Q, bifactor = TRUE, v0 = .001)
#' special_effects(fit)
#'
#' @seealso [vbfa()], and the `bifactor` vignette for the full workflow.
#' @export
special_effects <- function(object, tau = 0.5) {
  if (!inherits(object, "vbfa"))
    stop("special_effects() needs a vbfa fit.", call. = FALSE)
  if (length(tau) != 1L || !is.finite(tau) || tau < 0 || tau > 1)
    stop("tau must be a single number in [0, 1].", call. = FALSE)
  Q   <- object$Q
  Lam <- object$Lam
  J   <- nrow(Q); P <- ncol(Q)

  ## ---- locate the general and group columns ---------------------------
  if (isTRUE(object$bifactor)) {
    gcol <- 1L
    src  <- "bifactor"
  } else {
    if (!isTRUE(object$orthogonal))
      stop("Not a bifactor fit: the factors are oblique. Fit with ",
           "vbfa(..., bifactor = TRUE).", call. = FALSE)
    full <- which(apply(Q == 1, 2, all))
    if (length(full) != 1L)
      stop("Not recognizably a bifactor fit: expected exactly one ",
           "all-specified column in Q (found ", length(full), "). Fit with ",
           "vbfa(..., bifactor = TRUE) to declare the structure.",
           call. = FALSE)
    gcol <- full
    src  <- "legacy"
  }
  gc <- setdiff(seq_len(P), gcol)
  if (length(gc) < 1L)
    stop("No group factors found alongside the general column.", call. = FALSE)

  ## effective inclusion probabilities (specified -> 1, fixed zero -> 0)
  pi_eff <- object$pi
  pi_eff[Q == 1] <- 1; pi_eff[Q == 0] <- 0

  ## ---- group membership per item ---------------------------------------
  grp <- rep(NA_integer_, J)          # index into gc
  crossloaded <- integer(0)
  for (j in seq_len(J)) {
    act <- which(pi_eff[j, gc] > tau)
    if (length(act) == 1L) grp[j] <- act
    else if (length(act) > 1L) {
      crossloaded <- c(crossloaded, j)
      anc <- act[Q[j, gc[act]] == 1]
      grp[j] <- if (length(anc)) anc[1] else act[which.max(abs(Lam[j, gc[act]]))]
    }
  }
  unassigned <- which(is.na(grp))

  ## ---- ratios, per-group summaries, higher-order parameters ------------
  b_gen <- Lam[, gcol]
  b_grp <- ifelse(is.na(grp), NA_real_, Lam[cbind(seq_len(J), gc[grp])])
  no_general <- which(Q[, gcol] == 0)          # (S-1)-style: no ratio defined
  ratio <- abs(b_grp) / abs(b_gen)
  ratio[no_general] <- NA_real_

  K <- length(gc)
  gnames <- colnames(Q)[gc]
  if (is.null(gnames)) gnames <- paste0("group", seq_len(K))
  n_it <- mr <- sr <- cv <- gam <- D <- rep(NA_real_, K)
  for (k in seq_len(K)) {
    r <- ratio[!is.na(grp) & grp == k & !is.na(ratio)]
    n_it[k] <- length(r)
    if (length(r) >= 1L) mr[k] <- mean(r)
    if (length(r) >= 2L) { sr[k] <- stats::sd(r); cv[k] <- sr[k] / mr[k] }
    if (!is.na(mr[k])) gam[k] <- 1 / sqrt(1 + mr[k]^2)
    ## Eq. 16 (Zhang & Chen, 2024): average squared loading on the group
    ## factor; specified loadings always count, unspecified ones only when
    ## their inclusion probability exceeds tau.
    col <- gc[k]
    inc <- (Q[, col] == 1) | (Q[, col] == -1 & object$pi[, col] > tau)
    if (any(inc)) D[k] <- sum(Lam[inc, col]^2) / sum(inc)
  }
  ok <- !is.na(cv)
  mean_cv <- if (any(ok)) sum(n_it[ok] * cv[ok]) / sum(n_it[ok]) else NA_real_
  lambda <- abs(b_gen) / gam[grp]
  lambda[no_general] <- NA_real_

  out <- list(
    groups = data.frame(group = gnames, n_items = as.integer(n_it),
                        mean_ratio = mr, sd_ratio = sr, cv = cv,
                        gamma = gam, D = D, row.names = NULL),
    mean_cv = mean_cv,
    items = data.frame(item = seq_len(J),
                       group = ifelse(is.na(grp), NA_character_, gnames[grp]),
                       b_gen = b_gen, b_grp = b_grp, ratio = ratio,
                       lambda = lambda, row.names = NULL),
    gamma = stats::setNames(gam, gnames),
    unassigned = unassigned, crossloaded = crossloaded,
    no_general = no_general,
    tau = tau, source = src, call = match.call())
  class(out) <- "special_effects"
  out
}

#' @rdname special_effects
#' @param x A `special_effects` object.
#' @param digits Number of decimals in the printed table.
#' @param ... Unused.
#' @export
print.special_effects <- function(x, digits = 3, ...) {
  cat(sprintf("Special effects of a bifactor fit (%d group factors%s; tau = %g)\n",
              nrow(x$groups),
              if (identical(x$source, "legacy")) ", hand-built design" else "",
              x$tau))
  tab <- x$groups
  tab$mean_ratio <- round(tab$mean_ratio, digits)
  tab$sd_ratio   <- round(tab$sd_ratio, digits)
  tab$cv         <- round(tab$cv, digits)
  tab$gamma      <- round(tab$gamma, digits)
  tab$D          <- round(tab$D, digits)
  print(tab, row.names = FALSE)
  cat(sprintf("Item-weighted mean CV of the group/general loading ratio: %s\n",
              format(round(x$mean_cv, digits))))
  cat("  (near-zero CV = Schmid-Leiman proportional = the fit re-expresses a\n",
      "  higher-order/testlet model, and gamma/lambda are faithful; values\n",
      "  under ~.1 are conventionally read that way -- suggestive, not a test.\n",
      "  With clearly non-constant ratios the higher-order parameters are a\n",
      "  lossy summary of a genuinely richer bifactor.)\n", sep = "")
  if (length(x$no_general))
    cat(sprintf("Note: %d item(s) carry no general loading (general design 0): %s\n",
                length(x$no_general), paste(x$no_general, collapse = ", ")))
  if (length(x$unassigned))
    cat(sprintf("Note: %d item(s) have no active group loading and are excluded from the ratio summaries: %s\n",
                length(x$unassigned), paste(x$unassigned, collapse = ", ")))
  if (length(x$crossloaded))
    cat(sprintf("Note: %d item(s) are active on more than one group factor (assigned to anchor/largest): %s\n",
                length(x$crossloaded), paste(x$crossloaded, collapse = ", ")))
  invisible(x)
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(vbpm)

## ----fit----------------------------------------------------------------------
sim <- sim_fa(N = 500, K = 3, ipf = 6, lam = .7, lac = .3, rseed = 1)
Y <- sim$dat

## an AZ (anchor-zero) design: two anchors per factor, each specified (1) on
## its own factor and fixed to zero on the other two; every non-anchor row is
## left entirely to the data
Q <- matrix(-1L, ncol(Y), 3)
for (k in 1:3) {
  a <- which(rep(1:3, each = 6) == k)[1:2]
  Q[a, ] <- 0L
  Q[a, k] <- 1L
}

fit <- vbfa(Y, Q)
fit

## ----access-------------------------------------------------------------------
round(fit$Lam[1:6, ], 2)   # posterior mean loadings
round(fit$pi[1:6, ], 2)    # PIPs of the unspecified entries
round(fit$Phi, 2)          # factor correlations (oblique by default)

## ----recovery-----------------------------------------------------------------
active <- (Q == 1) | (Q == -1 & fit$pi >= .5)
table(truth = sim$MLA != 0, active = active)

## ----vbfit--------------------------------------------------------------------
round(fit_stats(fit), 3)

## ----rank-adjustment, eval=FALSE----------------------------------------------
# fit_stats(fit, rank_adjust = TRUE, rank_max_J = 100)

## ----ld-----------------------------------------------------------------------
simLD <- sim_fa(N = 500, K = 3, ipf = 6, lam = .7, lac = .3, ecr = .3,
                rseed = 2)
fLD <- vbfa(simLD$dat, Q, ld = TRUE, max_it = 300, tolVal = 1e-3)

## the largest recovered residual edges, vs the planted pairs
Poff <- abs(fLD$Psi); Poff[lower.tri(Poff, diag = TRUE)] <- 0
which(Poff >= sort(Poff, decreasing = TRUE)[3], arr.ind = TRUE)
simLD$ofd_ind

## ----ld-restricted------------------------------------------------------------
J <- ncol(simLD$dat)
Qe <- matrix(-1L, J, J)
Qe[1:3, 1:3] <- 1L    # a known testlet: freely estimated among these 3 items
Qe[4:6, 7:9] <- 0L    # a block known to be residually independent
Qe[7:9, 4:6] <- 0L
isSymmetric(unname(Qe))

fRestricted <- vbfa(simLD$dat, Q, ld = TRUE, Qe = Qe, max_it = 300,
                    tolVal = 1e-3)

round(fRestricted$Psi[1:3, 1:3], 3)   # freely estimated: off-diagonal is not forced
round(fRestricted$Psi[4:6, 7:9], 3)   # fixed absent: driven to (numerical) zero

## ----ld-compare---------------------------------------------------------------
## fully exploratory loadings isolate the comparison to the residual side
Qexp <- matrix(-1L, J, 3)

fDiag <- vbfa(simLD$dat, Qexp)
fLDc  <- vbfa(simLD$dat, Qexp, ld = TRUE, max_it = 300, tolVal = 1e-3)

round(rbind(diagonal = fit_stats(fDiag)[c("BIC", "RMSEA")],
            ld       = fit_stats(fLDc)[c("BIC", "RMSEA")]), 3)

## ----ld-control---------------------------------------------------------------
fCtrl <- vbfa(simLD$dat, Q, ld = TRUE, max_it = 300, tolVal = 1e-3,
              ld_control = list(xi0 = c(0.1, 0.5, 1)))
fCtrl$converged   # TRUE if the final v0 stage met the tolerance

## ----mar-sim------------------------------------------------------------------
simM <- sim_fa(N = 400, K = 3, ipf = 6, lam = .7, lac = .3, rseed = 1)
Ym0  <- simM$dat

## items 5 and 11 go missing depending on the OBSERVED value of anchor items
## 1 and 7 (higher values make missingness more likely); the anchors
## themselves stay fully observed. This is MAR, not MCAR: the probability of
## missingness varies systematically with observed, not missing, data.
set.seed(42)
Ymar <- Ym0
p5  <- ifelse(Ym0[, 1] > stats::median(Ym0[, 1]), .40, .05)
p11 <- ifelse(Ym0[, 7] > stats::median(Ym0[, 7]), .40, .05)
Ymar[stats::rbinom(nrow(Ym0), 1, p5)  == 1, 5]  <- NA
Ymar[stats::rbinom(nrow(Ym0), 1, p11) == 1, 11] <- NA
sum(is.na(Ymar))

fMar <- vbfa(Ymar, Q)
fMar$preprocess$n_missing
round(fMar$Lam[c(1, 5, 7, 11), ], 2)   # loadings recovered despite the missingness

## ----mar-sim-clean------------------------------------------------------------
fClean <- vbfa(Ym0, Q)
max(abs(fMar$Lam - fClean$Lam))    # largest discrepancy
mean(abs(fMar$Lam - fClean$Lam))   # typical discrepancy is much smaller

## ----nlsy---------------------------------------------------------------------
data(nlsy27)
Yn <- as.matrix(nlsy27$dat)
dim(Yn)
sum(is.na(Yn))            # incomplete cells, handled in-loop

fn <- vbfa(Yn, nlsy27$Q)
fn
fn$preprocess$n_missing   # recorded on the fit
round(fn$Lam, 2)

## ----nlsy-cc------------------------------------------------------------------
nrow(Yn) - sum(stats::complete.cases(Yn))   # respondents listwise deletion drops

## ----softq--------------------------------------------------------------------
## a toy graded membership matrix for the simulated items
set.seed(3)
S <- matrix(runif(18 * 3, 0, .2), 18, 3)          # baseline noise
S[cbind(1:18, rep(1:3, each = 6))] <- runif(18, .55, .95)  # true memberships

Qsoft <- matrix(0L, 18, 3)          # S < .20 stays 0
Qsoft[S >= .20 & S < .80] <- -1L
Qsoft[S >= .80] <- 1L
table(Qsoft)

fsoft <- vbfa(Y, Qsoft)
fsoft$converged

#' National Longitudinal Survey of Youth 1997 (27 items)
#'
#' Responses of 3,458 individuals to 27 mixed-type items, with 1.12% missing
#' data. Carried over from the LAWBL package (which `vbpm` supersedes) for
#' empirical illustration of the estimators.
#'
#' @format A list with three components:
#' \describe{
#'   \item{dat}{A `3458 x 27` data frame of item responses (17 polytomous
#'     items, 10 continuous), containing `NA`s.}
#'   \item{Q}{A `27 x 3` initial design matrix with two to three specified
#'     loadings per factor, in the `-1/0/1` coding used by [vbfa()].}
#'   \item{cati}{Indices of the 17 polytomous items.}
#' }
#'
#' @source Bureau of Labor Statistics, U.S. Department of Labor. National
#'   Longitudinal Survey of Youth 1997 cohort. Previously distributed in the
#'   LAWBL package (Chen, 2022).
#'
#' @examples
#' data(nlsy27)
#' dim(nlsy27$dat)
#' nlsy27$Q
"nlsy27"

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

my_quic_cpp <- function(J, S, Rho, eps, max_iter_quic) {
    .Call(`_vbpm_my_quic_cpp`, J, S, Rho, eps, max_iter_quic)
}

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @useDynLib vbpm, .registration = TRUE
#' @importFrom Rcpp evalCpp
#' @importFrom stats cor cov2cor plogis runif sd rbinom setNames
## usethis namespace: end
NULL

## Regularized variational Bayes for the extended MIMIC model.
##
## Ported from the estimator accompanying Jin & Chen (2025, Multivariate
## Behavioral Research). The variational updates are unchanged from that
## implementation; the packaging changes are input validation, namespace
## imports, a staged (warm-started) v0 path, a convergence flag, an optional
## standardization step, and vbfa()'s sign convention on the output. The
## defaults (scalar v0 = 0.001, standardize = FALSE) reproduce the published
## estimator exactly.

#' Regularized variational Bayes for extended MIMIC models
#'
#' Fits a partially confirmatory multiple-indicators multiple-causes (MIMIC)
#' model by a variational Bayes EM algorithm, with continuous spike-and-slab
#' regularization on **both** the measurement and the structural part. Items
#' load on factors through a measurement design matrix `Q_A`, while covariates
#' predict the factors through a structural design matrix `Q_B`; entries of
#' either that are left unspecified are selected from the data.
#'
#' This is the MIMIC member of the `vbpm` family and is the companion of
#' [vbfa()], which fits the measurement part alone. The two share the `-1/0/1`
#' design-matrix convention and the spike-and-slab formulation.
#'
#' @param Y Numeric `N x J` data matrix (observations by items). Missing values
#'   in `Y` are supported: because the residual covariance is diagonal, a
#'   missing response is replaced in-loop by its conditional expectation
#'   `eta_i A'`, with the conditional variance carried into the residual sum
#'   of squares. Items or rows with no observed response are rejected.
#' @param X Numeric `N x P` matrix of observed covariates (predictors of the
#'   factors), with the same number of rows as `Y`. Missing covariates are
#'   **not** supported: `X` is conditioned on rather than modelled, so
#'   imputing it would require a distributional assumption the MIMIC model
#'   does not make.
#' @param Q_A Integer `J x K` measurement design matrix: `1` = specified
#'   (anchored) loading, `0` = fixed zero, `-1` = unspecified (selected by
#'   spike-and-slab). `K` is the number of factors.
#' @param Q_B Integer `K x P` structural design matrix using the same coding,
#'   for the regression of the `K` factors on the `P` covariates.
#' @param v0 Spike variance. A scalar gives a single fixed spike; a decreasing
#'   vector gives a warm-started regularization path, each stage initialized at
#'   the previous stage's solution (Rockova & George, 2018). The same schedule
#'   is applied to the measurement and structural parts. `v0 = 0.001` (the
#'   default) is the fixed spike of the published estimator.
#' @param standardize Logical; standardize `Y` and `X` before fitting. The
#'   default `FALSE` is the behaviour of the published estimator. Note that
#'   [vbfa()] always standardizes internally; set `standardize = TRUE` for the
#'   analogous behaviour here.
#' @param max_it Maximum number of variational iterations *per stage*.
#' @param convChk Logical; print per-iteration convergence information via
#'   [message()]. Default `FALSE` (quiet).
#' @param tolVal Convergence tolerance on the maximum absolute change in the
#'   weighted residual vector.
#'
#' @return An object of class `c("vbmimic", "vbpm_fit")` -- a named list (see
#'   [vbpm_fit]) with components:
#'   \describe{
#'     \item{A}{`J x K` posterior mean measurement loadings.}
#'     \item{B}{`K x P` posterior mean structural coefficients.}
#'     \item{pi_A, pi_B}{Posterior inclusion probabilities for the unspecified
#'       entries of `Q_A` and `Q_B` (specified entries are held at 1, fixed
#'       zeros at 0).}
#'     \item{Q_A, Q_B}{The design matrices the model was fit with.}
#'     \item{eta}{`N x K` posterior mean factor scores.}
#'     \item{Phi}{`K x K` factor correlation matrix implied by `Sig`.}
#'     \item{Sig, U}{Factor covariance matrix and its inverse.}
#'     \item{V}{Length-`J` residual precisions.}
#'     \item{rho, theta}{Inclusion rates for the measurement and structural
#'       parts.}
#'     \item{A_var, B_var}{Posterior variances of the loadings and structural
#'       coefficients.}
#'     \item{iter, flag, time}{Total iterations, `1` if the final stage
#'       converged within `max_it` (else `0`), and elapsed time.}
#'     \item{ELBO}{`NA_real_`; see Note.}
#'     \item{standardize}{The `standardize` setting the model was fit with.}
#'     \item{path}{The `v0` schedule and per-stage iteration counts.}
#'     \item{preprocess}{`n_missing`, the response type, and `missing_mask` --
#'       stored only when the data contain missing values, `NULL` otherwise.}
#'   }
#'
#'   Every quantity appears exactly once: as of 0.7.0 the fit no longer
#'   carries `$coefficients`, `$design`, `$posterior`, or `$settings`, which
#'   duplicated the elements above.
#'
#' @note The evidence lower bound is **not** currently returned. The published
#'   algorithm converges on the weighted residual vector rather than on the
#'   bound, and the bound's implementable closed form has not been assembled.
#'   [fit_stats()] does accept a `vbmimic` fit, but returns a deliberately
#'   limited result: `n_active_coef` (soft-selected measurement and structural
#'   coefficients) plus typed `NA` for every covariance-based index and
#'   parameter count, with an explanatory `note` attribute. [pefa()] does not
#'   accept `vbmimic` fits at all: its sweep reports criterion gains across a
#'   factor-count window, which requires one comparable objective for every
#'   candidate in that window.
#'
#' @references
#' Jin, Y., & Chen, J. (2025). Regularized variational Bayesian approximations
#' for variable selection in extended multiple-indicators multiple-causes
#' models. *Multivariate Behavioral Research*.
#' \doi{10.1080/00273171.2025.2483253}
#'
#' Rockova, V., & George, E. I. (2018). The spike-and-slab LASSO.
#' *Journal of the American Statistical Association*, 113(521), 431-444.
#'
#' @seealso [vbfa()] for the measurement model alone, [sim_lvm()] to simulate
#'   MIMIC data.
#'
#' @examples
#' ## 3 factors, 18 items, 9 covariates in disjoint blocks of 3
#' B <- matrix(0, 3, 9)
#' for (k in 1:3) B[k, ((k - 1) * 3 + 1):(k * 3)] <- .3
#' sim <- sim_lvm(N = 300, K = 3, J = 18, P = 9, b = B, phx = 0, rseed = 1)
#' Y <- sim$dat[, 1:18]
#' X <- sim$dat[, 19:27]
#'
#' ## two anchored items per factor; the structural part left exploratory.
#' ## Anchoring at least one part is advisable: with both Q_A and Q_B fully
#' ## exploratory the solution can be rotationally ambiguous.
#' Q_A <- matrix(-1L, 18, 3)
#' for (k in 1:3) {
#'   a <- which(rep(1:3, each = 6) == k)[1:2]
#'   Q_A[a, ] <- 0L; Q_A[a, k] <- 1L
#' }
#' Q_B <- matrix(-1L, 3, 9)
#'
#' fit <- vbmimic(Y, X, Q_A, Q_B)
#' fit$flag           # 1 if converged
#' round(fit$B, 2)    # structural coefficients
#'
#' @export
vbmimic <- function(Y, X, Q_A, Q_B, v0 = 0.001, standardize = FALSE,
                    max_it = 1000, convChk = FALSE, tolVal = 1e-5) {

  call <- match.call()

  ## ---- input checks --------------------------------------------------
  ## Missing responses in Y are supported (see below). Missing covariates are
  ## not: X is conditioned on rather than modelled, so imputing it would
  ## require assuming a distribution for the covariates that the MIMIC model
  ## does not specify.
  if (anyNA(X))
    stop("X contains NA. vbmimic() does not support missing covariates: X is ",
         "conditioned on, not modelled. Impute or drop them beforehand.",
         call. = FALSE)

  Y   <- as.matrix(Y)
  X   <- as.matrix(X)
  Q_A <- as.matrix(Q_A)
  Q_B <- as.matrix(Q_B)

  if (nrow(X) != nrow(Y))
    stop("Y and X must have the same number of rows (observations).",
         call. = FALSE)
  if (nrow(Q_A) != ncol(Y))
    stop("nrow(Q_A) must equal ncol(Y) (the number of items).", call. = FALSE)
  if (ncol(Q_B) != ncol(X))
    stop("ncol(Q_B) must equal ncol(X) (the number of covariates).",
         call. = FALSE)
  if (nrow(Q_B) != ncol(Q_A))
    stop("nrow(Q_B) must equal ncol(Q_A) (the number of factors).",
         call. = FALSE)
  if (!all(Q_A %in% c(-1, 0, 1)))
    stop("Q_A entries must be in {-1, 0, 1}.", call. = FALSE)
  if (!all(Q_B %in% c(-1, 0, 1)))
    stop("Q_B entries must be in {-1, 0, 1}.", call. = FALSE)

  v0_seq <- as.numeric(v0)
  if (any(v0_seq <= 0)) stop("v0 must be positive.", call. = FALSE)
  n_stage <- length(v0_seq)

  J <- ncol(Y)
  N <- nrow(Y)
  K <- ncol(Q_A)
  P <- ncol(X)
  if (K < 2)
    stop("vbmimic() requires at least two factors (ncol(Q_A) >= 2).",
         call. = FALSE)

  ## ---- missing responses ----------------------------------------------
  ## The residual covariance of Y is diagonal here (V holds residual
  ## precisions), so the joint Gaussian conditional of vbfa() collapses to its
  ## diagonal case: E[y_ij | y_i,obs] = eta_i A_j'. The covariate information
  ## is already carried by eta_i, whose own update uses X and B, so no extra
  ## conditioning on X is required. The conditional variance of an imputed
  ## cell is 1 / V_j and is added to the residual sum of squares below.
  missing_mask <- is.na(Y)
  has_missing  <- any(missing_mask)
  if (has_missing) {
    if (any(colSums(!missing_mask) == 0L))
      stop("Every item needs at least one observed response.", call. = FALSE)
    if (any(rowSums(!missing_mask) == 0L))
      stop("Rows with all responses missing are not supported.", call. = FALSE)
  }
  miss_count <- colSums(missing_mask)

  if (standardize) {
    ## centre and scale from observed entries only
    center <- colMeans(Y, na.rm = TRUE)
    scale_ <- apply(Y, 2, stats::sd, na.rm = TRUE)
    Y <- sweep(sweep(Y, 2, center, "-"), 2, scale_, "/")
    X <- scale(X)
  }
  Y[missing_mask] <- 0   # replaced by the conditional mean in iteration 1

  ## ---- initialization -------------------------------------------------
  ## Deterministic: the estimator consumes no random numbers, so there is no
  ## seed argument (cf. vbfa()).
  mu.q.A <- mu.q.A_sq <- matrix(0.8, J, K)
  mu.q.A[Q_A == -1] <- mu.q.A_sq[Q_A == -1] <- 0.2
  mu.q.A[Q_A == 0]  <- mu.q.A_sq[Q_A == 0]  <- 0
  sigsq.q.A <- matrix(1, J, K)
  Alpha <- matrix(0.5, J, K)
  Alpha[Q_A == 0] <- 0
  Alpha[Q_A == 1] <- 1
  RHO <- 0.5

  mu.q.B <- mu.q.B_sq <- matrix(0.8, K, P)
  mu.q.B[Q_B == -1] <- mu.q.B_sq[Q_B == -1] <- 0.2
  mu.q.B[Q_B == 0]  <- mu.q.B_sq[Q_B == 0]  <- 0
  sigsq.q.B <- matrix(1, K, P)
  Beta <- matrix(0.5, K, P)
  Beta[Q_B == 0] <- 0
  Beta[Q_B == 1] <- 1
  Theta <- 0.5

  mu.F <- mu.F_sq <- matrix(1, N, K)
  sigsq.F <- matrix(1, N, K)
  V <- rep(1, J)
  U <- Sig <- diag(1, K, K)

  invH_a <- invH_b <- 1
  a_v <- 1
  b_v <- 0.1
  inv_va1 <- inv_vb1 <- 1      # slab precision (v1 = 1)
  a0 <- b0 <- 1

  ## ---- staged VB loop ---------------------------------------------------
  itNum <- 0
  flag <- 0
  stage_iters <- integer(n_stage)
  startTime <- Sys.time()

  for (stage in seq_len(n_stage)) {
    ## spike precision for this stage, shared by the measurement and
    ## structural parts
    inv_va0 <- inv_vb0 <- 1 / v0_seq[stage]
    if (convChk && n_stage > 1)
      message("Stage ", stage, "/", n_stage, ": v0 = ", signif(v0_seq[stage], 4))

    converged <- FALSE
    parVecHatPrev <- NULL
    stage_it <- 0
    flag <- 0        # reset per stage; reflects the FINAL stage on return

    while (!converged) {
      itNum <- itNum + 1
      stage_it <- stage_it + 1

      ## ---- factors ------------------------------------------------------
      for (k in 1:K) {
        sigsq.F[, k] <- 1 / (U[k, k] + sum(mu.q.A_sq[, k] * V))
        tem1 <- (Y - mu.F[, -k, drop = FALSE] %*% t(mu.q.A[, -k, drop = FALSE])) %*%
          matrix(mu.q.A[, k] * V, nrow = J, ncol = 1)
        tem2 <- X %*% matrix(mu.q.B[k, ], nrow = P, ncol = 1) * U[k, k]
        tem3 <- apply(t(t(mu.F[, -k, drop = FALSE]) * U[k, -k]), 1, sum)
        tem4 <- apply(t(t(X %*% t(mu.q.B[-k, , drop = FALSE])) * U[k, -k]), 1, sum)
        mu.F[, k] <- sigsq.F[, k] * (tem1 + tem2 - tem3 + tem4)
      }

      ## rescale factors to unit variance (identification)
      sdF <- apply(mu.F, 2, sd)
      mu.F <- t(t(mu.F) / sdF)
      mu.F_sq <- mu.F^2 + sigsq.F

      ## ---- missing responses: conditional mean and variance --------------
      ## Diagonal residual covariance => Equation (2) of the design note.
      ## V from the start of the sweep is used for the variance correction,
      ## as in vbfa()'s coordinate-ascent ordering.
      if (has_missing) {
        mean_now <- mu.F %*% t(mu.q.A)
        Y[missing_mask] <- mean_now[missing_mask]
        miss_var <- miss_count / V           # length J; 1/V_j per missing cell
      }

      ## ---- residual precisions and measurement loadings ------------------
      for (j in 1:J) {
        ## expressions kept in the published associativity so the fit is
        ## bit-identical to the reference implementation; do not "simplify"
        ## without re-running the equivalence test
        tem1 <- matrix(Y[, j], nrow = 1, ncol = N) %*%
          matrix(Y[, j], nrow = N, ncol = 1)
        tem2 <- Y[, j] %*% (mu.F %*% matrix(mu.q.A[j, ], nrow = K, ncol = 1))
        tem3 <- (mu.q.A[j, ]) %*% t(mu.F) %*% mu.F %*%
          matrix(mu.q.A[j, ], nrow = K, ncol = 1)
        sub <- sum(t(mu.F^2) * mu.q.A[j, ]^2)
        add <- sum(t(mu.F_sq) * (mu.q.A_sq[j, ]))
        tem4 <- (tem1 - 2 * tem2 + tem3 - sub + add)
        ## E[y_ij^2] for an imputed cell is (conditional mean)^2 + 1/V_j; the
        ## squared mean is already in tem1, so only the variance is added.
        if (has_missing) tem4 <- tem4 + miss_var[j]
        V[j] <- (N / 2 + a_v) / (tem4 / 2 + b_v)

        for (k in 1:K) {
          if (Q_A[j, k] == -1) {
            tem <- 1 / 2 * log(inv_va1 / inv_va0) + log(RHO / (1 - RHO)) +
              1 / 2 * mu.q.A_sq[j, k] * (inv_va0 - inv_va1)
            tem <- pmin(tem, 100)
            Alpha[j, k] <- exp(tem) / (1 + exp(tem))
            tem2b <- (1 - Alpha[j, k]) * inv_va0 + Alpha[j, k] * inv_va1
            sigsq.q.A[j, k] <- 1 / (V[j] * sum(mu.F_sq[, k]) + tem2b)
            sigsq.q.A[j, k] <- max(sigsq.q.A[j, k], 1e-6)
            sigsq.q.A[j, k] <- min(sigsq.q.A[j, k], 1e6)
            tem1b <- Y[, j] - (mu.q.A[j, -k]) %*% t(mu.F[, -k, drop = FALSE])
            mu.q.A[j, k] <- sigsq.q.A[j, k] * V[j] *
              (matrix(mu.F[, k], nrow = 1, ncol = N) %*%
                 matrix(tem1b, nrow = N, ncol = 1))
            mu.q.A_sq[j, k] <- mu.q.A[j, k]^2 + sigsq.q.A[j, k]
          } else if (Q_A[j, k] == 1) {
            sigsq.q.A[j, k] <- 1 / (V[j] * sum(mu.F_sq[, k]) + invH_a)
            sigsq.q.A[j, k] <- max(sigsq.q.A[j, k], 1e-6)
            sigsq.q.A[j, k] <- min(sigsq.q.A[j, k], 1e6)
            tem1b <- Y[, j] - (mu.q.A[j, -k]) %*% t(mu.F[, -k, drop = FALSE])
            mu.q.A[j, k] <- sigsq.q.A[j, k] * V[j] *
              (matrix(mu.F[, k], nrow = 1, ncol = N) %*%
                 matrix(tem1b, nrow = N, ncol = 1))
            mu.q.A_sq[j, k] <- mu.q.A[j, k]^2 + sigsq.q.A[j, k]
          }
        }
      }
      RHO <- (a0 + sum(Alpha[Q_A == -1])) / (sum(Q_A == -1) + a0 + b0)

      ## ---- structural coefficients ---------------------------------------
      for (k in 1:K) {
        for (p in 1:P) {
          if (Q_B[k, p] == -1 || Q_B[k, p] == 1) {
            if (Q_B[k, p] == -1) {
              tem <- 1 / 2 * log(inv_vb1 / inv_vb0) + log(Theta / (1 - Theta)) +
                1 / 2 * mu.q.B_sq[k, p] * (inv_vb0 - inv_vb1)
              tem <- pmin(tem, 100)
              Beta[k, p] <- exp(tem) / (1 + exp(tem))
              pen <- (1 - Beta[k, p]) * inv_vb0 + Beta[k, p] * inv_vb1
            } else {
              pen <- invH_b
            }
            xtx <- matrix(X[, p], nrow = 1, ncol = N) %*%
              matrix(X[, p], nrow = N, ncol = 1)
            sigsq.q.B[k, p] <- 1 / (U[k, k] * xtx + pen)
            sigsq.q.B[k, p] <- max(sigsq.q.B[k, p], 1e-6)
            sigsq.q.B[k, p] <- min(sigsq.q.B[k, p], 1e6)
            tem1 <- sum(matrix(X[, p], nrow = 1, ncol = N) %*% mu.F * U[k, ])
            tem2 <- sum((mu.q.B[, -p, drop = FALSE]) %*%
                          t(X[, -p, drop = FALSE]) %*%
                          matrix(X[, p], nrow = N, ncol = 1) * U[k, ])
            tem3 <- sum(U[k, -k] * mu.q.B[-k, p]) * xtx
            mu.q.B[k, p] <- sigsq.q.B[k, p] * (tem1 - tem2 - tem3)
            mu.q.B_sq[k, p] <- mu.q.B[k, p]^2 + sigsq.q.B[k, p]
          }
        }
      }
      Theta <- (a0 + sum(Beta[Q_B == -1])) / (sum(Q_B == -1) + a0 + b0)

      ## ---- factor covariance ----------------------------------------------
      for (k in 1:K) {
        rk <- mu.F[, k] - X %*% matrix(mu.q.B[k, ], nrow = P, ncol = 1)
        tem1 <- matrix(rk, nrow = 1, ncol = N) %*% matrix(rk, nrow = N, ncol = 1)
        tem2 <- sum(mu.F_sq[, k] - mu.F[, k]^2) +
          apply(X^2, 2, sum) %*% matrix(mu.q.B_sq[k, ] - mu.q.B[k, ]^2,
                                        nrow = P, ncol = 1)
        Sig[k, k] <- (tem1 + tem2) / N
      }
      for (k in 1:(K - 1)) {
        for (t in (k + 1):K) {
          rk <- mu.F[, k] - X %*% matrix(mu.q.B[k, ], nrow = P, ncol = 1)
          rt <- mu.F[, t] - X %*% matrix(mu.q.B[t, ], nrow = P, ncol = 1)
          Sig[k, t] <- Sig[t, k] <-
            matrix(rk, nrow = 1, ncol = N) %*% matrix(rt, nrow = N, ncol = 1) / N
        }
      }
      U <- solve(Sig)

      ## ---- convergence -----------------------------------------------------
      ## weighted absolute residuals, in the published form. `%*% diag(V)`
      ## builds a J x J matrix and is wasteful for large J, but t(t(R) * V) --
      ## although mathematically identical -- sums in a different order and
      ## breaks bit-equivalence with the reference. Kept deliberately.
      parVecHat <- c(t(t(abs(Y - mu.F %*% t(mu.q.A)) %*% diag(V))))

      if (stage_it > 1) {
        relErrTemp <- abs(parVecHat - parVecHatPrev)
        relErrTemp[is.nan(relErrTemp)] <- 0
        relErr <- max(relErrTemp)
        if (convChk)
          message("Iteration ", itNum, "; MFVB relative error: ",
                  signif(relErr, 5))
        if (relErr < tolVal) {
          converged <- TRUE
          flag <- 1
        }
      }

      if (stage_it >= max_it) {
        converged <- TRUE
        if (flag != 1)
          warning("Maximum number of MFVB iterations exceeded in stage ",
                  stage, ".", call. = FALSE)
      }
      parVecHatPrev <- parVecHat
    }
    stage_iters[stage] <- stage_it
  }

  ## ---- sign convention (matches vbfa()) ---------------------------------
  ## Factor signs are arbitrary; flip any factor whose mean loading is
  ## negative and propagate to the scores and the structural coefficients.
  mean_A <- apply(mu.q.A, 2, mean)
  sign_flag <- which(mean_A < 0)
  if (length(sign_flag) > 0) {
    sign_vec <- rep(1, K)
    sign_vec[sign_flag] <- -1
    mu.q.A <- t(t(mu.q.A) * sign_vec)
    mu.F   <- t(t(mu.F) * sign_vec)
    mu.q.B <- mu.q.B * sign_vec        # rows are factors
    Sig    <- Sig * outer(sign_vec, sign_vec)
    U      <- U   * outer(sign_vec, sign_vec)
  }

  endTime <- Sys.time()

  pi_A <- Alpha
  pi_B <- Beta

  out <- list(model = "vbmimic", call = call,
              nobs = N, nitem = J, nfactor = K,
              converged = isTRUE(flag == 1),
              A = mu.q.A, B = mu.q.B,
              pi_A = pi_A, pi_B = pi_B,
              Q_A = Q_A, Q_B = Q_B,
              eta = mu.F, Phi = cov2cor(Sig), Sig = Sig, U = U, V = V,
              rho = RHO, theta = Theta,
              A_var = sigsq.q.A, B_var = sigsq.q.B,
              iter = itNum, flag = flag,
              time = difftime(endTime, startTime, units = "secs"),
              ELBO = NA_real_, objective = NA_real_, objective_type = "none",
              standardize = standardize,
              ## stored only when something is missing; see vbfa()
              preprocess = list(missing_mask = if (has_missing) missing_mask else NULL,
                                n_missing = sum(missing_mask),
                                response_type = "continuous"),
              path = list(v0 = v0_seq, n_stage = n_stage,
                          stage_iters = stage_iters))
  class(out) <- c("vbmimic", "vbpm_fit")
  out
}

## Sign-invariant Equation-17 correspondence and the seven pair measurements
## for one ordered candidate pair, K_from = k < l = K_to (note sections 4-5).
##
## The caller guarantees group-block matrices: in bifactor mode the leading
## general column has already been removed, so every column index computed and
## reported here is a group-block index.  Every entry is finite, because the
## constructor errors on a malformed candidate before reaching this file.
##
## The assignment rule is minimum sign-aligned squared distance.  Congruence is
## an exact-tie breaker and a reported shape diagnostic, never the rule itself.
## There is no eligibility screen of any kind, so every column participates and
## a comparison always holds exactly k scored pairs -- which is why no branch
## below needs an empty-set guard.

## The largest power of two not exceeding `x`, and 1 for a nonpositive or
## nonfinite `x` so a caller always receives a usable positive divisor.  Every
## rescaling below divides by this rather than by a raw magnitude, because
## dividing a double by an exact power of two only shifts its exponent: the
## scaled arithmetic reproduces the unscaled result bit for bit wherever the
## unscaled one neither overflows nor underflows, so removing either moves no
## number an ordinary fixture pins.  log2()/floor() can only misjudge at the
## subnormal edge, where a zero or infinite divisor would be worse than an
## inexact one, so the value itself is the fallback.
.pefa_pow2_scale <- function(x) {
  if (!is.finite(x) || x <= 0) return(1)
  p <- 2^floor(log2(x))
  if (is.finite(p) && p > 0) p else x
}

## s(a, b) = 1 when a'b >= 0, else -1.  The +1 convention at a'b = 0 is the
## note's arbitrary choice and is what keeps a zero-norm column's aligned
## distance defined.  The dot product is read on copies divided by one common
## power of two: an unscaled a'b can reach Inf + (-Inf) = NaN, and `NaN >= 0`
## is NA, which would turn the branch itself into an error rather than merely a
## wrong sign.  A positive divisor cannot change a sign, and negating a whole
## column negates the sum exactly, which is the root of reflection invariance.
.pefa_sign <- function(a, b) {
  scale <- .pefa_pow2_scale(max(abs(a), abs(b)))
  if (sum((a / scale) * (b / scale)) >= 0) 1 else -1
}

## d(a, b) = ||a - s b||^2 expressed on the basis `scale`, that is, the true
## aligned distance divided by scale^2.  Competing targets are ranked on this
## value directly and never on a restored one.
.pefa_sse <- function(a, b, s, scale) {
  d <- a / scale - s * (b / scale)
  sum(d^2)
}

## phi(a, b) = |a'b| / (||a|| ||b||).  A zero-norm column returns NA_real_ as a
## value rather than being skipped: note 5.1 forbids concealing an undefined
## scored congruence behind na.rm, and the raw 0/0 would be NaN rather than the
## documented NA flavor.  Congruence is scale free, so each column is
## normalized by its own peak; sum(unit^2) is then at least 1, so neither norm
## can underflow and neither product can overflow.
.pefa_congruence <- function(a, b) {
  scale_a <- max(abs(a))
  scale_b <- max(abs(b))
  if (scale_a == 0 || scale_b == 0) return(NA_real_)
  unit_a <- a / scale_a
  unit_b <- b / scale_b
  value <- abs(sum(unit_a * unit_b)) /
    (sqrt(sum(unit_a^2)) * sqrt(sum(unit_b^2)))
  ## |cos| cannot exceed 1; only roundoff can push it past, and an unclamped
  ## 1 + 1e-16 would make two exactly collinear targets compare unequal in the
  ## tie rule, pre-empting the index rule that should settle them.
  min(1, value)
}

## Adjusted Rand index between two item partitions (note 5.2).
.pefa_ari <- function(p, q) {
  J <- length(p)
  ## One item is one trivially identical partition at both endpoints.
  if (J < 2L) return(1)
  c2 <- function(x) x * (x - 1) / 2
  tab <- table(p, q)
  A <- sum(c2(tab))
  B <- sum(c2(rowSums(tab)))
  C <- sum(c2(colSums(tab)))
  E <- B * C / c2(J)
  D <- (B + C) / 2 - E
  if (D == 0) {
    ## A zero denominator forces both endpoints into the same degenerate
    ## partition -- all singletons, or one cluster -- so agreement is the only
    ## reachable answer and the 0 branch is purely defensive.  Coding each
    ## partition by order of first appearance compares them as partitions
    ## rather than as label vectors.
    same <- identical(as.integer(factor(p, levels = unique(p))),
                      as.integer(factor(q, levels = unique(q))))
    return(if (same) 1 else 0)
  }
  ## Roundoff, not mathematics, is the only way out of [-1, 1].
  min(1, max(-1, (A - E) / D))
}

## The seven structural facts for one ordered candidate pair.
.pefa_compare_pair <- function(lam_from, lam_to, pip_from, pip_to,
                               reg_from, reg_to, K0) {
  K0 <- as.integer(K0)
  J <- nrow(lam_from)
  k <- ncol(lam_from)
  l <- ncol(lam_to)
  backbone <- seq_len(K0)
  src_free <- if (k > K0) seq.int(K0 + 1L, k) else integer(0)
  tgt_free <- if (l > K0) seq.int(K0 + 1L, l) else integer(0)

  ## Assignment (note 4).  Backbone column h takes target h by position; each
  ## exploratory source then selects its own minimizer independently, so reuse
  ## is permitted, recorded, and never repaired into a one-to-one map.
  assigned <- integer(length(src_free))
  for (pos in seq_along(src_free)) {
    a <- lam_from[, src_free[pos]]
    signs <- vapply(tgt_free, function(j) .pefa_sign(a, lam_to[, j]),
                    numeric(1))
    ## One common power-of-two basis for this source and every target it may
    ## take, so the whole competing field is ranked on the same footing (note
    ## section 4).  Both ends of the double range manufacture a false exact tie
    ## otherwise: two astronomically distant targets share an Inf, and two
    ## astronomically near ones share a 0, and either hands the choice to the
    ## tie rule, which then resolves on congruence or index a distance had
    ## already settled.  Dividing by an exact power of two only shifts
    ## exponents, so a scaled SSE reproduces its unscaled value bit for bit
    ## wherever that value is representable, and one positive divisor common to
    ## the whole field is order preserving -- the mathematical ranking, exact
    ## ties included, is the unscaled one.
    scale <- .pefa_pow2_scale(max(abs(a), abs(lam_to[, tgt_free])))
    sse <- vapply(seq_along(tgt_free), function(t) {
      .pefa_sse(a, lam_to[, tgt_free[t]], signs[t], scale)
    }, numeric(1))
    best <- which(sse == min(sse))
    if (length(best) > 1L) {
      ## An exact distance tie goes to the larger finite absolute congruence,
      ## a finite congruence outranks an undefined one, and the smaller
      ## original target index settles what is left.  which.max() returns the
      ## first maximum over an ascending target set, which is that index rule.
      ## No tolerance widens a near-tie into a tie.
      phi <- vapply(best, function(t) {
        .pefa_congruence(a, lam_to[, tgt_free[t]])
      }, numeric(1))
      phi[is.na(phi)] <- -Inf
      best <- best[which.max(phi)]
    }
    assigned[pos] <- tgt_free[best[1L]]
  }

  ## Exactly k scored pairs: K0 position matches plus one per exploratory
  ## source.  Sources are distinct by construction; targets need not be.
  pair_src <- c(backbone, src_free)
  pair_tgt <- c(backbone, assigned)
  a_mat <- lam_from[, pair_src, drop = FALSE]
  b_mat <- lam_to[, pair_tgt, drop = FALSE]

  ## One common power-of-two basis for every scored cell, so a column whose
  ## aligned difference leaves the double range before the division by sqrt(J)
  ## still reports the representable RMSD instead of Inf.  Pooling and the
  ## columnwise maximum come off the same scaled sums, which is what keeps
  ## rmsd <= rmsd_max in the arithmetic and not only on paper.
  scale <- .pefa_pow2_scale(max(abs(a_mat), abs(b_mat)))
  signs <- vapply(seq_len(k), function(p) {
    .pefa_sign(a_mat[, p], b_mat[, p])
  }, numeric(1))
  sse <- vapply(seq_len(k), function(p) {
    .pefa_sse(a_mat[, p], b_mat[, p], signs[p], scale)
  }, numeric(1))
  rmsd <- scale * sqrt(sum(sse) / (J * k))
  rmsd_max <- scale * sqrt(max(sse) / J)

  ## phi_min runs over every scored pair, dead source columns included, so it
  ## is a weakest-link statistic: read a low value next to ssl() before reading
  ## it as movement in the retained structure.
  phi <- vapply(seq_len(k), function(p) {
    .pefa_congruence(a_mat[, p], b_mat[, p])
  }, numeric(1))
  phi_min <- if (anyNA(phi)) NA_real_ else min(phi)

  ## ARI uses every endpoint group column and is independent of the
  ## correspondence.  ties.method = "first" over an ascending column set is the
  ## note's "exact loading ties go to the lower column index".
  ari <- .pefa_ari(max.col(abs(lam_from), ties.method = "first"),
                   max.col(abs(lam_to), ties.method = "first"))

  ## PIPs are probabilities and are never sign-transformed.  Only cells
  ## regularized at both endpoints under the correspondence are comparable;
  ## fixed cells and unselected target columns drop out.
  sel <- reg_from[, pair_src, drop = FALSE] &
    reg_to[, pair_tgt, drop = FALSE]
  pip_rmsd <- NA_real_
  if (any(sel)) {
    d <- pip_from[, pair_src, drop = FALSE][sel] -
      pip_to[, pair_tgt, drop = FALSE][sel]
    pip_rmsd <- sqrt(sum(d^2) / length(d))
  }

  ## Every exploratory target no exploratory source selected, including the
  ## surplus a collision leaves behind.  The target has more exploratory
  ## columns than the source, so the set is nonempty and the NA_real_ branch
  ## exists only to keep max() off an empty vector.
  unmatched <- setdiff(tgt_free, assigned)
  unmatched_ssl <- if (length(unmatched)) {
    max(colSums(lam_to[, unmatched, drop = FALSE]^2))
  } else {
    NA_real_
  }
  ## Reused targets, not reusing sources; backbone positions never reuse.
  collision <- anyDuplicated(assigned) > 0L

  list(phi_min = phi_min, rmsd = rmsd, rmsd_max = rmsd_max, ari = ari,
       pip_rmsd = pip_rmsd, unmatched_ssl = unmatched_ssl,
       collision = collision)
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(vbpm)

## ----population---------------------------------------------------------------
K <- 3
ipf <- 6
J <- K * ipf
cluster <- rep(seq_len(K), each = ipf)

set.seed(7)
mla <- matrix(0, J, K + 1)
mla[, 1] <- round(runif(J, .35, .75), 2)
for (k in seq_len(K)) {
  mla[cluster == k, k + 1] <- round(runif(ipf, .35, .70), 2)
}
while (max(rowSums(mla^2)) > .9) mla <- round(mla * .97, 3)

sim <- sim_fa(N = 800, mla = mla, phi = 0, rseed = 1)
Y <- sim$dat

## ----backbones----------------------------------------------------------------
Q0_AO <- matrix(-1L, J, 2)
for (k in 1:2) Q0_AO[which(cluster == k)[1:2], k] <- 1L

Q0_AZ <- Q0_AO
for (k in 1:2) {
  anchors <- which(cluster == k)[1:2]
  Q0_AZ[anchors, ] <- 0L
  Q0_AZ[anchors, k] <- 1L
}

Q0_show <- cbind(Q0_AO[1:8, ], Q0_AZ[1:8, ])
colnames(Q0_show) <- c("AO_1", "AO_2", "AZ_1", "AZ_2")
Q0_show

## ----step1--------------------------------------------------------------------
AO <- pefa(Q0_AO, Y, Kmin = 2, Kmax = 5, verbose = FALSE)
AZ <- pefa(Q0_AZ, Y, Kmin = 2, Kmax = 5, verbose = FALSE)

AO$sweep[, c("K", "ELBO", "AIC", "BIC", "SRMR", "CFI")]
AZ$sweep[, c("K", "ELBO", "AIC", "BIC", "SRMR", "CFI")]

## ----step1-transitions--------------------------------------------------------
AO$transitions[, c("K_from", "K_to", "ELBO_gain_pct", "BIC_gain_pct",
                   "phi_min", "rmsd_max", "unmatched_ssl", "collision")]
AZ$transitions[, c("K_from", "K_to", "ELBO_gain_pct", "BIC_gain_pct",
                   "phi_min", "rmsd_max", "unmatched_ssl", "collision")]

## ----step1-persistence--------------------------------------------------------
round(AO$persistence$phi, 3)
round(AZ$persistence$phi, 3)
ssl(AO)[c("3", "4")]
ssl(AZ)[c("3", "4")]

## ----step1-reading------------------------------------------------------------
## Analysis-owned profiles. ELBO and BIC each use the full fitted window,
## a 20% gain cutoff, sustain one, a post-peak strict crossing, and no fallback.
profile_specs <- data.frame(
  profile = c(".85/r1", ".80/r2", ".70/r3"),
  phi_cut = c(.85, .80, .70), r = 1:3,
  practical_default = c(FALSE, TRUE, FALSE)
)
profile_sources <- 2:3

read_sweep <- function(x, phi_cut, r, sources) {
  K <- as.integer(x$sweep$K)
  converged <- x$sweep$converged

  read_count <- function(score) {
    usable <- length(K) >= 2L && all(is.finite(K)) && all(diff(K) == 1L) &&
      length(score) == length(K) && all(is.finite(score)) &&
      is.logical(converged) && all(converged %in% TRUE)
    if (!usable) return(list(usable = FALSE, Khat = NA_integer_))
    gain <- diff(score)
    gain_max <- max(gain)
    if (!(gain_max > 0)) return(list(usable = TRUE, Khat = NA_integer_))
    peak <- max(which(gain == gain_max))
    hit <- which(seq_along(gain) > peak & gain < .20 * gain_max)
    list(usable = TRUE,
         Khat = if (length(hit)) K[hit[1L]] else NA_integer_)
  }

  ELBO <- read_count(x$sweep$ELBO)
  BIC <- read_count(-x$sweep$BIC)
  C20 <- sort(unique(c(ELBO$Khat, BIC$Khat)[
    is.finite(c(ELBO$Khat, BIC$Khat))]))

  phi <- x$persistence$phi
  collision <- x$persistence$collision
  conv <- setNames(converged, K)
  state <- vapply(sources, function(k) {
    targets <- k + seq_len(r)
    edge <- vapply(targets, function(target) {
      from <- as.character(k); to <- as.character(target)
      if (!from %in% rownames(phi) || !to %in% colnames(phi) ||
          !isTRUE(conv[[from]]) || !isTRUE(conv[[to]]) ||
          is.na(collision[from, to])) return(NA_integer_)
      if (isTRUE(collision[from, to])) return(0L)
      value <- phi[from, to]
      if (!is.finite(value)) NA_integer_ else as.integer(value >= phi_cut)
    }, integer(1))
    if (any(edge %in% 0L)) 0L else
      if (length(edge) && all(edge %in% 1L)) 1L else NA_integer_
  }, integer(1))

  persistent <- which(state == 1L)
  if (!length(state)) {
    resolution <- "unresolved"
    Kp <- NA_integer_
  } else if (!length(persistent)) {
    resolution <- if (all(state %in% 0L)) "nonpersistent" else "unresolved"
    Kp <- NA_integer_
  } else {
    top <- max(persistent)
    higher <- if (top < length(state)) state[(top + 1L):length(state)] else 0L
    resolution <- if (all(higher %in% 0L)) "persistent" else "unresolved"
    Kp <- if (resolution == "persistent") sources[top] else NA_integer_
  }

  hit_ELBO <- is.finite(Kp) && is.finite(ELBO$Khat) && ELBO$Khat == Kp
  hit_BIC <- is.finite(Kp) && is.finite(BIC$Khat) && BIC$Khat == Kp
  support <- if (hit_ELBO && hit_BIC) "both" else if (hit_ELBO) "ELBO only" else
    if (hit_BIC) "BIC only" else "none"
  any_count_usable <- ELBO$usable || BIC$usable
  layer <- if (resolution == "persistent" && Kp %in% C20) "L1" else
    if (resolution == "persistent" && any_count_usable) "L2" else
    if (resolution == "nonpersistent") "L3" else
      "unclassified"
  gap <- if (resolution == "persistent" && length(C20))
    min(abs(C20 - Kp)) else NA_real_

  list(Khat_ELBO = ELBO$Khat, Khat_BIC = BIC$Khat, C20 = C20,
       count_usable = c(ELBO = ELBO$usable, BIC = BIC$usable),
       support = support, Kp = Kp, source_state = setNames(state, sources),
       resolution = resolution, layer = layer, g_CP = gap)
}

profile_row <- function(x, backbone, spec) {
  z <- read_sweep(x, spec$phi_cut, spec$r, profile_sources)
  data.frame(backbone = backbone, profile = spec$profile,
             practical_default = spec$practical_default,
             Khat_ELBO = z$Khat_ELBO, Khat_BIC = z$Khat_BIC,
             C20 = if (length(z$C20)) paste0("{", paste(z$C20,
               collapse = ","), "}") else "{}",
             support = z$support, Kp = z$Kp, layer = z$layer,
             g_CP = z$g_CP, row.names = NULL)
}

sweeps <- list(AO = AO, AZ = AZ)
profiles <- do.call(rbind, lapply(names(sweeps), function(backbone)
  do.call(rbind, lapply(seq_len(nrow(profile_specs)), function(i)
    profile_row(sweeps[[backbone]], backbone, profile_specs[i, ])))))
profiles

## ----step2-design-------------------------------------------------------------
default_AO <- profiles[profiles$backbone == "AO" &
                       profiles$profile == ".80/r2", , drop = FALSE]
if (nrow(default_AO) != 1L || !default_AO$layer %in% c("L1", "L2"))
  stop("The declared profile did not deliver a Step-2 proposal.")
K_step2 <- as.integer(default_AO$Kp)
K0      <- ncol(AO$Q0)
Lam     <- AO$loadings[[as.character(K_step2)]]
d     <- unique(colSums(AO$Q0 == 1L))
if (length(d) != 1L || d[1] < 1L) {
  stop("Choose one marker depth before building Q2.")
}

Q2 <- cbind(AO$Q0, matrix(-1L, J, K_step2 - K0))
colnames(Q2) <- paste0("F", seq_len(K_step2))
if (K_step2 > K0) {
  for (k in (K0 + 1L):K_step2) {
    top <- head(order(-abs(Lam[, k]), seq_len(J)), d)
    Q2[top, k] <- 1L
  }
}

c(K_step2 = K_step2, K0 = K0, depth = d)
Q2[rowSums(Q2 == 1L) > 0L, , drop = FALSE]

## ----step2-fit----------------------------------------------------------------
f_obl <- vbfa(Y, Q2)
f_bif <- vbfa(Y, Q2, bifactor = TRUE)

idx <- c("RMSEA", "SRMR", "CFI", "TLI", "AIC", "BIC")
round(rbind(
  oblique  = fit_stats(f_obl)[idx],
  bifactor = fit_stats(f_bif)[idx]
), 3)

## ----step2-effects------------------------------------------------------------
special_effects(f_bif)

## ----bifactor-sweep-----------------------------------------------------------
BI <- pefa(Q0_AO, Y, Kmin = 2, Kmax = 5, bifactor = TRUE, verbose = FALSE)
BI$sweep[, c("K", "ELBO", "AIC", "BIC", "SRMR", "CFI")]
BI$transitions[, c("K_from", "K_to", "ELBO_gain_pct", "BIC_gain_pct",
                   "phi_min", "rmsd_max", "unmatched_ssl", "collision")]
round(BI$persistence$phi, 3)
round(BI$persistence$rmsd, 3)
ssl(BI)

---
title: "Bifactor Structures and PEFA Sweeps"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Bifactor Structures and PEFA Sweeps}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(vbpm)
```

This vignette covers the bifactor side of `vbpm`: what a bifactor model is,
how a `pefa()` sweep is read when the factor count is uncertain, and how a
later oblique-versus-bifactor comparison is built out of the sweep that
preceded it. The sweep object itself is walked in `vignette("pefa")`; nothing
here re-explains its tables.

## Three related structures

Suppose `J` items fall into `K` clusters.

**Correlated factors.** Each item loads on its cluster factor, and the cluster
factors correlate through $\Phi$.

**Orthogonal bifactor.** Every item loads on a general factor and on one group
factor; general and group factors are mutually orthogonal.

**Higher-order.** Items load on first-order factors, which load in turn on a
second-order factor. Its Schmid-Leiman representation is the orthogonal
bifactor with

$$
b^{gen}_j=\lambda_j\gamma_k,
\qquad
b^{grp}_j=\lambda_j\sqrt{1-\gamma_k^2}
$$

for item $j$ in cluster $k$, so $|b^{grp}_j|/|b^{gen}_j|$ is constant within a
cluster under exact higher-order proportionality. Testlet and method factors
use the same general-plus-special geometry under a different substantive
reading.

In `vbfa()` and `pefa()`, `bifactor = TRUE` makes `K` the number of **group**
factors: the design matrix supplies only the group block and the general column
is added internally, so a fitted matrix has `K + 1` columns. Every
between-candidate comparison a sweep performs removes that general column
first and works on the `K` group columns, which is what makes an ordinary and a
bifactor sweep of the same data structurally comparable.

## A bifactor population and two backbones

Three clusters of six items, each item loading on a general factor and on its
own group factor.

```{r population}
K <- 3
ipf <- 6
J <- K * ipf
cluster <- rep(seq_len(K), each = ipf)

set.seed(7)
mla <- matrix(0, J, K + 1)
mla[, 1] <- round(runif(J, .35, .75), 2)
for (k in seq_len(K)) {
  mla[cluster == k, k + 1] <- round(runif(ipf, .35, .70), 2)
}
while (max(rowSums(mla^2)) > .9) mla <- round(mla * .97, 3)

sim <- sim_fa(N = 800, mla = mla, phi = 0, rseed = 1)
Y <- sim$dat
```

Two anchors per backbone factor, expressed as the two designs `vbpm`
documentation uses throughout.

* **AO (anchor-only)** specifies each anchor's intended cell as `1` and leaves
  that anchor's cells on the *other* backbone columns unspecified (`-1`).
* **AZ (anchor-zero)** additionally fixes those cross-backbone anchor cells to
  `0`.

```{r backbones}
Q0_AO <- matrix(-1L, J, 2)
for (k in 1:2) Q0_AO[which(cluster == k)[1:2], k] <- 1L

Q0_AZ <- Q0_AO
for (k in 1:2) {
  anchors <- which(cluster == k)[1:2]
  Q0_AZ[anchors, ] <- 0L
  Q0_AZ[anchors, k] <- 1L
}

Q0_show <- cbind(Q0_AO[1:8, ], Q0_AZ[1:8, ])
colnames(Q0_show) <- c("AO_1", "AO_2", "AZ_1", "AZ_2")
Q0_show
```

AZ is the stronger claim: it asserts that an anchor of cluster 1 has no loading
on backbone factor 2. AO only asserts where the anchor *does* load. Both
backbones cover two of the three clusters, so the third has to be found by the
sweep.

## Step 1: ordinary oblique sweeps

The calls below use each function's documented defaults; no call in this
vignette passes `v0` or `max_it`. Both `pefa()` and `vbfa()` walk the same
four-step spike ladder `v0 = c(.01, .005, .002, .001)`. What differs is
`max_it`, and it is a cap **per ladder stage**, not per fit or per candidate:
10000 per stage in `pefa()` against 5000 per stage in a direct `vbfa()`, so
the worst-case totals over the four stages are 40000 and 20000. These are two
independent documented defaults, not one shared ceiling.

```{r step1}
AO <- pefa(Q0_AO, Y, Kmin = 2, Kmax = 5, verbose = FALSE)
AZ <- pefa(Q0_AZ, Y, Kmin = 2, Kmax = 5, verbose = FALSE)

AO$sweep[, c("K", "ELBO", "AIC", "BIC", "SRMR", "CFI")]
AZ$sweep[, c("K", "ELBO", "AIC", "BIC", "SRMR", "CFI")]
```

```{r step1-transitions}
AO$transitions[, c("K_from", "K_to", "ELBO_gain_pct", "BIC_gain_pct",
                   "phi_min", "rmsd_max", "unmatched_ssl", "collision")]
AZ$transitions[, c("K_from", "K_to", "ELBO_gain_pct", "BIC_gain_pct",
                   "phi_min", "rmsd_max", "unmatched_ssl", "collision")]
```

Both backbones put their only positive gain on the `2 -> 3` step and both put
`unmatched_ssl` near 2.6 there, which is a full column's worth against the
`ssl()` values below. Both minimize BIC at `K = 3` --- 35212.89 for AO and
35209.52 for AZ --- but only just: the `3 -> 4` BIC step gives back 0.5% (AO)
and 1.0% (AZ) of that first gain, so the criterion is nearly flat between three
and four factors. AIC dissents in both backbones, taking its minimum at `K = 4`
(35015.32 and 35015.94, against 35025.55 and 35022.18 at three).

```{r step1-persistence}
round(AO$persistence$phi, 3)
round(AZ$persistence$phi, 3)
ssl(AO)[c("3", "4")]
ssl(AZ)[c("3", "4")]
```

Persistence separates the two backbones a little more than the criteria do.
AO's three-factor columns survive into `K = 4` and `K = 5` at 0.989; AZ's
survive at 0.985, and AZ's `2 -> 4` and `2 -> 5` cells read 0.905 against AO's
0.925. In both backbones the fourth column is small --- `ssl()` puts it at 0.37
(AO) and 0.45 (AZ) against three columns above 2.4 --- so the four-factor
candidate is the three-factor one plus a thin extra column, not a different
structure.

```{r step1-reading}
## Analysis-owned profiles. ELBO and BIC each use the full fitted window,
## a 20% gain cutoff, sustain one, a post-peak strict crossing, and no fallback.
profile_specs <- data.frame(
  profile = c(".85/r1", ".80/r2", ".70/r3"),
  phi_cut = c(.85, .80, .70), r = 1:3,
  practical_default = c(FALSE, TRUE, FALSE)
)
profile_sources <- 2:3

read_sweep <- function(x, phi_cut, r, sources) {
  K <- as.integer(x$sweep$K)
  converged <- x$sweep$converged

  read_count <- function(score) {
    usable <- length(K) >= 2L && all(is.finite(K)) && all(diff(K) == 1L) &&
      length(score) == length(K) && all(is.finite(score)) &&
      is.logical(converged) && all(converged %in% TRUE)
    if (!usable) return(list(usable = FALSE, Khat = NA_integer_))
    gain <- diff(score)
    gain_max <- max(gain)
    if (!(gain_max > 0)) return(list(usable = TRUE, Khat = NA_integer_))
    peak <- max(which(gain == gain_max))
    hit <- which(seq_along(gain) > peak & gain < .20 * gain_max)
    list(usable = TRUE,
         Khat = if (length(hit)) K[hit[1L]] else NA_integer_)
  }

  ELBO <- read_count(x$sweep$ELBO)
  BIC <- read_count(-x$sweep$BIC)
  C20 <- sort(unique(c(ELBO$Khat, BIC$Khat)[
    is.finite(c(ELBO$Khat, BIC$Khat))]))

  phi <- x$persistence$phi
  collision <- x$persistence$collision
  conv <- setNames(converged, K)
  state <- vapply(sources, function(k) {
    targets <- k + seq_len(r)
    edge <- vapply(targets, function(target) {
      from <- as.character(k); to <- as.character(target)
      if (!from %in% rownames(phi) || !to %in% colnames(phi) ||
          !isTRUE(conv[[from]]) || !isTRUE(conv[[to]]) ||
          is.na(collision[from, to])) return(NA_integer_)
      if (isTRUE(collision[from, to])) return(0L)
      value <- phi[from, to]
      if (!is.finite(value)) NA_integer_ else as.integer(value >= phi_cut)
    }, integer(1))
    if (any(edge %in% 0L)) 0L else
      if (length(edge) && all(edge %in% 1L)) 1L else NA_integer_
  }, integer(1))

  persistent <- which(state == 1L)
  if (!length(state)) {
    resolution <- "unresolved"
    Kp <- NA_integer_
  } else if (!length(persistent)) {
    resolution <- if (all(state %in% 0L)) "nonpersistent" else "unresolved"
    Kp <- NA_integer_
  } else {
    top <- max(persistent)
    higher <- if (top < length(state)) state[(top + 1L):length(state)] else 0L
    resolution <- if (all(higher %in% 0L)) "persistent" else "unresolved"
    Kp <- if (resolution == "persistent") sources[top] else NA_integer_
  }

  hit_ELBO <- is.finite(Kp) && is.finite(ELBO$Khat) && ELBO$Khat == Kp
  hit_BIC <- is.finite(Kp) && is.finite(BIC$Khat) && BIC$Khat == Kp
  support <- if (hit_ELBO && hit_BIC) "both" else if (hit_ELBO) "ELBO only" else
    if (hit_BIC) "BIC only" else "none"
  any_count_usable <- ELBO$usable || BIC$usable
  layer <- if (resolution == "persistent" && Kp %in% C20) "L1" else
    if (resolution == "persistent" && any_count_usable) "L2" else
    if (resolution == "nonpersistent") "L3" else
      "unclassified"
  gap <- if (resolution == "persistent" && length(C20))
    min(abs(C20 - Kp)) else NA_real_

  list(Khat_ELBO = ELBO$Khat, Khat_BIC = BIC$Khat, C20 = C20,
       count_usable = c(ELBO = ELBO$usable, BIC = BIC$usable),
       support = support, Kp = Kp, source_state = setNames(state, sources),
       resolution = resolution, layer = layer, g_CP = gap)
}

profile_row <- function(x, backbone, spec) {
  z <- read_sweep(x, spec$phi_cut, spec$r, profile_sources)
  data.frame(backbone = backbone, profile = spec$profile,
             practical_default = spec$practical_default,
             Khat_ELBO = z$Khat_ELBO, Khat_BIC = z$Khat_BIC,
             C20 = if (length(z$C20)) paste0("{", paste(z$C20,
               collapse = ","), "}") else "{}",
             support = z$support, Kp = z$Kp, layer = z$layer,
             g_CP = z$g_CP, row.names = NULL)
}

sweeps <- list(AO = AO, AZ = AZ)
profiles <- do.call(rbind, lapply(names(sweeps), function(backbone)
  do.call(rbind, lapply(seq_len(nrow(profile_specs)), function(i)
    profile_row(sweeps[[backbone]], backbone, profile_specs[i, ])))))
profiles
```

ELBO and BIC are co-primary count readings here: each uses the 20% rule with
sustain one over all of `K = 2:5`, and finite readings form $C_{20}$. The
`support` column says descriptively whether both, only ELBO, only BIC, or
neither corroborates $K_p$; it does not rank the paths. The three structural
profiles are analysis-owned and scenario-bound. `.80/r2` was retained as the
practical default in one simulation study, not made a **vbpm** default.
The profiles share one declared source set, so the comparison is like-for-like;
here it is `2:3` because this demonstration window is short. Because this short window ends at five, `.70/r3` lacks the
`3 -> 6` edge and is therefore unclassified rather than silently searching a
smaller source universe. Fitting through `K = 6` would resolve that row.

Within either backbone, agreement among the simultaneously displayed resolved
profiles is **within-family robustness** to these declared cutoff/depth choices,
not evidence of population truth, validation, or a package-selected count.
Disagreement identifies sensitivity within the family; it does not say which
profile is true. The `.70/r3` row is also a concrete short-window unresolved
example, not structural nonpersistence.

L1 means that selected $K_p$ belongs to finite $C_{20}$. L2 means that selected
$K_p$ does not belong to $C_{20}$ while at least one count path is usable. L3
requires all declared sources to resolve nonpersistent regardless of count-path
usability; otherwise the row is unclassified. The displayed
$g_{CP}=\min_{\widehat K\in C_{20}}|\widehat K-K_p|$ is corroboration only,
never a decision criterion, and is undefined without a finite count. Here
$g_{CP}=0$ algebraically restates L1 membership; only a positive L2 gap has
descriptive magnitude. A horizon is always direct: $r=2$ asks separately for
`K -> K + 1` and `K -> K + 2`,
never for a chain through the intervening fit. A known collision vetoes its
edge, and the reader selects only the highest persistent source whose higher
declared sources have all resolved nonpersistent.

## Step 2: a comparison derived from Step 1

The two criterion paths supply descriptive finite set $C_{20}$. A persistence
profile separately supplies $K_p$ and the layer. Only L1 or L2 delivers a
candidate to Step 2, and then $K_{\mathrm{step2}}=K_p$. Thus
$K_{\mathrm{step2}}$ is literally $K_p$ conditional on L1/L2 delivery; it is
neither another selector nor the true population count. The group Q below is
derived from the AO candidate at that delivered count. All of this is ordinary
analysis-side code: `pefa()` neither performs it nor stores it.

Let `K0 = ncol(Q0)`. The **specification depth** is the number of markers in
each backbone column. Here both columns have depth two. If the depths differed,
the analyst would need to state a rule of their own or supply a substantive
`J x K_step2` group-loading Q.

The Step-2 design keeps `Q0` unchanged and appends one all-`-1` column for each
new factor. In every new column, the `d` largest absolute loadings become
markers (`1`). Exact ties go to the lower item index. There is no PIP or loading
cutoff, and the same item may mark more than one new factor.

```{r step2-design}
default_AO <- profiles[profiles$backbone == "AO" &
                       profiles$profile == ".80/r2", , drop = FALSE]
if (nrow(default_AO) != 1L || !default_AO$layer %in% c("L1", "L2"))
  stop("The declared profile did not deliver a Step-2 proposal.")
K_step2 <- as.integer(default_AO$Kp)
K0      <- ncol(AO$Q0)
Lam     <- AO$loadings[[as.character(K_step2)]]
d     <- unique(colSums(AO$Q0 == 1L))
if (length(d) != 1L || d[1] < 1L) {
  stop("Choose one marker depth before building Q2.")
}

Q2 <- cbind(AO$Q0, matrix(-1L, J, K_step2 - K0))
colnames(Q2) <- paste0("F", seq_len(K_step2))
if (K_step2 > K0) {
  for (k in (K0 + 1L):K_step2) {
    top <- head(order(-abs(Lam[, k]), seq_len(J)), d)
    Q2[top, k] <- 1L
  }
}

c(K_step2 = K_step2, K0 = K0, depth = d)
Q2[rowSums(Q2 == 1L) > 0L, , drop = FALSE]
```

The added column inherits the backbone depth of two and uses items 13 and 18,
both from the cluster the two-column backbone did not cover.

The same group Q now goes to both models. The bifactor general column is added
inside `vbfa()` and never participates in marker selection.

```{r step2-fit}
f_obl <- vbfa(Y, Q2)
f_bif <- vbfa(Y, Q2, bifactor = TRUE)

idx <- c("RMSEA", "SRMR", "CFI", "TLI", "AIC", "BIC")
round(rbind(
  oblique  = fit_stats(f_obl)[idx],
  bifactor = fit_stats(f_bif)[idx]
), 3)
```

The bifactor arm is ahead on every index, which is the right answer for this
population, but by very unequal margins: 71.2 AIC points and only 5.6 BIC
points. It is an operational comparison of two fitted models on one design, not
a test against an unrestricted alternative, and it is conditional on the count
and the markers that Step 1 delivered. A thin margin, or a margin that flips
when the Step-1 count changes, should be reported rather than resolved.

An analyst may instead supply a substantive `J x K_step2` group-loading Q of their
own, never one containing the bifactor general column. That is a different
analysis design: it should be labelled as such, recorded, validated for codes
and dimnames, and used identically in both arms. It is not a silent default and
it is not a `pefa()` decision.

`special_effects()` then describes the retained bifactor fit: within-cluster
Schmid-Leiman ratios and their CVs, the higher-order parameters those ratios
imply, and group effect sizes. Print the object rather than pulling out
columns --- the print method carries the reading rule the numbers need.

```{r step2-effects}
special_effects(f_bif)
```

The item-weighted mean CV is 0.297, roughly three times the ~.1 convention,
and that is the expected answer here: the population drew each item's general
and group loading independently, so the within-cluster ratios cannot be
proportional by construction. These data are genuinely bifactor, not a
higher-order model in Schmid-Leiman clothing. The `gamma` column is still
printed, but for this fit it is the deliberately lossy summary the print
method warns about and must not be reported as a set of second-order
loadings.

## The same window as a bifactor sweep

`pefa(..., bifactor = TRUE)` sweeps the number of group factors. The backbone
supplies group columns only, the general column is added internally, and the
comparator strips it before matching, so the correspondence and the seven pair
facts have exactly the definitions used above.

```{r bifactor-sweep}
BI <- pefa(Q0_AO, Y, Kmin = 2, Kmax = 5, bifactor = TRUE, verbose = FALSE)
BI$sweep[, c("K", "ELBO", "AIC", "BIC", "SRMR", "CFI")]
BI$transitions[, c("K_from", "K_to", "ELBO_gain_pct", "BIC_gain_pct",
                   "phi_min", "rmsd_max", "unmatched_ssl", "collision")]
round(BI$persistence$phi, 3)
round(BI$persistence$rmsd, 3)
ssl(BI)
```

In this example the bifactor sweep is much less sharp than the ordinary one.

* `ELBO_gain_pct` is `NA` for every row. ELBO is highest at the window's lower
  edge and falls at every step, so that path contains no positive gain and its
  whole percentage column is `NA` by construction. The declared reader above
  treats the complete finite raw ELBO path as usable but returns no finite
  `Khat_ELBO`; that convention belongs to the analysis, not the package.
  `BIC_gain_pct` does have a positive step, at `3 -> 4`, and normalizes to it.
* `phi_min` is 0.998 or better in every cell that involves `K = 2` and 0.990 at
  `4 -> 5`, then drops to 0.444 and 0.519 in the two cells that start at
  `K = 3`. No `rmsd_max` cell exceeds 0.122, so no column moved far in absolute
  terms; those two low congruences are the weakest-link statistic scoring a
  group column that has been emptied.
* `ssl()` says why. The general column holds 4.29-4.37 and the two backbone
  group columns hold about 2.09-2.25 at every candidate, while the third group
  column reaches only 0.27 at `K = 3` and collapses to the order of 1e-9 to
  1e-7 at `K = 4` and `K = 5`. With the general factor already carrying the
  shared variance, the third cluster's specific variance never accumulates into
  a group column here. The four- and five-factor candidates are the two-group
  solution again --- `$persistence$phi` reads 1.000 and 0.999 against `K = 2`
  --- which is also why the four-factor candidate's two empty columns pick the
  same target and raise `collision` at `4 -> 5`.

Read that as a property of this example --- this population, this two-column
backbone, this window, this sample size --- and not as a general ranking. A
bifactor sweep measures the same quantities as an ordinary one, and whether it
separates candidates as well depends on how much group-specific variance is
left once a general factor is in the model. A study that needs both readings
should run both on the same data, backbone, and window rather than assume they
must agree.

## References

* Schmid, J., & Leiman, J. M. (1957). The development of hierarchical factor
  solutions. *Psychometrika*, 22(1), 53-61.
* Yung, Y.-F., Thissen, D., & McLeod, L. D. (1999). On the relationship
  between the higher-order factor model and the hierarchical factor model.
  *Psychometrika*, 64(2), 113-128.
* Zhang, Y., & Chen, J. (2024). Accommodating and extending various models for
  special effects within the generalized partially confirmatory factor
  analysis framework. *Applied Psychological Measurement*, 48(4-5), 208-229.
  \doi{10.1177/01466216241261704}
* Chen, J., & Jin, Y. (2026). Recovering latent structures after variational
  Bayesian variable selection: Fit assessment and factor-number selection in
  partially exploratory factor analysis. *arXiv preprint* arXiv:2607.07159.

---
title: "MIMIC models with vbmimic: regularizing measurement and structure"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{MIMIC models with vbmimic: regularizing measurement and structure}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(vbpm)
```

A MIMIC (multiple-indicators multiple-causes) model adds a structural part to
the factor model: observed covariates $X$ predict the factors, which in turn
produce the item responses $Y$. `vbmimic()` implements the extended MIMIC model
of Jin & Chen (2025, *Multivariate Behavioral Research*), whose contribution is
to regularize **both** parts: spike-and-slab priors select unspecified entries
of the measurement design `Q_A` (items on factors, `J x K`) *and* of the
structural design `Q_B` (factors on covariates, `K x P`). Both matrices use the
same `-1/0/1` codes as `vbfa()`.

With the default settings, `vbmimic()` reproduces the published estimator
exactly.

## Simulating MIMIC data

`sim_lvm()` (absorbed from the LAWBL package) generates the data. Supply the
structural coefficients as a matrix — a sparse design where each factor is
predicted by its own covariates is both realistic and estimable:

```{r simulate}
B <- matrix(0, 3, 9)
for (k in 1:3) B[k, (k * 3 - 2):(k * 3)] <- .3   # 3 covariates per factor

sim <- sim_lvm(N = 500, K = 3, J = 18, P = 9, b = B, phx = 0, rseed = 1)
Y <- sim$dat[, 1:18]        # items first ...
X <- sim$dat[, 19:27]       # ... covariates last
```

(A scalar `b` would make *every* covariate predict *every* factor, which
quickly leaves no disturbance variance — `sim_lvm()` stops with an informative
error if the design is infeasible.)

## Anchor at least one part

One modelling caution, learned the empirical way: with **both** `Q_A` and
`Q_B` fully exploratory the model converges cleanly but the solution is
rotationally ambiguous — factors can merge or swap. Anchoring either part
resolves it. Here we anchor two items per factor in the measurement part and
leave the whole structural part to the data:

```{r design}
## Q_A is an AZ (anchor-zero) design: each anchor is specified (1) on its own
## factor and fixed to zero on the other two
Q_A <- matrix(-1L, 18, 3)
for (k in 1:3) {
  a <- which(rep(1:3, each = 6) == k)[1:2]
  Q_A[a, ] <- 0L
  Q_A[a, k] <- 1L
}
Q_B <- matrix(-1L, 3, 9)     # structural selection is the question
```

AZ (anchor-zero) and AO (anchor-only) are the two anchor conventions `vbpm`'s
documentation uses throughout: AO codes each anchor's intended cell `1` and
leaves its cells on the other factors `-1`, while AZ additionally fixes those
cells to `0`. AZ is the stronger claim, and it is what anchors the measurement
part here; `vignette("bifactor")` compares the two.

## Fit and read

```{r fit}
fit <- vbmimic(Y, X, Q_A, Q_B)
fit
```

The structural question — *which covariates predict which factors?* — is
answered by `B` and its posterior inclusion probabilities `pi_B`:

```{r structural}
round(fit$B, 2)
round(fit$pi_B, 2)
```

Compare PIP-based selection with the generating design:

```{r recovery}
selected <- fit$pi_B >= .5
table(truth = B != 0, selected = selected)
```

Read that table both ways. All nine generating coefficients are selected — no
false negatives — but so are two of the eighteen true zeros, and not
marginally: both sit at PIP 1.00 in the matrix above (`B[2, 3]` and
`B[3, 4]`, each a factor picking up a covariate from a neighbouring block).
Exploratory structural selection at this `N` is therefore not free of false
positives, and a PIP at the ceiling is not on its own evidence that a
coefficient is real. Magnitude is the useful second filter here: the two false
positives are estimated at -0.18 and -0.14, below every one of the nine true
coefficients, whose smallest estimate is 0.25 against a generating value of
0.30.

The measurement side reads exactly as in `vbfa()`:

```{r measurement}
round(fit$A, 2)[1:6, ]
round(fit$Phi, 2)            # factor correlations, from the disturbances
```

## Missing responses

`vbmimic()` accepts `NA` in `Y`; a handful of missing item responses is
enough to demonstrate it:

```{r missing}
Ym <- Y
Ym[cbind(1:20, rep(1:4, each = 5))] <- NA
sum(is.na(Ym))

fitm <- vbmimic(Ym, X, Q_A, Q_B)
fitm$preprocess$n_missing

## structural and measurement recovery are essentially unaffected
max(abs(fitm$B - fit$B))
max(abs(fitm$A - fit$A))
```

The residual covariance here is diagonal (`vbmimic()` has no LD branch), so a
missing response is replaced in-loop by its conditional mean `eta_i A'`, with
the conditional variance `1 / V_j` carried into the residual sum of squares —
the same conditional-moment logic as `vbfa()`'s diagonal case, simplified
because there is no cross-item residual covariance to condition on. As with
`vbfa()`, this in-loop treatment is valid under **missing at random (MAR)**
(Chen, 2021).

`X` must be **complete**. Covariates are conditioned on, not modelled, so
there is no distribution to impute them from:

```{r missing-x, error = TRUE}
Xm <- X; Xm[1, 1] <- NA
vbmimic(Y, Xm, Q_A, Q_B)
```

## Options worth knowing

* **`v0` path.** As in `vbfa()`, a decreasing vector gives a warm-started
  regularization path over both parts; the scalar default `0.001` is the
  published fixed spike.
* **`standardize`.** The default `FALSE` is the published estimator's
  behaviour. `vbfa()` by contrast always standardizes internally; set
  `standardize = TRUE` for the analogous behaviour.
* **Determinism.** Like `vbfa()`, the estimator consumes no random numbers —
  no seed argument, bit-identical reruns.
* **Fit statistics.** The published algorithm converges on a residual
  criterion, not an ELBO; `fit$ELBO` is `NA`. `fit_stats()` accepts a
  `vbmimic` fit (it dispatches to a dedicated method) but returns a
  deliberately limited result: `NA` for every SEM-like index (`RMSEA`, `BIC`,
  `ELBO`, `objective`, ...) and a single `n_active_coef` — the count of
  soft-selected measurement plus structural coefficients, not a full
  parameter count, and not to be fed to an information criterion. `pefa()`
  sweeps `vbfa()` from one backbone `Q0`; it does not sweep MIMIC models.

## References

* Jin, Y., & Chen, J. (2025). Regularized variational Bayesian approximations
  for variable selection in extended multiple-indicators multiple-causes
  models. *Multivariate Behavioral Research*.
  https://doi.org/10.1080/00273171.2025.2483253
* Chen, J. (2021). A generalized partially confirmatory factor analysis
  framework with mixed Bayesian Lasso methods. *Multivariate Behavioral
  Research*, 57(6), 879–894. https://doi.org/10.1080/00273171.2021.1925520
* Ročková, V., & George, E. I. (2018). The spike-and-slab LASSO. *Journal of
  the American Statistical Association*, 113(521), 431–444.
  https://doi.org/10.1080/01621459.2016.1260469

---
title: "PEFA: Factor-Count Evidence from a Fixed Window"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{PEFA: Factor-Count Evidence from a Fixed Window}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(vbpm)
```

`pefa()` fits one partially exploratory factor model for every integer `K` in a
fixed consecutive window and reports what it measured: a candidate table, the
adjacent criterion gains, and the direct structural comparison of every ordered
pair of candidates. It **chooses nothing**: there is no count, cut, threshold,
or stopping rule inside the package, and no argument that would add one.

This vignette walks one sweep end to end, deliberately an *overextraction*
example: the population has three factors and the window runs to seven, so the
point is what the three tables look like when the extra columns are surplus.
Bifactor sweeps and the step that follows a count are in
`vignette("bifactor")`.

## Decision inventory

Six things must be decided before a factor count comes out of a sweep. The
package decides none of them; each line below names the field that carries the
evidence.

* **The window.** `Kmin:Kmax` bounds every number under it; a reading at the
  edge of `$sweep$K` is a reading about the window.
* **The criterion.** `$sweep` carries `ELBO`, `AIC`, `BIC`, and four
  descriptive indices side by side and ranks none of them.
* **The count rule and its variant.** `$transitions$ELBO_gain_pct` and
  `BIC_gain_pct` are the raw material; the cut, the look-ahead depth, and the
  boundary fallback are yours.
* **The collision policy.** `$transitions$collision` and
  `$persistence$collision` record target reuse; whether reuse disqualifies a
  candidate is your call.
* **The persistence horizon.** A row of `$persistence$phi` says how far a
  candidate's columns survive; how far they must survive is not in the object.
* **The tie-break.** When two criterion paths disagree, nothing arbitrates:
  `$sweep` reports both.

## One sweep

Three factors correlated at 0.5, six items each with main loadings of 0.7, and
two cross-loadings of 0.3 per factor with alternating signs --- the `sim_fa()`
defaults. The backbone is **anchor-only (AO)**: two anchors per backbone factor
get their intended cell fixed to `1`, and every other cell stays `-1`
(regularized, otherwise unspecified).

```{r sweep}
sim <- sim_fa(N = 500, K = 3, ipf = 6, lam = .7, lac = .3, rseed = 11)
Y <- sim$dat
groups <- rep(1:3, each = 6)

Q0 <- matrix(-1L, ncol(Y), 2)
for (k in 1:2) Q0[which(groups == k)[1:2], k] <- 1L
Q0[1:8, ]

p <- pefa(Q0, Y, Kmin = 2, Kmax = 7, verbose = FALSE)
p
```

`persistence 3 x 6 x 6` in that header means three 6 x 6 matrices, one per
measure, not a three-dimensional array.

The backbone has `K0 = 2` columns, so the window may start no lower than 2.
For each candidate `K`, `pefa()` appends `K - K0` fully regularized columns to
`Q0` and fits that design with `vbfa()`; the first two group columns keep the
backbone's identity by position, the rest are exploratory.

The call omits `v0` and `max_it`, so every candidate is fitted at the
documented `pefa()` defaults: the four-step spike ladder
`c(.01, .005, .002, .001)` and `max_it = 10000`, a cap **per ladder stage**
rather than per fit. `vbfa()` defaults to the same ladder and differs only in
that cap, at 5000.

The object has exactly seven components and class `"pefa"`.

```{r object}
names(p)
class(p)
names(p$settings)
```

`$settings` holds the resolved fitting and fit-statistic controls only --- no
threshold or rule. Candidate matrices are stored in two K-named lists.

```{r matrices}
names(p$loadings)
names(p$pips)
round(p$loadings[["3"]], 2)
round(p$pips[["3"]][1:6, ], 2)
```

A PIP is the posterior probability that a regularized cell is selected. Fixed
backbone cells are not regularized, so their PIPs are structural rather than
informative.

## The candidate table

```{r sweep-table}
p$sweep
```

One row per `K`, always the same eleven columns. `AIC` and `BIC` are the
hard-selection values at the resolved `tau` and `t` is their effective
parameter count. `RMSEA`, `SRMR`, `CFI`, and `TLI` are descriptive and may be
`NA` when their own definition does not apply --- `RMSEA` and `TLI` need
positive degrees of freedom. `converged` reflects the last ladder stage only: a
candidate whose last stage ran to `max_it` is retained with `converged = FALSE`
and one aggregate warning, and `iter` totals the iterations over all four
stages. Every candidate converged here.

ELBO peaks at `K = 3` (-10936.65) and falls afterwards. BIC is minimized at the
same candidate, 21709.40, and rises at every later one. AIC is minimized there
as well, at 21519.83, and also rises monotonically --- at the default spike
ladder the three criteria agree on this data set. The descriptive indices point
the same way rather than adding discrimination: RMSEA (0.017) and SRMR (0.035)
are lowest at `K = 3`, and CFI and TLI peak there at 0.996 and 0.995.

## Adjacent transitions

```{r transitions}
p$transitions
```

One row per adjacent pair. `ELBO_gain_pct` and `BIC_gain_pct` are oriented so
that positive favors `K_to`: the raw gains are `ELBO_to - ELBO_from` and
`BIC_from - BIC_to`, and each is divided by the largest positive gain on its
own path and multiplied by 100. Here the largest positive gain on both paths is
the `2 -> 3` step, so that row is 100 and every later row is negative. If a
path has no positive gain at all, its whole percentage column is `NA`.

The remaining seven columns describe structural change. Backbone columns are
paired by position. Each exploratory source column is paired with the closest
target after allowing a whole-column sign flip. See `?pefa` for the exact
distance and tie rules.

The assignment is independent, not one-to-one, so two sources may pick the same
target. That reuse is a **collision**, recorded rather than repaired; the
`5 -> 6` and `6 -> 7` rows carry it here.

## Persistence, read directly

There is no accessor function. The three upper-triangular matrices over the
fitted `K` values are read with `$`: `phi` (holding `phi_min`), `rmsd` (holding
**`rmsd_max`**, the weakest-column value, not the pooled `rmsd` of
`$transitions`), and the logical collision mask.

```{r persistence}
names(p$persistence)
round(p$persistence$phi, 3)
round(p$persistence$rmsd, 3)
p$persistence$collision
```

Every cell above the diagonal is a **direct** comparison of its two endpoint
loading matrices, not a product of the adjacent steps between them: `3 -> 7`
compares the three- and seven-factor solutions themselves. The diagonal and
lower triangle are typed `NA`.

Read the `K = 3` row across: 0.996 against `K = 4`, 0.990 against `K = 5`,
0.989 against `K = 6`, and then 0.789 against `K = 7`. Because `phi_min` is a
weakest-link statistic, 0.989 says the *worst*-matched of the three columns
still has a near-copy three candidates on, with the weakest-column RMSD no
worse than 0.062. That is what a high persistence row buys: one number retires
"the structure moved when I added a column" for the whole range out to
`K = 6`. Only the seven-factor cell breaks the row, at RMSD 0.253.

The `K = 2` row never exceeds 0.052 anywhere, and the two-factor loading matrix
says why.

```{r k2}
round(p$loadings[["2"]][c(1, 7, 8, 13, 17, 18), ], 2)
```

Backbone columns are compared by *position*, and at `K = 2` the second backbone
column is not holding what it holds from `K = 3` on. Items 13-18 have taken it
over --- the three shown load 0.69, 0.70, and 0.67 on it --- while its own
anchors, items 7 and 8, sit at 0.02 and 0.08, absorbed into the first column
alongside cluster 1. Its congruence with the second column of the three-factor
solution is 0.041, and that one weak pair is the whole of `phi_min` for every
`2 -> K` cell. Specifying an anchor cell leaves it free to be estimated; it
does not force the column to be about that cluster when there are too few
columns to go round.

The collision mask is `TRUE` at `5 -> 6`, `5 -> 7`, and `6 -> 7`, where two or
more source columns selected the same target. Those cells sit at `phi_min`
0.998, 0.284, and 0.294 --- a first sign that a collision is not by itself bad
news about the structure.

## Column strength

`ssl()` is the descriptive companion to those numbers: `colSums(Lam^2)` for
every stored candidate matrix. It screens nothing and enters no correspondence
rule.

```{r ssl}
ssl(p)
```

This is what a *low* `phi_min` needs to be read against. `phi_min` is a
**weakest-link** statistic over *every* source column, including one the fit
has driven to near zero. Such a column has no shape: it matches whichever
target has the smallest norm, contributes an arbitrary congruence, and, because
two of them pick the same target, raises `collision`. The `4 -> 5` row is
exactly that case --- `phi_min` 0.656 from a source column of squared length
0.003 matching a target of 0.007 --- while its pooled `rmsd` is 0.030 and its
`ari` is 1: not one item changed its dominant column. The `5 -> 6` and `6 -> 7`
collisions come from the same near-empty columns.

`unmatched_ssl` is on the same scale: the largest sum of squared loadings among
the exploratory target columns that no exploratory source selected, the biggest
column the smaller candidate did *not* account for. Across the adjacent rows it
reads 3.049, 0.003, 0.007, 0.005, 0.642. At `2 -> 3` the surplus column carries
3.05, indistinguishable from the three genuine columns of the three-factor
solution (3.08, 3.14, 3.05): the two-factor candidate really was missing a
factor. At `3 -> 4`, `4 -> 5`, and `5 -> 6` it is under 0.01: each of those
candidates adds a column with nothing in it and leaves the three real columns
alone.

`6 -> 7` breaks that pattern, and `ssl(p)` shows why. The seven-factor candidate
does append two empty columns --- `F6` and `F7`, both at 0.003 --- but neither
of them is the *surplus* one. `unmatched_ssl` here is 0.642, which is exactly
`ssl(p)[["7"]]["F3"]`, and the third cluster has been broken into pieces of
1.853, 0.642, and 0.570.

```{r k7}
round(p$loadings[["7"]][13:18, 3:5], 2)
```

Items 17 and 18 have each been given a column of their own --- a genuinely
different structure, and what drives the `3 -> 7` congruence down to 0.789
after three cells at 0.989 or better. The `6 -> 7` value of 0.294 is a separate
matter: like `4 -> 5`, it is two empty columns being compared.

## What the three readings say here

The criterion path, the persistence row, and `unmatched_ssl` against `ssl()`
point the same way in this example.

* ELBO, AIC, and BIC all turn at `K = 3`, and every later adjacent gain is
  negative on both percentage paths.
* The three-factor columns persist into `K = 4`, `5`, and `6` at congruence
  0.989 or better: the extra columns are added beside that structure.
* The column added at `K = 3` carries 3.05, a full factor's worth; those added
  at `K = 4`, `5`, and `6` carry under 0.01 each.

The population had three factors, so the reading is right. **It is still the
reader's inference, not a package verdict.** The sentence "the structure is
three-dimensional and the later columns are surplus" was composed here, in the
vignette, out of six candidate rows, five transition rows, and three triangles.
Two of those numbers point the other way if read alone: `phi_min` is 0.656 at
`4 -> 5` and 0.294 at `6 -> 7`, structural instability until `ssl()` shows that
both belong to columns with nothing in them.

## Analysis-side dual-count and persistence profiles

The following compact reader illustrates one declared analysis. ELBO and BIC
are co-primary descriptive count paths: each uses a 20% gain cutoff, sustain
one, a strict scan after the last maximum gain, the **full fitted window**, and
no boundary fallback. Their finite readings form
$C_{20}=\{\widehat K_{ELBO},\widehat K_{BIC}\}$ after absent values are removed.
Neither path breaks a tie or outranks the other.

The same reader applies three scenario-bound minimum-congruence profiles:
`.85/r1`, `.80/r2`, and `.70/r3`, with `.80/r2` labelled the practical default
for the motivating study. Each source is compared directly with every endpoint
through `K + r`; no adjacent chain is substituted. A known collision vetoes
that edge. The common declared source set is `2:4`, chosen so the deepest
profile has all required endpoints in the fitted window. $K_p$ is the highest
persistent source only after every higher declared source has resolved
nonpersistent.

```{r reading}
profile_specs <- data.frame(
  profile = c(".85/r1", ".80/r2", ".70/r3"),
  phi_cut = c(.85, .80, .70), r = 1:3,
  practical_default = c(FALSE, TRUE, FALSE)
)
profile_sources <- 2:4

read_sweep <- function(x, phi_cut, r, sources) {
  K <- as.integer(x$sweep$K)
  converged <- x$sweep$converged

  read_count <- function(score) {
    usable <- length(K) >= 2L && all(is.finite(K)) && all(diff(K) == 1L) &&
      length(score) == length(K) && all(is.finite(score)) &&
      is.logical(converged) && all(converged %in% TRUE)
    if (!usable) return(list(usable = FALSE, Khat = NA_integer_))
    gain <- diff(score)
    gain_max <- max(gain)
    if (!(gain_max > 0)) return(list(usable = TRUE, Khat = NA_integer_))
    peak <- max(which(gain == gain_max))
    hit <- which(seq_along(gain) > peak & gain < .20 * gain_max)
    list(usable = TRUE,
         Khat = if (length(hit)) K[hit[1L]] else NA_integer_)
  }

  ELBO <- read_count(x$sweep$ELBO)
  BIC <- read_count(-x$sweep$BIC)
  C20 <- sort(unique(c(ELBO$Khat, BIC$Khat)[
    is.finite(c(ELBO$Khat, BIC$Khat))]))

  phi <- x$persistence$phi
  collision <- x$persistence$collision
  conv <- setNames(converged, K)
  state <- vapply(sources, function(k) {
    targets <- k + seq_len(r)
    edge <- vapply(targets, function(target) {
      from <- as.character(k); to <- as.character(target)
      if (!from %in% rownames(phi) || !to %in% colnames(phi) ||
          !isTRUE(conv[[from]]) || !isTRUE(conv[[to]]) ||
          is.na(collision[from, to])) return(NA_integer_)
      if (isTRUE(collision[from, to])) return(0L)
      value <- phi[from, to]
      if (!is.finite(value)) NA_integer_ else as.integer(value >= phi_cut)
    }, integer(1))
    if (any(edge %in% 0L)) 0L else
      if (length(edge) && all(edge %in% 1L)) 1L else NA_integer_
  }, integer(1))

  persistent <- which(state == 1L)
  if (!length(state)) {
    resolution <- "unresolved"
    Kp <- NA_integer_
  } else if (!length(persistent)) {
    resolution <- if (all(state %in% 0L)) "nonpersistent" else "unresolved"
    Kp <- NA_integer_
  } else {
    top <- max(persistent)
    higher <- if (top < length(state)) state[(top + 1L):length(state)] else 0L
    resolution <- if (all(higher %in% 0L)) "persistent" else "unresolved"
    Kp <- if (resolution == "persistent") sources[top] else NA_integer_
  }

  hit_ELBO <- is.finite(Kp) && is.finite(ELBO$Khat) && ELBO$Khat == Kp
  hit_BIC <- is.finite(Kp) && is.finite(BIC$Khat) && BIC$Khat == Kp
  support <- if (hit_ELBO && hit_BIC) "both" else if (hit_ELBO) "ELBO only" else
    if (hit_BIC) "BIC only" else "none"
  any_count_usable <- ELBO$usable || BIC$usable
  layer <- if (resolution == "persistent" && Kp %in% C20) "L1" else
    if (resolution == "persistent" && any_count_usable) "L2" else
    if (resolution == "nonpersistent") "L3" else
      "unclassified"
  gap <- if (resolution == "persistent" && length(C20))
    min(abs(C20 - Kp)) else NA_real_

  list(Khat_ELBO = ELBO$Khat, Khat_BIC = BIC$Khat, C20 = C20,
       count_usable = c(ELBO = ELBO$usable, BIC = BIC$usable),
       support = support, Kp = Kp, source_state = setNames(state, sources),
       resolution = resolution, layer = layer, g_CP = gap)
}

profile_row <- function(x, spec) {
  z <- read_sweep(x, spec$phi_cut, spec$r, profile_sources)
  data.frame(profile = spec$profile, practical_default = spec$practical_default,
             Khat_ELBO = z$Khat_ELBO, Khat_BIC = z$Khat_BIC,
             C20 = if (length(z$C20)) paste0("{", paste(z$C20,
               collapse = ","), "}") else "{}",
             support = z$support, Kp = z$Kp, layer = z$layer,
             g_CP = z$g_CP, row.names = NULL)
}

profile_results <- do.call(rbind, lapply(seq_len(nrow(profile_specs)),
  function(i) profile_row(p, profile_specs[i, ])))
profile_results

## A declared source at the upper edge cannot persist without its endpoint.
short_window <- read_sweep(p, phi_cut = .85, r = 1L, sources = 7L)
data.frame(profile = ".85/r1", declared_source = 7L,
           required_endpoint = 8L, resolution = short_window$resolution,
           layer = short_window$layer)
```

Here both gain paths read three, so $C_{20}=\{3\}$. The three structural rows
also select three and receive L1 with descriptive support from both criteria.
The support label records whether one or both count paths corroborate $K_p$; it
does not rank the paths or select the structure.

Agreement across all three declared profiles is **within-family robustness** to
these particular cutoff/depth combinations. It is not evidence that three is
the population truth, a validation result, or a package decision. Conversely,
profile disagreement would describe sensitivity within this family rather than
identify which row is true.

The final printed row is the short-window counterexample. Source 7 under
`.85/r1` needs the direct `7 -> 8` endpoint, but this fit stops at 7; its
persistence resolution and layer are therefore `unresolved` and `unclassified`,
not L3 and not a reason to shrink the declared source set.

L1 means that a selected $K_p$ belongs to finite $C_{20}$. L2 means that $K_p$
is selected but does not belong to $C_{20}$ while at least one count path is
usable. L3 requires every declared source to resolve nonpersistent regardless
of count-path usability. All remaining cases are unclassified. The displayed
$g_{CP}=\min_{\widehat K\in C_{20}}|\widehat K-K_p|$ is corroboration only,
never a persistence or delivery criterion, and is undefined when there is no
finite count reading. Algebraically, $g_{CP}=0$ merely restates the L1
membership condition $K_p\in C_{20}$; only positive gap magnitudes add a
descriptive measure of how far an L2 count disagreement lies.

If an analysis continues to Step 2, only an L1/L2 delivery is eligible and then
$K_{\mathrm{step2}}=K_p$. Thus the Step-2 count is $K_p$ conditional on
delivery, not another count selector; `vignette("bifactor")` shows the design
construction.

These definitions, including the three profiles and the practical-default
label, belong to this analysis rather than **vbpm**. A different declared
source set, horizon, cutoff, collision policy, or window can change the result,
and an unlabelled reading copied from a vignette can quietly become a de facto
default. Gain fractions are relative to the largest positive gain in the full
declared window. Widening that window can change the denominator and the
reading; **vbpm** 0.9.1 has no reassembly API, so a wider window requires
another complete `pefa()` call.

## Display methods

```{r summary}
summary(p)
```

`summary()` returns exactly `window`, `sweep`, `transitions`, `persistence`,
`ssl`, `settings`, and `nonconverged_K`. Its print method marks a persistence
value with `*` wherever the parallel collision cell is `TRUE`; that is display
formatting only, and the stored matrices are untouched.

The plot methods are descriptive trajectories, with no threshold line and no
marked count.

```{r plots, fig.width=7, fig.height=4.5}
plot(p, type = "objective", criterion = "ELBO")
plot(p, type = "gain", criterion = "BIC", pct = TRUE)
plot(p, type = "fit")
```

The ELBO path turns visibly at `K = 3`, and both fit panels turn with it. The
gain view is indexed by `K_to`, so its point at `x = 3` is the `2 -> 3` step,
not the three-factor candidate; `pct = TRUE` reads the stored percentage
column, and the default `pct = FALSE` recomputes the raw oriented gains from
`$sweep`.

## Scope

What `pefa()` returns is measurement. The named profiles above illustrate
analysis-owned readings; they do not become package selectors or defaults, and
there is no automatic window extension. A failed or
malformed candidate aborts the call with an error naming its `K`; a
nonconverged but well-formed one is retained and reported.

The reading above is a property of this simulated example --- 500 observations,
main loadings of 0.7, three clean clusters, an anchored backbone --- and not
evidence about how these tables behave at weaker signal.

## References

* Chen, J. (2023). Fully and partially exploratory factor analysis with
  bi-level Bayesian regularization. *Behavior Research Methods*, 55(4),
  2125-2142. \doi{10.3758/s13428-022-01884-7}
* Chen, J., & Jin, Y. (2026). Recovering latent structures after variational
  Bayesian variable selection: Fit assessment and factor-number selection in
  partially exploratory factor analysis. *arXiv preprint* arXiv:2607.07159.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{nlsy27}
\alias{nlsy27}
\title{National Longitudinal Survey of Youth 1997 (27 items)}
\format{
A list with three components:
\describe{
\item{dat}{A \verb{3458 x 27} data frame of item responses (17 polytomous
items, 10 continuous), containing \code{NA}s.}
\item{Q}{A \verb{27 x 3} initial design matrix with two to three specified
loadings per factor, in the \code{-1/0/1} coding used by \code{\link[=vbfa]{vbfa()}}.}
\item{cati}{Indices of the 17 polytomous items.}
}
}
\source{
Bureau of Labor Statistics, U.S. Department of Labor. National
Longitudinal Survey of Youth 1997 cohort. Previously distributed in the
LAWBL package (Chen, 2022).
}
\usage{
nlsy27
}
\description{
Responses of 3,458 individuals to 27 mixed-type items, with 1.12\% missing
data. Carried over from the LAWBL package (which \code{vbpm} supersedes) for
empirical illustration of the estimators.
}
\examples{
data(nlsy27)
dim(nlsy27$dat)
nlsy27$Q
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vbmimic.R
\name{vbmimic}
\alias{vbmimic}
\title{Regularized variational Bayes for extended MIMIC models}
\usage{
vbmimic(
  Y,
  X,
  Q_A,
  Q_B,
  v0 = 0.001,
  standardize = FALSE,
  max_it = 1000,
  convChk = FALSE,
  tolVal = 1e-05
)
}
\arguments{
\item{Y}{Numeric \verb{N x J} data matrix (observations by items). Missing values
in \code{Y} are supported: because the residual covariance is diagonal, a
missing response is replaced in-loop by its conditional expectation
\verb{eta_i A'}, with the conditional variance carried into the residual sum
of squares. Items or rows with no observed response are rejected.}

\item{X}{Numeric \verb{N x P} matrix of observed covariates (predictors of the
factors), with the same number of rows as \code{Y}. Missing covariates are
\strong{not} supported: \code{X} is conditioned on rather than modelled, so
imputing it would require a distributional assumption the MIMIC model
does not make.}

\item{Q_A}{Integer \verb{J x K} measurement design matrix: \code{1} = specified
(anchored) loading, \code{0} = fixed zero, \code{-1} = unspecified (selected by
spike-and-slab). \code{K} is the number of factors.}

\item{Q_B}{Integer \verb{K x P} structural design matrix using the same coding,
for the regression of the \code{K} factors on the \code{P} covariates.}

\item{v0}{Spike variance. A scalar gives a single fixed spike; a decreasing
vector gives a warm-started regularization path, each stage initialized at
the previous stage's solution (Rockova & George, 2018). The same schedule
is applied to the measurement and structural parts. \code{v0 = 0.001} (the
default) is the fixed spike of the published estimator.}

\item{standardize}{Logical; standardize \code{Y} and \code{X} before fitting. The
default \code{FALSE} is the behaviour of the published estimator. Note that
\code{\link[=vbfa]{vbfa()}} always standardizes internally; set \code{standardize = TRUE} for the
analogous behaviour here.}

\item{max_it}{Maximum number of variational iterations \emph{per stage}.}

\item{convChk}{Logical; print per-iteration convergence information via
\code{\link[=message]{message()}}. Default \code{FALSE} (quiet).}

\item{tolVal}{Convergence tolerance on the maximum absolute change in the
weighted residual vector.}
}
\value{
An object of class \code{c("vbmimic", "vbpm_fit")} -- a named list (see
\link{vbpm_fit}) with components:
\describe{
\item{A}{\verb{J x K} posterior mean measurement loadings.}
\item{B}{\verb{K x P} posterior mean structural coefficients.}
\item{pi_A, pi_B}{Posterior inclusion probabilities for the unspecified
entries of \code{Q_A} and \code{Q_B} (specified entries are held at 1, fixed
zeros at 0).}
\item{Q_A, Q_B}{The design matrices the model was fit with.}
\item{eta}{\verb{N x K} posterior mean factor scores.}
\item{Phi}{\verb{K x K} factor correlation matrix implied by \code{Sig}.}
\item{Sig, U}{Factor covariance matrix and its inverse.}
\item{V}{Length-\code{J} residual precisions.}
\item{rho, theta}{Inclusion rates for the measurement and structural
parts.}
\item{A_var, B_var}{Posterior variances of the loadings and structural
coefficients.}
\item{iter, flag, time}{Total iterations, \code{1} if the final stage
converged within \code{max_it} (else \code{0}), and elapsed time.}
\item{ELBO}{\code{NA_real_}; see Note.}
\item{standardize}{The \code{standardize} setting the model was fit with.}
\item{path}{The \code{v0} schedule and per-stage iteration counts.}
\item{preprocess}{\code{n_missing}, the response type, and \code{missing_mask} --
stored only when the data contain missing values, \code{NULL} otherwise.}
}

Every quantity appears exactly once: as of 0.7.0 the fit no longer
carries \verb{$coefficients}, \verb{$design}, \verb{$posterior}, or \verb{$settings}, which
duplicated the elements above.
}
\description{
Fits a partially confirmatory multiple-indicators multiple-causes (MIMIC)
model by a variational Bayes EM algorithm, with continuous spike-and-slab
regularization on \strong{both} the measurement and the structural part. Items
load on factors through a measurement design matrix \code{Q_A}, while covariates
predict the factors through a structural design matrix \code{Q_B}; entries of
either that are left unspecified are selected from the data.
}
\details{
This is the MIMIC member of the \code{vbpm} family and is the companion of
\code{\link[=vbfa]{vbfa()}}, which fits the measurement part alone. The two share the \code{-1/0/1}
design-matrix convention and the spike-and-slab formulation.
}
\note{
The evidence lower bound is \strong{not} currently returned. The published
algorithm converges on the weighted residual vector rather than on the
bound, and the bound's implementable closed form has not been assembled.
\code{\link[=fit_stats]{fit_stats()}} does accept a \code{vbmimic} fit, but returns a deliberately
limited result: \code{n_active_coef} (soft-selected measurement and structural
coefficients) plus typed \code{NA} for every covariance-based index and
parameter count, with an explanatory \code{note} attribute. \code{\link[=pefa]{pefa()}} does not
accept \code{vbmimic} fits at all: its sweep reports criterion gains across a
factor-count window, which requires one comparable objective for every
candidate in that window.
}
\examples{
## 3 factors, 18 items, 9 covariates in disjoint blocks of 3
B <- matrix(0, 3, 9)
for (k in 1:3) B[k, ((k - 1) * 3 + 1):(k * 3)] <- .3
sim <- sim_lvm(N = 300, K = 3, J = 18, P = 9, b = B, phx = 0, rseed = 1)
Y <- sim$dat[, 1:18]
X <- sim$dat[, 19:27]

## two anchored items per factor; the structural part left exploratory.
## Anchoring at least one part is advisable: with both Q_A and Q_B fully
## exploratory the solution can be rotationally ambiguous.
Q_A <- matrix(-1L, 18, 3)
for (k in 1:3) {
  a <- which(rep(1:3, each = 6) == k)[1:2]
  Q_A[a, ] <- 0L; Q_A[a, k] <- 1L
}
Q_B <- matrix(-1L, 3, 9)

fit <- vbmimic(Y, X, Q_A, Q_B)
fit$flag           # 1 if converged
round(fit$B, 2)    # structural coefficients

}
\references{
Jin, Y., & Chen, J. (2025). Regularized variational Bayesian approximations
for variable selection in extended multiple-indicators multiple-causes
models. \emph{Multivariate Behavioral Research}.
\doi{10.1080/00273171.2025.2483253}

Rockova, V., & George, E. I. (2018). The spike-and-slab LASSO.
\emph{Journal of the American Statistical Association}, 113(521), 431-444.
}
\seealso{
\code{\link[=vbfa]{vbfa()}} for the measurement model alone, \code{\link[=sim_lvm]{sim_lvm()}} to simulate
MIMIC data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/special_effects.R
\name{special_effects}
\alias{special_effects}
\alias{print.special_effects}
\title{Special-effect summaries of a bifactor fit}
\usage{
special_effects(object, tau = 0.5)

\method{print}{special_effects}(x, digits = 3, ...)
}
\arguments{
\item{object}{A fitted \code{\link[=vbfa]{vbfa()}} object with a bifactor structure (see
Details).}

\item{tau}{Hard-selection threshold on the posterior inclusion
probability (default \code{0.5}), used both for group-membership inference
and for the Eq.-16 effect sizes.}

\item{x}{A \code{special_effects} object.}

\item{digits}{Number of decimals in the printed table.}

\item{...}{Unused.}
}
\value{
An object of class \code{special_effects}: a list with
\item{groups}{A data frame with one row per group factor: \code{n_items}
(assigned items), \code{mean_ratio}, \code{sd_ratio}, \code{cv}, \code{gamma}
(second-order loading), and \code{D} (Eq.-16 effect size).}
\item{mean_cv}{The item-weighted mean CV across group factors.}
\item{items}{A data frame with one row per item: assigned \code{group}
(\code{NA} if unassigned), \code{b_gen}, \code{b_grp}, \code{ratio}, and \code{lambda} (the
first-order loading implied by the higher-order parameterization).}
\item{gamma}{Named vector of second-order loadings (same as in
\code{groups}).}
\item{unassigned, crossloaded, no_general}{Integer vectors of item
indices flagged by the membership rules above.}
\item{tau, source, call}{Bookkeeping (\code{source} is \code{"bifactor"} for
native fits, \code{"legacy"} for hand-built ones).}
}
\description{
Post-processes an orthogonal bifactor \code{\link[=vbfa]{vbfa()}} fit into the three
summaries used throughout the \code{bifactor} vignette: (i) the
\strong{Schmid-Leiman (SL) proportionality check} -- the within-group
coefficient of variation (CV) of the group-to-general loading ratio, per
group factor and item-weighted overall; (ii) the \strong{(approximate)
higher-order parameterization} -- second-order loadings
\eqn{\hat\gamma_k = 1/\sqrt{1 + \bar r_k^2}} from each group's mean ratio
\eqn{\bar r_k}, and first-order loadings
\eqn{\hat\lambda_j = |b^{gen}_j| / \hat\gamma_{k(j)}}; and (iii)
\strong{testlet/special effect sizes} by Eq. 16 of Zhang and Chen (2024): the
average squared loading on each group factor, counting specified loadings
always and unspecified ones only when their posterior inclusion
probability exceeds \code{tau}.
}
\details{
All output is \strong{descriptive}. Near-zero CVs signal SL proportionality --
the fitted bifactor is then (approximately) a higher-order model, and by
the same token a testlet model, so the higher-order parameters are a
faithful re-expression. Values near \code{.1} are conventionally read as
approximately higher-order, but the cutoff is suggestive, not a test; the
higher-order parameters and effect sizes are reported regardless, and
with clearly non-constant ratios they are a deliberately lossy summary of
a genuinely richer bifactor. The ratio inherits the sampling softness of
the general/group split, so CVs are noisy below (roughly) a few thousand
observations -- see the \code{bifactor} vignette for demonstrations, caveats,
and the covariance geometry behind all three summaries.

\strong{Which fits are accepted.} A fit with \code{bifactor = TRUE} is used as-is
(general column first, group columns after). A legacy hand-built fit is
also accepted when it has \code{orthogonal = TRUE} and exactly one
all-specified column in \code{Q}, which is taken as the general factor. Group
membership of each item is the group column where it is anchored
(\code{Q == 1}) or, for unspecified rows, its single active group column
(inclusion probability above \code{tau}). Items with no active group loading
are excluded from the ratio summaries and reported; items active on
several group columns are assigned to the anchored one (else the largest
absolute loading) and reported. Ratios use absolute loadings; items whose
general loading is fixed at zero (bifactor-(S-1)-style designs) carry no
ratio and are reported.
}
\examples{
## higher-order data: the ratios are near-constant and gamma is recovered
sim <- sim_fa(N = 500, K = 3, ipf = 6, lam = .8, gamma = c(.8, .7, .6),
              rseed = 11)
Q <- matrix(-1L, 18, 3)
for (k in 1:3) { a <- which(rep(1:3, each = 6) == k)[1:2]
                 Q[a, ] <- 0L; Q[a, k] <- 1L }
fit <- vbfa(sim$dat, Q, bifactor = TRUE, v0 = .001)
special_effects(fit)

}
\references{
Schmid, J., & Leiman, J. M. (1957). The development of hierarchical
factor solutions. \emph{Psychometrika}, 22(1), 53-61.

Yung, Y.-F., Thissen, D., & McLeod, L. D. (1999). On the relationship
between the higher-order factor model and the hierarchical factor model.
\emph{Psychometrika}, 64(2), 113-128.

Zhang, Y., & Chen, J. (2024). Accommodating and extending various models
for special effects within the generalized partially confirmatory factor
analysis framework. \emph{Applied Psychological Measurement}, 48(4-5),
208-229. \doi{10.1177/01466216241261704}
}
\seealso{
\code{\link[=vbfa]{vbfa()}}, and the \code{bifactor} vignette for the full workflow.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pefa.R
\name{pefa}
\alias{pefa}
\title{Partially exploratory factor analysis over a factor-count window}
\usage{
pefa(
  Q0,
  Y,
  Kmin,
  Kmax,
  bifactor = FALSE,
  general = 1,
  v0 = c(0.01, 0.005, 0.002, 0.001),
  max_it = 10000,
  convChk = FALSE,
  tolVal = 1e-04,
  tau = 0.5,
  rank_adjust = FALSE,
  rank_max_J = 100,
  verbose = TRUE
)
}
\arguments{
\item{Q0}{A \code{J} by \code{K0} backbone with entries \code{1}, \code{0}, or \code{-1}. Zero
columns (\code{K0 = 0}, fully exploratory) are allowed.}

\item{Y}{A numeric \code{N} by \code{J} matrix of observed values, optionally with
literal \code{NA}; \code{NaN} and infinite values are rejected. Every row and every
item needs an observed value, and every item a positive finite SD.}

\item{Kmin, Kmax}{Inclusive window, whole numbers with
\verb{1 <= Kmin <= Kmax <= .Machine$integer.max} and \code{Kmin >= ncol(Q0)}.}

\item{bifactor}{Fit an orthogonal bifactor parameterization.}

\item{general}{Scalar or length-\code{J} general design in \verb{\{-1, 0, 1\}} with at
least one \code{1}, used only when \code{bifactor = TRUE}.}

\item{v0, max_it, convChk, tolVal}{Optimizer controls passed to \code{\link[=vbfa]{vbfa()}}:
finite positive \code{v0}, positive whole \code{max_it}, \code{TRUE}/\code{FALSE} \code{convChk},
and finite \code{tolVal >= 0}.}

\item{tau}{Loading-inclusion threshold in \verb{[0, 1]}, used by \code{\link[=fit_stats]{fit_stats()}}.}

\item{rank_adjust, rank_max_J}{Rank-adjustment controls for \code{\link[=fit_stats]{fit_stats()}}.
\code{rank_adjust = TRUE} requires \code{J <= rank_max_J} and the suggested
\code{numDeriv} package.}

\item{verbose}{Print one progress line per candidate.}
}
\value{
An object of class \code{"pefa"}; see \emph{The returned object} below.
}
\description{
Fits \code{\link[=vbfa]{vbfa()}} for every integer \code{K} in a fixed consecutive window, holding
the partially specified backbone \code{Q0} fixed and appending regularized
exploratory columns. It reports candidate summaries and between-candidate
loading measurements.
}
\details{
\code{pefa()} \strong{chooses nothing}: no factor count, threshold, persistence
horizon, collision policy, or stopping rule. Those belong to the analysis
that consumes the object.

Ordinary candidates are oblique with diagonal residuals. In bifactor mode
\code{K} counts group factors and \code{\link[=vbfa]{vbfa()}} adds one general column, so a stored
matrix has \code{K + 1} columns; every between-candidate comparison removes that
column first and works on the \code{K} group columns.

One private comparator produces every between-candidate number, and both
\code{transitions} and \code{persistence} are read from its stored results. The
accessors do no matching of their own: \code{\link[=ssl]{ssl()}}, \code{print()}, \code{summary()}, and
\code{plot()} only describe what is already stored, and none of them rematches
columns or applies an analysis rule. There is no \code{persistence()} accessor:
\code{x$persistence$phi}, \verb{$rmsd}, and \verb{$collision} are already the three
stored matrices.
}
\section{The returned object}{

A list of class \code{"pefa"} with exactly these components, in order.
\describe{
\item{\code{sweep}}{One row per candidate, in column order. \code{K} is integer;
\code{ELBO}, \code{AIC}, and \code{BIC} are required finite doubles; \code{RMSEA}, \code{SRMR},
\code{CFI}, and \code{TLI} are descriptive doubles; \code{t} is a double; \code{iter} is a
nonnegative integer, or \code{NA_integer_} when a well-formed engine result
reports none; \code{converged} is a nonmissing logical. \code{AIC} and \code{BIC} are
hard-selection values at the resolved \code{tau}, and \code{t} is their effective
parameter count -- a finite, nonnegative, whole-valued count stored as
a double -- rank-adjusted when \code{rank_adjust = TRUE}. Each descriptive
index is \code{NA_real_} exactly when its own definition does not apply, and
is never silently substituted.}
\item{\code{transitions}}{One row per adjacent pair: the integers \code{K_from} and
\code{K_to}, the doubles \code{ELBO_gain_pct} and \code{BIC_gain_pct}, then the seven
structural facts -- the doubles \code{phi_min}, \code{rmsd}, \code{rmsd_max}, \code{ari},
\code{pip_rmsd}, \code{unmatched_ssl} and the logical \code{collision}. Positive gains
favor \code{K_to}. Each percentage divides by the largest positive gain on
its own path, or is \code{NA_real_} when no positive gain exists. Raw
criteria stay in \code{sweep}.}
\item{\code{persistence}}{Three upper-triangular \code{W} by \code{W} matrices over the
fitted \code{K} values, named and ordered \code{phi}, \code{rmsd}, \code{collision}: \code{phi}
holds \code{phi_min}, \code{rmsd} holds \strong{\code{rmsd_max}}, and \code{collision} is the
logical reuse mask. Diagonal and lower triangle are \code{NA}. Every cell is
a direct endpoint comparison, never a chain through intervening
candidates.}
\item{\code{loadings}, \code{pips}}{Named \code{K}-indexed lists of candidate matrices
with items in rows and factors in columns: \code{J} by \code{K}, or \code{J} by
\code{K + 1} with the general column first in bifactor mode. Loadings are
finite and PIPs lie in \verb{[0, 1]}.}
\item{\code{Q0}, \code{settings}}{The canonical \code{J} by \code{K0} integer backbone, and
the resolved controls \code{bifactor}, \code{general}, \code{v0}, \code{max_it}, \code{convChk},
\code{tolVal}, \code{tau}, \code{rank_adjust}, \code{rank_max_J}, in that order. \code{general}
is \code{NULL} outside bifactor mode. \code{settings} records fitting and
fit-statistic controls only: no \code{verbose}, and no analysis rule.}
}

The loading and PIP matrices and the criteria \code{ELBO}, \code{AIC}, \code{BIC}, and \code{t}
are required. A fitting failure or a malformed required quantity aborts the
call with an error naming the \code{K}; there is no status taxonomy and no
partially populated object. A candidate that merely failed to converge is
retained with \code{converged = FALSE}, and one aggregate warning names every
affected \code{K}.
}

\section{Loading correspondence}{

For source column \code{a} and target column \code{b}, let \code{s = 1} when \verb{a'b >= 0} and
\code{-1} otherwise, and let \verb{d(a, b) = ||a - s b||^2}. Backbone columns pair by
position. Each exploratory source independently selects the exploratory
target minimizing \code{d}, which is Equation 17 of Chen (2023) made invariant to
whole-column reflection. Ties break by larger finite absolute Tucker
congruence, then by smaller target index.

The assignment is independent, not one-to-one, so two sources may select the
same target. That reuse is a \strong{collision}: it is recorded rather than
repaired, the pair keeps every otherwise defined measurement, and the target
left unselected still counts toward \code{unmatched_ssl}. There is no salience
screen: every column participates, and a comparison always contains exactly
\code{K_from} scored pairs.

A zero-norm column is no exception. Its aligned distance is defined, so it
takes part in the assignment and in \code{rmsd} and \code{rmsd_max}, but its Tucker
congruence is \code{0/0} and therefore undefined. An undefined congruence is
reported as \code{NA_real_} and makes that comparison's \code{phi_min} \code{NA_real_};
it is never dropped by an \code{na.rm}.

Because \code{phi_min} is a minimum over \emph{every} column, an overextracted
candidate carrying a near-empty column will report a low \code{phi_min} and often
\code{collision = TRUE}. Read that with \code{\link[=ssl]{ssl()}}: it usually means the candidate
contains a column with nothing in it, not that the retained structure moved.
}

\section{Identification}{

\code{pefa()} imposes no upper bound on \code{K} relative to \code{J}. The Ledermann bound
\verb{K <= (2J + 1 - sqrt(8J + 1)) / 2} orients the choice of \code{Kmax}, but it is
derived for an unrestricted factor model and does not constrain a
regularized partially confirmatory candidate, so applying it as a package
check would be wrong.

Beyond that bound expect candidates that do not converge and fit indices
that become \code{NA}: \code{RMSEA} and \code{TLI} are undefined at non-positive degrees of
freedom. Both are reported evidence about the window rather than failures.
\code{rank_adjust = TRUE} is the sharper diagnostic, since it counts the rank of
the Jacobian of the unique covariance elements and therefore tests local
identification directly; that count cannot exceed \code{J(J + 1)/2}, so its
degrees of freedom saturate at zero instead of going negative.
}

\section{Removed in 0.9.0}{

\code{select_K_elbow()}, the \code{pefa()} arguments \code{cuts}, \code{sustain}, and
\code{stability_eps}, and the \verb{$selected_K} and \verb{$boundary} components existed
in 0.8.3 and are gone. That is an intentional pre-1.0 breaking change rather
than a deprecation cycle: the package owns no count rule, so it can own
neither a selector nor a stability cutoff. The replacement is ordinary
analysis code over the stored path, as in the last chunk of the Examples,
and such code has to say which variant it is, because look-ahead and
boundary-fallback variants disagree on the same path. \code{NEWS.md} lists the
removals in full.
}

\examples{
\donttest{
sim <- sim_fa(N = 300, K = 3, ipf = 4, lam = .7, lac = .3, rseed = 1)
Q0 <- matrix(-1L, ncol(sim$dat), 2)
groups <- rep(1:3, each = 4)
for (k in 1:2) Q0[which(groups == k)[1:2], k] <- 1L

fit <- pefa(Q0, sim$dat, 2, 4, verbose = FALSE)
fit$sweep
fit$transitions[, c("K_from", "K_to", "ELBO_gain_pct", "phi_min")]
fit$persistence$phi
ssl(fit)

## A count rule is the analysis's, never the package's. This one stops at
## the first ELBO gain under 20\% and looks no further ahead; a
## full-look-ahead variant can answer differently on the same path.
g <- fit$transitions$ELBO_gain_pct
fit$transitions$K_from[which(g < 20)[1]]
}

}
\references{
Chen, J., & Jin, Y. (2026). Recovering latent structures after variational
Bayesian variable selection: Fit assessment and factor-number selection in
partially exploratory factor analysis. \emph{arXiv preprint} arXiv:2607.07159.

Chen, J. (2023). Fully and partially exploratory factor analysis with
bi-level Bayesian regularization. \emph{Behavior Research Methods}, 55(4),
2125--2142. \doi{10.3758/s13428-022-01884-7}
}
\seealso{
\code{\link[=vbfa]{vbfa()}}, \code{\link[=fit_stats]{fit_stats()}}, \code{\link[=ssl]{ssl()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_stats.R
\name{fit_stats}
\alias{fit_stats}
\alias{fit_stats.vbfa}
\alias{fit_stats.vbmimic}
\alias{fit_stats.vbpm_fit}
\title{Fit statistics for vbpm models}
\usage{
fit_stats(object, ...)

\method{fit_stats}{vbfa}(
  object,
  Y = NULL,
  Q = object$Q,
  tau = 0.5,
  gamma = 0.5,
  orthogonal = NULL,
  rank_adjust = FALSE,
  rank_max_J = 100,
  ...
)

\method{fit_stats}{vbmimic}(object, Y = NULL, Q = NULL, tau = 0.5, ...)

\method{fit_stats}{vbpm_fit}(object, ...)
}
\arguments{
\item{object}{A fitted \code{vbpm_fit} object.}

\item{...}{Further arguments passed to methods. Unused by the methods
supplied with the package.}

\item{Y}{The \verb{N x J} data matrix the model was fit to.}

\item{Q}{The \verb{J x K} loading design matrix used in the fit.}

\item{tau}{Hard-selection threshold on the posterior inclusion probability
(default \code{0.5}).}

\item{gamma}{Reserved (extended BIC tuning); currently unused in the
returned vector.}

\item{orthogonal}{Logical; whether the fit is an orthogonal (bifactor)
model, in which case factor correlations are not free parameters. The
default \code{NULL} reads the setting from the fit object itself, which is the
recommended use. Supplying a value that contradicts the fit raises an
error rather than silently miscounting parameters.}

\item{rank_adjust}{Logical; if \code{TRUE}, compute the effective parameter count
as the rank of the Jacobian of the model-implied covariance. This explicit
opt-in requires the suggested \code{numDeriv} package. The default \code{FALSE}
uses the nominal parameter count.}

\item{rank_max_J}{Largest number of observed variables for which an explicit
Jacobian-rank calculation is allowed (default \code{100}). If \code{rank_adjust = TRUE} and \code{J} exceeds this value, the method errors rather than silently
substituting the nominal count.}
}
\value{
An object of class \code{vbpm_fit_stats} (a named numeric vector) with
\code{t_nom}, \code{t}, \code{t_S} (parameter counts), \code{RMSEA}, \code{SRMR}, \code{CFI}, \code{TLI},
\code{AIC}, \code{BIC} (hard selection), \code{AIC_S}, \code{BIC_S} (soft selection), \code{ELBO},
and \code{objective}. The last is the model's comparability score: equal to
\code{ELBO} for a diagonal fit, and the terminal VECM objective under local
dependence, where \code{ELBO} is \code{NA}. It is comparable only across fits
sharing the \code{objective_type} attribute, which is also attached.

The \code{vbmimic} method returns the same shape with \code{n_active_coef} in place
of a full soft count and typed \code{NA} elsewhere; see \code{\link[=vbmimic]{vbmimic()}}.
}
\description{
Computes structural-equation-model fit statistics for a fitted \code{\link[=vbfa]{vbfa()}}
object under hard selection (an unspecified loading counts as active when its
posterior inclusion probability is at least \code{tau}). Absolute indices
(RMSEA, SRMR, CFI, TLI) and information criteria (AIC, BIC) are returned,
together with the model's ELBO.
}
\examples{
sim <- sim_fa(N = 300, K = 3, ipf = 6, lam = .7, lac = .3, rseed = 1)
Q <- matrix(-1L, ncol(sim$dat), 3)
for (k in 1:3) { a <- which(rep(1:3, each = 6) == k)[1:2]
                 Q[a, ] <- 0L; Q[a, k] <- 1L }
fit <- vbfa(sim$dat, Q)
round(fit_stats(fit), 3)

\donttest{
## bifactor: orthogonal is read from the fit object, so the same call works
Qb <- cbind(1L, Q)
fb <- vbfa(sim$dat, Qb, orthogonal = TRUE)
round(fit_stats(fb), 3)
}

}
\references{
Chen, J., & Jin, Y. (2026). Recovering latent structures after variational
Bayesian variable selection: Fit assessment and factor-number selection in
partially exploratory factor analysis. \emph{arXiv preprint} arXiv:2607.07159.

Jin, Y., & Chen, J. (2025). Regularized variational approximation for
partially confirmatory factor analysis. \emph{Structural Equation Modeling: A
Multidisciplinary Journal}, 32(3), 437-449.
\doi{10.1080/10705511.2024.2432612}
}
\seealso{
\code{\link[=vbfa]{vbfa()}}, \code{\link[=pefa]{pefa()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vbpm_fit.R
\name{print.vbpm_fit}
\alias{print.vbpm_fit}
\title{Compact display of a fitted vbpm model}
\usage{
\method{print}{vbpm_fit}(x, tau = 0.5, ...)
}
\arguments{
\item{x}{A fit returned by \code{\link[=vbfa]{vbfa()}} or \code{\link[=vbmimic]{vbmimic()}}.}

\item{tau}{Threshold on the posterior inclusion probability used only for
the "active loadings" line (default \code{0.5}, matching \code{\link[=fit_stats]{fit_stats()}}).}

\item{...}{Ignored (present for compatibility with the \code{\link[=print]{print()}} generic).}
}
\value{
\code{x}, invisibly (the R convention for print methods, so \code{print(fit)}
can be used in pipes without losing the object).
}
\description{
Prints a short summary -- model type, dimensions, convergence, timing, and
selection results -- instead of dumping the full list. Access the parts as
usual (\code{fit$Lam} etc.; see \link{vbpm_fit}); \code{str(fit)} shows everything.
}
\examples{
sim <- sim_fa(N = 300, K = 3, ipf = 6, lam = .7, lac = .3, rseed = 1)
Q <- matrix(-1L, ncol(sim$dat), 3)
for (k in 1:3) { a <- which(rep(1:3, each = 6) == k)[1:2]
                 Q[a, ] <- 0L; Q[a, k] <- 1L }
fit <- vbfa(sim$dat, Q)
fit               # dispatches here: compact summary, not a list dump
fit$Lam[1:3, ]    # the object is still an ordinary list underneath

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vbfa.R
\name{vbfa}
\alias{vbfa}
\title{Variational Bayes partially confirmatory factor analysis}
\usage{
vbfa(
  Y,
  Q,
  ld = FALSE,
  Qe = NULL,
  orthogonal = FALSE,
  bifactor = FALSE,
  general = 1,
  v0 = c(0.01, 0.005, 0.002, 0.001),
  max_it = 5000,
  convChk = FALSE,
  tolVal = 1e-04,
  ld_control = list()
)
}
\arguments{
\item{Y}{Numeric \verb{N x J} data matrix (observations by items). Standardized
internally. Missing continuous responses are handled by deterministic
Gaussian conditional-moment augmentation inside the variational loop.}

\item{Q}{Integer \verb{J x K} design matrix for the loadings: \code{1} = specified
(anchored) loading, \code{0} = fixed zero, \code{-1} = unspecified (estimated by
spike-and-slab). \code{K} is the number of factors.}

\item{ld}{Logical switch for local-dependence (sparse residual) estimation.
\code{FALSE} (default) fits diagonal residuals and ignores \code{Qe} (with a warning
if one is supplied). \code{TRUE} estimates a sparse residual precision by a
graphical spike-and-slab prior solved by QUIC, with the search space given
by \code{Qe}.}

\item{Qe}{Residual design matrix (\verb{J x J}), read only when \code{ld = TRUE}.
Entries \code{1} (freely estimated residual dependence), \code{-1} (uncertain;
selected by spike-and-slab), \code{0} (shrunk to zero). Symmetric; its diagonal
is ignored. The default \code{NULL} means fully exploratory local dependence,
i.e. \code{Qe = matrix(-1, J, J)}.}

\item{orthogonal}{Logical. If \code{TRUE}, the factor correlation is fixed at
the identity (a plain orthogonal factor model; also the engine under a
hand-built bifactor design). Default \code{FALSE} (oblique; factor
correlation estimated). For bifactor models prefer \code{bifactor = TRUE},
which implies orthogonality and keeps \code{K} counting group factors.}

\item{bifactor}{Logical. If \code{TRUE}, fit an orthogonal bifactor model with
\code{Q} supplying only the \strong{group} design (\verb{J x K} for \code{K} group factors):
the general column is added internally according to \code{general}, so the
user-facing \code{K} counts group factors and is directly comparable to an
oblique \code{K}-factor model. Implies orthogonal factors --
\code{bifactor = TRUE} overrides \code{orthogonal = FALSE} (with a message).
Supplying a \code{Q} that already contains an all-specified column is an
error: with \code{bifactor = TRUE} the general column must come from
\code{general}, not from \code{Q}.}

\item{general}{Design of the general column when \code{bifactor = TRUE}: a
scalar or length-\code{J} vector with the usual codes (\code{1} specified, \code{0}
fixed zero, \code{-1} unspecified). The default \code{1} gives the standard
bifactor (every item loads on the general factor). A vector with \code{0}
entries yields bifactor-(S-1)-style designs in which some items carry
no general loading; at least one entry must be \code{1}.}

\item{v0}{Loading-spike variance. A decreasing \strong{vector} gives a
warm-started dynamic regularization path (each stage runs to convergence
and warm-starts the next); a \strong{scalar} gives a single fixed spike. The
default path ends at \code{0.001}; a scalar \code{v0 = 0.001} reproduces the
fixed-spike estimator exactly.

For a \strong{gentler path that aids convergence}, pass a finer, longer,
log-spaced descent starting from a larger \code{v0}, e.g.
\code{v0 = exp(seq(log(0.05), log(0.001), length.out = 10))}. Each stage is
then a small perturbation of the previous one, so its warm start is a good
initialization and it converges quickly and stably. Prefer this when a fit
fails to converge within \code{max_it}, when the loading solution looks
unstable, or for near-singular / strongly regularized models: it navigates
the non-convex spike-and-slab landscape better than a single hard fit, and
is usually preferable to (and often faster than) simply raising \code{max_it}.
Keep the endpoint at your target; log spacing suits the multiplicative
spike.}

\item{max_it}{Maximum variational iterations \strong{per stage}. Default \code{5000}.}

\item{convChk}{Logical; report per-iteration relative error via
\code{\link[=message]{message()}} (so it can be silenced with \code{\link[=suppressMessages]{suppressMessages()}}). Default
\code{FALSE} (quiet).}

\item{tolVal}{Convergence tolerance on the maximum relative error.}

\item{ld_control}{A named list of local-dependence controls, read only when
\code{ld = TRUE}: \code{xi0} (spike-penalty path, units of \code{N}; default
\code{seq(0.1, 1, length.out = 5)}), \code{xi1} (slab penalty, units of \code{N};
default \code{0.01}), \code{diag_penalty} (\code{1} penalizes the precision diagonal with
\code{xi1}, \code{0} leaves it unpenalized), \code{quic_eps}, \code{quic_max_it}, and the Beta
prior parameters \code{a1}, \code{b1} on the local-dependence proportion.}
}
\value{
An object of class \code{c("vbfa", "vbpm_fit")} -- a named list (see
\link{vbpm_fit} for what the class does and does not change) with elements:
\describe{
\item{Lam}{\verb{J x K} posterior mean loadings.}
\item{pi}{Posterior inclusion probabilities of the unspecified loadings
(specified entries are reported as coded in \code{Q}).}
\item{eta}{\verb{N x K} posterior mean factor scores.}
\item{Phi}{Factor correlation matrix; the identity when
\code{orthogonal = TRUE}.}
\item{PsiInv}{Residual precisions; \code{NULL} when \code{ld = TRUE}.}
\item{ELBO}{Evidence lower bound; \code{NA} when \code{ld = TRUE}.}
\item{rho}{Posterior inclusion rate of the unspecified loadings.}
\item{Lam_var}{Posterior variances of the loadings.}
\item{eta_cov}{Posterior covariance of the factor scores.}
\item{Phi_inv_mean}{Posterior mean of the factor precision,
E[Phi^-1].}
\item{iter, flag, time}{Total iterations, \code{1} if the final stage
converged within \code{max_it} (else \code{0}), and elapsed time.}
\item{Q, Qe, orthogonal, ld}{The design matrices and settings the model
was fit with (\code{Qe} is \code{NULL} when \code{ld = FALSE}); read by \code{\link[=fit_stats]{fit_stats()}}
and \code{\link[=print.vbpm_fit]{print.vbpm_fit()}} so they never need repeating.}
\item{bifactor, n_general, general}{Bifactor bookkeeping: whether the
fit declared a bifactor structure, how many general columns it has,
and the general design (\code{NULL} unless \code{bifactor = TRUE}).}
\item{Psi, W, q_star, xi_star, tau}{Local-dependence results (\code{NULL}
when \code{ld = FALSE}): residual precision, its inverse, residual-edge
inclusion probabilities, edge shrinkage, and the LD proportion.}
\item{path}{The \code{v0} schedule (and \code{xi0} under LD) with per-stage
iteration counts.}
\item{preprocess}{Centering and scaling used, \code{n_missing}, the response
type, and \code{missing_mask} -- the mask is stored only when the data
actually contain missing values, and is \code{NULL} otherwise.}
\item{sample_cov}{The sample correlation matrix the fit statistics use.}
}

Every quantity appears exactly once: as of 0.7.0 the fit no longer
carries \verb{$coefficients}, \verb{$design}, \verb{$posterior}, or \verb{$settings}, which
duplicated the elements above.
}
\description{
Fits a partially confirmatory factor-analytic measurement model by
regularized mean-field variational Bayes. Loadings on \emph{specified} factors
are anchored by the design matrix \code{Q}; loadings on \emph{unspecified} factors are
selected from the data through continuous spike-and-slab priors. Optional
features are a warm-started dynamic regularization path, an orthogonal
bifactor parameterization, and sparse residual (local-dependence)
estimation.
}
\section{Notes}{

The diagonal-residual estimator is \strong{deterministic} (fixed initialization,
no random numbers), so results are reproducible without a seed; there is no
\code{rseed} argument. (The local-dependence branch's QUIC solver shuffles its
active set with the RNG, but that is a convex solve, so the shuffle order
does not change the solution.)

Missing continuous responses are supported (see \code{Y}). Support for
categorical or mixed responses is not yet implemented.

Under local dependence \code{ELBO} is \code{NA} by design, and this is not a
placeholder: the residual precision and the mixing proportion are point
updates rather than variational factors, so no single joint mean-field
bound over all unknowns is defined. The quantities that \emph{are} defined are
returned under their own names --- \code{objective} (the terminal VECM objective,
with \code{objective_type = "vecm"}) and \code{ELBO_conditional} (the conditional
bound given the terminal precision and mixing proportion). Use \code{objective}
for model comparison, and only across fits sharing one \code{objective_type}.
}

\examples{
## partially confirmatory: two anchors per factor, the rest exploratory
sim <- sim_fa(N = 200, K = 3, ipf = 6, lam = .7, lac = .3, rseed = 1)
Q <- matrix(-1L, ncol(sim$dat), 3)
for (k in 1:3) { a <- which(rep(1:3, each = 6) == k)[1:2]; Q[a, ] <- 0; Q[a, k] <- 1 }
fit <- vbfa(sim$dat, Q)
fit$flag                    # 1 if converged
round(fit$Lam, 2)           # loadings
round(fit$pi, 2)            # posterior inclusion probabilities

\donttest{
## orthogonal bifactor, preferred form: Q holds only the GROUP design and
## K counts group factors (comparable to an oblique K-factor model)
fb <- vbfa(sim$dat, Q, bifactor = TRUE, v0 = .001)  # fixed spike: fast demo
fb                          # prints "1 general + 3 group factors"
fb$Phi                      # identity by construction
## equivalent long form: vbfa(sim$dat, cbind(1L, Q), orthogonal = TRUE)

## local dependence: simulate correlated residuals, then set ld = TRUE
simLD <- sim_fa(N = 300, K = 3, ipf = 6, lam = .7, lac = .3, ecr = .3,
                rseed = 2)
fLD <- vbfa(simLD$dat, Q, ld = TRUE, max_it = 300,
            tolVal = 1e-3)
## largest recovered residual edges vs the planted ones
Poff <- abs(fLD$Psi); Poff[lower.tri(Poff, diag = TRUE)] <- 0
which(Poff >= sort(Poff, decreasing = TRUE)[3], arr.ind = TRUE)
simLD$ofd_ind

## empirical illustration: NLSY 1997, 27 items treated as continuous.
## A 500-row subset keeps the example quick; the vignette fits all 3,458.
data(nlsy27)
Yn <- as.matrix(nlsy27$dat)[1:500, ]
sum(is.na(Yn))            # incomplete cells, imputed in-loop
fn <- vbfa(Yn, nlsy27$Q)
round(fn$Lam, 2)
}

}
\references{
Chen, J., Guo, Z., Zhang, L., & Pan, J. (2021). A partially confirmatory
approach to scale development with the Bayesian Lasso. \emph{Psychological
Methods}, 26(2), 210-235. \doi{10.1037/met0000293}

Chen, J. (2021). A generalized partially confirmatory factor analysis
framework with mixed Bayesian Lasso methods. \emph{Multivariate Behavioral
Research}, 57(6), 879-894. \doi{10.1080/00273171.2021.1925520}

Jin, Y., & Chen, J. (2025). Regularized variational approximation for
partially confirmatory factor analysis. \emph{Structural Equation Modeling: A
Multidisciplinary Journal}, 32(3), 437-449.
\doi{10.1080/10705511.2024.2432612}

Jin, Y., Chen, J., Yan, Z., & Zhang, Y. (2026). Sparse residual estimation
in partially confirmatory factor analysis. \emph{PsyArXiv preprint}.
\doi{10.31234/osf.io/dehtv_v2} --- the warm-started \code{v0} path and the
\code{ld = TRUE} sparse-residual (local dependence) branch.

Rockova, V., & George, E. I. (2018). The spike-and-slab LASSO.
\emph{Journal of the American Statistical Association}, 113(521), 431-444.
\doi{10.1080/01621459.2016.1260469}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vbpm_fit.R
\name{coef.vbpm_fit}
\alias{coef.vbpm_fit}
\title{Extract coefficients from a fitted vbpm model}
\usage{
\method{coef}{vbpm_fit}(object, ...)
}
\arguments{
\item{object}{A fit returned by \code{\link[=vbfa]{vbfa()}} or \code{\link[=vbmimic]{vbmimic()}}.}

\item{...}{Ignored.}
}
\value{
A matrix (vbfa) or a list of two matrices (vbmimic).
}
\description{
For a \code{\link[=vbfa]{vbfa()}} fit, returns the loading matrix. For a \code{\link[=vbmimic]{vbmimic()}} fit,
returns a list with the measurement loadings \code{A} and the structural
coefficients \code{B}.
}
\examples{
sim <- sim_fa(N = 300, K = 3, ipf = 6, lam = .7, lac = .3, rseed = 1)
Q <- matrix(-1L, ncol(sim$dat), 3)
for (k in 1:3) { a <- which(rep(1:3, each = 6) == k)[1:2]
                 Q[a, ] <- 0L; Q[a, k] <- 1L }
fit <- vbfa(sim$dat, Q)
round(coef(fit), 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pefa_methods.R
\name{print.pefa}
\alias{print.pefa}
\alias{summary.pefa}
\alias{print.summary.pefa}
\alias{plot.pefa}
\title{Methods for a PEFA factor-count sweep}
\usage{
\method{print}{pefa}(x, ...)

\method{summary}{pefa}(object, ...)

\method{print}{summary.pefa}(x, digits = 3, ...)

\method{plot}{pefa}(
  x,
  type = c("objective", "gain", "fit"),
  criterion = c("ELBO", "AIC", "BIC"),
  pct = FALSE,
  ...
)
}
\arguments{
\item{...}{Further arguments; for \code{plot()}, graphical arguments passed to
the first base plot in the selected display.}

\item{object, x}{A result returned by \code{\link[=pefa]{pefa()}}; for
\code{print.summary.pefa()}, the object returned by \code{summary()}.}

\item{digits}{Number of decimals used when printing measured quantities.}

\item{type}{Display: criterion trajectories (\code{"objective"}), one gain
trajectory (\code{"gain"}), or descriptive absolute and incremental fit
trajectories (\code{"fit"}).}

\item{criterion}{Criterion path. \code{"ELBO"}, \code{"AIC"}, or \code{"BIC"} for
\code{type = "objective"}; \code{"ELBO"} or \code{"BIC"} for \code{type = "gain"}, which has
no AIC path. Ignored when \code{type = "fit"}.}

\item{pct}{Gain view only. \code{FALSE} (the default) computes the raw oriented
gains \code{ELBO_to - ELBO_from} or \code{BIC_from - BIC_to} from \verb{$sweep};
\code{TRUE} reads the stored \code{ELBO_gain_pct} or \code{BIC_gain_pct} column of
\verb{$transitions}, which is \code{NA} for a path with no positive gain.}
}
\value{
\code{print()} and \code{plot()} return their input invisibly.
\code{summary()} returns a list of class \code{"summary.pefa"} with exactly the
components \code{window} (named integer \code{Kmin}, \code{Kmax}, \code{K0}), \code{sweep},
\code{transitions}, \code{persistence}, \code{ssl}, \code{settings}, and \code{nonconverged_K}
(an integer vector of the K values whose candidate did not converge,
\code{integer(0)} when all converged).
}
\description{
Descriptive readers for the object returned by \code{\link[=pefa]{pefa()}}. They report what
was fitted and measured and nothing else: no method selects, recommends, or
marks a factor count, and none draws a threshold, because the package owns
no count rule. Count rules, persistence thresholds, and collision policies
belong to the analysis that consumes this object.
}
\seealso{
\code{\link[=pefa]{pefa()}}, \code{\link[=ssl]{ssl()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sim_lvm.R
\name{sim_lvm}
\alias{sim_lvm}
\title{Simulate data from a latent variable model}
\usage{
sim_lvm(
  N = 1000,
  mla = NULL,
  K = 3,
  J = 18,
  cpf = 0,
  lam = 0.7,
  lac = 0.3,
  phi = 0.3,
  ph12 = -1,
  phx = NULL,
  ecr = 0,
  P = 0,
  b = 0.3,
  K1 = 0,
  ph1 = 0.2,
  b1 = 0.3,
  phd = 0,
  ilvl = NULL,
  cati = NULL,
  noc = c(4),
  misp = 0,
  ome_out = FALSE,
  necw = K,
  necb = K,
  add_ind = c(),
  add_la = 0.5,
  add_phi = 0,
  zero_it = 0,
  rseed = 333,
  digits = 4
)
}
\arguments{
\item{N}{Sample size.}

\item{mla}{Population loading matrix. If supplied, \code{K}, \code{J} and \code{cpf} are
taken from it.}

\item{K}{Number of factors (used when \code{mla = NULL}). Must divide \code{J}.}

\item{J}{Number of items (used when \code{mla = NULL}).}

\item{cpf}{Number of cross-loadings per factor (used when \code{mla = NULL}).}

\item{lam}{Value of the primary loadings.}

\item{lac}{Value of the cross-loadings.}

\item{phi}{Homogeneous correlation between any two factors.}

\item{ph12}{Correlation between factors 1 and 2, when it differs from \code{phi};
\code{-1} (default) to use \code{phi}.}

\item{phx}{Correlation among the observed covariates. Defaults to \code{phi},
which is LAWBL's behaviour; set \code{phx = 0} for uncorrelated covariates.}

\item{ecr}{Residual correlation (local dependence).}

\item{P}{Number of observed predictors (MIMIC model); \code{0} for none.}

\item{b}{Coefficients of the observed predictors: a scalar, or a
\verb{(K - K1) x P} matrix.}

\item{K1}{Number of latent predictors (MIMIC / structural model).}

\item{ph1}{Correlation among the latent predictors.}

\item{b1}{Coefficients of the latent predictors: a scalar, or a
\verb{(K - K1) x K1} matrix.}

\item{phd}{Correlation among the structural disturbances (the residuals of
the regression of factors on predictors). \code{0} (default) gives independent
disturbances, which is LAWBL's behaviour; a positive value reproduces
designs with a freely correlated factor residual covariance. Ignored when
\code{P = 0} and \code{K1 = 0}.}

\item{ilvl}{Levels of every item, length \code{J + P}; values below 2 mean
continuous. \code{NULL} to use \code{cati}/\code{noc} instead.}

\item{cati}{Polytomous items by index; \code{NULL} for none, \code{-1} for all. Used
only when \code{ilvl = NULL}.}

\item{noc}{Number of categories for the polytomous items.}

\item{misp}{Proportion of missingness, applied to the \code{J} items only (never
to the covariate block).}

\item{ome_out}{Return the factor scores.}

\item{necw}{Number of within-factor local dependencies.}

\item{necb}{Number of between-factor local dependencies.}

\item{add_ind}{Indices of items loading on an additional minor factor.}

\item{add_la}{Loading value on the additional minor factor.}

\item{add_phi}{Correlation between the additional minor factor and the
others.}

\item{zero_it}{Number of surplus items with zero loadings.}

\item{rseed}{Random seed.}

\item{digits}{Significant digits used while the function is running.}
}
\value{
A list with the data (\code{dat}; items first, then any covariates),
the loading matrix (\code{MLA}), factor correlations (\code{PHI}), residual
covariance (\code{PSX}), the indices of the local-dependence pairs
(\code{ofd_ind}), and, when predictors are present, the coefficient matrices
(\code{mb}, \code{mb1}) and covariate correlation (\code{PHX}).
}
\description{
Generates factor-analytic or item-response data with optional observed and
latent predictors (MIMIC / structural models), cross-loadings, local
dependence, categorical or mixed response formats, and missingness.
}
\details{
\code{sim_lvm()} and \code{\link[=sim_fa]{sim_fa()}} are complementary rather than interchangeable.
Both accept a population loading matrix directly (\code{mla}, identical
semantics in either function); where they differ is what drives the
pattern-based path when \code{mla = NULL}, and what else each one models:

\describe{
\item{\code{sim_lvm()}}{is the \emph{model-side} generator. Use it when the design
involves \strong{predictors} -- observed covariates (\code{P}, \code{b}) or latent
predictors (\code{K1}, \code{ph1}, \code{b1}) -- or \strong{mixed response formats} per
item (\code{ilvl}). This is the generator for \code{\link[=vbmimic]{vbmimic()}}.}
\item{\code{\link[=sim_fa]{sim_fa()}}}{is the \emph{structure-side} generator. Use it when the
question is about the loading pattern itself: items per factor (\code{ipf}),
alternating-sign cross-loadings (\code{alt_sign}), or minor factors
(\code{minor}, definition-based per Auerswald & Moshagen, 2019). This is the
generator for \code{\link[=vbfa]{vbfa()}}.}
}

The two generators share \code{ph12} for the factor 1-2 correlation. \code{ph1}
exists only in \code{sim_lvm()}, where it means something different: the
correlation \strong{among latent predictors} (\code{K1}), not the factor 1-2
correlation.
}
\examples{
## MIMIC data: 18 items on 3 factors, 9 covariates in disjoint blocks of 3.
## Supply `b` as a matrix for a sparse structural design; a scalar `b` makes
## every covariate predict every factor, which quickly leaves no disturbance
## variance (the function stops if it does).
B <- matrix(0, 3, 9)
for (k in 1:3) B[k, ((k - 1) * 3 + 1):(k * 3)] <- .3

s <- sim_lvm(N = 500, K = 3, J = 18, P = 9, b = B, phx = 0, rseed = 1)
dim(s$dat)      # 500 x 27 -- items first, covariates last
s$mb            # the generating structural coefficients

## correlated disturbances (unrestricted factor residual covariance)
s2 <- sim_lvm(N = 500, K = 3, J = 18, P = 9, b = B, phx = 0, phd = .3,
              rseed = 1)
round(s2$PHI, 2)

## matrix-driven: supply the population loading matrix directly
mla <- matrix(0, 12, 2)
mla[1:6, 1] <- .7; mla[7:12, 2] <- c(.8, .7, .6, .7, .8, .6)
s3 <- sim_lvm(N = 300, mla = mla, phi = .3, rseed = 4)
s3$MLA                     # the generating matrix, as supplied

}
\references{
Chen, J. (2022). \emph{LAWBL: Latent (variable) analysis with Bayesian
learning} (R package version 1.5.0).

Jin, Y., & Chen, J. (2025). Regularized variational Bayesian approximations
for variable selection in extended multiple-indicators multiple-causes
models. \emph{Multivariate Behavioral Research}.
\doi{10.1080/00273171.2025.2483253}
}
\seealso{
\code{\link[=sim_fa]{sim_fa()}} for the structure-side generator, \code{\link[=vbmimic]{vbmimic()}} and
\code{\link[=vbfa]{vbfa()}} for the estimators.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vbpm_fit.R
\name{vbpm_fit}
\alias{vbpm_fit}
\alias{vbpm_fit-class}
\title{The vbpm_fit class: what it is and why it exists}
\description{
Every estimator in \code{vbpm} returns its results with an S3 class attached:
\code{\link[=vbfa]{vbfa()}} returns \code{c("vbfa", "vbpm_fit")} and \code{\link[=vbmimic]{vbmimic()}} returns
\code{c("vbmimic", "vbpm_fit")}. This page explains what that means in
practice, because "S3 class" sounds like more machinery than it is.
}
\section{What an S3 class actually is}{

An S3 class is nothing more than a \strong{label attached to an ordinary R
object}. A \code{vbpm_fit} is still a plain named list -- the class attribute
changes none of its contents and none of the ways you already use it:
\itemize{
\item \code{fit$Lam}, \code{fit$eta}, \code{fit$ELBO}, ... work exactly as before;
\item \code{names(fit)} lists every component;
\item \code{str(fit)} shows the full structure;
\item \code{unclass(fit)} strips the label and gives back the bare list, should you
ever want it.
}

The label does exactly two things. First, generic functions such as
\code{\link[=print]{print()}} can now \strong{dispatch}: typing \code{fit} at the console no longer
floods the screen with every matrix in the object, but shows a compact
summary (model type, dimensions, convergence, ELBO, active loadings).
Second, functions receiving a fit can \strong{recognize what it is}: this is
how \code{\link[=fit_stats]{fit_stats()}} knows a bifactor fit from an oblique one without being told
(the fit carries its own \code{orthogonal} and \code{ld} settings), and how it sends a
\code{\link[=vbmimic]{vbmimic()}} fit to its own method. That method, \code{fit_stats.vbmimic()},
returns a populated but deliberately limited result -- \code{NA} for every
SEM-like index, plus a single \code{n_active_coef} count of soft-selected
measurement and structural coefficients -- because the MIMIC objective and
joint covariance derivation those indices need are not available yet. It is
the family fallback \code{fit_stats.vbpm_fit()} that stops instead of computing
something meaningless, and it applies to classes with no method of their
own.
}

\section{Why the family shares a parent class}{

The specific class (\code{"vbfa"}, \code{"vbmimic"}) comes first, the family class
\code{"vbpm_fit"} second. Methods written for \code{vbpm_fit} therefore apply to
every model in the family unless a model overrides them -- one \code{print}
method serves all estimators, and future family members (e.g. further
variational models) inherit sensible behaviour on day one.
}

\section{Why two classes rather than one class plus a model field}{

A natural-looking alternative is a single \code{"vbpm_fit"} class with a field
saying which model it is. The field exists -- every fit carries
\verb{$model} (\code{"vbfa"} or \code{"vbmimic"}), and it is the right thing to \emph{read}
when a script wants to branch on model type. But it cannot replace the
class vector, because the class vector is what R's method dispatch runs
on. With the two-element class, model-specific behaviour is added by
writing a method (\code{fit_stats.vbmimic()}), the family default by writing
\code{fit_stats.vbpm_fit()}, and a future model (say \code{vblvm}) joins by
returning \code{c("vblvm", "vbpm_fit")} and adding only the methods where it
differs -- including from outside this package. With a single class, every
generic would have to contain a hand-written \code{switch(object$model, ...)}
that this package alone can extend, and \code{inherits(fit, "vbfa")} would stop
working. This is the same convention base R uses: a \code{glm} object has class
\code{c("glm", "lm")}, not class \code{"lm"} with a family field. So: dispatch on the
classes, read \verb{$model} for display and bookkeeping.
}

\section{What is deliberately NOT hidden}{

Some R packages wrap results in opaque objects whose internals are
discouraged territory. \code{vbpm} does the opposite: the list components are
documented in each estimator's help page (\code{?vbfa}, \code{?vbmimic}) and are
part of the public API. The class adds convenience on top; it takes
nothing away.
}

\seealso{
\code{\link[=vbfa]{vbfa()}}, \code{\link[=vbmimic]{vbmimic()}}, \code{\link[=print.vbpm_fit]{print.vbpm_fit()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pefa_methods.R
\name{ssl}
\alias{ssl}
\title{Sums of squared loadings for a PEFA sweep}
\usage{
ssl(x)
}
\arguments{
\item{x}{A result returned by \code{\link[=pefa]{pefa()}}.}
}
\value{
A list with one element per candidate, named as \code{x$loadings} is
(the candidate \code{K} as a character string). Each element is the double
vector \code{colSums(L^2)} of that candidate's loading matrix and keeps the
stored loading-column names when the matrix has them. In bifactor mode
each vector covers all \code{K + 1} stored columns with the general factor
first, so the group-block scale used by every between-candidate
comparison is \code{ssl(x)[[k]][-1]}.
}
\description{
Descriptive column sizes for every candidate in a \code{\link[=pefa]{pefa()}} sweep:
\code{colSums(Lam^2)} on the stored loading matrix of each candidate. They are
not Gram eigenvalues, they take part in no correspondence or selection
rule, and no cutoff is applied to them. \code{summary()} calls this function, so
there is one definition.
}
\details{
A low \code{phi_min} in \verb{$transitions} or \verb{$persistence$phi} is a weakest-link
reading over every source column, including one the fit has driven to
nearly zero. Such a column matches the smallest-norm target and carries an
essentially arbitrary congruence. \code{ssl()} is what separates "this candidate
contains a column with nothing in it" from "the retained structure moved".
}
\seealso{
\code{\link[=pefa]{pefa()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vbpm-package.R
\docType{package}
\name{vbpm-package}
\alias{vbpm}
\alias{vbpm-package}
\title{vbpm: Variational Bayes Psychometric Models}
\description{
Variational Bayes estimation for a family of psychometric measurement models. Two models are provided. Variational Bayes factor analysis (vbfa) is a regularized partially confirmatory factor model spanning the confirmatory-exploratory continuum via spike-and-slab priors on the loadings (Chen, Guo, Zhang, and Pan, 2021 \doi{10.1037/met0000293}; Chen, 2023 \doi{10.3758/s13428-022-01884-7}; Jin and Chen, 2025 \doi{10.1080/10705511.2024.2432612}), with an optional dynamic (warm-started) regularization path, an orthogonal bifactor parameterization, and optional sparse residual (local dependence) estimation through a graphical spike-and-slab prior solved by QUIC (Jin, Chen, Yan, and Zhang, 2026 \doi{10.31234/osf.io/dehtv_v2}). Regularized MIMIC (vbmimic) extends this to multiple-indicators multiple-causes models, placing spike-and-slab priors on both the measurement and the structural part (Jin and Chen, 2025 \doi{10.1080/00273171.2025.2483253}). Companion tools compute SEM-like fit statistics, and sweep a factor-count window to report candidate fit, criterion, and between-candidate loading-correspondence measurements without selecting a count (Chen and Jin, 2026 \doi{10.48550/arXiv.2607.07159}). Data generators for either model family are also provided.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/Jinsong-Chen/vbpm}
  \item Report bugs at \url{https://github.com/Jinsong-Chen/vbpm/issues}
}

}
\author{
\strong{Maintainer}: Jinsong Chen \email{jinsong.chen@live.com} (\href{https://orcid.org/0000-0002-0157-5469}{ORCID})

Authors:
\itemize{
  \item Jinsong Chen \email{jinsong.chen@live.com} (\href{https://orcid.org/0000-0002-0157-5469}{ORCID})
  \item Yi Jin (\href{https://orcid.org/0000-0002-8604-5992}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sim_fa.R
\name{sim_fa}
\alias{sim_fa}
\title{Simulating data with Latent Variable Modeling}
\usage{
sim_fa(
  N = 1000,
  mla = NULL,
  K = 3,
  ipf = 8,
  cpf = 2,
  lam = 0.7,
  lac = 0.3,
  alt_sign = TRUE,
  phi = 0.5,
  ph12 = -1,
  gamma = NULL,
  ecr = 0,
  ome_out = FALSE,
  cati = NULL,
  noc = c(4),
  misp = 0,
  rseed = 333,
  necw = K,
  necb = K,
  add_ind = c(),
  add_la = 0,
  add_phi = 0,
  add1_ind = c(),
  add1_la = 0,
  add1_phi = 0,
  zero_it = 0,
  minor = NULL,
  digits = 4
)
}
\arguments{
\item{N}{Sample size.}

\item{mla}{Population loading matrix (J x K). If supplied, the pattern
arguments (\code{ipf}, \code{cpf}, \code{lam}, \code{lac}, \code{alt_sign},
\code{add_*}, \code{minor}, \code{zero_it}) are ignored and \code{K},
\code{J} are taken from it -- the same convention as \code{\link[=sim_lvm]{sim_lvm()}}.}

\item{K}{Number of factors.}

\item{ipf}{Items per factor.}

\item{cpf}{Cross-loadings per factor.}

\item{lam}{Number of formal iterations for posterior sampling.}

\item{lac}{Number of iterations to update the sampling information.}

\item{alt_sign}{Logical; alternate the sign of the cross-loadings across the
cross-loading items (default \code{TRUE}).}

\item{phi}{Homogeneous correlations between any two factors.}

\item{ph12}{Correlation between factor 1 and 2 (if it's different from
\code{phi}). Shared with \code{\link[=sim_lvm]{sim_lvm()}}, which uses \code{ph1} for a
different purpose (correlation among latent predictors); see \code{\link[=sim_lvm]{sim_lvm()}}
for that distinction.}

\item{gamma}{Optional second-order loadings for direct \strong{higher-order}
generation: a scalar or length-\code{K} vector in \eqn{(0, 1)}. When
supplied, each group factor \eqn{\eta_k} is driven by a general factor
\eqn{\xi} through \eqn{\eta_k = \gamma_k \xi + \zeta_k}, and the data are
generated from the equivalent Schmid-Leiman orthogonal bifactor matrix:
general loadings \eqn{\lambda_j \gamma_k} and group loadings
\eqn{\lambda_j \sqrt{1 - \gamma_k^2}} (a testlet model is the same model).
The returned \code{MLA} is this \eqn{J \times (K+1)} bifactor matrix
(general factor first), \code{PHI} is the identity, and the first-order
matrix and \code{gamma} are returned as \code{MLA1} and \code{gamma}.
Incompatible with \code{mla}, \code{cpf > 0}, \code{add_ind},
\code{add1_ind}, \code{minor}, and \code{zero_it}; \code{phi}/\code{ph12}
are ignored (the SL solution is orthogonal by construction).}

\item{ecr}{Residual correlation (local dependence).}

\item{ome_out}{Output factor score or not.}

\item{cati}{The set of categorical (polytomous) items in sequence number (i.e., 1 to \eqn{J});
\code{NULL} for no and -1 for all (default is \code{NULL}).}

\item{noc}{Number of categories for categorical items}

\item{misp}{Proportion of missingness.}

\item{rseed}{An integer for the random seed.}

\item{necw}{Number of within-factor local dependence.}

\item{necb}{Number of between-factor local dependence.}

\item{add_ind}{(Additional) minor factor with cross-loadings.}

\item{add_la}{Value of cross-loadings on (Additional) minor factor.}

\item{add_phi}{Correlations between (Additional) minor factor and other factors.}

\item{add1_ind}{A second (additional) minor factor with cross-loadings,
given as item indices; \code{c()} for none.}

\item{add1_la}{Value of the cross-loadings on the second minor factor.}

\item{add1_phi}{Correlations between the second minor factor and other
factors.}

\item{zero_it}{Surplus items with zero loading.}

\item{minor}{Optional character vector requesting orthogonal minor factors
loading on every indicator, each \code{"weak"} or \code{"moderate"}
(Auerswald & Moshagen, 2019); \code{NULL} for none.}

\item{digits}{Number of significant digits to print when printing numeric values.}
}
\value{
An object of class \code{list} containing the data, loading, and factorial correlation matrix.
}
\description{
\code{sim_fa} can simulate data based on factor analysis or
item response models with different response formats (continuous or categorical),
loading patterns and residual covariance (local dependence) structures.

\code{sim_fa()} and \code{\link[=sim_lvm]{sim_lvm()}} are complementary rather than
interchangeable: \code{sim_fa()} is the \emph{structure-side} generator, driven
by the loading pattern itself -- items per factor (\code{ipf}), alternating-sign
cross-loadings (\code{alt_sign}), or minor factors (\code{minor}, definition-based
per Auerswald & Moshagen, 2019) -- and is the generator for \code{\link[=vbfa]{vbfa()}}.
\code{\link[=sim_lvm]{sim_lvm()}} is the \emph{model-side} generator, adding observed/latent
predictors (\code{P}/\code{b}, \code{K1}/\code{ph1}/\code{b1}) and mixed response formats (\code{ilvl}),
and is the generator for \code{\link[=vbmimic]{vbmimic()}}. Both generators also accept a
population loading matrix directly (\code{mla}) with identical semantics.
}
\examples{

# for continuous data with cross-loadings and local dependence effect .3
out <- sim_fa(N=1000,K=3,ipf=6,lam = .7, lac=.3,ecr=.3)
summary(out$dat)
out$MLA
out$ofd_ind

# for categorical data with cross-loadings .4 and 10\% missingness
out <- sim_fa(N=1000,K=3,ipf=6,lam = .7, lac=.4,cati=-1,noc=4,misp=.1)
summary(out$dat)
out$MLA
out$ofd_ind

## matrix-driven: supply the population loading matrix directly
mla <- matrix(0, 12, 2)
mla[1:6, 1] <- .7; mla[7:12, 2] <- c(.8, .7, .6, .7, .8, .6)
s <- sim_fa(N = 300, mla = mla, phi = .3, rseed = 4)
s$MLA                      # the generating matrix, as supplied

## higher-order (testlet) data, generated directly: 3 group factors driven
## by one general factor with second-order loadings .8, .7, .6
s <- sim_fa(N = 500, K = 3, ipf = 6, lam = .7, gamma = c(.8, .7, .6))
round(s$MLA, 2)            # J x 4 Schmid-Leiman bifactor matrix (general first)
s$gamma                    # the second-order loadings

}
